/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <string.h>
#include <stdbool.h>

#include "bluetooth-api.h"
#include "bt-internal-types.h"

#include "bt-common.h"
#include "bt-request-sender.h"
#include "bt-event-handler.h"

GVariant *__g_variant_new_array_split(char *str)
{
	GVariantBuilder b;
	g_variant_builder_init(&b, (const GVariantType*) "a{s}");
	char *pos = str;
	char *item;

	while ((item = strsep(&pos, ",")))
		g_variant_builder_add(&b, "{s}", item);

	return g_variant_builder_end(&b);
}

BT_EXPORT_API int bluetooth_map_client_init(void)
{
	BT_DBG("bluetooth_map_client_init");
	bt_user_info_t *user_info;

	user_info = _bt_get_user_data(BT_COMMON);
	retv_if(user_info->cb == NULL, BLUETOOTH_ERROR_INTERNAL);

	return _bt_register_event(BT_MAP_CLIENT_EVENT, user_info->cb, user_info->user_data);
}

BT_EXPORT_API int bluetooth_map_client_deinit(void)
{
	BT_DBG("bluetooth_map_client_deinit");
	return _bt_unregister_event(BT_MAP_CLIENT_EVENT);
}

BT_EXPORT_API int bluetooth_map_client_create_session(
	bt_map_client_session_info_s *session)
{
	BT_DBG("bluetooth_map_client_create_session");
	int result = BLUETOOTH_ERROR_INTERNAL;

	BT_CHECK_ENABLED(return);
	BT_CHECK_PARAMETER(session->remote_address, return);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	g_array_append_vals(in_param1, session->remote_address, strlen(session->remote_address)+1);

	result = _bt_send_request(BT_OBEX_SERVICE, BT_MAP_CREATE_SESSION,
		in_param1, in_param2, in_param3, in_param4, &out_param);

	if (result == BLUETOOTH_ERROR_NONE) {
		if (out_param->len > 0) {
			session->session_path = strdup(&g_array_index(out_param, char, 0));
			BT_DBG("session id: %s", session->session_path);
		} else {
			BT_ERR("out_param length is 0!!");
			return BLUETOOTH_ERROR_INTERNAL;
		}
	}
	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);
	return result;
}

BT_EXPORT_API int bluetooth_map_client_destroy_session(
	bt_map_client_session_info_s *session)
{
	BT_DBG("bluetooth_map_client_destroy_session");
	int result = BLUETOOTH_ERROR_INTERNAL;

	BT_CHECK_ENABLED(return);
	BT_CHECK_PARAMETER(session->session_path, return);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	g_array_append_vals(in_param1, session->session_path, strlen(session->session_path)+1);

	result = _bt_send_request(BT_OBEX_SERVICE, BT_MAP_DESTROY_SESSION,
		in_param1, in_param2, in_param3, in_param4, &out_param);

	if (result == BLUETOOTH_ERROR_NONE) {
		BT_DBG("session (%s) was destroyed", session->session_path);
		free(session->remote_address);
		free(session->session_path);
		free(session);
	}
	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);
	return result;
}

BT_EXPORT_API int bluetooth_map_client_set_folder(
	bt_map_client_session_info_s *session,
	const char *name)
{
	BT_DBG("bluetooth_map_client_set_folder");
	int result = BLUETOOTH_ERROR_INTERNAL;

	BT_CHECK_ENABLED(return);
	BT_CHECK_PARAMETER(session->session_path, return);
	BT_CHECK_PARAMETER(name, return);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	g_array_append_vals(in_param1, session->session_path, strlen(session->session_path) + 1);
	g_array_append_vals(in_param2, name, strlen(name) + 1);

	result = _bt_send_request(BT_OBEX_SERVICE, BT_MAP_SET_FOLDER,
		in_param1, in_param2, in_param3, in_param4, &out_param);

	if (result != BLUETOOTH_ERROR_NONE) {
		BT_ERR("bluetooth_map_client_set_folder failed");
	}

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	return result;
}

BT_EXPORT_API int bluetooth_map_client_list_folders(
		bt_map_client_session_info_s *session,
		bt_map_client_list_folders_filter_t *filter)
{
	BT_DBG("bluetooth_map_list_folders");
	int result = 0;

	BT_CHECK_ENABLED(return);
	BT_CHECK_PARAMETER(session->session_path, return);

	bt_user_info_t *user_info = _bt_get_user_data(BT_COMMON);
	retv_if(user_info->cb == NULL, BLUETOOTH_ERROR_INTERNAL);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	GVariantBuilder b;
	g_variant_builder_init(&b, (const GVariantType*) "a{sv}");
	if (filter->offset > -1)
		g_variant_builder_add(&b, "{sv}", "Offset", g_variant_new_uint16(filter->offset));
	if (filter->max_count > -1)
		g_variant_builder_add(&b, "{sv}", "MaxCount", g_variant_new_uint16(filter->max_count));
	GVariant *filter_variant = g_variant_builder_end(&b);
	char *filter_serialized = g_variant_print(filter_variant, TRUE);

	g_array_append_vals(in_param1, session->session_path, strlen(session->session_path)+1);
	g_array_append_vals(in_param2, filter_serialized, strlen(filter_serialized)+1);

	result = _bt_send_request_async(BT_OBEX_SERVICE, BT_MAP_LIST_FOLDERS,
			in_param1, in_param2, in_param3, in_param4, user_info->cb, user_info->user_data);

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	free(filter_serialized);

	return result;
}

BT_EXPORT_API int bluetooth_map_client_list_filter_fields(bt_map_client_session_info_s *session)
{
	BT_DBG("bluetooth_map_client_list_filter_fields");
	int result = BLUETOOTH_ERROR_NONE;
	bt_user_info_t *user_info;

	BT_CHECK_ENABLED(return);
	BT_CHECK_PARAMETER(session->session_path, return);

	user_info = _bt_get_user_data(BT_COMMON);
	retv_if(user_info->cb == NULL, BLUETOOTH_ERROR_INTERNAL);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	g_array_append_vals(in_param1, session->session_path, strlen(session->session_path)+1);

	result = _bt_send_request_async(BT_OBEX_SERVICE, BT_MAP_LIST_FILTER_FIELDS,
		in_param1, in_param2, in_param3, in_param4,
		user_info->cb, user_info->user_data);

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	return result;
}

BT_EXPORT_API int bluetooth_map_client_list_messages(bt_map_client_session_info_s *session,
			const char *folder,
			bt_map_client_list_messages_filter_t *filter)
{
	BT_DBG("bluetooth_map_client_list_messages");
	int result = 0;

	BT_CHECK_ENABLED(return);
	BT_CHECK_PARAMETER(session->session_path, return);
	BT_CHECK_PARAMETER(folder, return);
	BT_CHECK_PARAMETER(filter, return);

	bt_user_info_t *user_info = _bt_get_user_data(BT_COMMON);
	retv_if(user_info->cb == NULL, BLUETOOTH_ERROR_INTERNAL);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	GVariantBuilder b;
	g_variant_builder_init(&b, (const GVariantType*) "a{sv}");
	if (filter->offset > -1)
		g_variant_builder_add(&b, "{sv}", "Offset", g_variant_new_uint16(filter->offset));
	if (filter->max_count > -1)
		g_variant_builder_add(&b, "{sv}", "MaxCount", g_variant_new_uint16(filter->max_count));
	if (filter->subject_length > -1)
		g_variant_builder_add(&b, "{sv}", "SubjectLength", g_variant_new_byte(filter->subject_length));
	if (filter->fields != NULL)
		g_variant_builder_add(&b, "{sv}", "Fields", __g_variant_new_array_split(filter->fields));
	if (filter->types != NULL)
		g_variant_builder_add(&b, "{sv}", "Types", __g_variant_new_array_split(filter->types));
	if (filter->period_begin != NULL)
		g_variant_builder_add(&b, "{sv}", "PeriodBegin", g_variant_new_string(filter->period_begin));
	if (filter->period_end != NULL)
		g_variant_builder_add(&b, "{sv}", "PeriodEnd", g_variant_new_string(filter->period_end));
	if (filter->is_read > -1)
		g_variant_builder_add(&b, "{sv}", "Read", g_variant_new_boolean(filter->is_read == 1 ? TRUE : FALSE));
	if (filter->recipient != NULL)
		g_variant_builder_add(&b, "{sv}", "Recipient", g_variant_new_string(filter->recipient));
	if (filter->sender != NULL)
		g_variant_builder_add(&b, "{sv}", "Sender", g_variant_new_string(filter->sender));
	if (filter->is_priority > -1)
		g_variant_builder_add(&b, "{sv}", "Priority", g_variant_new_boolean(filter->is_priority == 1 ? TRUE : FALSE));
	GVariant *filter_variant = g_variant_builder_end(&b);
	char *filter_serialized = g_variant_print(filter_variant, TRUE);

	g_array_append_vals(in_param1, session->session_path, strlen(session->session_path)+1);
	g_array_append_vals(in_param2, folder, strlen(folder)+1);
	g_array_append_vals(in_param3, filter_serialized, strlen(filter_serialized)+1);

	result = _bt_send_request_async(BT_OBEX_SERVICE, BT_MAP_LIST_MESSAGES,
			in_param1, in_param2, in_param3, in_param4, user_info->cb, user_info->user_data);

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	free(filter_serialized);

	return result;
}

BT_EXPORT_API int bluetooth_map_client_update_inbox(bt_map_client_session_info_s *session)
{
	BT_DBG("bluetooth_map_client_update_inbox");
	int result = BLUETOOTH_ERROR_INTERNAL;

	BT_CHECK_ENABLED(return);
	BT_CHECK_PARAMETER(session->session_path, return);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	g_array_append_vals(in_param1, session->session_path, strlen(session->session_path)+1);

	result = _bt_send_request(BT_OBEX_SERVICE, BT_MAP_UPDATE_INBOX,
		in_param1, in_param2, in_param3, in_param4, &out_param);

	if (result != BLUETOOTH_ERROR_NONE) {
		BT_ERR("bluetooth_map_client_update_inbox failed");
	}

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	return result;
}

BT_EXPORT_API int bluetooth_map_client_push_message(bt_map_client_session_info_s *session,
			const char *source_file,
			const char *folder,
			bt_map_client_push_message_args_t *args)
{
	BT_DBG("Entered bluetooth_map_client_push_message");
	int result = 0;

	BT_CHECK_ENABLED(return);
	BT_CHECK_PARAMETER(session, return);
	BT_CHECK_PARAMETER(session->session_path, return);
	BT_CHECK_PARAMETER(source_file, return);
	BT_CHECK_PARAMETER(folder, return);

	bt_user_info_t *user_info = _bt_get_user_data(BT_COMMON);
	retv_if(user_info->cb == NULL, BLUETOOTH_ERROR_INTERNAL);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	GVariantBuilder b;
	g_variant_builder_init(&b, (const GVariantType*) "a{sv}");
	if (args->is_transparent > -1)
		g_variant_builder_add(&b, "{sv}", "Transparent", g_variant_new_boolean(args->is_transparent == 1 ? TRUE : FALSE));
	if (args->is_retry > -1)
		g_variant_builder_add(&b, "{sv}", "Retry", g_variant_new_boolean(args->is_retry == 1 ? TRUE : FALSE));
	if (args->charset != NULL)
		g_variant_builder_add(&b, "{sv}", "Charset", g_variant_new_string(args->charset));
	GVariant *args_variant = g_variant_builder_end(&b);
	char *args_serialized = g_variant_print(args_variant, TRUE);

	g_array_append_vals(in_param1, session->session_path, strlen(session->session_path)+1);
	g_array_append_vals(in_param2, source_file, strlen(source_file)+1);
	g_array_append_vals(in_param3, folder, strlen(folder)+1);
	g_array_append_vals(in_param4, args_serialized, strlen(args_serialized)+1);

	result = _bt_send_request_async(BT_OBEX_SERVICE, BT_MAP_PUSH_MESSAGE,
			in_param1, in_param2, in_param3, in_param4, user_info->cb, user_info->user_data);

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	return result;
}

BT_EXPORT_API int bluetooth_map_client_get_message(bt_map_client_session_info_s *session,
			const bt_map_client_message_object_t message_object,
			const char *target_file,
			bool attachment)
{
	BT_DBG("Entered bluetooth_map_client_get_message");
	int result = 0;

	BT_CHECK_ENABLED(return);
	BT_CHECK_PARAMETER(session, return);
	BT_CHECK_PARAMETER(session->session_path, return);
	BT_CHECK_PARAMETER(message_object, return);
	BT_CHECK_PARAMETER(target_file, return);

	bt_user_info_t *user_info = _bt_get_user_data(BT_COMMON);
	retv_if(user_info->cb == NULL, BLUETOOTH_ERROR_INTERNAL);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	// TODO session currently is not used, but should be valid
	//g_array_append_vals(in_param1, session->session_path, strlen(session->session_path)+1);
	g_array_append_vals(in_param2, message_object, strlen(message_object)+1);
	g_array_append_vals(in_param3, target_file, strlen(target_file)+1);
	g_array_append_vals(in_param4, &attachment, sizeof(attachment));

	result = _bt_send_request_async(BT_OBEX_SERVICE, BT_MAP_GET_MESSAGE,
			in_param1, in_param2, in_param3, in_param4, user_info->cb, user_info->user_data);

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	return result;
}
