/*
 * Copyright (c) 2011 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "bluetooth-api.h"
#include "bt-internal-types.h"

#include "bt-common.h"
#include "bt-request-sender.h"
#include "bt-event-handler.h"

static int _bluetooth_handle_get_len(const char *str)
{
	int i;
	for (i = 0; str && str[i] != '\0'; i++);
	return i;
}

BT_EXPORT_API int bluetooth_tds_provider_register(void)
{
	int result = BLUETOOTH_ERROR_INTERNAL;

	BT_CHECK_ENABLED_LE(return);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);
	BT_DBG("");

	result = _bt_send_request(BT_BLUEZ_SERVICE, BT_TDS_PROVIDER_REGISTER,
		in_param1, in_param2, in_param3, in_param4, &out_param);

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	return result;
}

BT_EXPORT_API int bluetooth_tds_provider_unregister(void)
{
	int result = BLUETOOTH_ERROR_INTERNAL;

	BT_CHECK_ENABLED_LE(return);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);
	BT_DBG("");

	result = _bt_send_request(BT_BLUEZ_SERVICE, BT_TDS_PROVIDER_UNREGISTER,
		in_param1, in_param2, in_param3, in_param4, &out_param);

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	return result;
}

BT_EXPORT_API int bluetooth_set_tds_provider_manuf_data(unsigned char *buf, unsigned int length)
{
	int result;
	bluetooth_advertising_data_t manuf_data;

	BT_CHECK_ENABLED_LE(return);

	retv_if(length > 0 && NULL == buf, BLUETOOTH_ERROR_INVALID_PARAM);
	retv_if(length > BLUETOOTH_ADVERTISING_DATA_LENGTH_MAX, BLUETOOTH_ERROR_INVALID_PARAM);

	memcpy(manuf_data.data, buf, length);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	g_array_append_vals(in_param1, &length, sizeof(unsigned int));
	g_array_append_vals(in_param2, &manuf_data, sizeof(bluetooth_advertising_data_t));

	result = _bt_send_request(BT_BLUEZ_SERVICE, BT_TDS_PROVIDER_SET_MANUF_DATA,
			in_param1, in_param2, in_param3, in_param4, &out_param);

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	return result;
}

BT_EXPORT_API int bluetooth_tds_provider_create(int transport, unsigned int handle)
{
	int result = BLUETOOTH_ERROR_INTERNAL;

	BT_CHECK_ENABLED_LE(return);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);
	BT_DBG("");

	g_array_append_vals(in_param1, &handle, sizeof(unsigned int));
	g_array_append_vals(in_param2, &transport, sizeof(unsigned int));

	result = _bt_send_request(BT_BLUEZ_SERVICE, BT_TDS_PROVIDER_CREATE,
		in_param1, in_param2, in_param3, in_param4, &out_param);

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	return result;
}

BT_EXPORT_API int bluetooth_tds_provider_destroy(unsigned int handle)
{
	int result = BLUETOOTH_ERROR_INTERNAL;

	BT_CHECK_ENABLED_LE(return);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);
	BT_DBG("");

	g_array_append_vals(in_param1, &handle, sizeof(unsigned int));

	result = _bt_send_request(BT_BLUEZ_SERVICE, BT_TDS_PROVIDER_DESTROY,
		in_param1, in_param2, in_param3, in_param4, &out_param);

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	return result;
}

BT_EXPORT_API int bluetooth_set_tds_provider_transport_data(unsigned int tds_handle,
		bluetooth_tds_transport_state_t state, unsigned char *buf, unsigned int length)
{
	int result;
	bluetooth_tds_data_t tds_data;

	BT_CHECK_ENABLED_LE(return);

	retv_if(NULL == buf, BLUETOOTH_ERROR_INVALID_PARAM);
	retv_if(length == 0 || length > BLUETOOTH_TDS_DATA_LENGTH_MAX,
			BLUETOOTH_ERROR_INVALID_PARAM);
	tds_data.length = length;
	memcpy(tds_data.data, buf, length);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	g_array_append_vals(in_param1, &tds_handle, sizeof(unsigned int));
	g_array_append_vals(in_param2, &state, sizeof(int));
	g_array_append_vals(in_param3, &tds_data, sizeof(bluetooth_tds_data_t));

	result = _bt_send_request(BT_BLUEZ_SERVICE, BT_TDS_PROVIDER_SET_TRANSPORT_DATA,
			in_param1, in_param2, in_param3, in_param4, &out_param);

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	return result;
}

BT_EXPORT_API int bluetooth_send_tds_activation_response(bluetooth_device_address_t *device_address,
	unsigned int tds_handle, int response, unsigned char *buf, unsigned int length)
{
	int result;
	bluetooth_tds_data_t tds_data;

	BT_CHECK_PARAMETER(device_address, return);
	BT_CHECK_ENABLED_LE(return);

	retv_if(length > 0 && NULL == buf, BLUETOOTH_ERROR_INVALID_PARAM);
	tds_data.length = length;
	memcpy(tds_data.data, buf, length);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	g_array_append_vals(in_param1, &tds_handle, sizeof(unsigned int));
	g_array_append_vals(in_param2, &response, sizeof(int));
	g_array_append_vals(in_param3, device_address, sizeof(bluetooth_device_address_t));
	g_array_append_vals(in_param4, &tds_data, sizeof(bluetooth_tds_data_t));

	result = _bt_send_request(BT_BLUEZ_SERVICE, BT_TDS_SEND_ACTIVATION_RESPONSE,
			in_param1, in_param2, in_param3, in_param4, &out_param);

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	return result;
}

BT_EXPORT_API int bluetooth_tds_read_transport_data(const bluetooth_device_address_t *device_address,
							const char *handle)
{
	char *path;
	int path_len = 0;
	bt_user_info_t *user_info;
	int result = BLUETOOTH_ERROR_INTERNAL;
	BT_DBG("+");

	BT_CHECK_PARAMETER(device_address, return);
	BT_CHECK_PARAMETER(handle, return);
	BT_CHECK_ENABLED_LE(return);

	user_info = _bt_get_user_data(BT_COMMON);
	retv_if(user_info->cb == NULL, BLUETOOTH_ERROR_INTERNAL);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	path = g_strdup(handle);
	path_len = _bluetooth_handle_get_len(path);
	g_array_append_vals(in_param1, path, path_len);
	g_free(path);

	g_array_append_vals(in_param2, device_address,
			sizeof(bluetooth_device_address_t));

	result = _bt_send_request_async(BT_BLUEZ_SERVICE, BT_TDS_READ_TRANSPORT_DATA,
			in_param1, in_param2, in_param3, in_param4,
			user_info->cb, user_info->user_data);

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	BT_DBG("-");
	return result;
}

BT_EXPORT_API int bluetooth_tds_enable_control_point(const bluetooth_device_address_t *device_address,
							const char *handle)
{
	char *path;
	int path_len = 0;
	bt_user_info_t *user_info;
	int result = BLUETOOTH_ERROR_INTERNAL;
	BT_DBG("+");

	BT_CHECK_PARAMETER(device_address, return);
	BT_CHECK_PARAMETER(handle, return);
	BT_CHECK_ENABLED_LE(return);

	user_info = _bt_get_user_data(BT_COMMON);
	retv_if(user_info->cb == NULL, BLUETOOTH_ERROR_INTERNAL);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	path = g_strdup(handle);
	path_len = _bluetooth_handle_get_len(path);
	g_array_append_vals(in_param1, path, path_len);
	g_free(path);

	g_array_append_vals(in_param2, device_address,
			sizeof(bluetooth_device_address_t));

	result = _bt_send_request_async(BT_BLUEZ_SERVICE, BT_TDS_ENABLE_CONTROL_POINT,
			in_param1, in_param2, in_param3, in_param4,
			user_info->cb, user_info->user_data);

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	BT_DBG("-");
	return result;
}

BT_EXPORT_API int bluetooth_tds_activate_control_point(const bluetooth_device_address_t *device_address,
					const char *handle, unsigned char *buf,
					int length)
{
	char *path;
	int path_len = 0;
	bt_user_info_t *user_info;
	int result = BLUETOOTH_ERROR_INTERNAL;
	bluetooth_control_point_data_t data;
	BT_DBG("+");

	BT_CHECK_PARAMETER(device_address, return);
	BT_CHECK_PARAMETER(handle, return);
	BT_CHECK_PARAMETER(buf, return);
	BT_CHECK_ENABLED_LE(return);

	user_info = _bt_get_user_data(BT_COMMON);
	retv_if(user_info->cb == NULL, BLUETOOTH_ERROR_INTERNAL);

	BT_INIT_PARAMS();
	BT_ALLOC_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	memset(&data, 0x00, sizeof(bluetooth_control_point_data_t));

	data.length = length;
	if (length > 0)
		memcpy(data.data, buf, length);

	path = g_strdup(handle);
	path_len = _bluetooth_handle_get_len(path);

	/*Fill parameters*/
	g_array_append_vals(in_param1, path, path_len);

	g_array_append_vals(in_param2, device_address,
			sizeof(bluetooth_device_address_t));

	g_array_append_vals(in_param3, &data, sizeof(bluetooth_control_point_data_t));

	result = _bt_send_request_async(BT_BLUEZ_SERVICE, BT_TDS_ACTIVATE_CONTROL_POINT,
			in_param1, in_param2, in_param3, in_param4,
			user_info->cb, user_info->user_data);

	BT_FREE_PARAMS(in_param1, in_param2, in_param3, in_param4, out_param);

	g_free(path);
	BT_DBG("-");
	return result;
}
