/*
 * Bluetooth HAL
 *
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact:  Nilesh Trimbake <t.shripati@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdbool.h>
#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <dlog.h>

#include "bt-hal.h"
#include "bt-hal-log.h"
#include "bt-hal-msg.h"
#include "bt-hal-utils.h"
#include "bt-hal-a2dp-sink-dbus-handler.h"

static const btav_callbacks_t *bt_hal_a2dp_sink_cbacks;

static bool interface_ready(void)
{
	return bt_hal_a2dp_sink_cbacks != NULL;
}

static bt_status_t a2dp_sink_connect(bt_bdaddr_t *bd_addr)
{
	DBG("");
	return _bt_hal_dbus_handler_a2dp_sink_connect(bd_addr);
}

static bt_status_t a2dp_sink_disconnect(bt_bdaddr_t *bd_addr)
{
	DBG("");
	return _bt_hal_dbus_handler_a2dp_sink_disconnect(bd_addr);
}

static void __bt_hal_handle_a2dp_sink_conn_state(void *buf, uint16_t len)
{
	struct hal_ev_a2dp_conn_state *ev = buf;

	if (bt_hal_a2dp_sink_cbacks->connection_state_cb)
		bt_hal_a2dp_sink_cbacks->connection_state_cb(ev->state, (bt_bdaddr_t *) ev->bdaddr);
}

static void __bt_hal_handle_a2dp_sink_audio_conn_state(void *buf, uint16_t len)
{
	struct hal_ev_a2dp_audio_state *ev = buf;

	if (bt_hal_a2dp_sink_cbacks->audio_state_cb)
		bt_hal_a2dp_sink_cbacks->audio_state_cb(ev->state, (bt_bdaddr_t *) ev->bdaddr);
}

static void __bt_hal_handle_a2dp_sink_events(int message, void *buf, uint16_t len)
{
	DBG("+");
	if (!interface_ready())
		return;
	switch (message) {
	case HAL_EV_A2DP_SOURCE_CONN_STATE:
		DBG("Event: HAL_EV_A2DP_SOURCE_CONN_STATE");
		__bt_hal_handle_a2dp_sink_conn_state(buf, len);
		break;
	case HAL_EV_A2DP_AUDIO_STATE:
		DBG("Event: HAL_EV_A2DP_AUDIO_STATE");
		__bt_hal_handle_a2dp_sink_audio_conn_state(buf, len);
		break;
	default:
		DBG("Event Currently not handled!!");
		break;
	}
	DBG("-");
}

static bt_status_t init(btav_callbacks_t* callbacks)
{
	DBG("");

	if (interface_ready())
		return BT_STATUS_DONE;

	if (BT_STATUS_SUCCESS != _bt_hal_dbus_handler_enable_a2dp_sink()) {
		ERR("_bt_hal_dbus_handler_enable_a2dp_sink failed");
		return BT_STATUS_FAIL;
	}

	bt_hal_a2dp_sink_cbacks = callbacks;
	DBG("Register A2DP Sink events callback function");
	_bt_hal_register_a2dp_sink_dbus_handler_cb(__bt_hal_handle_a2dp_sink_events);
	_bt_hal_register_event_handler_cb(HAL_A2DP_SNK, __bt_hal_handle_a2dp_sink_events);
	return BT_STATUS_SUCCESS;
}

static void cleanup(void)
{
	DBG("");

	if (!interface_ready())
		return;

	bt_hal_a2dp_sink_cbacks = NULL;
	_bt_hal_unregister_event_handler_cb(HAL_A2DP_SNK);
}

static btav_interface_t a2dp_sink_if = {
	.size = sizeof(a2dp_sink_if),
	.init = init,
	.connect = a2dp_sink_connect,
	.disconnect = a2dp_sink_disconnect,
	.cleanup = cleanup
};

btav_interface_t *bt_get_a2dp_sink_interface(void)
{
	DBG("Get A2DP Sink Profile Interface");
	return &a2dp_sink_if;
}
