/*
 * BLUETOOTH HAL
 *
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact:  Anupam Roy <anupam.r@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <malloc.h>

#include <vconf.h>
#include <bundle.h>
#include <bundle_internal.h>
#include <eventsystem.h>

#include <glib.h>
#include <gio/gio.h>
#include <dlog.h>
#include <vconf.h>

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/mman.h>

/* BT HAL Headers */
#include "bt-hal.h"
#include "bt-hal-log.h"
#include "bt-hal-msg.h"
#include "bt-hal-internal.h"
#include "bt-hal-utils.h"
#include "bt-hal-event-receiver.h"
#include "bt-hal-dbus-common-utils.h"

#include "bt-hal-adapter-dbus-handler.h"
#include "bt-hal-rfcomm-dbus-handler.h"
#include "bt-hal-event-receiver.h"

#include <bt-hal-agent.h>
#include <bt-hal-gap-agent.h>
#include <bt-hal-dbus-common-utils.h>

/* TODO_40 : 4.0 merge  */
#include "bt-internal-types.h"

#define BT_HAL_AGENT_AUTO_PAIR_BLACKLIST_FILE (APP_SYSCONFDIR"/auto-pair-blacklist")
#define BT_HAL_AGENT_NEW_LINE "\r\n"
#define BUF_SIZE                256
#define PAGE_SIZE               (1 << 12)
#define _ALIGN_UP(addr, size)    (((addr)+((size)-1))&(~((size)-1)))
#define _ALIGN_DOWN(addr, size) ((addr)&(~((size)-1)))
#define PAGE_ALIGN(addr)        _ALIGN_DOWN(addr, PAGE_SIZE)
#define BT_HAL_PIN_MAX_LENGTH 16
#define BT_HAL_PASSKEY_MAX_LENGTH 4
#define BT_HAL_LOWER_ADDRESS_LENGTH 9
#define BT_HAL_AGENT_SYSPOPUP_TIMEOUT_FOR_MULTIPLE_POPUPS 200
#define BT_HAL_AGENT_SYSPOPUP_MAX_ATTEMPT 3

#define G_VARIANT_UNREF(variant) \
	g_variant_unref(variant); \
	variant = NULL
#define BT_HAL_MAX_EVENT_STR_LENGTH 50

static void *adapter_agent = NULL;
static gboolean passkey_display_notification = FALSE;
static gboolean passkey_confirm_notification = FALSE;
static gboolean passkey_notification = FALSE;
static gboolean pincode_notification = FALSE;

/* Forward delcaration */
static void __bt_hal_send_pin_request_event(const gchar *address, const gchar *name,
		uint32_t cod);
static gboolean __bt_hal_pincode_request(GapAgentPrivate *agent, GDBusProxy *device);
static gboolean __bt_hal_display_request(GapAgentPrivate *agent, GDBusProxy *device,
		guint passkey);
static gboolean __bt_hal_passkey_request(GapAgentPrivate *agent, GDBusProxy *device);
static gboolean __bt_hal_confirm_request(GapAgentPrivate *agent, GDBusProxy *device,
		guint passkey);
static gboolean __bt_hal_authorize_request(GapAgentPrivate *agent, GDBusProxy *device,
								const char *uuid);
static gboolean __bt_hal_authorize_cancel_request(GapAgentPrivate *agent, const char *address);
static gboolean __bt_hal_pairing_cancel_request(GapAgentPrivate *agent, const char *address);
static GVariant *__bt_hal_service_getall(GDBusProxy *device, const char *interface);
static void __bt_hal_agent_release_memory(void);
static inline void stack_trim(void);

static gboolean __bt_hal_device_is_hid_keyboard(unsigned int dev_class);
static int __bt_hal_device_generate_passkey(char *passkey, int size);
static gboolean __bt_hal_device_is_auto_response(uint32_t dev_class,
		const gchar *address, const gchar *name);
static gboolean __bt_hal_device_is_device_blacklisted(const char *address, const char *name);
static gboolean __bt_hal_find_device_by_address_exactname(char *buffer,
		const char *address);
static gboolean __bt_hal_find_device_by_partial_name(char *buffer,
		const char *partial_name);
static void __bt_hal_send_ssp_request_events(const gchar *address, const gchar *name,
		guint passkey, uint32_t cod, unsigned char variant);
static void __bt_hal_send_authorize_request_event(const gchar *address, const char *uuid);

void* _bt_hal_create_agent(const char *path, gboolean adapter)
{
	GAP_AGENT_FUNC_CB func_cb;
	GDBusProxy *adapter_proxy;
	GapAgentPrivate *agent;

	DBG("+");
	adapter_proxy = _bt_hal_get_adapter_proxy();
	if (!adapter_proxy)
		return NULL;

	func_cb.pincode_func = __bt_hal_pincode_request;
	func_cb.display_func = __bt_hal_display_request;
	func_cb.passkey_func = __bt_hal_passkey_request;
	func_cb.confirm_func = __bt_hal_confirm_request;
	func_cb.authorize_func = __bt_hal_authorize_request;
	func_cb.pairing_cancel_func = __bt_hal_pairing_cancel_request;
	func_cb.authorization_cancel_func = __bt_hal_authorize_cancel_request;

	/* Allocate memory*/
	agent = g_new0(GapAgentPrivate, 1);

	_gap_agent_setup_dbus(agent, &func_cb, path, adapter_proxy);

	if (adapter) {
		if (!_gap_agent_register(agent)) {
			ERR("gap agent registration failed!");
			_bt_hal_destroy_agent(agent);
			agent = NULL;
		}
	}
	DBG("-");
	return agent;
}

void _bt_hal_destroy_agent(void *agent)
{
	DBG("+");
	if (!agent)
		return;

	_gap_agent_reset_dbus((GapAgentPrivate *)agent);

	g_free(agent);
	DBG("-");
}

gboolean _bt_hal_agent_is_canceled(void)
{
	void *agent = _bt_hal_get_adapter_agent();
	if (!agent)
		return FALSE;

	return _gap_agent_is_canceled(agent);
}

int _bt_hal_agent_reply_cancellation(void)
{
	void *agent = _bt_hal_get_adapter_agent();
	if (!agent)
		return BT_STATUS_FAIL;

	if (gap_agent_reply_confirmation(agent, GAP_AGENT_CANCEL, NULL) != TRUE) {
		ERR("Fail to reply agent");
		return BT_STATUS_FAIL;
	}
	DBG("gap agent cancellation done successfully!");
	return BT_STATUS_SUCCESS;

}

void _bt_hal_agent_set_canceled(gboolean value)
{
	void *agent = _bt_hal_get_adapter_agent();
	if (!agent)
		return;

	return _gap_agent_set_canceled(agent, value);
}

void _bt_hal_initialize_adapter_agent(void)
{
	adapter_agent = _bt_hal_create_agent(BT_HAL_ADAPTER_AGENT_PATH, TRUE);
	if (!adapter_agent) {
		ERR("Fail to register agent");
		return;
	}
}

void _bt_hal_destroy_adapter_agent(void)
{
	if (adapter_agent)
		_bt_hal_destroy_agent(adapter_agent);
	adapter_agent = NULL;
}

void* _bt_hal_get_adapter_agent(void)
{
	return adapter_agent;
}

void _bt_hal_enable_gap_auth_notifications(unsigned int type, gboolean enable)
{
	INFO("type: %d, enable: %d", type, enable);

	switch (type) {
	case BT_PASSKEY_CONFIRMATION:
		/* Note: Currently not used for notification sending, should be used when required */
		passkey_confirm_notification = enable;
		break;
	case BT_PASSKEY_DISPLAY:
		passkey_display_notification = enable;
		break;
	case BT_PASSKEY_ENTRY:
		/* Note: Currently not used for notification sending, should be used when required */
		passkey_notification = enable;
		break;
	case BT_PINCODE_ENTRY:
		pincode_notification = enable;
		break;
	default:
		ERR("Unknown type: %d", type);
	}
}

static void __bt_hal_send_authorize_request_event(const gchar *address, const char *uuid)
{
	struct hal_ev_authorize_request ev;
	memset(&ev, 0, sizeof(ev));

	DBG("Remote Device address [%s]", address);

	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);

	/* UUID received in authorization request (In case of HFP unit, UUID received from Bluez
	   would be of that of Audio Gateway UUID (0x111f) */
	ev.service_id = _bt_convert_uuid_string_to_service_id(uuid);

	handle_stack_msg event_cb = _bt_hal_get_stack_message_handler();
	if (event_cb) {
		DBG("Sending AUTHORIZE REQUEST");
		event_cb(HAL_EV_AUTHORIZE_REQUEST, (void*)&ev, sizeof(ev));
	}

	DBG("-");
}

#ifdef TIZEN_BT_HAL
static void __bt_hal_send_rfcomm_authorize_request_event(const gchar *address, const char *uuid)
{
	struct hal_ev_sock_conn_auth ev;

	DBG("Remote Device address [%s], uuid: %s", address, uuid);

	memset(&ev, 0, sizeof(ev));
	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
	_bt_hal_convert_uuid_string_to_type(ev.uuid, uuid);

	handle_stack_msg event_cb = _bt_hal_get_stack_message_handler();
	if (event_cb) {
		DBG("Sending Socket AUTHORIZE REQUEST");
		event_cb(HAL_EV_SOCK_AUTHORIZE_REQUEST, (void*)&ev, sizeof(ev));
	}

	DBG("-");
}
#endif

/* Legacy Pairing */
static void __bt_hal_send_pin_request_event(const gchar *address, const gchar *name,
		uint32_t cod)
{
	struct hal_ev_pin_request ev;
	memset(&ev, 0, sizeof(ev));

	DBG("Remote Device address [%s]", address);
	DBG("Remote Device Name [%s]", name);
	DBG("Remote Device COD [%u]", cod);

	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);

	memcpy(ev.name, name, strlen(name)+1);
	ev.class_of_dev = cod;

	handle_stack_msg event_cb = _bt_hal_get_stack_message_handler();
	if (event_cb) {
		DBG("Sending PIN REQUEST");
		event_cb(HAL_EV_PIN_REQUEST, (void*)&ev, sizeof(ev));
	}

	DBG("-");
}


void __bt_hal_get_auth_info(GVariant *reply, char *auth_info)
{
	int cursor;
	GVariant *tmp_value;
	char *manufacturer_data = NULL;
	int manufacturer_data_len;
	gboolean is_alias_set;
	GVariantIter *value_iter;
	guint8 m_value;
	int i = 0;

	tmp_value = g_variant_lookup_value(reply, "IsAliasSet",
								G_VARIANT_TYPE_BOOLEAN);
	if (tmp_value) {
		is_alias_set = g_variant_get_boolean(tmp_value);
		g_variant_unref(tmp_value);
	} else {
		is_alias_set = FALSE;
	}
	if (is_alias_set == FALSE) {
		tmp_value = g_variant_lookup_value(reply, "LagacyManufacturerDataLen",
								G_VARIANT_TYPE_UINT16);
		if (tmp_value) {
			manufacturer_data_len = g_variant_get_uint16(tmp_value);
			if (manufacturer_data_len >
					MAX_MANUFACTURE_LEN) {
				ERR("manufacturer_data_len is too long");
				manufacturer_data_len = MAX_MANUFACTURE_LEN;
			}
			g_variant_unref(tmp_value);
		} else
			manufacturer_data_len = 0;

		tmp_value = g_variant_lookup_value(reply, "LagacyManufacturerData",
								G_VARIANT_TYPE_ARRAY);
		if (tmp_value) {
			if ((manufacturer_data_len == 0) ||
					manufacturer_data_len != g_variant_get_size(tmp_value)) {
				ERR("manufacturer data length doesn't match");
				manufacturer_data_len = 0;
				manufacturer_data = NULL;
			} else {
				manufacturer_data = g_malloc0(manufacturer_data_len);
				g_variant_get(tmp_value, "ay", &value_iter);
				while (g_variant_iter_loop(value_iter, "y", &m_value))
					manufacturer_data[i++] = m_value;
			}
			g_variant_unref(tmp_value);
		} else {
			INFO("manufacture data is not a G_VARIANT_TYPE_ARRAY ");
			manufacturer_data_len = 0;
			manufacturer_data = NULL;
		}
		/*minimum Size of the samsung specific manufacturer data is greater than 30 */
		if (manufacturer_data_len < 30) {
			g_free(manufacturer_data);
			return;
		}
		if (manufacturer_data[0] != 0x00 || manufacturer_data[1] != 0x75) {
			DBG("This is not a samsung specific manufaturer data");
			g_free(manufacturer_data);
			return;
		}

		/* 2  samsung (0x00 0x75) + 1 (control and version) + 1 (service ID) +
		1 (discovery version) + 1 (associated service ID)
		2 (Proxamity and locality) + 2 (Device type and icon) */

		cursor = 10;

		memcpy(auth_info, &(manufacturer_data[cursor]), 5);
	}
	 g_free(manufacturer_data);
}

static gboolean __bt_hal_pincode_request(GapAgentPrivate *agent, GDBusProxy *device)
{
	uint32_t device_class;
	const gchar *address;
	unsigned char auth_info[5] = {0, };
	const gchar *name;
	GVariant *reply = NULL;
	GVariant *reply_temp = NULL;
	GVariant *tmp_value;
	DBG("+");

	reply_temp = __bt_hal_service_getall(device, BT_HAL_DEVICE_INTERFACE);

	if (reply_temp == NULL) {
		gap_agent_reply_pin_code(agent, GAP_AGENT_REJECT, "",
				NULL);
		goto done;
	}

	g_variant_get(reply_temp, "(@a{sv})", &reply); /* Format of reply a{sv}*/

	tmp_value = g_variant_lookup_value(reply, "Class", G_VARIANT_TYPE_UINT32);
	g_variant_get(tmp_value, "u", &device_class);
	G_VARIANT_UNREF(tmp_value);

	tmp_value = g_variant_lookup_value(reply, "Address", G_VARIANT_TYPE_STRING);
	g_variant_get(tmp_value, "s", &address);
	G_VARIANT_UNREF(tmp_value);

	tmp_value = g_variant_lookup_value(reply, "Name", G_VARIANT_TYPE_STRING);
	g_variant_get(tmp_value, "s", &name);
	G_VARIANT_UNREF(tmp_value);

	if (!name)
		name = address;

	if (headed_plugin_info->plugin_headed_enabled) {
		__bt_hal_get_auth_info(reply, (char *)auth_info);
		if (__bt_hal_device_is_hid_keyboard(device_class)) {
			char str_passkey[BT_HAL_PASSKEY_MAX_LENGTH + 1] = { 0 };

			DBG("Device is HID Keyboard");
			if (__bt_hal_device_generate_passkey(str_passkey,
						BT_HAL_PASSKEY_MAX_LENGTH) != 0) {
				gap_agent_reply_pin_code(agent, GAP_AGENT_REJECT,
						"", NULL);
				goto done;
			}

			gap_agent_reply_pin_code(agent, GAP_AGENT_ACCEPT,
					str_passkey, NULL);

			DBG("Launch BT Syspopup");
			headed_plugin_info->headed_plugin->bt_launch_system_popup(
					BT_HAL_AGENT_EVENT_KEYBOARD_PASSKEY_REQUEST,
					name, auth_info, str_passkey, NULL, _gap_agent_get_path(agent));
		} else if (__bt_hal_device_is_auto_response(device_class, address, name)) {
			DBG("Device is of Type Auto response, send event to HAL");
			__bt_hal_send_pin_request_event(address, name, device_class);
		} else if (pincode_notification) {
			DBG("pincode_notification is enabled, send event to HAL");
			__bt_hal_send_pin_request_event(address, name, device_class);
		} else {
			DBG("Device is not of Auto response class, Show PIN Entry");
			headed_plugin_info->headed_plugin->bt_launch_system_popup(
					BT_HAL_AGENT_EVENT_PIN_REQUEST, name, auth_info,
					NULL, NULL, _gap_agent_get_path(agent));
		}
	} else {
		INFO("Plugin Headed not Enabled");
		__bt_hal_send_pin_request_event(address, name, device_class);
	}

done:
	g_variant_unref(reply);
	g_variant_unref(reply_temp);
	__bt_hal_agent_release_memory();
	DBG("-");

	return TRUE;
}


/* BT_SSP_VARIANT_PASSKEY_CONFIRMATION */
/* BT_SSP_VARIANT_PASSKEY_NOTIFICATION */
/* BT_SSP_VARIANT_PASSKEY_ENTRY */
/* BT_SSP_VARIANT_CONSENT */

static void __bt_hal_send_ssp_request_events(const gchar *address,
		const gchar *name,
		guint passkey,
		uint32_t cod,
		unsigned char variant)
{
	struct hal_ev_ssp_request ev;
	memset(&ev, 0, sizeof(ev));
	DBG("sizeof ev [%zu]", sizeof(ev));

	DBG("Remote Device address [%s]", address);
	DBG("Remote Device Name [%s]", name);
	DBG("Remote Device passkey [%d]", passkey);
	DBG("Remote Device pairing variant [0x%x]", variant);
	DBG("Remote Device cod [%d]", cod);

	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);

	memcpy(ev.name, name, strlen(name)+1);
	ev.class_of_dev = cod;
	ev.pairing_variant = variant;
	ev.passkey = passkey;

	handle_stack_msg event_cb = _bt_hal_get_stack_message_handler();
	if (event_cb) {
		DBG("Sending SSP type [%d]", variant);
		event_cb(HAL_EV_SSP_REQUEST, (void*)&ev, sizeof(ev));
	}

	DBG("-");
}

/* SSP */
static gboolean __bt_hal_display_request(GapAgentPrivate *agent, GDBusProxy *device,
		guint passkey)
{
	const gchar *address;
	const gchar *name;
	unsigned char auth_info[5] = {0, };
	char *str_passkey;
	uint32_t device_class;
	GVariant *reply = NULL;
	GVariant *reply_temp = NULL;
	GVariant *tmp_value = NULL;
	DBG("+");

	reply_temp = __bt_hal_service_getall(device, BT_HAL_DEVICE_INTERFACE);
	if (reply_temp == NULL) {
		gap_agent_reply_pin_code(agent, GAP_AGENT_REJECT, "",
				NULL);
		goto done;
	}

	g_variant_get(reply_temp, "(@a{sv})", &reply); /* Format of reply a{sv}*/

	tmp_value = g_variant_lookup_value(reply, "Address", G_VARIANT_TYPE_STRING);
	g_variant_get(tmp_value, "s", &address);
	G_VARIANT_UNREF(tmp_value);
	if (!address) {
		gap_agent_reply_pin_code(agent, GAP_AGENT_REJECT, "", NULL);
		goto done;
	}

	tmp_value = g_variant_lookup_value(reply, "Name", G_VARIANT_TYPE_STRING);
	g_variant_get(tmp_value, "s", &name);
	G_VARIANT_UNREF(tmp_value);

	tmp_value = g_variant_lookup_value(reply, "Class", G_VARIANT_TYPE_UINT32);
	g_variant_get(tmp_value, "u", &device_class);
	G_VARIANT_UNREF(tmp_value);

	if (!name)
		name = address;

	str_passkey = g_strdup_printf("%d", passkey);

	__bt_hal_get_auth_info(reply, (char *)auth_info);

	DBG("KEYBOARD_PASSKEY_REQUEST");

	if (passkey_display_notification) {
		DBG("passkey_display_notification is enabled, send event to HAL");
		__bt_hal_send_ssp_request_events(address, name, passkey, device_class,
				BT_SSP_VARIANT_PASSKEY_NOTIFICATION);
	} else if (headed_plugin_info->plugin_headed_enabled) {
		INFO("Plugin Headed Enabled");
		headed_plugin_info->headed_plugin->bt_launch_system_popup(BT_HAL_AGENT_EVENT_KEYBOARD_PASSKEY_REQUEST, name,
				auth_info, str_passkey, NULL, _gap_agent_get_path(agent));
	} else {
		INFO("Plugin Headed not Enabled");
		__bt_hal_send_ssp_request_events(address, name, passkey, device_class,
				BT_SSP_VARIANT_PASSKEY_NOTIFICATION);
	}

	g_free(str_passkey);
done:
	g_variant_unref(reply);
	g_variant_unref(reply_temp);
	__bt_hal_agent_release_memory();

	DBG("-");
	return TRUE;
}

/* SSP */
static gboolean __bt_hal_passkey_request(GapAgentPrivate *agent, GDBusProxy *device)
{
	const gchar *address;
	const gchar *name;
	unsigned char auth_info[5] = {0, };
	uint32_t device_class;
	GVariant *reply = NULL;
	GVariant *reply_temp = NULL;
	GVariant *tmp_value;
	DBG("+");

	reply_temp = __bt_hal_service_getall(device, BT_HAL_DEVICE_INTERFACE);

	if (reply_temp == NULL) {
		gap_agent_reply_pin_code(agent, GAP_AGENT_REJECT, "",
				NULL);
		goto done;
	}

	g_variant_get(reply_temp, "(@a{sv})", &reply); /* Format of reply a{sv}*/

	tmp_value = g_variant_lookup_value(reply, "Address", G_VARIANT_TYPE_STRING);
	g_variant_get(tmp_value, "s", &address);
	G_VARIANT_UNREF(tmp_value);
	if (!address) {
		gap_agent_reply_pin_code(agent, GAP_AGENT_REJECT, "", NULL);
		goto done;
	}

	tmp_value = g_variant_lookup_value(reply, "Name", G_VARIANT_TYPE_STRING);
	g_variant_get(tmp_value, "s", &name);
	G_VARIANT_UNREF(tmp_value);

	tmp_value = g_variant_lookup_value(reply, "Class", G_VARIANT_TYPE_UINT32);
	g_variant_get(tmp_value, "u", &device_class);
	G_VARIANT_UNREF(tmp_value);

	if (!name)
		name = address;

	DBG("PASSKEY_REQUEST");

	__bt_hal_get_auth_info(reply, (char *)auth_info);

	if (headed_plugin_info->plugin_headed_enabled) {
		headed_plugin_info->headed_plugin->bt_launch_system_popup(BT_HAL_AGENT_EVENT_PASSKEY_REQUEST, name, auth_info,
				NULL, NULL, _gap_agent_get_path(agent));
	} else {
		__bt_hal_send_ssp_request_events(address, name, 0, device_class,
				BT_SSP_VARIANT_PASSKEY_ENTRY);
	}

done:
	g_variant_unref(reply);
	g_variant_unref(reply_temp);
	__bt_hal_agent_release_memory();

	DBG("-");
	return TRUE;
}

/*ssp*/
static gboolean __bt_hal_confirm_request(GapAgentPrivate *agent, GDBusProxy *device,
		guint passkey)
{
	const gchar *address;
	const gchar *name;
	char str_passkey[7];
	uint32_t device_class;
	GVariant *reply_temp = NULL;
	GVariant *reply = NULL;
	GVariant *tmp_value;
	DBG("+ passkey[%.6d]", passkey);
	DBG("Agent Path [%s]", agent->path);

	snprintf(str_passkey, sizeof(str_passkey), "%.6d", passkey);

	reply_temp = __bt_hal_service_getall(device, BT_HAL_DEVICE_INTERFACE);

	if (reply_temp == NULL) {
		ERR("####Device doesn't exist####");
		gap_agent_reply_pin_code(agent, GAP_AGENT_REJECT, "",
				NULL);
		goto done;
	}
	g_variant_get(reply_temp, "(@a{sv})", &reply); /* Format of reply a{sv}*/

	tmp_value = g_variant_lookup_value(reply, "Address", G_VARIANT_TYPE_STRING);
	g_variant_get(tmp_value, "s", &address);
	G_VARIANT_UNREF(tmp_value);

	tmp_value = g_variant_lookup_value(reply, "Name", G_VARIANT_TYPE_STRING);
	g_variant_get(tmp_value, "s", &name);
	G_VARIANT_UNREF(tmp_value);

	tmp_value = g_variant_lookup_value(reply, "Class", G_VARIANT_TYPE_UINT32);
	g_variant_get(tmp_value, "u", &device_class);
	G_VARIANT_UNREF(tmp_value);

	if (!name)
		name = address;

	if (headed_plugin_info->plugin_headed_enabled) {
		unsigned char auth_info[5] = {0, };

		DBG("LAUNCH SYSPOPUP");
		DBG("Name [%s]", name);
		DBG("Passkey [%s]", str_passkey);

		__bt_hal_get_auth_info(reply, (char *)auth_info);
		headed_plugin_info->headed_plugin->bt_launch_system_popup(BT_HAL_AGENT_EVENT_PASSKEY_CONFIRM_REQUEST, name,
				auth_info, str_passkey, NULL,
				_gap_agent_get_path(agent));
	} else {
		DBG("Headless Confirmation");
		__bt_hal_send_ssp_request_events(address, name, passkey,
				device_class, BT_SSP_VARIANT_PASSKEY_CONFIRMATION);
	}

done:
	g_variant_unref(reply);
	g_variant_unref(reply_temp);
	__bt_hal_agent_release_memory();
	DBG("-");
	return TRUE;
}

static gboolean __bt_hal_authorize_request(GapAgentPrivate *agent, GDBusProxy *device,
		const char *uuid)
{
	const gchar *address;
	const gchar *name;
	gboolean trust;
	gboolean paired;
	GVariant *reply = NULL;
	GVariant *reply_temp = NULL;
	GVariant *tmp_value;

	DBG("Authorize Request from Bluez Stack: UUID [%s]", uuid);

	reply_temp = __bt_hal_service_getall(device, BT_HAL_DEVICE_INTERFACE);
	if (reply_temp == NULL) {
		gap_agent_reply_authorize(agent, GAP_AGENT_REJECT, NULL);
		goto done;
	}

	g_variant_get(reply_temp, "(@a{sv})", &reply); /* Format of reply a{sv}*/

	tmp_value = g_variant_lookup_value(reply, "Address", G_VARIANT_TYPE_STRING);
	g_variant_get(tmp_value, "s", &address);
	G_VARIANT_UNREF(tmp_value);
	if (!address) {
		gap_agent_reply_authorize(agent, GAP_AGENT_REJECT, NULL);
		goto done;
	}

	tmp_value = g_variant_lookup_value(reply, "Alias", G_VARIANT_TYPE_STRING);
	g_variant_get(tmp_value, "s", &name);
	G_VARIANT_UNREF(tmp_value);
	if (!name)
		name = address;

	tmp_value = g_variant_lookup_value(reply, "Trusted", G_VARIANT_TYPE_BOOLEAN);
	g_variant_get(tmp_value, "b", &trust);
	G_VARIANT_UNREF(tmp_value);

	tmp_value = g_variant_lookup_value(reply, "Paired", G_VARIANT_TYPE_BOOLEAN);
	g_variant_get(tmp_value, "b", &paired);
	G_VARIANT_UNREF(tmp_value);
	if ((paired == FALSE) && (trust == FALSE)) {
		ERR("No paired & No trusted device");
		gap_agent_reply_authorize(agent, GAP_AGENT_REJECT, NULL);
		goto done;
	}

	INFO("Authorization request for device [%s] Service:[%s]\n", address, uuid);

	if (trust) {
		INFO("Trusted device, so authorize\n");
		gap_agent_reply_authorize(agent, GAP_AGENT_ACCEPT, NULL);
		goto done;
	} else {
		INFO("Device is not Trusted, so prompt user to accept or reject authorization \n");
	}

	if (!strcasecmp(uuid, BT_HAL_HFP_AUDIO_GATEWAY_UUID) ||
			!strcasecmp(uuid, BT_HAL_HSP_AUDIO_GATEWAY_UUID) ||
			!strcasecmp(uuid, BT_HAL_HFP_HF_UUID) ||
			!strcasecmp(uuid, BT_HAL_HSP_HS_UUID) ||
			!strcasecmp(uuid, BT_HAL_A2DP_PROFILE_UUID) ||
			!strcasecmp(uuid, BT_HAL_HID_UUID) ||
			!strcasecmp(uuid, BT_HAL_HID_DEVICE_UUID) ||
			!strcasecmp(uuid, BT_HAL_SAP_UUID_OLD) ||
			!strcasecmp(uuid, BT_HAL_SAP_UUID_NEW) ||
#ifdef TIZEN_BT_HAL
			!strcasecmp(uuid, BT_HAL_IOTIVITY_UUID) ||
#endif
			!strcasecmp(uuid, BT_HAL_AVRCP_TARGET_UUID)) {
		DBG("Authorization for UUID: %s, send event", uuid);
		__bt_hal_send_authorize_request_event(address, uuid);
		goto done;
	}

#ifdef TIZEN_BT_HAL
	if (!strcasecmp(uuid, BT_HAL_OPP_UUID)) {
		if (NULL != _bt_hal_gap_agent_find_osp_server_by_type(
				agent->osp_servers, BT_OSP_SERVER_OBEX)) {
			INFO("OSP server for OPP found, send event");
			__bt_hal_send_authorize_request_event(address, uuid);
			goto done;
		} else {
			INFO("Native OPP server, auto accept");
			gap_agent_reply_authorize(agent,
					GAP_AGENT_ACCEPT, NULL);
			goto done;
		}
	}

	if (_is_rfcomm_server_uuid(uuid)) {
		__bt_hal_send_rfcomm_authorize_request_event(address, uuid);
		goto done;
	}
#endif

	if (trust || !headed_plugin_info->plugin_headed_enabled) {
		INFO("Trusted or Headless device, so authorize\n");
		gap_agent_reply_authorize(agent,
				GAP_AGENT_ACCEPT, NULL);

		goto done;
	}

	if (headed_plugin_info->plugin_headed_enabled) {
		unsigned char auth_info[5] = {0, };
		int request_type = BT_HAL_AGENT_EVENT_AUTHORIZE_REQUEST;

		__bt_hal_get_auth_info(reply, (char *)auth_info);

		if (!strcasecmp(uuid, BT_HAL_OPP_UUID))
			request_type = BT_HAL_AGENT_EVENT_EXCHANGE_REQUEST;
		else if (!strcasecmp(uuid, BT_HAL_PBAP_UUID))
			request_type = BT_HAL_AGENT_EVENT_PBAP_REQUEST;
		else if (!strcasecmp(uuid, BT_HAL_MAP_UUID))
			request_type = BT_HAL_AGENT_EVENT_MAP_REQUEST;

		headed_plugin_info->headed_plugin->bt_launch_system_popup(request_type, name, auth_info, NULL, NULL,
					_gap_agent_get_path(agent));
	}

done:
	g_variant_unref(reply);
	g_variant_unref(reply_temp);
	__bt_hal_agent_release_memory();

	DBG("-");
	return TRUE;
}

static gboolean __bt_hal_authorize_cancel_request(GapAgentPrivate *agent,
						const char *address)
{
	DBG("On Going Authorization is cancelled by remote [%s]", address);
	gap_agent_reply_authorize(agent, GAP_AGENT_CANCEL, NULL);

	if (headed_plugin_info->plugin_headed_enabled)
		headed_plugin_info->headed_plugin->bt_destroy_popup_all();

	__bt_hal_agent_release_memory();
	return TRUE;
}

static gboolean __bt_hal_pairing_cancel_request(GapAgentPrivate *agent, const char *address)
{
	DBG("On Going Pairing is cancelled by remote [%s]", address);

	if (headed_plugin_info->plugin_headed_enabled)
		headed_plugin_info->headed_plugin->bt_destroy_popup_all();

	__bt_hal_agent_release_memory();
	return TRUE;
}

static gboolean __bt_hal_device_is_hid_keyboard(unsigned int dev_class)
{
	switch ((dev_class & 0x1f00) >> 8) {
	case 0x05:
		switch ((dev_class & 0xc0) >> 6) {
		case 0x01:
			/* input-keyboard" */
			return TRUE;
		}
		break;
	}

	return FALSE;
}

static int __bt_hal_device_generate_passkey(char *passkey, int size)
{
	int i;
	ssize_t len;
	int random_fd;
	unsigned int value = 0;

	if (passkey == NULL)
		return -1;

	if (size <= 0)
		return -1;

	random_fd = open("/dev/urandom", O_RDONLY);

	if (random_fd < 0)
		return -1;

	for (i = 0; i < size; i++) {
		len = read(random_fd, &value, sizeof(value));
		if (len > 0)
			passkey[i] = '0' + (value % 10);
	}

	close(random_fd);

	DBG("passkey: %s", passkey);

	return 0;
}

static gboolean __bt_hal_find_device_by_address_exactname(char *buffer,
		const char *address)
{
	char *pch;
	char *last;

	pch = strtok_r(buffer, "= ,", &last);

	if (pch == NULL)
		return FALSE;

	while ((pch = strtok_r(NULL, ",", &last))) {
		if (0 == g_strcmp0(pch, address)) {
			DBG("Match found\n");
			return TRUE;
		}
	}
	return FALSE;
}

static gboolean __bt_hal_find_device_by_partial_name(char *buffer,
		const char *partial_name)
{
	char *pch;
	char *last;

	pch = strtok_r(buffer, "= ,", &last);

	if (pch == NULL)
		return FALSE;

	while ((pch = strtok_r(NULL, ",", &last))) {
		if (g_str_has_prefix(partial_name, pch)) {
			DBG("Match found\n");
			return TRUE;
		}
	}
	return FALSE;
}

static gboolean __bt_hal_device_is_device_blacklisted(const char *address,
		const char *name)
{
	char *buffer;
	char **lines;
	int i;
	FILE *fp;
	long size;
	size_t result;

	DBG("+");

	fp = fopen(BT_HAL_AGENT_AUTO_PAIR_BLACKLIST_FILE, "r");

	if (fp == NULL) {
		ERR("Unable to open blacklist file");
		return FALSE;
	}

	fseek(fp, 0, SEEK_END);
	size = ftell(fp);
	if (size <= 0) {
		DBG("size is not a positive number");
		fclose(fp);
		return FALSE;
	}

	rewind(fp);

	buffer = g_malloc0(sizeof(char) * size);
	/* Fix : NULL_RETURNS */
	if (buffer == NULL) {
		ERR("Fail to allocate memory");
		fclose(fp);
		return FALSE;
	}
	result = fread((char *)buffer, 1, size, fp);
	fclose(fp);
	if (result != size) {
		ERR("Read Error");
		g_free(buffer);
		return FALSE;
	}

	DBG("Buffer = %s", buffer);

	lines = g_strsplit_set(buffer, BT_HAL_AGENT_NEW_LINE, 0);
	g_free(buffer);

	if (lines == NULL) {
		ERR("No lines in the file");
		return FALSE;
	}

	for (i = 0; lines[i] != NULL; i++) {
		if (g_str_has_prefix(lines[i], "AddressBlacklist"))
			if (__bt_hal_find_device_by_address_exactname(
						lines[i], address))
				goto done;
		if (g_str_has_prefix(lines[i], "ExactNameBlacklist"))
			if (__bt_hal_find_device_by_address_exactname(
						lines[i], name))
				goto done;
		if (g_str_has_prefix(lines[i], "PartialNameBlacklist"))
			if (__bt_hal_find_device_by_partial_name(lines[i],
						name))
				goto done;
		if (g_str_has_prefix(lines[i], "KeyboardAutoPair"))
			if (__bt_hal_find_device_by_address_exactname(
						lines[i], address))
				goto done;
	}
	g_strfreev(lines);
	DBG("-");
	return FALSE;
done:
	DBG("Found the device");
	g_strfreev(lines);
	return TRUE;
}

static gboolean __bt_hal_device_is_auto_response(uint32_t dev_class,
		const gchar *address, const gchar *name)
{
	gboolean is_headset = FALSE;
	gboolean is_mouse = FALSE;
	char lap_address[BT_HAL_LOWER_ADDRESS_LENGTH];

	DBG("bt_agent_is_headset_class, %d +", dev_class);

	if (address == NULL)
		return FALSE;

	switch ((dev_class & 0x1f00) >> 8) {
	case 0x04:
		switch ((dev_class & 0xfc) >> 2) {
		case 0x01:
		case 0x02:
			/* Headset */
			is_headset = TRUE;
			break;
		case 0x06:
			/* Headphone */
			is_headset = TRUE;
			break;
		case 0x0b:      /* VCR */
		case 0x0c:      /* Video Camera */
		case 0x0d:      /* Camcorder */
			break;
		default:
			/* Other audio device */
			is_headset = TRUE;
			break;
		}
		break;
	case 0x05:
		switch (dev_class & 0xff) {
		case 0x80:  /* 0x80: Pointing device(Mouse) */
			is_mouse = TRUE;
			break;

		case 0x40: /* 0x40: input device (BT keyboard) */

			/* Get the LAP(Lower Address part) */
			g_strlcpy(lap_address, address, sizeof(lap_address));

			/* Need to Auto pair the blacklisted Keyboard */
			if (__bt_hal_device_is_device_blacklisted(lap_address, name) != TRUE) {
				DBG("Device is not black listed\n");
				return FALSE;
			} else {
				ERR("Device is black listed\n");
				return TRUE;
			}
		}
	}

	if ((!is_headset) && (!is_mouse))
		return FALSE;

	/* Get the LAP(Lower Address part) */
	g_strlcpy(lap_address, address, sizeof(lap_address));

	DBG("Device address = %s\n", address);
	DBG("Address 3 byte = %s\n", lap_address);

	if (__bt_hal_device_is_device_blacklisted(lap_address, name)) {
		ERR("Device is black listed\n");
		return FALSE;
	}

	return TRUE;
}

static GVariant *__bt_hal_service_getall(GDBusProxy *device, const char *interface)
{
	GError *error = NULL;
	GVariant *reply;
	DBG("+");
	reply = g_dbus_proxy_call_sync(device,
			"GetAll", g_variant_new("(s)", interface),
			G_DBUS_CALL_FLAGS_NONE, -1,
			NULL, &error);
	if (reply == NULL) {
		ERR("GetAll dBUS-RPC failed");
		if (error) {
			ERR("D-Bus API failure: errCode[%x], message[%s]",
					error->code, error->message);
			g_clear_error(&error);
		}
		return NULL;
	}

	DBG("-");
	return reply;
}

static void __bt_hal_agent_release_memory(void)
{
	/* Release Malloc Memory*/
	malloc_trim(0);

	/* Release Stack Memory*/
	stack_trim();
}

static inline void stack_trim(void)
{
#ifdef STACK_FLUSH
	unsigned int sp;
	char buf[BUF_SIZE];
	FILE *file;
	unsigned int stacktop;
	int found = 0;

	asm volatile ("mov %0,sp " : "=r"(sp));

	sprintf(buf, "/proc/%d/maps", getpid());
	file = fopen(buf, "r");
	while (fgets(buf, BUF_SIZE, file) != NULL) {
		if (strstr(buf, "[stack]")) {
			found = 1;
			break;
		}
	}
	fclose(file);

	if (found) {
		sscanf(buf, "%x-", &stacktop);
		if (madvise((void *)PAGE_ALIGN(stacktop), PAGE_ALIGN(sp) - stacktop,
					MADV_DONTNEED) < 0)
			perror("stack madvise fail");
	}
#endif
}
