/*
 * BLUETOOTH HAL
 *
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: Anupam Roy <anupam.r@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

#include <glib.h>
#include <gio/gio.h>
#include <dlog.h>
#include <vconf.h>

#include "bt-hal-av-dbus-handler.h"
#include "bt-hal-dbus-common-utils.h"
#include "bt-hal-internal.h"


static handle_stack_msg event_cb = NULL;

/* To send stack event to hal-av handler */
void _bt_hal_register_av_dbus_handler_cb(handle_stack_msg cb)
{
	event_cb = cb;
}

/* To send stack event to hal-av handler */
void _bt_hal_unregister_av_dbus_handler_cb()
{
	event_cb = NULL;
}

static void __bt_av_connect_cb(GDBusProxy *proxy, GAsyncResult *res,
		gpointer user_data)
{
	GError *g_error = NULL;
	struct hal_ev_a2dp_conn_state ev;
	GVariant *reply = NULL;
	char *address = user_data;
	int result = BT_STATUS_SUCCESS;

	DBG("+");

	reply = g_dbus_proxy_call_finish(proxy, res, &g_error);
	g_object_unref(proxy);
	if (reply == NULL) {
		ERR("A2DP Connect Dbus Call Error");
		if (g_error) {
			ERR("Error: %s\n", g_error->message);
			g_clear_error(&g_error);
		}
		result = BT_STATUS_FAIL;
	}
	g_variant_unref(reply);

	DBG("Address: %s", address);
	/*
	 * If result is success, AV Connecting event will be triggered
	 * from here.
	 */
	if (result == BT_STATUS_SUCCESS) {
		/* Prepare to send AV connecting event */
		memset(&ev, 0, sizeof(ev));
		ev.state = HAL_EV_A2DP_STATE_CONNECTING;
		_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
		if (!event_cb)
			ERR("A2DP dbus handler callback not registered");
		else
			event_cb(HAL_EV_A2DP_CONN_STATE, (void *)&ev, sizeof(ev));
	} else {
		/* Prepare to send AV disconnected event */
		memset(&ev, 0, sizeof(ev));
		ev.state = HAL_EV_A2DP_STATE_DISCONNECTED;
		_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
		if (!event_cb)
			ERR("HF DBUS handler callback not registered");
		else
			event_cb(HAL_EV_A2DP_CONN_STATE, (void *)&ev, sizeof(ev));
	}

	g_free(address);
}

bt_status_t _bt_hal_dbus_handler_av_connect(bt_bdaddr_t *bd_addr)
{
	char *address;
	GDBusConnection *conn;
	int ret;

	if (!bd_addr) {
		ERR("bd_addr is NULL, return");
		return BT_STATUS_PARM_INVALID;
	}

	conn = _bt_hal_get_system_gconn();
	if (!conn) {
		ERR("_bt_hal_get_system_gconn returned NULL, return");
		return BT_STATUS_FAIL;
	}

	address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE * sizeof(char));
	if (!address) {
		ERR("Memory allocation failed");
		return BT_STATUS_NOMEM;
	}
	_bt_hal_convert_addr_type_to_string(address, bd_addr->address);

	ret = _bt_hal_connect_profile(address, A2DP_SINK_UUID,
			__bt_av_connect_cb, address);

	if (ret != BT_HAL_ERROR_NONE) {
		ERR("_bt_hal_connect_profile(A2DP) Error");
		g_free(address);
		return BT_STATUS_FAIL;
	}
	return BT_STATUS_SUCCESS;
}

static void __bt_av_disconnect_cb(GDBusProxy *proxy, GAsyncResult *res,
		gpointer user_data)
{
	GError *g_error = NULL;
	struct hal_ev_a2dp_conn_state ev;
	GVariant *reply = NULL;
	char *address = user_data;
	int result = BT_STATUS_SUCCESS;

	DBG("+");

	reply = g_dbus_proxy_call_finish(proxy, res, &g_error);
	g_object_unref(proxy);
	if (reply == NULL) {
		ERR("A2DP Disconnect Dbus Call Error");
		if (g_error) {
			ERR("Error: %s\n", g_error->message);
			g_clear_error(&g_error);
		}
		result = BT_STATUS_FAIL;
	}
	g_variant_unref(reply);

	if (result != BT_STATUS_FAIL)
		DBG("A2DP Disconnect successful for Device: %s", address);
	else
		DBG("A2DP Disconnect un-successful for Device: %s", address);

	if (result == BT_STATUS_FAIL) {
		/* Prepare to send AV Disconnected event */
		memset(&ev, 0, sizeof(ev));
		_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
		ev.state = HAL_EV_A2DP_STATE_DISCONNECTED;
		if (!event_cb)
			ERR("HF DBUS handler callback not registered");
		else
			event_cb(HAL_EV_A2DP_CONN_STATE, (void *)&ev, sizeof(ev));
	} else {
		/* Prepare to send AV Disconnecting event */
		memset(&ev, 0, sizeof(ev));
		_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
		ev.state = HAL_EV_A2DP_STATE_DISCONNECTING;
		if (!event_cb)
			ERR("HF DBUS handler callback not registered");
		else
			event_cb(HAL_EV_A2DP_CONN_STATE, (void *)&ev, sizeof(ev));
	}
	g_free(address);
	DBG("-");
}

bt_status_t _bt_hal_dbus_handler_av_disconnect(bt_bdaddr_t *bd_addr)
{
	char *address;
	GDBusConnection *conn;
	int ret;

	if (!bd_addr) {
		ERR("bd_addr is NULL, return");
		return BT_STATUS_PARM_INVALID;
	}

	conn = _bt_hal_get_system_gconn();
	if (!conn) {
		ERR("_bt_hal_get_system_gconn returned NULL, return");
		return BT_STATUS_FAIL;
	}

	address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE * sizeof(char));
	if (!address) {
		ERR("Memory allocation failed");
		return BT_STATUS_NOMEM;
	}
	_bt_hal_convert_addr_type_to_string(address, bd_addr->address);

	ret = _bt_hal_disconnect_profile(address, A2DP_SINK_UUID,
			__bt_av_disconnect_cb, address);

	if (ret != BT_HAL_ERROR_NONE) {
		ERR("_bt_hal_connect_profile(AV) Error");
		return BT_STATUS_FAIL;
	}

	return BT_STATUS_SUCCESS;
}

bt_status_t _bt_hal_dbus_handler_enable_a2dp_source(void)
{
	GDBusProxy *proxy;
	GVariant *ret;
	GError *error = NULL;
	gchar *adapter_path;
	GDBusConnection *gconn;

	DBG("+");

	if (_bt_hal_is_service_enabled(A2DP_SOURCE_UUID)) {
		ERR("Audio sink role already enabled, return");
		return BT_STATUS_SUCCESS;
	}

	gconn = _bt_hal_get_system_gconn();
	if (!gconn) {
		ERR("_bt_hal_get_system_gconn returned NULL, return");
		return BT_STATUS_FAIL;
	}

	adapter_path = _bt_hal_get_adapter_path();
	if (!adapter_path) {
		ERR("adapter_path is NULL");
		return BT_STATUS_FAIL;
	}

	proxy = g_dbus_proxy_new_sync(gconn, G_DBUS_PROXY_FLAGS_NONE, NULL,
			BT_HAL_BLUEZ_NAME, adapter_path, BT_HAL_MEDIA_INTERFACE, NULL, &error);
	g_free(adapter_path);
	if (proxy == NULL) {
		ERR("Unable to create proxy");

		if (error) {
			ERR("Error: %s", error->message);
			g_clear_error(&error);
		}
		return BT_STATUS_FAIL;
	}

	ret = g_dbus_proxy_call_sync(proxy, "SelectRole", g_variant_new("(s)", "source"),
			G_DBUS_CALL_FLAGS_NONE, -1, NULL, &error);
	g_object_unref(proxy);
	if (ret == NULL) {
		ERR("Call SelectRole Failed");
		if (error) {
			ERR("errCode[%x], message[%s]", error->code, error->message);
			g_clear_error(&error);
		}
		return BT_STATUS_FAIL;
	}

	g_variant_unref(ret);
	DBG("-");
	return BT_STATUS_SUCCESS;
}
