/*
 *
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: Nilesh Trimbake <t.shripati@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

#include <glib.h>
#include <gio/gio.h>
#include <dlog.h>
#include <vconf.h>

#include "bt-hal-avrcp-ctrl-dbus-handler.h"
#include "bt-hal-dbus-common-utils.h"
#include "bt-hal-internal.h"

static bt_player_settinngs_t repeat_status[] = {
	{ REPEAT_MODE_OFF, "off" },
	{ REPEAT_SINGLE_TRACK, "singletrack" },
	{ REPEAT_ALL_TRACK, "alltracks" },
	{ REPEAT_GROUP, "group" },
	{ REPEAT_INVALID, "" }
};

static bt_player_settinngs_t equalizer_status[] = {
	{ EQUALIZER_OFF, "off" },
	{ EQUALIZER_ON, "on" },
	{ EQUALIZER_INVALID, "" },
};

static bt_player_settinngs_t scan_status[] = {
	{ SCAN_MODE_OFF, "off" },
	{ SCAN_ALL_TRACK, "alltracks" },
	{ SCAN_GROUP, "group" },
	{ SCAN_INVALID, "" },
};

static bt_player_settinngs_t shuffle_settings[] = {
	{ SHUFFLE_MODE_OFF, "off" },
	{ SHUFFLE_ALL_TRACK, "alltracks" },
	{ SHUFFLE_GROUP, "group" },
	{ SHUFFLE_INVALID, "" }
};

static handle_stack_msg event_cb = NULL;

/* To send stack event to hal-avrcp-ctrl handler */
void _bt_hal_register_avrcp_ctrl_dbus_handler_cb(handle_stack_msg cb)
{
	event_cb = cb;
}

/* To send stack event to hal-avrcp-ctrl handler */
void _bt_hal_unregister_avrcp_ctrl_dbus_handler_cb()
{
	event_cb = NULL;
}

static void __bt_avrcp_target_connect_cb(GDBusProxy *proxy, GAsyncResult *res,
		gpointer user_data)
{
	struct hal_ev_avrcp_ctrl_conn_state ev;
	char *address = user_data;
	int result = BT_STATUS_SUCCESS;
	GVariant *reply = NULL;
	GError *g_error = NULL;

	DBG("+");

	reply = g_dbus_proxy_call_finish(proxy, res, &g_error);
	g_object_unref(proxy);
	if (reply == NULL) {
		ERR("AVRCP Target Connect Dbus Call Error");
		if (g_error) {
			ERR("Error: %s\n", g_error->message);
			g_clear_error(&g_error);
		}
		result = BT_STATUS_FAIL;
	}
	g_variant_unref(reply);

	DBG("Address: %s", address);
	/*
	 * If result is success, Remote device connected event will be triggered
	 * automatically from stack, so return from here.
	 */
	if (result == BT_STATUS_SUCCESS)
		goto done;

	/* Prepare to send AVRCP Target connection state event */
	memset(&ev, 0, sizeof(ev));
	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
	ev.state = HAL_AVRCP_CTRL_STATE_DISCONNECTED;
	if (!event_cb)
		ERR("AVRCP CTRL DBUS handler callback not registered");
	else
		event_cb(HAL_EV_AVRCP_CTRL_CONN_STATE, (void *)&ev, sizeof(ev));

done:
	g_free(address);
}

bt_status_t _bt_hal_dbus_handler_avrcp_ctrl_connect(bt_bdaddr_t *bd_addr)
{
	char *address;
	int ret;
	GDBusConnection *conn;

	if (!bd_addr) {
		ERR("bd_addr is NULL, return");
		return BT_STATUS_PARM_INVALID;
	}

	conn = _bt_hal_get_system_gconn();
	if (!conn) {
		ERR("_bt_hal_get_system_gconn returned NULL, return");
		return BT_STATUS_FAIL;
	}

	address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE * sizeof(char));
	if (!address) {
		ERR("Memory allocation failed");
		return BT_STATUS_NOMEM;
	}
	_bt_hal_convert_addr_type_to_string(address, bd_addr->address);

	ret = _bt_hal_connect_profile(address, AVRCP_TARGET_UUID,
			__bt_avrcp_target_connect_cb, address);

	if (ret != BT_HAL_ERROR_NONE) {
		ERR("_bt_hal_connect_profile(AVRCP Target) Error");
		g_free(address);
		return BT_STATUS_FAIL;
	}

	return BT_STATUS_SUCCESS;
}

static void __bt_avrcp_target_disconnect_cb(GDBusProxy *proxy, GAsyncResult *res,
		gpointer user_data)
{
	struct hal_ev_avrcp_ctrl_conn_state ev;
	char *address = user_data;
	int result = BT_STATUS_SUCCESS;
	GError *g_error = NULL;
	GVariant *reply = NULL;

	DBG("+");

	reply = g_dbus_proxy_call_finish(proxy, res, &g_error);
	g_object_unref(proxy);
	if (reply == NULL) {
		ERR("AVRCP Target Disconnect Dbus Call Error");
		if (g_error) {
			ERR("Error: %s\n", g_error->message);
			g_clear_error(&g_error);
		}
		result = BT_STATUS_FAIL;
	}
	g_variant_unref(reply);

	if (result != BT_STATUS_FAIL) {
		DBG("AVRCP Target Disconnect successful for Device: %s", address);
		g_free(address);
		return;
	} else {
		DBG("AVRCP Target Disconnect un-successful for Device: %s", address);
	}

	/*
	 * Prepare to send AVRCP Target connection state event indicating
	 * disconnection failed.
	 */
	memset(&ev, 0, sizeof(ev));
	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
	ev.state = HAL_AVRCP_CTRL_STATE_CONNECTED;
	if (!event_cb)
		ERR("AVRCP Target DBUS handler callback not registered");
	else
		event_cb(HAL_EV_AVRCP_CTRL_CONN_STATE, (void *)&ev, sizeof(ev));
	g_free(address);
	DBG("-");
}

bt_status_t _bt_hal_dbus_handler_avrcp_ctrl_disconnect(bt_bdaddr_t *bd_addr)
{
	char *address;
	int ret;
	GDBusConnection *conn;

	if (!bd_addr) {
		ERR("bd_addr is NULL, return");
		return BT_STATUS_PARM_INVALID;
	}

	conn = _bt_hal_get_system_gconn();
	if (!conn) {
		ERR("_bt_hal_get_system_gconn returned NULL, return");
		return BT_STATUS_FAIL;
	}

	address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE * sizeof(char));
	if (!address) {
		ERR("Memory allocation failed");
		return BT_STATUS_NOMEM;
	}
	_bt_hal_convert_addr_type_to_string(address, bd_addr->address);

	ret = _bt_hal_disconnect_profile(address, AVRCP_TARGET_UUID,
			__bt_avrcp_target_disconnect_cb, address);

	if (ret != BT_HAL_ERROR_NONE) {
		ERR("_bt_hal_connect_profile(AVRCP Target) Error");
		g_free(address);
		return BT_STATUS_FAIL;
	}

	return BT_STATUS_SUCCESS;
}

static int __bt_media_send_control_msg(bt_bdaddr_t *bd_addr, const char *name)
{
	GVariant *reply = NULL;
	GError *err = NULL;
	GDBusProxy *proxy = NULL;

	if (name == NULL) {
		ERR("Command is NULL");
		return BT_STATUS_FAIL;
	}
	DBG("Command [%s]", name);
	proxy = _bt_hal_get_avrcp_ctrl_proxy(bd_addr);
	if (proxy == NULL) {
		ERR("Unable to allocate new proxy \n");
		if (err) {
			ERR("%s", err->message);
			g_clear_error(&err);
		}
		return BT_STATUS_FAIL;
	}
	reply = g_dbus_proxy_call_sync(proxy, name, NULL,
			G_DBUS_CALL_FLAGS_NONE, -1, NULL, &err);

	if (!reply) {
		ERR("Error returned in method call");
		if (err) {
			ERR("%s", err->message);
			g_clear_error(&err);
		}
		return BT_STATUS_FAIL;
	}
	g_variant_unref(reply);
	DBG("-");
	return BT_STATUS_SUCCESS;
}

bt_status_t _bt_hal_dbus_handler_avrcp_ctrl_cmd(bt_bdaddr_t *bd_addr, uint8_t key_code, uint8_t key_state)
{	int ret;
	DBG("+");

	switch (key_code) {
		case PLAY: {
			ret = __bt_media_send_control_msg(bd_addr, "Play");
			break;
		}
		case PAUSE: {
			ret = __bt_media_send_control_msg(bd_addr, "Pause");
			break;
		}
		case STOP: {
			ret = __bt_media_send_control_msg(bd_addr, "Stop");
			break;
		}
		case NEXT: {
			ret = __bt_media_send_control_msg(bd_addr, "Next");
			break;
		}
		case PREVIOUS: {
			ret = __bt_media_send_control_msg(bd_addr, "Previous");
			break;
		}
		case PRESS_FAST_FORWARD: {
			ret = __bt_media_send_control_msg(bd_addr, "PressFastForward");
			break;
		}
		case RELEASE_FAST_FORWARD: {
			ret = __bt_media_send_control_msg(bd_addr, "ReleaseFastForward");
			break;
		}
		case PRESS_REWIND: {
			ret = __bt_media_send_control_msg(bd_addr, "PressRewind");
			break;
		}
		case RELEASE_REWIND: {
			ret = __bt_media_send_control_msg(bd_addr, "ReleaseRewind");
			break;
		}
		case VOLUME_UP: {
			ret = __bt_media_send_control_msg(bd_addr, "VolumeUp");
			break;
		}
		case VOLUME_DOWN: {
			ret = __bt_media_send_control_msg(bd_addr, "VolumeDown");
			break;
		}
		default: {
			DBG("Invalid Type\n");
			return BT_STATUS_FAIL;
		}
	}
	DBG("-");
	return ret;
}
static const char *__bt_media_type_to_str(int type)
{
	switch (type) {
	case BTRC_PLAYER_ATTR_EQUALIZER:
		return "Equalizer";
	case BTRC_PLAYER_ATTR_REPEAT:
		return "Repeat";
	case BTRC_PLAYER_ATTR_SHUFFLE:
		return "Shuffle";
	case BTRC_PLAYER_ATTR_SCAN:
		return "Scan";
	default:
		return NULL;
	}

	return NULL;
}

#if 0
static const char *__bt_transport_type_to_str(int type)
{
	switch (type) {
	case BTRC_TRANSPORT_ATTR_DELAY:
		return "Delay";
	case BTRC_TRANSPORT_ATTR_VOLUME:
		return "Volume";
	default:
		return NULL;
	}

	return NULL;
}
#endif

bt_status_t _bt_hal_dbus_handler_avrcp_ctrl_set_property(bt_bdaddr_t *bd_addr, uint8_t type, uint8_t value)
{
	GVariant *reply, *param = NULL;
	GError *err = NULL;
	GDBusProxy *proxy = NULL;

	DBG("+");
	switch (type) {
	case BTRC_PLAYER_ATTR_EQUALIZER:
		param = g_variant_new("s", equalizer_status[value].property);
		INFO("equalizer_status %s", equalizer_status[value].property);
		break;
	case BTRC_PLAYER_ATTR_REPEAT:
		param = g_variant_new("s", repeat_status[value].property);
		INFO("repeat_status %s", repeat_status[value].property);
		break;
	case BTRC_PLAYER_ATTR_SHUFFLE:
		param = g_variant_new("s", shuffle_settings[value].property);
		INFO("shuffle_settings %s", shuffle_settings[value].property);
		break;
	case BTRC_PLAYER_ATTR_SCAN:
		param = g_variant_new("s", scan_status[value].property);
		INFO("scan_status %s", scan_status[value].property);
		break;
	default:
		ERR("Invalid property type: %d", type);
		return BT_STATUS_FAIL;
	}

	proxy = _bt_hal_get_avrcp_ctrl_properties_proxy(bd_addr);
	if (proxy == NULL) {
		ERR("Unable to allocate new proxy \n");
		if (err) {
			ERR("%s", err->message);
			g_clear_error(&err);
		}
		return BT_STATUS_FAIL;
	}

	reply = g_dbus_proxy_call_sync(proxy,
			"Set", g_variant_new("(ssv)",
				BT_HAL_PLAYER_CONTROL_INTERFACE,
				__bt_media_type_to_str(type), param),
			G_DBUS_CALL_FLAGS_NONE, -1, NULL, &err);
	g_object_unref(proxy);
	if (!reply) {
		ERR("Can't get managed objects");
		if (err) {
			ERR("SetProperty Fail: %s", err->message);
			g_clear_error(&err);
			return BT_STATUS_FAIL;
		}
	}

	g_variant_unref(reply);
	DBG("-");
	return BT_STATUS_SUCCESS;
}

bt_status_t _bt_hal_dbus_handler_avrcp_transport_set_property(bt_bdaddr_t *bd_addr, int type, unsigned int value)
{
	GVariant *reply, *param = NULL;
	GError *err = NULL;
	GDBusProxy *proxy = NULL;
	uint16_t property_level = (uint16_t)value;

	DBG("+");
	switch (type) {
	case BTRC_TRANSPORT_ATTR_DELAY:
		param = g_variant_new("q", property_level);
		INFO("delay level %d", property_level);
		break;
	case BTRC_TRANSPORT_ATTR_VOLUME:
		param = g_variant_new("q", property_level);
		INFO("volume level %d", property_level);
		break;
	default:
		ERR("Invalid property type: %d", type);
		return BT_STATUS_FAIL;
	}

	proxy = _bt_hal_get_avrcp_transport_properties_proxy(bd_addr);
	if (proxy == NULL) {
		ERR("Unable to allocate new proxy \n");
		if (err) {
			ERR("%s", err->message);
			g_clear_error(&err);
		}
		return BT_STATUS_FAIL;
	}

	reply = g_dbus_proxy_call_sync(proxy,
			"Set", g_variant_new("(ssv)",
				BT_HAL_MEDIATRANSPORT_INTERFACE,
				__bt_media_type_to_str(type), param),
			G_DBUS_CALL_FLAGS_NONE, -1, NULL, &err);
	g_object_unref(proxy);
	if (!reply) {
		ERR("Can't get managed objects");
		if (err) {
			ERR("SetProperty Fail: %s", err->message);
			g_clear_error(&err);
			return BT_STATUS_FAIL;
		}
	}

	g_variant_unref(reply);
	DBG("-");
	return BT_STATUS_SUCCESS;
}
