/*
 * Bluetooth-frwk
 *
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact:  Nilesh Trimbake <t.shripati@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <stdbool.h>
#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <dlog.h>

#include "bt-hal.h"
#include "bt-hal-log.h"
#include "bt-hal-msg.h"
#include "bt-hal-utils.h"
#include "bt-hal-avrcp-ctrl-dbus-handler.h"

#define AVRC_MAX_NUM_MEDIA_ATTR_ID               7

static const btrc_ctrl_callbacks_t *bt_hal_avrcp_ctrl_cbacks;

static bool interface_ready(void)
{
	return bt_hal_avrcp_ctrl_cbacks != NULL;
}

static bt_status_t avrcp_ctrl_connect(bt_bdaddr_t *bd_addr)
{
	DBG("");
	return _bt_hal_dbus_handler_avrcp_ctrl_connect(bd_addr);
}

static bt_status_t avrcp_ctrl_disconnect(bt_bdaddr_t *bd_addr)
{
	DBG("");
	return _bt_hal_dbus_handler_avrcp_ctrl_disconnect(bd_addr);
}

static bt_status_t avrcp_ctrl_send_pass_through_cmd(bt_bdaddr_t *bd_addr, uint8_t key_code, uint8_t key_state)
{
	DBG("");
	return _bt_hal_dbus_handler_avrcp_ctrl_cmd(bd_addr, key_code, key_state);
}

static bt_status_t avrcp_ctrl_set_player_app_setting_cmd(bt_bdaddr_t *bd_addr,
				uint8_t num_attrib, uint8_t* type, uint8_t* value)
{
	int i = 0;
	int ret = BT_STATUS_SUCCESS;

	DBG("");
	for (i = 0; i < num_attrib; i++)
		ret = _bt_hal_dbus_handler_avrcp_ctrl_set_property(bd_addr, type[i], value[i]);

	return ret;
}

static bt_status_t avrcp_transport_setting_cmd(bt_bdaddr_t *bd_addr,
				uint8_t num_attrib, int *type, unsigned int *value)
{
	int i = 0;
	int ret = BT_STATUS_SUCCESS;

	DBG("");
	for (i = 0; i < num_attrib; i++)
		ret = _bt_hal_dbus_handler_avrcp_transport_set_property(bd_addr, type[i], value[i]);

	return ret;
}

static void __bt_hal_handle_avrcp_ctrl_conn_state(void *buf, uint16_t len)
{
	struct hal_ev_avrcp_ctrl_conn_state *ev = buf;

	if (bt_hal_avrcp_ctrl_cbacks->connection_state_cb) {
		if (ev->state == HAL_AVRCP_CTRL_STATE_CONNECTED)
			bt_hal_avrcp_ctrl_cbacks->connection_state_cb(
					TRUE, TRUE, (bt_bdaddr_t *) ev->bdaddr);
		else
			bt_hal_avrcp_ctrl_cbacks->connection_state_cb(
					FALSE, FALSE, (bt_bdaddr_t *) ev->bdaddr);
	}
}

btrc_play_status_t __get_play_status(int status)
{
	btrc_play_status_t play_status;
	play_status = BTRC_PLAYSTATE_ERROR;

	switch (status) {
	case BTRC_PLAYSTATE_STOPPED:
		play_status = BTRC_PLAYSTATE_STOPPED;
		break;
	case BTRC_PLAYSTATE_PLAYING:
		play_status = BTRC_PLAYSTATE_PLAYING;
		break;
	case BTRC_PLAYSTATE_PAUSED:
		play_status = BTRC_PLAYSTATE_PAUSED;
		break;
	case BTRC_PLAYSTATE_FWD_SEEK:
		play_status = BTRC_PLAYSTATE_FWD_SEEK;
		break;
	case BTRC_PLAYSTATE_REV_SEEK:
		play_status = BTRC_PLAYSTATE_REV_SEEK;
		break;
	default:
		DBG("Wrong play status received:");
		break;
	}
	return play_status;
}

static void __bt_hal_handle_avrcp_ctrl_playstatus_changed_event(void *buf, uint16_t len)
{
	struct hal_ev_play_status_changed *ev = buf;

	if (bt_hal_avrcp_ctrl_cbacks->play_position_changed_cb)
		bt_hal_avrcp_ctrl_cbacks->play_status_changed_cb((bt_bdaddr_t *)ev->bdaddr, ev->status);
}

static void __bt_hal_handle_avrcp_ctrl_playposition_changed_event(void *buf, uint16_t len)
{
	struct hal_ev_play_position *ev = buf;

	if (bt_hal_avrcp_ctrl_cbacks->play_position_changed_cb)
		bt_hal_avrcp_ctrl_cbacks->play_position_changed_cb((bt_bdaddr_t *)ev->bdaddr, ev->len, ev->pos);

}

static void __bt_hal_handle_avrcp_ctrl_player_setting_changed(void *buf, uint16_t len)
{
	struct hal_ev_player_setting *ev = buf;
	btrc_player_settings_t player_setting;

	memset(&player_setting, 0, sizeof(player_setting));
	player_setting.num_attr = ev->num_attr;
	memcpy(player_setting.attr_ids, ev->attr_ids, sizeof(player_setting.attr_ids));
	memcpy(player_setting.attr_values, ev->attr_values, sizeof(player_setting.attr_values));

	if (bt_hal_avrcp_ctrl_cbacks->playerapplicationsetting_changed_cb)
		bt_hal_avrcp_ctrl_cbacks->playerapplicationsetting_changed_cb(
				(bt_bdaddr_t *)ev->bdaddr, &player_setting);
}

static void __bt_hal_handle_avrcp_ctrl_pass_cmd_res_event(void *buf, uint16_t len)
{
	struct hal_ev_pass_cmd_rsp *ev = (struct hal_ev_pass_cmd_rsp *)buf;

	if (bt_hal_avrcp_ctrl_cbacks->passthrough_rsp_cb)
		bt_hal_avrcp_ctrl_cbacks->passthrough_rsp_cb(NULL,
			ev->key_code, ev->key_state);
}

static void __bt_hal_handle_avrcp_ctrl_set_app_setting_res_event(void *buf, uint16_t len)
{
	struct hal_ev_set_player_app_setting_rsp *ev = buf;

	if (bt_hal_avrcp_ctrl_cbacks->setplayerappsetting_rsp_cb)
		bt_hal_avrcp_ctrl_cbacks->setplayerappsetting_rsp_cb((bt_bdaddr_t *)ev->bdaddr, ev->resp);
}

static void __bt_hal_handle_avrcp_ctrl_track_changed_event(void *buf, uint16_t len)
{
	struct hal_ev_track_changed *ev = buf;
	btrc_element_attr_val_t attr[HAL_MAX_ATTR_NUM];
	int i;

	for (i = 0; i < ev->num_attr; i++) {
		attr[i].attr_id = ev->attr[i].attr_id;
		memcpy(attr[i].text, ev->attr[i].text, sizeof(ev->attr[i].text));
	}

	DBG("call track_changed_cb");
	if (bt_hal_avrcp_ctrl_cbacks->track_changed_cb)
		bt_hal_avrcp_ctrl_cbacks->track_changed_cb(
				(bt_bdaddr_t *)ev->bdaddr, ev->num_attr, attr);
}

static void __bt_hal_handle_avrcp_ctrl_event(int message, void *buf, uint16_t len)
{
	DBG("+");
	if (!interface_ready())
		return;
	switch (message) {
	case HAL_EV_AVRCP_CTRL_CONN_STATE:
		DBG("Event: HAL_EV_AVRCP_CTRL_CONN_STATE");
		__bt_hal_handle_avrcp_ctrl_conn_state(buf, len);
		break;
	case HAL_EV_AVRCP_CTRL_PASS_THROUGH_RSP:
		DBG("Event: HAL_EV_AVRCP_CTRL_PASS_THROUGH_RSP");
		__bt_hal_handle_avrcp_ctrl_pass_cmd_res_event(buf, len);
		break;
	case HAL_EV_AVRCP_CTRL_SET_PLAYER_APP_SETTING_RSP:
		DBG("Event: HAL_EV_AVRCP_CTRL_SET_PLAYER_APP_SETTING_RSP");
		__bt_hal_handle_avrcp_ctrl_set_app_setting_res_event(buf, len);
		break;
	case HAL_EV_AVRCP_CTRL_PLAYER_APP_SETTING_CHANGED:
		DBG("Event: HAL_EV_AVRCP_CTRL_PLAYER_APP_SETTING_CHANGED");
		__bt_hal_handle_avrcp_ctrl_player_setting_changed(buf, len);
		break;
	case HAL_EV_AVRCP_CTRL_TRACK_CHANGED:
		DBG("Event: HAL_EV_AVRCP_CTRL_TRACK_CHANGED");
		__bt_hal_handle_avrcp_ctrl_track_changed_event(buf, len);
		break;
	case HAL_EV_AVRCP_CTRL_PLAY_POSITION_CHANGED:
		DBG("Event: HAL_EV_AVRCP_CTRL_PLAY_POSITION_CHANGED");
		__bt_hal_handle_avrcp_ctrl_playposition_changed_event(buf, len);
		break;
	case HAL_EV_AVRCP_CTRL_PLAY_STATUS_CHANGED:
		DBG("Event: HAL_EV_AVRCP_CTRL_PLAY_STATUS_CHANGED");
		__bt_hal_handle_avrcp_ctrl_playstatus_changed_event(buf, len);
		break;
	default:
		DBG("Event Currently not handled!!");
		break;
	}

	DBG("-");
}

static bt_status_t init(btrc_ctrl_callbacks_t* callbacks)
{
	DBG("");

	if (interface_ready())
		return BT_STATUS_DONE;

	bt_hal_avrcp_ctrl_cbacks = callbacks;
	DBG("Register AVRCP Ctroller events callback function");
	_bt_hal_register_avrcp_ctrl_dbus_handler_cb(__bt_hal_handle_avrcp_ctrl_event);
	_bt_hal_register_event_handler_cb(HAL_AVRCP_CTRL, __bt_hal_handle_avrcp_ctrl_event);
	return BT_STATUS_SUCCESS;
}

static void cleanup(void)
{
	DBG("");

	if (!interface_ready())
		return;

	_bt_hal_unregister_event_handler_cb(HAL_AVRCP_CTRL);
	bt_hal_avrcp_ctrl_cbacks = NULL;
}

static btrc_ctrl_interface_t avrcp_ctrl_if = {
	.size = sizeof(avrcp_ctrl_if),
	.init = init,
	.connect = avrcp_ctrl_connect,
	.disconnect = avrcp_ctrl_disconnect,
	.send_pass_through_cmd = avrcp_ctrl_send_pass_through_cmd,
	.set_player_app_setting_cmd = avrcp_ctrl_set_player_app_setting_cmd,
	.set_transport_setting_cmd = avrcp_transport_setting_cmd,
	.cleanup = cleanup
};

btrc_ctrl_interface_t *bt_get_avrcp_ctrl_interface(void)
{
	DBG("Get AVRCP Controller Profile Interface");
	return &avrcp_ctrl_if;
}
