/*
 * Bluetooth-frwk
 *
 * Copyright (c) 2000 - 2011 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact:  Atul Kumar Rai <a.rai@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdbool.h>
#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <dlog.h>

#include "bt-hal.h"
#include "bt-hal-log.h"
#include "bt-hal-msg.h"
#include "bt-hal-utils.h"

#include "bt-hal-event-receiver.h"
#include "bt-hal-avrcp-tg-dbus-handler.h"

static const btrc_callbacks_t *bt_hal_avrcp_tg_cbacks;

static bool interface_ready(void)
{
	return bt_hal_avrcp_tg_cbacks != NULL;
}

bt_status_t get_play_status_rsp(bt_bdaddr_t *bd_addr, btrc_play_status_t play_status,
			uint32_t song_len, uint32_t song_pos)
{
	return BT_STATUS_UNSUPPORTED;
}

bt_status_t list_player_app_attr_rsp(bt_bdaddr_t *bd_addr, int num_attr,
			btrc_player_attr_t *p_attrs)
{
	return BT_STATUS_UNSUPPORTED;
}

bt_status_t list_player_app_value_rsp(bt_bdaddr_t *bd_addr, int num_val, uint8_t *p_vals)
{
	return BT_STATUS_UNSUPPORTED;
}

bt_status_t get_player_app_value_rsp(bt_bdaddr_t *bd_addr, btrc_player_settings_t *p_vals)
{
	return BT_STATUS_UNSUPPORTED;
}

bt_status_t get_player_app_attr_text_rsp(bt_bdaddr_t *bd_addr, int num_attr,
		btrc_player_setting_text_t *p_attrs)
{
	return BT_STATUS_UNSUPPORTED;
}

bt_status_t get_player_app_value_text_rsp(bt_bdaddr_t *bd_addr, int num_val,
		btrc_player_setting_text_t *p_vals)
{
	return BT_STATUS_UNSUPPORTED;
}

bt_status_t get_element_attr_rsp(bt_bdaddr_t *bd_addr, uint8_t num_attr,
		btrc_element_attr_val_t *p_attrs)
{
	if (num_attr > 0 && !p_attrs)
		return BT_STATUS_PARM_INVALID;

	return _bt_hal_dbus_handler_avrcp_tg_set_track_info(num_attr, p_attrs);
}

bt_status_t set_player_app_value_rsp(bt_bdaddr_t *bd_addr, btrc_status_t rsp_status)
{
	return BT_STATUS_UNSUPPORTED;
}

gboolean __fetch_track_info(gpointer data)
{
	if (bt_hal_avrcp_tg_cbacks->get_element_attr_cb) {
		uint8_t num_attr;
		btrc_media_attr_t attrs[] = {
			BTRC_MEDIA_ATTR_TITLE, BTRC_MEDIA_ATTR_ARTIST,
			BTRC_MEDIA_ATTR_ALBUM, BTRC_MEDIA_ATTR_TRACK_NUM,
			BTRC_MEDIA_ATTR_NUM_TRACKS, BTRC_MEDIA_ATTR_GENRE,
			BTRC_MEDIA_ATTR_PLAYING_TIME};

		num_attr = sizeof(attrs) / sizeof(btrc_media_attr_t);

		bt_hal_avrcp_tg_cbacks->get_element_attr_cb(num_attr, attrs, NULL);
	}

	return FALSE;
}

bt_status_t register_notification_rsp(btrc_event_id_t event_id,
		btrc_notification_type_t type,
		btrc_register_notification_t *p_param)
{
	if (BTRC_NOTIFICATION_TYPE_INTERIM == type)
		return BT_STATUS_SUCCESS;

	switch (event_id) {
	case BTRC_EVT_PLAY_STATUS_CHANGED:
		_bt_hal_dbus_handler_avrcp_tg_play_status_changed(p_param->play_status);
		break;
	case BTRC_EVT_TRACK_CHANGE:
		g_idle_add(__fetch_track_info, NULL);
		break;
	case BTRC_EVT_PLAY_POS_CHANGED:
		_bt_hal_dbus_handler_avrcp_tg_play_pos_changed(p_param->song_pos);
		break;
	case BTRC_EVT_APP_SETTINGS_CHANGED:
		_bt_hal_dbus_handler_avrcp_tg_app_setting_changed(&(p_param->player_setting));
		break;
	default:
		ERR("Unknown event Id: %d", event_id);
	}

	return BT_STATUS_SUCCESS;
}

bt_status_t set_volume(uint8_t volume)
{
	return BT_STATUS_UNSUPPORTED;
}

bt_status_t set_addressed_player_rsp(bt_bdaddr_t *bd_addr, btrc_status_t rsp_status)
{
	return BT_STATUS_UNSUPPORTED;
}

bt_status_t set_browsed_player_rsp(bt_bdaddr_t *bd_addr, btrc_status_t rsp_status,
		uint32_t num_items, uint16_t charset_id, uint8_t folder_depth,
		btrc_br_folder_name_t *p_folders)
{
	return BT_STATUS_UNSUPPORTED;
}

bt_status_t get_folder_items_list_rsp(bt_bdaddr_t *bd_addr, btrc_status_t rsp_status,
		uint16_t uid_counter, uint8_t num_items, btrc_folder_items_t *p_items)
{
	return BT_STATUS_UNSUPPORTED;
}

bt_status_t change_path_rsp(bt_bdaddr_t *bd_addr, btrc_status_t rsp_status,
		uint32_t num_items)
{
	return BT_STATUS_UNSUPPORTED;
}

bt_status_t get_item_attr_rsp(bt_bdaddr_t *bd_addr, btrc_status_t rsp_status,
		uint8_t num_attr, btrc_element_attr_val_t *p_attrs)
{
	return BT_STATUS_UNSUPPORTED;
}

bt_status_t play_item_rsp(bt_bdaddr_t *bd_addr, btrc_status_t rsp_status)
{
	return BT_STATUS_UNSUPPORTED;
}

bt_status_t get_total_num_of_items_rsp(bt_bdaddr_t *bd_addr, btrc_status_t rsp_status,
		uint32_t uid_counter, uint32_t num_items)
{
	return BT_STATUS_UNSUPPORTED;
}

bt_status_t search_rsp(bt_bdaddr_t *bd_addr, btrc_status_t rsp_status, uint32_t uid_counter,
		uint32_t num_items)
{
	return BT_STATUS_UNSUPPORTED;
}

bt_status_t add_to_now_playing_rsp(bt_bdaddr_t *bd_addr, btrc_status_t rsp_status)
{
	return BT_STATUS_UNSUPPORTED;
}

static gboolean __bt_avrcp_register_notifications(gpointer user_data)
{
	bt_bdaddr_t *bdaddr = user_data;

	if (!bt_hal_avrcp_tg_cbacks->register_notification_cb)
		return FALSE;

	bt_hal_avrcp_tg_cbacks->register_notification_cb(BTRC_EVT_PLAY_STATUS_CHANGED, 0, bdaddr);
	bt_hal_avrcp_tg_cbacks->register_notification_cb(BTRC_EVT_TRACK_CHANGE, 0, bdaddr);
	bt_hal_avrcp_tg_cbacks->register_notification_cb(BTRC_EVT_PLAY_POS_CHANGED, 1, bdaddr);
	bt_hal_avrcp_tg_cbacks->register_notification_cb(BTRC_EVT_APP_SETTINGS_CHANGED, 0, bdaddr);

	return FALSE;
}

static void __handle_player_property(void *buf, uint16_t len)
{
	btrc_player_settings_t player_setting;
	struct hal_ev_avrcp_tg_player_property *ev = buf;

	if (!bt_hal_avrcp_tg_cbacks->set_player_app_value_cb)
		return;

	memset(&player_setting, 0, sizeof(btrc_player_settings_t));
	player_setting.num_attr = 1;

	switch (ev->prop_type) {
	case HAL_AVRCP_TG_PLAYER_PROP_REPEAT:
		player_setting.attr_ids[0] = BTRC_PLAYER_ATTR_REPEAT;
		player_setting.attr_values[0] = ev->value;
		break;
	case HAL_AVRCP_TG_PLAYER_PROP_SHUFFLE:
		player_setting.attr_ids[0] = BTRC_PLAYER_ATTR_SHUFFLE;
		player_setting.attr_values[0] = ev->value;
		break;
	case HAL_AVRCP_TG_PLAYER_PROP_SCAN:
		player_setting.attr_ids[0] = BTRC_PLAYER_ATTR_SCAN;
		player_setting.attr_values[0] = ev->value;
		break;
	case HAL_AVRCP_TG_PLAYER_PROP_EQUALIZER:
		player_setting.attr_ids[0] = BTRC_PLAYER_ATTR_EQUALIZER;
		player_setting.attr_values[0] = ev->value;
		break;
	default:
		INFO("Player property not handled: %d", ev->prop_type);
		return;
	}

	bt_hal_avrcp_tg_cbacks->set_player_app_value_cb(&player_setting, NULL);
}

static void __bt_hal_handle_avrcp_tg_events(int message, void *buf, uint16_t len)
{
	DBG("+");

	switch (message) {
#ifdef TIZEN_BT_HAL
	case HAL_EV_AVRCP_TG_CONN_STATE:
	{
		struct hal_ev_avrcp_tg_conn_state *ev = buf;

		if (!bt_hal_avrcp_tg_cbacks->connection_state_cb)
			break;

		bt_hal_avrcp_tg_cbacks->connection_state_cb(
				(bt_bdaddr_t *)ev->bdaddr, ev->state);
		/*
		 * To retrieve various player properties, callbacks are invoked from bluedroid to
		 * upper layers (in our case OAL). But in case of BlueZ, these properties are
		 * cached in stack and replied from within. For this, BlueZ relies on property
		 * changed signals from upper layer, to keep its cache updated. To handle the above
		 * case in BlueZ HAL, we register notifications (internal to HAL) and send property
		 * changed signals to bluez on receiving notifications from upper layer (OAL).
		 */
		if (HAL_AVRCP_TG_STATE_CONNECTED == ev->state)
			g_idle_add(__bt_avrcp_register_notifications, ev->bdaddr);
		break;
	}
	case HAL_EV_AVRCP_TG_DELAY_CHANGE:
	{
		DBG("avrcp tg delay change");
		struct hal_ev_avrcp_tg_delay_changed *ev = buf;
		bt_hal_avrcp_tg_cbacks->delay_change_cb(
				ev->value, (bt_bdaddr_t *)ev->bdaddr);
		break;
	}
#endif
	case HAL_EV_AVRCP_TG_SET_PLAYER_PROPERTY:
		__handle_player_property(buf, len);
		break;
	default:
		ERR("Unknown event: 0x%.2X", message);
	}

	DBG("-");
}

#ifdef TIZEN_BT_HAL
bt_status_t hal_avrcp_tg_connect(bt_bdaddr_t *rem_addr)
{
	return _bt_hal_dbus_handler_avrcp_tg_connect(rem_addr);
}

bt_status_t hal_avrcp_tg_disconnect(bt_bdaddr_t *rem_addr)
{
	return _bt_hal_dbus_handler_avrcp_tg_disconnect(rem_addr);
}
#endif

static bt_status_t init(btrc_callbacks_t *callbacks)
{
	int ret;

	DBG("");

	if (interface_ready())
		return BT_STATUS_DONE;

	if (!callbacks)
		return BT_STATUS_PARM_INVALID;

	/* Register AVRCP Media player */
	ret = _bt_hal_dbus_handler_register_media_player();
	if (BT_STATUS_SUCCESS != ret)
		ERR("_bt_hal_dbus_handler_register_media_player failed");

	/* Register AVRCP event callback */
	_bt_hal_register_avrcp_tg_dbus_handler_cb(
			__bt_hal_handle_avrcp_tg_events);
	_bt_hal_register_event_handler_cb(HAL_AVRCP_TG,
			__bt_hal_handle_avrcp_tg_events);

	bt_hal_avrcp_tg_cbacks = callbacks;

	return BT_STATUS_SUCCESS;
}

static void cleanup(void)
{
	DBG("");

	if (!interface_ready())
		return;

	/* Un-register AVRCP Media player */
	_bt_hal_dbus_handler_unregister_media_player();

	/* Un-register AVRCP event callback */
	_bt_hal_unregister_avrcp_tg_dbus_handler_cb();
	_bt_hal_unregister_event_handler_cb(HAL_AVRCP_TG);

	bt_hal_avrcp_tg_cbacks = NULL;
}

static btrc_interface_t avrcp_tg_if = {
	.size = sizeof(avrcp_tg_if),
	.init = init,
#ifdef TIZEN_BT_HAL
	.connect = hal_avrcp_tg_connect,
	.disconnect = hal_avrcp_tg_disconnect,
#endif
	.get_play_status_rsp = get_play_status_rsp,
	.list_player_app_attr_rsp = list_player_app_attr_rsp,
	.list_player_app_value_rsp = list_player_app_value_rsp,
	.get_player_app_attr_text_rsp = get_player_app_attr_text_rsp,
	.get_player_app_value_text_rsp = get_player_app_value_text_rsp,
	.get_element_attr_rsp = get_element_attr_rsp,
	.set_player_app_value_rsp = set_player_app_value_rsp,
	.register_notification_rsp = register_notification_rsp,
	.set_volume = set_volume,
	.set_addressed_player_rsp = set_addressed_player_rsp,
	.set_browsed_player_rsp = set_browsed_player_rsp,
	.get_folder_items_list_rsp = get_folder_items_list_rsp,
	.change_path_rsp = change_path_rsp,
	.get_item_attr_rsp = get_item_attr_rsp,
	.play_item_rsp = play_item_rsp,
	.get_total_num_of_items_rsp = get_total_num_of_items_rsp,
	.search_rsp = search_rsp,
	.add_to_now_playing_rsp = add_to_now_playing_rsp,
	.cleanup = cleanup
};

btrc_interface_t *bt_get_avrcp_tg_interface(void)
{
	return &avrcp_tg_if;
}
