/*
 * BLUETOOTH HAL
 *
 * Copyright (c) 2015 -2016 Samsung Electronics Co., Ltd All Rights Reserved.
 *
 * Contact: Anupam Roy <anupam.r@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <glib.h>
#include <string.h>
#include <dlog.h>
#include <vconf.h>

#include <glib.h>
#include <gio/gio.h>
#include <vconf.h>

/* BT HAL Headers */
#include "bt-hal.h"
#include "bt-hal-log.h"
#include "bt-hal-msg.h"
#include "bt-hal-internal.h"
#include "bt-hal-event-receiver.h"
#include "bt-hal-dbus-common-utils.h"
#include "bt-hal-agent.h"
#include "bt-hal-adapter-le.h"
#include "bt-hal-gatt-server.h"
#include "bt-hal-gatt-client.h"

#define BASELEN_PROP_CHANGED (sizeof(struct hal_ev_adapter_props_changed) \
		+ sizeof(struct hal_property))

/*TODO: Basic filters are currently added,
  Need to add different event filters like HID,
  Device etc in subsequent patches */

/* Global variables and structures */
static GDBusConnection *manager_conn;
static handle_stack_msg event_cb = NULL;
static handle_stack_msg hid_event_cb = NULL;
static handle_stack_msg av_event_cb = NULL;
static handle_stack_msg a2dp_sink_event_cb = NULL;
static handle_stack_msg hf_event_cb = NULL;
static handle_stack_msg avrcp_ctrl_event_cb = NULL;
static handle_stack_msg avrcp_tg_event_cb = NULL;
static handle_stack_msg gatt_event_cb = NULL;
static guint event_id;

/*State Management sepration Control for Adapter and LE */
static gboolean is_adapter_activating = FALSE;
static gboolean is_le_activating = FALSE;

typedef struct {
	gchar* sender_name;
	gchar* object_path;
	gchar* interface_name;
	gchar* signal_name;
	GVariant *parameters;
} bt_hal_main_event_data_t;

/* Forward declarations */
static gboolean __bt_hal_event_manager(gpointer param);
static int __bt_hal_register_service_event(GDBusConnection *g_conn, int event_type);
static int __bt_hal_register_manager_subscribe_signal(GDBusConnection *conn, int subscribe);
static int __bt_hal_register_device_subscribe_signal(GDBusConnection *conn, int subscribe);
static int __bt_hal_register_input_subscribe_signal(GDBusConnection *conn, int subscribe);

static int __bt_hal_parse_event(GVariant *msg);
static int __bt_hal_get_owner_info(GVariant *msg, char **name, char **previous, char **current);

static void __bt_hal_handle_property_changed_event(GVariant *msg, const char *object_path);
static void __bt_hal_adapter_property_changed_event(GVariant *msg);
static  void __bt_hal_manager_event_filter(GDBusConnection *connection, const gchar *sender_name,
			const gchar *object_path, const gchar *interface_name, const gchar *signal_name,
			GVariant *parameters, gpointer user_data);
static int __bt_hal_initialize_manager_receiver(void);
static gboolean __bt_hal_parse_interface(GVariant *msg);
static void __bt_hal_handle_device_event(GVariant *value, GVariant *parameters);
static gboolean __bt_hal_parse_device_properties(GVariant *item);
static gboolean __bt_hal_discovery_finished_cb(gpointer user_data);
static void __bt_hal_device_property_changed_event(GVariant *msg, const char *path);
static void __bt_hal_dbus_device_found_properties(const char *device_path);
static void __bt_hal_device_properties_lookup(GVariant *result, char *address);
static void __bt_hal_handle_device_specific_events(GVariant *msg, const char *member, const char *path);
static void __bt_hal_send_device_acl_connection_state_event(int status, gboolean connected, const char *address);
static void __bt_hal_handle_input_event(GVariant *msg, const char *path);
static void __bt_hal_send_av_connection_state_event(gboolean connected, const char *address);
static void __bt_hal_send_a2dp_sink_connection_state_event(gboolean connected, const char *address);
static void __bt_hal_send_avrcp_ctrl_connection_state_event(gboolean connected, const char *address);
static void __bt_hal_handle_avrcp_ctrl_events(GVariant *msg, const char *member, const char *path);
static void __bt_hal_handle_avrcp_transport_events(GVariant *msg, const char *member, const char *path);

static void __bt_hal_send_device_trust_state_event(gboolean is_trusted, const char *address);
static int __bt_hal_register_audio_subscribe_signal(GDBusConnection *conn, int subscribe);
static void __bt_hal_handle_headset_events(GVariant *msg, const char *member, const char *path);
static void __bt_hal_send_hf_audio_connection_state_event(gboolean connected, const char *address);
static void __bt_hal_send_hf_connection_state_event(gboolean connected, const char *address);
static void __bt_hal_send_device_trusted_profile_changed_event(uint32_t trust_val, const char *address);
static void __bt_hal_handle_adv_report(GVariant *msg, const char *path);

static gboolean __bt_hal_discovery_finished_cb(gpointer user_data)
{
	event_id = 0;
	DBG("+");
	struct hal_ev_discovery_state_changed ev;
	ev.state = HAL_DISCOVERY_STATE_STOPPED;
	event_cb(HAL_EV_DISCOVERY_STATE_CHANGED, &ev, sizeof(ev));
	DBG("-");

	return FALSE;
}

static int __bt_hal_parse_event(GVariant *msg)
{
	GVariantIter iter;
	GVariant *child;
	char *interface_name = NULL;
	GVariant *inner_iter = NULL;

	g_variant_iter_init(&iter, msg);

	while ((child = g_variant_iter_next_value(&iter))) {
		g_variant_get(child, "{&s@a{sv}}", &interface_name, &inner_iter);
		if (g_strcmp0(interface_name,
					BT_HAL_DEVICE_INTERFACE) == 0) {
			DBG("__bt_hal_parse_event: Interface: BT_HAL_DEVICE_INTERFACE");
			g_variant_unref(inner_iter);
			g_variant_unref(child);
			return BT_HAL_DEVICE_EVENT;
		} else if (g_strcmp0(interface_name,
					BT_HAL_MEDIATRANSPORT_INTERFACE) == 0) {
			DBG("__bt_hal_parse_event: Interface: BT_HAL_MEDIATRANSPORT_INTERFACE");
			g_variant_unref(inner_iter);
			g_variant_unref(child);
			return BT_HAL_MEDIA_TRANSFER_EVENT;
		} else if (g_strcmp0(interface_name,
					BT_HAL_PLAYER_CONTROL_INTERFACE) == 0) {
			DBG("__bt_hal_parse_event: Interface: BT_HAL_PLAYER_CONTROL_INTERFACE");
			g_variant_unref(inner_iter);
			g_variant_unref(child);
			return BT_HAL_AVRCP_CONTROL_EVENT;
		}
		g_variant_unref(inner_iter);
		g_variant_unref(child);
	}

	return 0;
}

static int __bt_hal_get_owner_info(GVariant *msg, char **name, char **previous, char **current)
{
	g_variant_get(msg, "(sss)", name, previous, current);
	return BT_HAL_ERROR_NONE;
}

int __bt_insert_hal_properties(void *buf, uint8_t type, uint16_t len, const void *val)
{	struct hal_property *prop = buf;

	prop->type = type;
	prop->len = len;

	if (len)
		memcpy(prop->val, val, len);

	return sizeof(*prop) + len;
}

handle_stack_msg _bt_hal_get_stack_message_handler(void)
{
	return event_cb;
}

static void __bt_hal_adapter_property_changed_event(GVariant *msg)
{
	GVariantIter value_iter;
	GVariant *value = NULL;
	GDBusProxy *adapter_proxy;
	GError *err = NULL;
	char *key = NULL;
	g_variant_iter_init(&value_iter, msg);

	/* Buffer and propety count management */
	uint8_t buf[BT_HAL_MAX_PROPERTY_BUF_SIZE];
	struct hal_ev_adapter_props_changed *ev = (void*) buf;
	size_t size = 0;
	const gchar *address = NULL;
	gchar *name = NULL;
	unsigned int cod = 0;
	gboolean discoverable;
	gboolean connectable;
	unsigned int scan_mode = BT_SCAN_MODE_NONE;
	unsigned int disc_timeout;
	const gchar *version;
	gboolean ipsp_initialized;
	gboolean powered;
	gboolean pairable;
	unsigned int pairable_timeout;
	gboolean scan_mode_property_update = FALSE;
	gboolean is_discovering;
	gboolean is_le_discovering;

	memset(buf, 0, sizeof(buf));
	size = sizeof(*ev);
	ev->num_props = 0;
	ev->status = BT_STATUS_SUCCESS;

	DBG("+");

	while (g_variant_iter_loop(&value_iter, "{sv}", &key, &value)) {
		if (!g_strcmp0(key, "Address")) {
			uint8_t bdaddr[6];

			address = g_variant_get_string(value, NULL);
			DBG("##Address [%s]", address);
			_bt_hal_convert_addr_string_to_type(bdaddr, address);
			size += __bt_insert_hal_properties(buf + size,
					HAL_PROP_ADAPTER_ADDR, sizeof(bdaddr), bdaddr);
			ev->num_props++;
		} else if (!g_strcmp0(key, "Alias")) {
			g_variant_get(value, "&s", &name);
			DBG("##Alias [%s] ", name);
			size += __bt_insert_hal_properties(buf + size,
					HAL_PROP_ADAPTER_NAME, strlen(name) + 1, name);
			ev->num_props++;
		} else if (!g_strcmp0(key, "Class")) {
			cod = g_variant_get_uint32(value);
			DBG("##Class [%d]", cod);
			size += __bt_insert_hal_properties(buf + size,
					HAL_PROP_ADAPTER_CLASS, sizeof(unsigned int), &cod);
			ev->num_props++;
		} else if (!g_strcmp0(key, "Discoverable")) {
			discoverable = g_variant_get_boolean(value);
			DBG("##Discoverable [%d]", discoverable);
			if (discoverable)
				scan_mode = BT_SCAN_MODE_CONNECTABLE_DISCOVERABLE;
			else
				scan_mode = BT_SCAN_MODE_CONNECTABLE;
			scan_mode_property_update = TRUE;
		} else if (!g_strcmp0(key, "DiscoverableTimeout")) {
			disc_timeout = g_variant_get_uint32(value);
			DBG("##Discoverable Timeout [%d]", disc_timeout);
			size += __bt_insert_hal_properties(buf + size,
					HAL_PROP_ADAPTER_DISC_TIMEOUT, sizeof(unsigned int), &disc_timeout);
			ev->num_props++;
		} else if (!g_strcmp0(key, "Connectable")) {
			connectable = g_variant_get_boolean(value);
			DBG("##Connectable [%d]", connectable);
			if (!connectable)
				scan_mode = BT_SCAN_MODE_NONE;
			else if (scan_mode == BT_SCAN_MODE_NONE)
				scan_mode = BT_SCAN_MODE_CONNECTABLE;
			scan_mode_property_update = TRUE;
		} else if (!g_strcmp0(key, "Version")) {
			version = g_variant_get_string(value, NULL);
			DBG("##Version [%s]", version);
			size += __bt_insert_hal_properties(buf + size,
					HAL_PROP_ADAPTER_VERSION, strlen(version) + 1, version);
			ev->num_props++;
		} else if (!g_strcmp0(key, "Name")) {
			g_variant_get(value, "&s", &name);
			DBG("##Name [%s]", name);
			size += __bt_insert_hal_properties(buf + size,
					HAL_PROP_ADAPTER_NAME, strlen(name) + 1, name);
			ev->num_props++;
		} else if (!g_strcmp0(key, "Powered")) {
			powered = g_variant_get_boolean(value);
			DBG("##Powered = %d", powered);
			/* TODO: Need to check this operation!! */
			if (powered == FALSE) {
				DBG("###### Adapter Powered Down ######");
				struct hal_ev_adapter_state_changed ev;
				ev.state = HAL_POWER_OFF;
				event_cb(HAL_EV_ADAPTER_STATE_CHANGED, &ev, sizeof(ev));
#ifdef TIZEN_BT_HAL
				struct hal_ev_le_state_changed le_ev;
				le_ev.state = HAL_POWER_OFF;
				event_cb(HAL_EV_LE_STATE_CHANGED, &le_ev, sizeof(le_ev));
#endif
				/* Destroy Agent */
				_bt_hal_destroy_adapter_agent();
			} else {
				DBG("###### Adapter Powered Up ######");
				if (_bt_hal_get_adapter_request_state()) {
					DBG("Sending STATE CHANGE EVENT for Adapter... ");
					_bt_hal_set_adapter_request_state(FALSE);
					struct hal_ev_adapter_state_changed ev;
					ev.state = HAL_POWER_ON;
					event_cb(HAL_EV_ADAPTER_STATE_CHANGED, &ev, sizeof(ev));
				}
#ifdef TIZEN_BT_HAL
				if (_bt_hal_get_le_request_state()) {
					DBG("Sending STATE CHANGE EVENT for LE... ");
					_bt_hal_set_le_request_state(FALSE);
					struct hal_ev_le_state_changed ev;
					ev.state = HAL_POWER_ON;
					event_cb(HAL_EV_LE_STATE_CHANGED, &ev, sizeof(ev));
				}
#endif
				/* Create Agent */
				_bt_hal_initialize_adapter_agent();
			}

		} else if (!g_strcmp0(key, "Pairable")) {
			pairable = g_variant_get_boolean(value);
			DBG("##Pairable [%d]", pairable);
		} else if (!g_strcmp0(key, "PairableTimeout")) {
			pairable_timeout = g_variant_get_uint32(value);
			DBG("##Pairable Timeout = %d", pairable_timeout);
		} else if (!g_strcmp0(key, "UUIDs")) {
			char **uuid_value;
			int uuid_count = 0;
			gsize size1 = 0;
			int i = 0;
			size1 = g_variant_get_size(value);
			int num_props_tmp = ev->num_props;
			if (size1 > 0) {
				uuid_value = (char **)g_variant_get_strv(value, &size1);
				for (i = 0; uuid_value[i] != NULL; i++)
					uuid_count++;
				/* UUID collection */
				uint8_t uuids[BT_HAL_STACK_UUID_SIZE * uuid_count];
				for (i = 0; uuid_value[i] != NULL; i++) {
					char *uuid_str = NULL;
					uint8_t uuid[BT_HAL_STACK_UUID_SIZE];
					uuid_str = g_strdup(uuid_value[i]);
					DBG("##UUID string [%s]\n", uuid_str);
					_bt_hal_convert_uuid_string_to_type(uuid, uuid_str);
					memcpy(uuids+i*BT_HAL_STACK_UUID_SIZE, uuid, BT_HAL_STACK_UUID_SIZE);
					g_free(uuid_str);
				}
				size += __bt_insert_hal_properties(buf + size, HAL_PROP_ADAPTER_UUIDS,
						(BT_HAL_STACK_UUID_SIZE * uuid_count),
						uuids);
				ev->num_props = num_props_tmp + 1;
				g_free(uuid_value);
			}
		} else if (!g_strcmp0(key, "Discovering")) {
			is_discovering = g_variant_get_boolean(value);
			DBG("##Discovering = [%d]", is_discovering);

			if (is_discovering == FALSE) {
				DBG("###### Adapter Has stopped Discovering ######");
				/* In Tizen Bluez, this actually does not mean Discovery is stopped
				   in Bluez. Tizen Bluez sends this event after a certain timeout,
				   Therefore, we must forecefully call StopDiscovery to stop discovery in BlueZ */
				if (event_id > 0)
					continue;

				adapter_proxy = _bt_hal_get_adapter_proxy();

				if (adapter_proxy == NULL)
					continue;

				/* Need to stop searching */
				DBG("Event though Bluez reported DIscovering stopped, we force stop Discovery ");
				g_dbus_proxy_call_sync(adapter_proxy, "StopDiscovery",
						NULL,
						G_DBUS_CALL_FLAGS_NONE,
						DBUS_TIMEOUT, NULL,
						&err);
				if (err) {
					ERR("Dbus Error : %s", err->message);

					/* This error is thrown by Bluez, as Discovery is already stopped.
					   Discovery is stopped if user cancels on going discovery.
					   In order to maintain correct state of Bluetooth Discovery state,
					   simply send Discovery stopped event to HAL user */
					struct hal_ev_discovery_state_changed ev;
					ev.state = HAL_DISCOVERY_STATE_STOPPED;
					event_cb(HAL_EV_DISCOVERY_STATE_CHANGED, &ev, sizeof(ev));
					g_clear_error(&err);
					continue;

				} else {
					event_id = g_timeout_add(BT_HAL_DISCOVERY_FINISHED_DELAY,
							(GSourceFunc)__bt_hal_discovery_finished_cb, NULL);
				}

			} else {
				DBG("###### Adapter Has started Discovering ######");
				struct hal_ev_discovery_state_changed ev;
				ev.state = HAL_DISCOVERY_STATE_STARTED;
				event_cb(HAL_EV_DISCOVERY_STATE_CHANGED, &ev, sizeof(ev));
			}

		} else if (!g_strcmp0(key, "LEDiscovering")) {
			GVariant *result;

			is_le_discovering = g_variant_get_boolean(value);
			DBG("##LE Discovering = [%d]", is_le_discovering);

			if (is_le_discovering)
				continue;

			adapter_proxy = _bt_hal_get_adapter_proxy();
			if (adapter_proxy == NULL) {
				ERR("adapter_proxy == NULL");
				continue;
			}

			/* Need to stop searching */
			result = g_dbus_proxy_call_sync(adapter_proxy, "StopLEDiscovery",
					NULL, G_DBUS_CALL_FLAGS_NONE,
					DBUS_TIMEOUT, NULL, &err);
			if (!result) {
				ERR("Error occured in Proxy call");
				if (err) {
					ERR("(Error: %s)", err->message);
					g_clear_error(&err);
				}
				continue;
			}
			g_variant_unref(result);
		} else if (!g_strcmp0(key, "Modalias")) {
			char *modalias = NULL;
			g_variant_get(value, "s", &modalias);
			DBG("##Adapter ModAlias [%s]", modalias);
		} else if (!g_strcmp0(key, "SupportedLEFeatures")) {
			DBG("##LE Supported features");
			char *name = NULL;
			char *val = NULL;
			GVariantIter *iter = NULL;
			g_variant_get(value, "as", &iter);
			bt_local_le_features_t le_features;
			gboolean le_features_present = FALSE;

			if (iter == NULL)
				continue;

			memset(&le_features, 0x00, sizeof(le_features));

			while (g_variant_iter_next(iter, "&s", &name) &&
					g_variant_iter_next(iter, "&s", &val)) {
				DBG("name = %s, Value = %s", name, val);
				if (FALSE == _bt_hal_update_le_feature_support(name, val, &le_features))
					ERR("Failed to update LE feature (name = %s, value = %s)", name, val);
				else
					le_features_present = TRUE;
			}

			g_variant_iter_free(iter);
			if (le_features_present) {
				size += __bt_insert_hal_properties(buf + size,
						HAL_PROP_ADAPTER_LOCAL_LE_FEAT, sizeof(le_features), &le_features);
				ev->num_props++;
			} else {
				DBG("le supported features values are NOT provided by Stack");
			}
		} else if (!g_strcmp0(key, "IpspInitStateChanged")) {
			g_variant_get(value, "b" , &ipsp_initialized);
			DBG("##IPSP Initialized = %d", ipsp_initialized);
		} else {
			ERR("Unhandled Property:[%s]", key);
		}
	}

	if (scan_mode_property_update) {
		size += __bt_insert_hal_properties(buf + size,
				HAL_PROP_ADAPTER_SCAN_MODE, sizeof(int), &scan_mode);
		ev->num_props++;
	}


	if (size > 2) {
		DBG("Send Adapter properties changed event to HAL user, Num Prop [%d] total size [%zd]", ev->num_props, size);
		event_cb(HAL_EV_ADAPTER_PROPS_CHANGED, buf, size);
	}

	DBG("-");
}

void _bt_hal_handle_adapter_event(GVariant *msg, const char *member)
{
	DBG("+");

	if (member == NULL)
		return;

	if (strcasecmp(member, "DeviceCreated") == 0) {
		DBG("DeviceCreated: Unhandled");
	} else if (strcasecmp(member, "InterfacesRemoved") == 0) {
		DBG("InterfacesRemoved: Unhandled");
	} else if (strcasecmp(member, "AdvertisingEnabled") == 0) {
		DBG("AdvertisingEnabled");
		DBG("Advertising Enabled");
		int slot_id;
		gboolean status = FALSE;
		g_variant_get(msg, "(ib)", &slot_id, &status);
		DBG("Advertising Enabled : slot_id [%d]  status [%d]", slot_id, status);
		/* Send event to application */
		_bt_hal_set_advertising_status(slot_id, status);
	} else if (strcasecmp(member, "RssiEnabled") == 0) {
		struct hal_ev_rssi_monitor_state_changed ev;
		gboolean status = FALSE;
		char *address = NULL;
		int link_type;

		g_variant_get(msg, "(sib)", &address, &link_type, &status);
		DBG("RSSI monitoring %s for %s",
				(status ? "Enabled" : "Disabled"), address);

		_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
		ev.link_type = link_type;
		ev.state = (status ? HAL_RSSI_MONITORING_ENABLED : HAL_RSSI_MONITORING_DISABLED);
		if (!event_cb)
			ERR("event_cb is NULL");
		else
			event_cb(HAL_EV_RSSI_MONITOR_STATE_CHANGED, &ev, sizeof(ev));

		g_free(address);
	} else if (strcasecmp(member, "RssiAlert") == 0) {
		struct hal_ev_rssi_alert_recieved ev;
		int alert_type;
		int rssi_dbm;
		int link_type;
		char *address = NULL;

		g_variant_get(msg, "(siii)", &address, &link_type, &alert_type, &rssi_dbm);
		DBG("RSSI Alert: [Address %s LinkType %d] [Type %d DBM %d]",
				address, alert_type, rssi_dbm, link_type);

		_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
		ev.link_type = link_type;
		ev.alert_type = alert_type;
		ev.rssi = rssi_dbm;

		if (!event_cb)
			ERR("event_cb is NULL");
		else
			event_cb(HAL_EV_RSSI_ALERT_RECIEVED, &ev, sizeof(ev));

		g_free(address);
	} else if (strcasecmp(member, "RawRssi") == 0) {
		struct hal_ev_raw_rssi_recieved ev;
		int rssi_dbm;
		int link_type;
		char *address = NULL;

		g_variant_get(msg, "(sii)", &address, &link_type, &rssi_dbm);
		DBG("Raw RSSI: [Address %s] [Link Type %d][RSSI DBM %d]",
				address, link_type, rssi_dbm);

		_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
		ev.link_type = link_type;
		ev.rssi = rssi_dbm;

		if (!event_cb)
			ERR("event_cb is NULL");
		else
			event_cb(HAL_EV_RAW_RSSI_RECIEVED, &ev, sizeof(ev));

		g_free(address);
	} else if (strcasecmp(member, BT_HAL_HARDWARE_ERROR) == 0) {
		DBG("BT Hardware Error: Unhandled");
	} else if (strcasecmp(member, BT_HAL_TX_TIMEOUT_ERROR) == 0) {
		DBG("BT TX Timeout Error: Unhandled");

	}
	DBG("-");
}

static gboolean __bt_hal_parse_device_properties(GVariant *item)
{
	GVariantIter iter;
	gchar *key;
	GVariant *val;
	gsize len = 0;
	if (!item)
		return FALSE;
	DBG("+");

	/* Buffer and propety count management */
	uint8_t buf[BT_HAL_MAX_PROPERTY_BUF_SIZE];
	struct hal_ev_device_found *ev = (void *) buf;
	size_t size = 0;
	memset(buf, 0, sizeof(buf));
	size = sizeof(*ev);
	ev->num_props = 0;

	g_variant_iter_init(&iter, item);
	while (g_variant_iter_loop(&iter, "{sv}", &key, &val)) {

		if (strcasecmp(key, "Address") == 0)  {

			char * address = NULL;
			address = g_variant_dup_string(val, &len);
			uint8_t bdaddr[6];
			_bt_hal_convert_addr_string_to_type(bdaddr, address);

			size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_ADDR,
					sizeof(bdaddr), bdaddr);

			ev->num_props++;
			DBG("Device address [%s] property Num [%d]", address, ev->num_props);
			g_free(address);
		} else if (strcasecmp(key, "Class") == 0) {
			unsigned int class = g_variant_get_uint32(val);
			size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_CLASS,
					sizeof(unsigned int), &class);
			ev->num_props++;
			DBG("Device class [%d] Property num [%d]", class, ev->num_props);
		} else if (strcasecmp(key, "name") == 0) {
			char *name = g_variant_dup_string(val, &len);
			if (name) {
				size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_NAME,
						strlen(name) + 1, name);
				ev->num_props++;
				DBG("Device Name [%s] Property num [%d]", name, ev->num_props);
			}
			g_free(name);
		} else if (strcasecmp(key, "Connected") == 0) {
			unsigned int connected = g_variant_get_byte(val);

			size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_CONNECTED,
					sizeof(unsigned int), &connected);
			ev->num_props++;
			DBG("Device connected [%u] Property num [%d]", connected,  ev->num_props);
		} else if (strcasecmp(key, "paired") == 0) {
			uint8_t paired = (g_variant_get_boolean(val) ? 1 : 0);
			size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_PAIRED,
					sizeof(uint8_t), &paired);
			ev->num_props++;
			DBG("Device Paired [%d] Property num [%d]", paired, ev->num_props);
		} else if (strcasecmp(key, "Trusted") == 0) {
			uint8_t trust = (g_variant_get_boolean(val) ? 1 : 0);
			size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_TRUSTED,
					sizeof(uint8_t), &trust);
			ev->num_props++;
			DBG("Device trusted [%d] Property num [%d]", trust, ev->num_props);
		} else if (strcasecmp(key, "RSSI") == 0) {
			int rssi = g_variant_get_int16(val);
			size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_RSSI,
					sizeof(int), &rssi);
			ev->num_props++;
			DBG("Device RSSI [%d] Property num [%d]", rssi, ev->num_props);
		} else if (strcasecmp(key, "LastAddrType") == 0) {
			/* TODO: To be handled later*/
		} else if (!g_strcmp0(key, "IsAliasSet")) {
			uint8_t is_alias_set = (g_variant_get_boolean(val) ? 1 : 0);
			DBG("IsAliasSet: %s", (is_alias_set ? "TRUE" : "FALSE"));
			size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_IS_ALIAS_SET,
					sizeof(uint8_t), &is_alias_set);
			ev->num_props++;
		} else if (strcasecmp(key, "UUIDs") == 0) {
			char **uuid_value;
			int uuid_count = 0;
			gsize size1 = 0;
			int i = 0;
			size1 = g_variant_get_size(val);
			DBG("UUID count from size  [%zu]\n", size1);
			int num_props_tmp = ev->num_props;

			if (size1 > 0) {
				uuid_value = (char **)g_variant_get_strv(val, &size1);
				for (i = 0; uuid_value[i] != NULL; i++)
					uuid_count++;
				DBG("UUID count [%d]\n", uuid_count);
				/* UUID collection */
				uint8_t uuids[BT_HAL_STACK_UUID_SIZE * uuid_count];

				for (i = 0; uuid_value[i] != NULL; i++) {

					char *uuid_str = NULL;
					uint8_t uuid[BT_HAL_STACK_UUID_SIZE];
					memset(uuid, 0x00, BT_HAL_STACK_UUID_SIZE);

					DBG("UUID string from Bluez [%s]\n", uuid_value[i]);
					uuid_str = g_strdup(uuid_value[i]);
					DBG("UUID string [%s]\n", uuid_str);
					_bt_hal_convert_uuid_string_to_type(uuid, uuid_str);
					memcpy(uuids+i*BT_HAL_STACK_UUID_SIZE, uuid, BT_HAL_STACK_UUID_SIZE);
					g_free(uuid_str);
				}

				size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_UUIDS,
						(BT_HAL_STACK_UUID_SIZE * uuid_count),
						uuids);
				ev->num_props = num_props_tmp + 1;
				g_free(uuid_value);
			}

		} else if (strcasecmp(key, "LegacyManufacturerDataLen") == 0) {
			/* TODO: To be handled later*/
		} else if (strcasecmp(key, "LegacyManufacturerData") == 0) {
			/* TODO: To be handled later*/
		} else {
			ERR("Unhandled Property:[%s]", key);
		}
	}
	DBG("-");

	if (size > 1) {
		DBG("Send Device found event to HAL user, Num Prop [%d] total size [%zd]", ev->num_props, size);
		event_cb(HAL_EV_DEVICE_FOUND, (void*) buf, size);
	}

	return TRUE;
}

static void __bt_hal_handle_avrcp_tg_events(GVariant *msg, const char *path)
{
	int state;
	GVariantIter value_iter;
	char *property = NULL;
	char *address;
	GVariant *val = NULL;
	GVariant *child = NULL;

	g_variant_iter_init(&value_iter, msg);
	while ((child = g_variant_iter_next_value(&value_iter))) {
		g_variant_get(child, "{sv}", &property, &val);
		INFO("Property %s", property);
		if (strcasecmp(property, "Connected") == 0) {
			struct hal_ev_avrcp_tg_conn_state ev;

			gboolean connected = FALSE;

			g_variant_get(val, "b", &connected);

			state = connected ? HAL_AVRCP_TG_STATE_CONNECTED :
				HAL_AVRCP_TG_STATE_DISCONNECTED;

			address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE);

			_bt_hal_convert_device_path_to_address(path, address);

			DBG("connected: %d", connected);
			DBG("address: %s", address);

			/* Prepare to send AVRCP Target connection state event */
			memset(&ev, 0, sizeof(ev));
			_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
			ev.state = state;
			if (!avrcp_tg_event_cb)
				ERR("AVRCP target DBUS handler callback not registered");
			else
				avrcp_tg_event_cb(HAL_EV_AVRCP_TG_CONN_STATE, (void *)&ev, sizeof(ev));
			g_free(address);
		}
		g_free(property);
		g_variant_unref(child);
		g_variant_unref(val);
	}

	DBG("-");
}

static void __bt_hal_handle_property_changed_event(GVariant *msg, const char *object_path)
{
	char *interface_name = NULL;
	GVariant *val = NULL;
	DBG("+");

	g_variant_get(msg, "(&s@a{sv}@as)", &interface_name, &val, NULL);

	if (!interface_name) {
		DBG("Failed to get interface name");

		if (val)
			g_variant_unref(val);
		return;
	}

	if (strcasecmp(interface_name, BT_HAL_ADAPTER_INTERFACE) == 0) {
		DBG("Event: Property Changed: Interface: BT_HAL_ADAPTER_INTERFACE");
		__bt_hal_adapter_property_changed_event(val);
	} else if (strcasecmp(interface_name, BT_HAL_DEVICE_INTERFACE) == 0) {
		DBG("Event: Property Changed: Interface: BT_HAL_DEVICE_INTERFACE");
		__bt_hal_device_property_changed_event(val, object_path);
	} else if (strcasecmp(interface_name, BT_HAL_OBEX_TRANSFER_INTERFACE) == 0) {
		DBG("Event: Property Changed: Interface: BT_HAL_OBEX_TRANSFER_INTERFACE");
		/* TODO: Handle event */
	} else if (strcasecmp(interface_name, BT_HAL_MEDIA_CONTROL_INTERFACE) == 0) {
		DBG("Event: Property Changed: Interface: BT_HAL_MEDIA_CONTROL_INTERFACE");
		/* Handle AVRCP target event */
		__bt_hal_handle_avrcp_tg_events(val, object_path);
	} else if (strcasecmp(interface_name, BT_HAL_PLAYER_CONTROL_INTERFACE) == 0) {
		DBG("Event: Property Changed: Interface: BT_HAL_PLAYER_CONTROL_INTERFACE");
		__bt_hal_handle_avrcp_ctrl_events(val, NULL, object_path);
	} else if (strcasecmp(interface_name, BT_HAL_MEDIATRANSPORT_INTERFACE) == 0) {
		DBG("Event: Property Changed: Interface: BT_HAL_MEDIATRANSPORT_INTERFACE");
		__bt_hal_handle_avrcp_transport_events(val, NULL, object_path);
	} else if (strcasecmp(interface_name, BT_HAL_NETWORK_CLIENT_INTERFACE) == 0) {
		DBG("Event: Property Changed: Interface: BT_HAL_NETWORK_CLIENT_INTERFACE");
		/* TODO: Handle event */
	} else if (strcasecmp(interface_name, BT_HAL_GATT_CHAR_INTERFACE) == 0) {
		DBG("Event: Property Changed: Interface: BT_HAL_GATT_CHAR_INTERFACE");
		/* TODO: Handle event */
	} else if (strcasecmp(interface_name, BT_HAL_INPUT_INTERFACE) == 0) {
		DBG("Event: Property Changed: Interface: BT_HAL_INPUT_INTERFACE");
		__bt_hal_handle_input_event(val, object_path);
	}
	g_variant_unref(val);
}

static void __bt_hal_handle_device_event(GVariant *value, GVariant *parameters)
{
	DBG("+");

	if (__bt_hal_parse_interface(parameters) == FALSE) {
		ERR("Fail to parse the properies");
		g_variant_unref(value);
		return;
	}

	DBG("-");
}

static void __bt_hal_send_hid_connection_state_event(
		gboolean connected, char *address)
{
	struct hal_ev_hidhost_conn_state ev;

	ev.state = (connected == TRUE) ?
		HAL_HIDHOST_STATE_CONNECTED :
		HAL_HIDHOST_STATE_DISCONNECTED;

	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);

	if (!hid_event_cb)
		ERR("HID event handler not registered");
	else
		hid_event_cb(HAL_EV_HIDHOST_CONN_STATE, &ev, sizeof(ev));
}

static void __bt_hal_handle_input_event(GVariant *msg, const char *path)
{
	gboolean property_flag = FALSE;
	GVariantIter value_iter;
	char *property = NULL;
	GVariant *child = NULL, *val = NULL;

	DBG("+");
	g_variant_iter_init(&value_iter, msg);
	while ((child = g_variant_iter_next_value(&value_iter))) {
		g_variant_get(child, "{sv}", &property, &val);

		if (property == NULL)
			return;

		if (strcasecmp(property, "Connected") == 0) {
			char *address;

			g_variant_get(val, "b", &property_flag);
			address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE);
			_bt_hal_convert_device_path_to_address(path, address);
			__bt_hal_send_hid_connection_state_event(property_flag, address);
			g_free(address);
		}
		g_free(property);
		g_variant_unref(val);
		g_variant_unref(child);
	}

	DBG("-");
}

static gboolean __bt_hal_parse_interface(GVariant *msg)
{
	char *path = NULL;
	GVariant *optional_param;
	GVariantIter iter;
	GVariant *child;
	char *interface_name = NULL;
	GVariant *inner_iter = NULL;
	g_variant_get(msg, "(&o@a{sa{sv}})",
			&path, &optional_param);
	g_variant_iter_init(&iter, optional_param);

	while ((child = g_variant_iter_next_value(&iter))) {
		g_variant_get(child, "{&s@a{sv}}", &interface_name, &inner_iter);
		if (g_strcmp0(interface_name, BT_HAL_DEVICE_INTERFACE) == 0) {
			DBG("Found a device: %s", path);
			if (__bt_hal_parse_device_properties(inner_iter) == FALSE) {
				g_variant_unref(inner_iter);
				g_variant_unref(child);
				g_variant_unref(optional_param);
				ERR("Fail to parse the properies");
				return FALSE;
			} else {
				g_variant_unref(inner_iter);
				g_variant_unref(child);
				g_variant_unref(optional_param);
				return TRUE;
			}
		}
		g_variant_unref(inner_iter);
		g_variant_unref(child);
	}

	g_variant_unref(optional_param);

	return FALSE;
}

void __bt_hal_handle_gatt_char_event(GVariant *parameters, const char *signal_name)
{
	DBG("+");

	if (signal_name == NULL)
		return;

	if (strcasecmp(signal_name, "GattValueChanged") == 0) {
		DBG("GattValueChanged event received");

		int result = 0;
		const char *char_handle = NULL;
		GVariant *char_value_var = NULL;
		int len = 0;
		char *char_value = NULL;

		g_variant_get(parameters, "(i&s@ay)", &result, &char_handle, &char_value_var);
		DBG("char handle: %s", char_handle);

		len = g_variant_get_size(char_value_var);
		if (len > 0)
			char_value = (char *)g_variant_get_data(char_value_var);

		_bt_hal_handle_gattc_value_changed_event(result, char_handle, char_value, len);

		g_variant_unref(char_value_var);
	}
}


static gboolean __bt_hal_event_manager(gpointer data)
{
	bt_hal_event_type_t bt_event = 0x00;
	GVariant *value;
	char *obj_path = NULL;

	bt_hal_main_event_data_t *param = (bt_hal_main_event_data_t*)data;
	if (strcasecmp(param->signal_name, "InterfacesAdded") == 0) {

		/*TODO: Handle Interfaces Added Signal from stack */
		DBG("Manager Event: Signal Name: InterfacesAdded");

		g_variant_get(param->parameters, "(&o@a{sa{sv}})", &obj_path, &value);

		if (obj_path == NULL) {
			DBG("obj_path is NULL");
			return FALSE;
		}

		if (strcasecmp(obj_path, BT_HAL_BLUEZ_HCI_PATH) == 0) {
			/* TODO: Handle adapter added */
			DBG("Manager Event: Signal Name: InterfiacesAdded: Adapter added in bluetoothd: path [hci0]");
		} else {
			bt_event = __bt_hal_parse_event(value);
			if (bt_event == BT_HAL_DEVICE_EVENT) {
				DBG("Device path : %s ", obj_path);
				__bt_hal_handle_device_event(value, param->parameters);
			} else if (bt_event == BT_HAL_AVRCP_CONTROL_EVENT) {
				DBG("Device path : %s ", obj_path);
				_bt_hal_set_control_device_path(obj_path);
			}
		}
		g_variant_unref(value);

	} else if (strcasecmp(param->signal_name, "InterfacesRemoved") == 0) {
		char *str;
		GVariantIter *iter;

		/*TODO: Handle Interfaces Removed Signal from stack */
		DBG("Manager Event: Signal Name: InterfacesRemoved");

		g_variant_get(param->parameters, "(&oas)", &obj_path, &iter);
		DBG("Device path : %s ", obj_path);
		while (g_variant_iter_loop(iter, "s", &str)) {
			if (g_strcmp0(str, BT_HAL_PLAYER_CONTROL_INTERFACE) == 0)
				_bt_hal_remove_control_device_path(obj_path);
		}
		g_variant_iter_free(iter);
	} else if (strcasecmp(param->signal_name, "NameOwnerChanged") == 0) {
		char *name = NULL;
		char *previous = NULL;
		char *current = NULL;

		/* TODO: Handle Name Owener changed Signal */
		if (__bt_hal_get_owner_info(param->parameters, &name, &previous, &current)) {
			DBG("Fail to get the owner info");
			return FALSE;
		}
		if (current && *current != '\0') {
			g_free(name);
			g_free(previous);
			g_free(current);
			return FALSE;
		}
		if (strcasecmp(name, BT_HAL_BLUEZ_NAME) == 0) {
			DBG("Bluetoothd is terminated");

			/* TODO: Handle Bluetoothd terminating scenario */
			_bt_hal_le_deinit();
		}
		INFO("Name Owner changed [%s]", name);
		g_free(name);
		g_free(previous);
		g_free(current);

	} else if (g_strcmp0(param->interface_name, BT_HAL_PROPERTIES_INTERFACE) == 0) {
		DBG("Manager Event: Interface Name: BT_HAL_PROPERTIES_INTERFACE");
		__bt_hal_handle_property_changed_event(param->parameters, param->object_path);
	} else if (g_strcmp0(param->interface_name, BT_HAL_ADAPTER_INTERFACE) == 0) {
		DBG("Manager Event: Interface Name: BT_HAL_ADAPTER_INTERFACE");
		_bt_hal_handle_adapter_event(param->parameters, param->signal_name);
	} else if (g_strcmp0(param->interface_name, BT_HAL_INPUT_INTERFACE) == 0) {
		DBG("Manager Event: Interface Name: BT_HAL_INPUT_INTERFACE");
		__bt_hal_handle_input_event(param->parameters, param->object_path);
	} else if (g_strcmp0(param->interface_name, BT_HAL_NETWORK_SERVER_INTERFACE) == 0) {
		/* TODO: Handle Network Server events from stack */
		DBG("Manager Event: Interface Name: BT_HAL_NETWORK_SERVER_INTERFACE");
	} else if (g_strcmp0(param->interface_name, BT_HAL_HEADSET_INTERFACE) == 0) {
		DBG("Manager Event: Interface Name: BT_HAL_HEADSET_INTERFACE");
		__bt_hal_handle_headset_events(param->parameters, param->signal_name, param->object_path);
	} else if (g_strcmp0(param->interface_name, BT_HAL_SINK_INTERFACE) == 0) {
		/* TODO: Handle Sink interface events from stack */
		DBG("Manager Event: Interface Name:BT_HAL_SINK_INTERFACE");
	} else if (g_strcmp0(param->interface_name, BT_HAL_AGENT_INTERFACE) == 0) {
		/* TODO: Handle Agent events from stack */
		DBG("Manager Event: Interface Name:BT_HAL_AGENT_INTERFACE");
	} else if (g_strcmp0(param->interface_name, BT_HAL_DEVICE_INTERFACE) == 0) {
		DBG("Manager Event: Interface Name:BT_HAL_DEVICE_INTERFACE");
		__bt_hal_handle_device_specific_events(param->parameters, param->signal_name, param->object_path);
	} else if (g_strcmp0(param->interface_name, BT_HAL_PLAYER_CONTROL_INTERFACE) == 0) {
		DBG("Manager Event: Interface Name: BT_HAL_PLAYER_CONTROL_INTERFACE");
		__bt_hal_handle_avrcp_ctrl_events(param->parameters, param->signal_name, param->object_path);
	} else if (g_strcmp0(param->interface_name, BT_HAL_MEDIATRANSPORT_INTERFACE) == 0) {
		DBG("Manager Event: Interface Name: BT_HAL_MEDIATRANSPORT_INTERFACE");
		__bt_hal_handle_avrcp_transport_events(param->parameters, param->signal_name, param->object_path);
	} else if (g_strcmp0(param->interface_name, BT_HAL_GATT_CHAR_INTERFACE) == 0) {
		DBG("Manager Event: Interface Name: BT_HAL_GATT_CHAR_INTERFACE");
		__bt_hal_handle_gatt_char_event(param->parameters, param->signal_name);
	}


	/* Free data */
	g_free(param->sender_name);
	g_free(param->object_path);
	g_free(param->interface_name);
	g_free(param->signal_name);
	g_variant_unref(param->parameters);

	return FALSE;
}

static  void __bt_hal_manager_event_filter(GDBusConnection *connection,
		const gchar *sender_name,
		const gchar *object_path,
		const gchar *interface_name,
		const gchar *signal_name,
		GVariant *parameters,
		gpointer user_data)
{
	if (signal_name == NULL)
		return;

	bt_hal_main_event_data_t *param = g_new0(bt_hal_main_event_data_t, 1);
	param->sender_name = g_strdup(sender_name);
	param->object_path = g_strdup(object_path);
	param->interface_name = g_strdup(interface_name);
	param->signal_name = g_strdup(signal_name);
	param->parameters = g_variant_ref(parameters);

	g_idle_add(__bt_hal_event_manager, (gpointer)param);
	return;
}

static void __bt_hal_handle_headset_events(GVariant *msg, const char *member, const char *path)
{
	gboolean property_flag = FALSE;
	char *property = NULL;
	GVariant *value = NULL;
	g_variant_get(msg, "(sv)", &property, &value);

	if (property == NULL)
		return;

	DBG("Property = %s \n", property);
	/* We allow only 1 headset connection (HSP or HFP)*/
	if (strcasecmp(property, "Connected") == 0) {
		char *address;
		g_variant_get(value, "b", &property_flag);

		address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE);

		/* Fix : NULL_RETURNS */
		if (address == NULL)
			return;

		_bt_hal_convert_device_path_to_address(path, address);
		__bt_hal_send_hf_audio_connection_state_event(property_flag, address);
		g_free(address);
	} else if (strcasecmp(property, "State") == 0) {
		char *state = NULL;

		g_variant_get(value, "s", &state);

		/* This code assumes we support only 1 headset connection */
		/* Need to use the headset list, if we support multi-headsets */
		if (strcasecmp(state, "Playing") == 0) {
			DBG("Playing: Sco Connected");
		} else if (strcasecmp(state, "connected") == 0 ||
				strcasecmp(state, "disconnected") == 0) {
			if (strcasecmp(state, "connected") == 0)
				DBG("Sco Connected");
			else
				DBG("Sco Disconnected");
		} else {
			ERR("Not handled state - %s", state);
			g_free(state);
			return;
		}
		g_free(state);
	} else if (strcasecmp(property, "SpeakerGain") == 0) {
		char *address;

		address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE);

		_bt_hal_convert_device_path_to_address(path, address);
		INFO("Speaker Gain for address [%s]", address);
		/* TODO Handle event sending to HAL */

		g_free(address);
	} else if (strcasecmp(property, "MicrophoneGain") == 0) {
		char *address;

		address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE);

		_bt_hal_convert_device_path_to_address(path, address);
		INFO("Microphone Gain for address [%s]", address);
		/* TODO Handle event sending to HAL */

		g_free(address);
	}

	if (property)
		g_free(property);
	g_variant_unref(value);
}

static int __bt_hal_register_manager_subscribe_signal(GDBusConnection *conn,
		int subscribe)
{
	if (conn == NULL)
		return -1;

	static int subs_interface_added_id = -1;
	static int subs_interface_removed_id = -1;
	static int subs_name_owner_id = -1;
	static int subs_property_id = -1;
	static int subs_adapter_id = -1;

	INFO_C("+");

	if (subscribe) {
		if (subs_interface_added_id == -1) {
			subs_interface_added_id = g_dbus_connection_signal_subscribe(conn,
					NULL, BT_HAL_MANAGER_INTERFACE,
					BT_HAL_INTERFACES_ADDED, NULL, NULL, 0,
					__bt_hal_manager_event_filter,
					NULL, NULL);
		}
		if (subs_interface_removed_id == -1) {
			subs_interface_removed_id = g_dbus_connection_signal_subscribe(conn,
					NULL, BT_HAL_MANAGER_INTERFACE,
					BT_HAL_INTERFACES_REMOVED, NULL, NULL, 0,
					__bt_hal_manager_event_filter,
					NULL, NULL);
		}
		if (subs_name_owner_id == -1) {
			subs_name_owner_id = g_dbus_connection_signal_subscribe(conn,
					NULL, BT_HAL_FREEDESKTOP_INTERFACE,
					BT_HAL_NAME_OWNER_CHANGED, NULL, NULL, 0,
					__bt_hal_manager_event_filter,
					NULL, NULL);
		}
		if (subs_property_id == -1) {
			subs_property_id = g_dbus_connection_signal_subscribe(conn,
					NULL, BT_HAL_PROPERTIES_INTERFACE,
					BT_HAL_PROPERTIES_CHANGED, NULL, NULL, 0,
					__bt_hal_manager_event_filter,
					NULL, NULL);
		}
		if (subs_adapter_id == -1) {
			subs_adapter_id = g_dbus_connection_signal_subscribe(conn,
					NULL, BT_HAL_ADAPTER_INTERFACE,
					NULL, NULL, NULL, 0,
					__bt_hal_manager_event_filter,
					NULL, NULL);
		}
	} else {
		if (subs_interface_added_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_interface_added_id);
			subs_interface_added_id = -1;
		}
		if (subs_interface_removed_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_interface_removed_id);
			subs_interface_removed_id = -1;
		}
		if (subs_name_owner_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_name_owner_id);
			subs_name_owner_id = -1;
		}
		if (subs_property_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_property_id);
			subs_property_id = -1;
		}
		if (subs_adapter_id == -1) {
			g_dbus_connection_signal_unsubscribe(conn, subs_adapter_id);
			subs_adapter_id = -1;
		}
	}

	INFO_C("-");
	return 0;
}

static int __bt_hal_register_device_subscribe_signal(GDBusConnection *conn,
		int subscribe)
{
	static int subs_device_id = -1;

	DBG("+");
	if (conn == NULL)
		return -1;

	if (subscribe) {
		if (subs_device_id == -1) {
			subs_device_id = g_dbus_connection_signal_subscribe(conn,
					NULL, BT_HAL_DEVICE_INTERFACE,
					NULL, NULL, NULL, 0,
					__bt_hal_manager_event_filter,
					NULL, NULL);
		}
	} else {
		if (subs_device_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_device_id);
			subs_device_id = -1;
		}
	}

	DBG("-");
	return 0;
}

static int __bt_hal_register_input_subscribe_signal(GDBusConnection *conn, int subscribe)
{
	static int subs_input_id = -1;

	DBG("+");

	if (conn == NULL)
		return -1;

	if (subscribe) {
		if (subs_input_id == -1) {
			subs_input_id = g_dbus_connection_signal_subscribe(conn,
					NULL, BT_HAL_INPUT_INTERFACE,
					NULL, NULL, NULL, 0,
					__bt_hal_manager_event_filter,
					NULL, NULL);
		}
	} else {
		if (subs_input_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_input_id);
			subs_input_id = -1;
		}
	}

	DBG("-");

	return 0;
}

static int __bt_hal_register_gatt_subscribe_signal(GDBusConnection *conn,
		int subscribe)
{
	static int subs_gatt_id = -1;

	DBG("+");

	if (subscribe) {
		if (subs_gatt_id == -1) {
			subs_gatt_id = g_dbus_connection_signal_subscribe(conn,
					NULL, BT_HAL_GATT_CHAR_INTERFACE,
					NULL, NULL, NULL, 0,
					__bt_hal_manager_event_filter,
					NULL, NULL);
		}
	} else {
		if (subs_gatt_id == -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_gatt_id);
			subs_gatt_id = -1;
		}
	}

	return BT_HAL_ERROR_NONE;
}



static int __bt_hal_register_service_event(GDBusConnection *g_conn, int event_type)
{
	DBG("+");

	if (g_conn == NULL)
		return  BT_HAL_ERROR_INTERNAL;

	/* TODO: Add more events in subsequent patches */
	switch (event_type) {
	case BT_HAL_MANAGER_EVENT:
		__bt_hal_register_manager_subscribe_signal(g_conn, TRUE);
		break;
	case BT_HAL_DEVICE_EVENT:
		__bt_hal_register_device_subscribe_signal(g_conn, TRUE);
		break;
	case BT_HAL_HID_EVENT:
		__bt_hal_register_input_subscribe_signal(g_conn, TRUE);
		break;
	case BT_HAL_HEADSET_EVENT:
		 __bt_hal_register_audio_subscribe_signal(g_conn, TRUE);
		break;
	case BT_HAL_GATT_EVENT:
		__bt_hal_register_gatt_subscribe_signal(g_conn, TRUE);
		break;
	default:
		INFO_C("Register Event: event_type [%d]", event_type);
		return BT_HAL_ERROR_NOT_SUPPORT;
	}

	return BT_HAL_ERROR_NONE;
}

static int __bt_hal_register_audio_subscribe_signal(GDBusConnection *conn,
		int subscribe)
{
	if (conn == NULL)
		return -1;

	static int subs_headset_id = -1;
	static int subs_sink_id = -1;

	if (subscribe) {
		if (subs_headset_id == -1) {
			subs_headset_id = g_dbus_connection_signal_subscribe(conn,
					NULL, BT_HAL_HEADSET_INTERFACE,
					NULL, NULL, NULL, 0,
					__bt_hal_manager_event_filter,
					NULL, NULL);
		}
		if (subs_sink_id == -1) {
			subs_sink_id = g_dbus_connection_signal_subscribe(conn,
					NULL, BT_HAL_SINK_INTERFACE,
					NULL, NULL, NULL, 0,
					__bt_hal_manager_event_filter,
					NULL, NULL);
		}
	} else {
		if (subs_headset_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_headset_id);
			subs_headset_id = -1;
		}
		if (subs_sink_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_sink_id);
			subs_sink_id = -1;
		}
	}
	return 0;
}

static int __bt_hal_initialize_manager_receiver(void)
{
	DBG("+");

	GError *error = NULL;

	if (manager_conn == NULL) {
		manager_conn =  g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);
		if (error != NULL) {
			ERR_C("ERROR: Can't get on system bus [%s]", error->message);
			g_clear_error(&error);
		}
		if (manager_conn == NULL)
			goto fail;
	}

	if (__bt_hal_register_service_event(manager_conn,
				BT_HAL_MANAGER_EVENT) != BT_HAL_ERROR_NONE)
		goto fail;
	if (__bt_hal_register_service_event(manager_conn,
				BT_HAL_DEVICE_EVENT) != BT_HAL_ERROR_NONE)
		goto fail;
	if (__bt_hal_register_service_event(manager_conn,
				BT_HAL_HID_EVENT) != BT_HAL_ERROR_NONE)
		goto fail;
	if (__bt_hal_register_service_event(manager_conn,
				BT_HAL_HEADSET_EVENT) != BT_HAL_ERROR_NONE)
		goto fail;
	if (__bt_hal_register_service_event(manager_conn,
				BT_HAL_GATT_EVENT) != BT_HAL_ERROR_NONE)
		goto fail;
	return BT_HAL_ERROR_NONE;
fail:
	if (manager_conn) {
		g_object_unref(manager_conn);
		manager_conn = NULL;
	}

	DBG("-");

	return BT_HAL_ERROR_INTERNAL;
}

/* To receive the event from bluez */
int _bt_hal_initialize_event_receiver(handle_stack_msg cb)
{
	int result;
	DBG("+");

	if (!cb)
		return BT_HAL_ERROR_INVALID_PARAM;

	result = __bt_hal_initialize_manager_receiver();

	DBG("Manager event receiver initialization result [%d]", result);
	if (result != BT_HAL_ERROR_NONE)
		return result;

	/*TODO: Initialize Obexd Event receiver */

	event_cb = cb;
	DBG("-");

	return BT_HAL_ERROR_NONE;
}

static void __bt_hal_device_property_changed_event(GVariant *msg, const char *path)
{
	GVariantIter value_iter;
	GVariant *value = NULL;
	char *key = NULL;
	g_variant_iter_init(&value_iter, msg);
	DBG("+");

	while (g_variant_iter_loop(&value_iter, "{sv}", &key, &value)) {
		if (!g_strcmp0(key, "Connected")) {
			guint connected = 0;
			g_variant_get(value, "i", &connected);
			DBG("Device property changed : Connected [%d]", connected);
		} else if (!g_strcmp0(key, "RSSI")) {
			DBG("Device property changed : RSSI");
			__bt_hal_dbus_device_found_properties(path);
		} else if (!g_strcmp0(key, "GattConnected")) {
			DBG("Device property changed : GattConnected");
			gboolean gatt_connected = FALSE;
			g_variant_get(value, "b", &gatt_connected);
			char *address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE);
			_bt_hal_convert_device_path_to_address(path, address);
			DBG("@@gatt_connected: %d", gatt_connected);
			DBG("@@address: %s", address);
			_bt_hal_gatt_connected_state_event(gatt_connected, address);
			_bt_hal_handle_gattc_connected_event(address, gatt_connected);
			g_free(address);
		} else if (!g_strcmp0(key, "Paired")) {
			gboolean paired = FALSE;
			struct hal_ev_bond_state_changed ev;
			char address[BT_HAL_ADDRESS_STRING_SIZE];

			g_variant_get(value, "b", &paired);
			DBG("Device property changed : Paired = %s", (paired ? "TRUE" : "FALSE"));

			_bt_hal_agent_set_canceled(FALSE);
			_bt_hal_convert_device_path_to_address(path, address);

			/* Prepare to send event to HAL bluetooth */
			ev.status = BT_STATUS_SUCCESS;
			ev.state = paired ? BT_BOND_STATE_BONDED : BT_BOND_STATE_NONE;
			_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);

			if (!event_cb) {
				ERR("Bluetooth HAL event handler not registered");
			} else {
				DBG("Sending HAL_EV_BOND_STATE_CHANGED event");
				event_cb(HAL_EV_BOND_STATE_CHANGED, (void*)&ev, sizeof(ev));
			}
		} else if (!g_strcmp0(key, "LegacyPaired")) {
			DBG("Device property changed : LegacyPaired");
		} else if (!g_strcmp0(key, "Trusted")) {
			DBG("Device property changed : Trusted");
			gboolean trusted = FALSE;
			gchar *address = NULL;
			g_variant_get(value, "b", &trusted);
			address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE);

			_bt_hal_convert_device_path_to_address(path, address);
			DBG("Device [%s] trusted: [%d]", address, trusted);

			__bt_hal_send_device_trust_state_event(trusted, address);
			g_free(address);
		} else if (!g_strcmp0(key, "IpspConnected")) {
			DBG("Device property changed : IpspConnected");
		} else if (!g_strcmp0(key, "IpspInitStateChanged")) {
			DBG("Device property changed : IpspInitStateChanged");
		} else if (!g_strcmp0(key, "TrustedProfiles")) {
			int trust_val = 0;
			char address[BT_HAL_ADDRESS_STRING_SIZE];

			g_variant_get(value, "u", &trust_val);
			_bt_hal_convert_device_path_to_address(path, address);
			DBG("Address: %s, TrustedProfiles: 0x%X", address, trust_val);
			__bt_hal_send_device_trusted_profile_changed_event(trust_val, address);
		} else {
			ERR("Unhandled Property:[%s]", key);
		}
	}
	DBG("-");
}

static void __bt_hal_dbus_device_found_properties(const char *device_path)
{
	char *address;
	GError *error = NULL;
	GDBusProxy *device_proxy;
	GDBusConnection *conn;
	GVariant *result;
	DBG("+");

	if (!device_path) {
		ERR("Invalid device path");
		return;
	}

	conn = _bt_hal_get_system_gconn();
	if (!conn) {
		ERR("_bt_hal_get_system_gconn failed");
		return;
	}

	device_proxy = g_dbus_proxy_new_sync(conn, G_DBUS_PROXY_FLAGS_NONE,
			NULL,
			BT_HAL_BLUEZ_NAME,
			device_path,
			BT_HAL_PROPERTIES_INTERFACE,
			NULL, NULL);

	if (!device_proxy) {
		ERR("Error creating device_proxy");
		return;
	}

	result = g_dbus_proxy_call_sync(device_proxy,
			"GetAll",
			g_variant_new("(s)", BT_HAL_DEVICE_INTERFACE),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);
	if (!result) {
		ERR("Error occured in Proxy call");
		if (error != NULL) {
			ERR("Error occured in Proxy call (Error: %s)", error->message);
			g_clear_error(&error);
		}
		g_object_unref(device_proxy);
		return;
	}

	address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE);
	_bt_hal_convert_device_path_to_address(device_path, address);

	__bt_hal_device_properties_lookup(result, address);

	g_object_unref(device_proxy);
	g_free(address);

	DBG("-");
}

static void __bt_hal_device_properties_lookup(GVariant *result, char *address)
{
	/* Buffer and propety count management */
	uint8_t buf[BT_HAL_MAX_PROPERTY_BUF_SIZE];
	struct hal_ev_device_found *ev = (void *) buf;
	size_t size = 0;
	memset(buf, 0, sizeof(buf));
	size = sizeof(*ev);
	ev->num_props = 0;

	GVariant *tmp_value;
	GVariant *value;
	gchar *name;
	gchar *manufacturer_data = NULL;
	DBG("+");

	if (result != NULL) {
		g_variant_get(result , "(@a{sv})", &value);
		g_variant_unref(result);

		/* Alias */
		tmp_value = g_variant_lookup_value(value, "Alias", G_VARIANT_TYPE_STRING);

		g_variant_get(tmp_value, "s", &name);

		g_variant_unref(tmp_value);
		if (name != NULL) {
			DBG_SECURE("Alias Name [%s]", name);
			size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_NAME,
					strlen(name) + 1, name);
			ev->num_props++;
			DBG("Device Name [%s] Property num [%d]", name, ev->num_props);
		} else {
			/* Name */
			tmp_value = g_variant_lookup_value(value, "Name", G_VARIANT_TYPE_STRING);
			g_variant_get(tmp_value, "s", &name);
			g_variant_unref(tmp_value);

			size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_NAME,
					strlen(name) + 1, name);
			ev->num_props++;
			DBG("Device Name [%s] Property num [%d]", name, ev->num_props);
			g_variant_get(tmp_value, "s", &name);
		}

		/* Class */
		tmp_value = g_variant_lookup_value(value, "Class", G_VARIANT_TYPE_UINT32);
		unsigned int class = tmp_value ? g_variant_get_uint32(tmp_value) : 0;
		size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_CLASS,
				sizeof(unsigned int), &class);
		ev->num_props++;
		if (tmp_value)
			g_variant_unref(tmp_value);


		/* Connected */
		tmp_value = g_variant_lookup_value(value, "Connected",  G_VARIANT_TYPE_BOOLEAN);
		unsigned int connected = tmp_value ? g_variant_get_boolean(tmp_value) : 0;
		size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_CONNECTED,
				sizeof(unsigned int), &connected);
		ev->num_props++;
		DBG("Device connected [%u] Property num [%d]", connected,  ev->num_props);
		if (tmp_value)
			g_variant_unref(tmp_value);

		/* Trust */
		tmp_value = g_variant_lookup_value(value, "Trusted",  G_VARIANT_TYPE_BOOLEAN);
		uint8_t trust = tmp_value ? (g_variant_get_boolean(tmp_value) ? 1 : 0) : 0;
		size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_TRUSTED,
				sizeof(uint8_t), &trust);
		ev->num_props++;
		DBG("Device trusted [%d] Property num [%d]", trust, ev->num_props);
		if (tmp_value)
			g_variant_unref(tmp_value);

		/* Paired */
		tmp_value = g_variant_lookup_value(value, "Paired",  G_VARIANT_TYPE_BOOLEAN);
		uint8_t paired = tmp_value ? (g_variant_get_boolean(tmp_value) ? 1 : 0) : 0;

		size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_PAIRED,
				sizeof(uint8_t), &paired);
		ev->num_props++;
		DBG("Device Paired [%d] Property num [%d]", paired, ev->num_props);
		if (tmp_value)
			g_variant_unref(tmp_value);

		/* RSSI*/
		tmp_value = g_variant_lookup_value(value, "RSSI", G_VARIANT_TYPE_INT32);
		int rssi = tmp_value ? g_variant_get_int32(tmp_value) : 0;
		size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_RSSI,
				sizeof(int), &rssi);
		ev->num_props++;
		DBG("Device RSSI [%d] Property num [%d]", rssi, ev->num_props);
		if (tmp_value)
			g_variant_unref(tmp_value);

		/* Last Addr Type */
		tmp_value = g_variant_lookup_value(value, "LastAddrType", G_VARIANT_TYPE_UINT32);
		unsigned int addr_type = tmp_value ? g_variant_get_uint32(tmp_value) : 0;
		if (tmp_value)
			g_variant_unref(tmp_value);
		DBG("Device Last Address Type [0x%x]", addr_type);

		/* Is Alias Set */
		tmp_value = g_variant_lookup_value(value, "IsAliasSet", G_VARIANT_TYPE_UINT32);
		uint8_t is_alias_set = tmp_value ? (g_variant_get_boolean(tmp_value) ? 1 : 0) : 0;
		size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_IS_ALIAS_SET,
					sizeof(uint8_t), &is_alias_set);
		ev->num_props++;
		DBG("IsAliasSet: [%s], Property num [%d]", (is_alias_set ? "TRUE" : "FALSE"), ev->num_props);
		if (tmp_value)
			g_variant_unref(tmp_value);

		/* UUID's */
		tmp_value = g_variant_lookup_value(value, "UUIDs", G_VARIANT_TYPE_STRING_ARRAY);
		gsize uuid_count = g_variant_get_size(tmp_value);
		char **uuid_value = g_variant_dup_strv(tmp_value, &uuid_count);
		{
			/* UUID collection */
			int i;
#ifdef __TEST_
			int z;
#endif
			int num_props_tmp = ev->num_props;

			uint8_t uuids[BT_HAL_STACK_UUID_SIZE * uuid_count];

			for (i = 0; uuid_value[i] != NULL; i++) {

				char *uuid_str = NULL;
				uint8_t uuid[BT_HAL_STACK_UUID_SIZE];
				memset(uuid, 0x00, BT_HAL_STACK_UUID_SIZE);

				DBG("UUID string from Bluez [%s]\n", uuid_value[i]);
				uuid_str = g_strdup(uuid_value[i]);
				DBG("UUID string [%s]\n", uuid_str);

				_bt_hal_convert_uuid_string_to_type(uuid, uuid_str);
#ifdef __TEST_
				for (z = 0; z < 16; z++)
					DBG("[0x%x]", uuid[z]);
#endif

				memcpy(uuids+i*BT_HAL_STACK_UUID_SIZE, uuid, BT_HAL_STACK_UUID_SIZE);
				g_free(uuid_str);
			}

			size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_UUIDS,
					(BT_HAL_STACK_UUID_SIZE * uuid_count),
					uuids);
			ev->num_props = num_props_tmp + 1;
			g_free(uuid_value);
		}
		g_variant_unref(tmp_value);

		/* LegacyManufacturerDataLen */
		tmp_value = g_variant_lookup_value(value, "LegacyManufacturerDataLen", G_VARIANT_TYPE_UINT32);
		unsigned int manufacturer_data_len = tmp_value ? g_variant_get_uint32(tmp_value) : 0;
		if (manufacturer_data_len > BT_HAL_MANUFACTURER_DATA_LENGTH_MAX) {
			ERR("manufacturer_data_len is too long(len = %d)", manufacturer_data_len);
			manufacturer_data_len = BT_HAL_MANUFACTURER_DATA_LENGTH_MAX;
		}
		if (tmp_value)
			g_variant_unref(tmp_value);
		/*size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_MANUFACTURER_DATA_LEN,
		  sizeof(unsigned int), &manufacturer_data_len);
		  ev->num_props++;*/
		DBG("Device Legacy Manufacturer data length [%u]", manufacturer_data_len);

		/* ManufacturerData */
		tmp_value = g_variant_lookup_value(value, "LegacyManufacturerData", G_VARIANT_TYPE_BYTESTRING);
		manufacturer_data = value ? (gchar *)g_variant_get_bytestring(tmp_value) : NULL;
		if (manufacturer_data) {
			if (manufacturer_data_len > 0) {
				//size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_MANUFACTURER_DATA,
				size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_BLE_ADV_DATA,
						manufacturer_data_len, manufacturer_data);
				ev->num_props++;
			}
		}
		if (tmp_value)
			g_variant_unref(tmp_value);

		/* Address */
		uint8_t bdaddr[6];
		_bt_hal_convert_addr_string_to_type(bdaddr, address);
		size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_ADDR,
				sizeof(bdaddr), bdaddr);
		ev->num_props++;
		DBG("Device address [%s] property Num [%d]", address, ev->num_props);

		g_free(name);
		g_variant_unref(value);
	} else {
		ERR("result  is NULL\n");
	}
	if (size > 1) {
		DBG("Send Device found event to HAL user, Num Prop [%d] total size [%zd]", ev->num_props, size);
		event_cb(HAL_EV_DEVICE_FOUND, (void*) buf, size);
	}
	DBG("-");
}

static void __bt_hal_send_device_acl_connection_state_event(int status, gboolean connected, const char *address)
{
	DBG("+");
	struct hal_ev_acl_state_changed ev;

	ev.status = status;
	ev.state = (connected == TRUE) ?
		HAL_ACL_STATE_CONNECTED :
		HAL_ACL_STATE_DISCONNECTED;

	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);

	if (!event_cb)
		ERR("Bluetooth HAL event handler not registered");
	else
		event_cb(HAL_EV_ACL_STATE_CHANGED, &ev, sizeof(ev));
	DBG("-");
}

static void __bt_hal_send_device_le_connection_state_event(int status, gboolean connected, const char *address)
{
	DBG("+");
	struct hal_ev_le_conn_state_changed ev;

	ev.status = status;
	ev.state = (connected == TRUE) ?
		HAL_LE_STATE_CONNECTED :
		HAL_LE_STATE_DISCONNECTED;

	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);

	if (!event_cb)
		ERR("Bluetooth HAL event handler not registered");
	else
		event_cb(HAL_EV_LE_CONN_STATE_CHANGED, &ev, sizeof(ev));
	DBG("-");
}

static void __bt_hal_send_device_trust_state_event(gboolean is_trusted,
				const char *address)
{
	struct hal_ev_device_trust_state_changed ev;
	DBG("+");

	ev.trust = (is_trusted == TRUE) ?
		HAL_DEVICE_TRUSTED :
		HAL_DEVICE_UNTRUSTED;

	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);

	if (!event_cb)
		ERR("Bluetooth HAL event handler not registered");
	else
		event_cb(HAL_EV_DEVICE_TRUST_CHANGED, &ev, sizeof(ev));
	DBG("-");
}

static void __bt_hal_send_device_trusted_profile_changed_event(
		uint32_t trust_val, const char *address)
{
	struct hal_ev_device_trusted_profiles_changed ev;
	DBG("+");

	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
	ev.trust_val = trust_val;

	if (!event_cb)
		ERR("Bluetooth HAL event handler not registered");
	else
		event_cb(HAL_EV_DEVICE_TRUSTED_PROFILES_CHANGED, &ev, sizeof(ev));
	DBG("-");
}

static void __bt_hal_handle_device_specific_events(GVariant *msg, const char *member,
			const char *path)
{
	char *address;
	const char *property = NULL;
	if (path == NULL)
		return;

	if (strcasecmp(member, "PropertyChanged") == 0) {
		g_variant_get(msg, "(s)", &property);
		if (property == NULL)
			return;
		if (strcasecmp(property, "GattConnected") == 0) {
			INFO("GATT Connected");
			gboolean connected = FALSE;
			char *address;
			address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE);

			_bt_hal_convert_device_path_to_address(path, address);
			g_variant_get(msg, "(b)", &connected);

			INFO("Connected device address[%s] connnected[%d]", address, connected);
			g_free(address);
		} else if (strcasecmp(property, "Paired") == 0) {
			gboolean paired = FALSE;
			struct hal_ev_bond_state_changed ev;
			char address[BT_HAL_ADDRESS_STRING_SIZE];

			g_variant_get(msg, "(b)", &paired);
			DBG("Device property changed : Paired = %s", (paired ? "TRUE" : "FALSE"));

			_bt_hal_agent_set_canceled(FALSE);
			_bt_hal_convert_device_path_to_address(path, address);

			/* Prepare to send event to HAL bluetooth */
			ev.status = BT_STATUS_SUCCESS;
			ev.state = paired ? BT_BOND_STATE_BONDED : BT_BOND_STATE_NONE;
			_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);

			if (!event_cb) {
				ERR("Bluetooth HAL event handler not registered");
			} else {
				DBG("Sending HAL_EV_BOND_STATE_CHANGED event");
				event_cb(HAL_EV_BOND_STATE_CHANGED, (void*)&ev, sizeof(ev));
			}
		} else if (strcasecmp(property, "UUIDs") == 0) {
			/* TODO */
		}
	} else if (strcasecmp(member, "DeviceConnected") == 0) {
		unsigned char addr_type = 0;

		g_variant_get(msg, "(y)", &addr_type);

		address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE);
		_bt_hal_convert_device_path_to_address(path, address);

		DBG("Member: [%s]", member);
		ERR_C("Connected [%s] [%s]", !addr_type ? "BREDR" : "LE", address);
		if (!addr_type)
			__bt_hal_send_device_acl_connection_state_event(BT_STATUS_SUCCESS, TRUE, address);
		else
			__bt_hal_send_device_le_connection_state_event(BT_STATUS_SUCCESS, TRUE, address);
		g_free(address);
	} else if (strcasecmp(member, "Disconnected") == 0) {
		unsigned char disc_reason = 0;
		unsigned char addr_type = 0;
		char *name = NULL;

		g_variant_get(msg, "(yy&s)", &addr_type, &disc_reason, &name);

		address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE);
		_bt_hal_convert_device_path_to_address(path, address);

		DBG("Member: [%s]", member);

		ERR_C("DisConnected [%s] [%s]", !addr_type ? "BREDR" : "LE", address);
		DBG("Disconnected Reason [%d : %s]", disc_reason, _bt_hal_convert_disc_reason_to_string(disc_reason));
		DBG("Name: %s", name);
		if (!addr_type)
			__bt_hal_send_device_acl_connection_state_event(_bt_hal_convert_disc_reason_to_status(disc_reason), FALSE, address);
		else
			__bt_hal_send_device_le_connection_state_event(_bt_hal_convert_disc_reason_to_status(disc_reason), FALSE, address);
		g_free(address);
	} else if (strcasecmp(member, "ProfileStateChanged") == 0) {
		int state = 0;
		char *profile_uuid = NULL;

		g_variant_get(msg, "(si)", &profile_uuid, &state);
		address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE);
		_bt_hal_convert_device_path_to_address(path, address);

		DBG("Address: %s", address);
		DBG("Profile UUID: %s", profile_uuid);
		DBG("State: %d", state);
		if (strncmp(profile_uuid, HID_UUID, strlen(HID_UUID)) == 0) {
			if (state == BT_HAL_PROFILE_STATE_CONNECTED)
				__bt_hal_send_hid_connection_state_event(TRUE, address);
			else if (state == BT_HAL_PROFILE_STATE_DISCONNECTED)
				__bt_hal_send_hid_connection_state_event(FALSE, address);
			else
				DBG("Profile state: %d", state);

		} else if ((strncmp(profile_uuid, A2DP_SINK_UUID, strlen(A2DP_SINK_UUID)) == 0)) {
			if (state == BT_HAL_PROFILE_STATE_CONNECTED) {
				DBG("A2DP Profile state changed: BT_PROFILE_STATE_CONNECTED");
				__bt_hal_send_av_connection_state_event(TRUE, address);
			} else if (state == BT_HAL_PROFILE_STATE_DISCONNECTED) {
				DBG("A2DP Profile state changed: BT_PROFILE_STATE_DISCONNECTED");
				__bt_hal_send_av_connection_state_event(FALSE, address);
			} else if (state == BT_HAL_PROFILE_STATE_DISCONNECTING) {
				DBG("A2DP Profile state changed: BT_PROFILE_STATE_DISCONNECTING");
			} else if (state == BT_HAL_PROFILE_STATE_CONNECTING) {
				DBG("A2DP Profile state changed: BT_PROFILE_STATE_CONNECTING");
			} else {
				ERR("A2DP Profile state: Invalid");
			}
		} else if ((strncmp(profile_uuid, A2DP_SOURCE_UUID, strlen(A2DP_SOURCE_UUID)) == 0)) {
			if (state == BT_HAL_PROFILE_STATE_CONNECTED) {
				DBG("A2DP Sink Profile state changed: BT_HAL_PROFILE_STATE_CONNECTED");
				__bt_hal_send_a2dp_sink_connection_state_event(TRUE, address);
			} else if (state == BT_HAL_PROFILE_STATE_DISCONNECTED) {
				DBG("A2DP Sink Profile state changed: BT_HAL_PROFILE_STATE_DISCONNECTED");
				__bt_hal_send_a2dp_sink_connection_state_event(FALSE, address);
			} else if (state == BT_HAL_PROFILE_STATE_DISCONNECTING) {
				DBG("A2DP Sink Profile state changed: BT_HAL_PROFILE_STATE_DISCONNECTING");
			} else if (state == BT_HAL_PROFILE_STATE_CONNECTING) {
				DBG("A2DP Sink Profile state changed: BT_HAL_PROFILE_STATE_CONNECTING");
			}
		} else if (strncmp(profile_uuid, HFP_HF_UUID, strlen(HFP_HF_UUID)) == 0) {
			if (state == BT_HAL_PROFILE_STATE_CONNECTING)
				DBG("HFP Profile state changed: BT_PROFILE_STATE_CONNECTING");
			else if (state == BT_HAL_PROFILE_STATE_CONNECTED) {
				DBG("HFP Profile state changed: BT_PROFILE_STATE_CONNECTED");
				__bt_hal_send_hf_connection_state_event(TRUE, address);
			} else if (state == BT_HAL_PROFILE_STATE_DISCONNECTED) {
				DBG("HFP Profile state changed: BT_PROFILE_STATE_DISCONNECTED");
				__bt_hal_send_hf_connection_state_event(FALSE, address);
			} else if (state == BT_HAL_PROFILE_STATE_DISCONNECTING) {
				DBG("HFP Profile state changed: BT_PROFILE_STATE_DISCONNECTING");
			} else {
				ERR("HFP Profile state: Invalid");
			}
		} else if ((strncmp(profile_uuid, AVRCP_TARGET_UUID, strlen(AVRCP_TARGET_UUID)) == 0)) {
			if (state == BT_HAL_PROFILE_STATE_CONNECTED) {
				DBG("AVRCP Controller Profile state changed: BT_HAL_PROFILE_STATE_CONNECTED");
				__bt_hal_send_avrcp_ctrl_connection_state_event(TRUE, address);
			} else if (state == BT_HAL_PROFILE_STATE_DISCONNECTED) {
				DBG("AVRCP Controller Profile state changed: BT_HAL_PROFILE_STATE_DISCONNECTED");
				__bt_hal_send_avrcp_ctrl_connection_state_event(FALSE, address);
			} else if (state == BT_HAL_PROFILE_STATE_DISCONNECTING) {
				DBG("AVRCP Controller Profile state changed: BT_HAL_PROFILE_STATE_DISCONNECTING");
			} else if (state == BT_HAL_PROFILE_STATE_CONNECTING) {
				DBG("AVRCP Controller Profile state changed: BT_HAL_PROFILE_STATE_CONNECTING");
			}
		} else {
			DBG("Profile[%s] State changed status [%d] ", profile_uuid, state);
		}
		g_free(address);
		g_free(profile_uuid);
	} else if (strcasecmp(member, "AdvReport") == 0) {
		DBG("Member: [%s]", member);
		__bt_hal_handle_adv_report(msg, path);
	}
}

static void __bt_hal_handle_adv_report(GVariant *msg, const char *path)
{
	uint8_t buf[BT_HAL_MAX_PROPERTY_BUF_SIZE];
	struct hal_ev_gatt_client_scan_result *ev = (void *)buf;
	size_t size = 0;
	char *address = NULL;
	GVariant *value = NULL;
	char *buffer = NULL;
	int data_len = 0;
	int buffer_len = 0;
	uint8_t addr_type = 0;
	uint8_t adv_type = 0;
	int rssi = 0;

	if (!gatt_event_cb)
		return;

	memset(buf, 0, sizeof(buf));
	size = sizeof(*ev);

	g_variant_get(msg, "(&syyii@ay)", &address, &addr_type,
			&adv_type, &rssi, &data_len, &value);

	buffer_len = g_variant_get_size(value);
	if (buffer_len > 0)
		buffer = (char *)g_variant_get_data(value);

	if (data_len != buffer_len) {
		ERR("Unexpected: buffer_len: %d, data_len: %d",
				buffer_len, data_len);
		data_len = buffer_len;
	}

	DBG("Address: %s, len: %d, rssi: %d, addr_type: 0x%02X, adv_type: 0x%02X",
			address, data_len, rssi, addr_type, adv_type);

	_bt_hal_convert_addr_string_to_type(ev->bd_addr, address);
	ev->addr_type = addr_type;
	ev->adv_type = adv_type;
	ev->rssi = rssi;
	ev->len = data_len;
	memcpy(ev->adv_data, buffer, data_len);
	size += data_len;

	DBG("Send le scan result event to HAL, size: [%zd]", size);
	gatt_event_cb(HAL_EV_GATT_CLIENT_SCAN_RESULT, buf, size);
	g_variant_unref(value);
}

/* AVRCP Controller Role(Remote:AVRCP Target) Events */
static void __bt_hal_send_avrcp_ctrl_connection_state_event(gboolean connected, const char *address)
{
	DBG("+");
	struct hal_ev_avrcp_ctrl_conn_state ev;

	if (connected == TRUE)
		INFO("AVRCP(Controller) Profile Connected for address [%s]", address);
	else
		INFO("AVRCP(Controller) Profile DisConnected for address [%s]", address);

	if (connected == TRUE)
		ev.state = HAL_AVRCP_CTRL_STATE_CONNECTED;
	else
		ev.state = HAL_AVRCP_CTRL_STATE_DISCONNECTED;
	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);

	if (!a2dp_sink_event_cb)
		ERR("AV event handler not registered");
	else
		avrcp_ctrl_event_cb(HAL_EV_AVRCP_CTRL_CONN_STATE, &ev, sizeof(ev));
}

static int __bt_media_attr_to_type(const char *str)
{
	if (!strcasecmp(str, "Equalizer"))
		return HAL_PLAYER_ATTR_EQUALIZER;
	else if (!strcasecmp(str, "Repeat"))
		return HAL_PLAYER_ATTR_REPEAT;
	else if (!strcasecmp(str, "Shuffle"))
		return HAL_PLAYER_ATTR_SHUFFLE;
	else if (!strcasecmp(str, "Scan"))
		return HAL_PLAYER_ATTR_SCAN;
	else
		return 0;
}


static int __bt_hal_play_status_str_to_type(const char *value)
{
	if (!strcmp(value, "stopped"))
		return HAL_PLAYSTATE_STOPPED;
	else if (!strcmp(value, "playing"))
		return HAL_PLAYSTATE_PLAYING;
	else if (!strcmp(value, "paused"))
		return HAL_PLAYSTATE_PAUSED;
	else if (!strcmp(value, "forward-seek"))
		return HAL_PLAYSTATE_FWD_SEEK;
	else if (!strcmp(value, "reverse-seek"))
		return HAL_PLAYSTATE_REV_SEEK;
	else
		return HAL_PLAYSTATE_ERROR;
}

static void __bt_avrcp_control_parse_properties(struct hal_ev_track_changed *ev, GVariant *item)
{
	GVariant *value = NULL;
	GVariantIter iter;
	char *value_string = NULL;
	const char *key = NULL;
	int i = 0;

	if (ev == NULL)
		return;

	g_variant_iter_init(&iter, item);
	while (g_variant_iter_loop(&iter, "{sv}", &key, &value)) {
		if (strcasecmp(key, "Title") == 0) {
			value_string = (char *)g_variant_get_string(value, NULL);
			DBG("Value : %s ", value_string);
			ev->attr[i].attr_id = HAL_MEDIA_ATTR_ID_TITLE;
			g_strlcpy((gchar *)ev->attr[i].text, value_string, HAL_MAX_ATTR_STR_LEN);
			i++;
		} else if (strcasecmp(key, "Artist") == 0) {
			value_string = (char *)g_variant_get_string(value, NULL);
			DBG("Value : %s ", value_string);
			ev->attr[i].attr_id = HAL_MEDIA_ATTR_ID_ARTIST;
			g_strlcpy((gchar *)ev->attr[i].text, value_string, HAL_MAX_ATTR_STR_LEN);
			i++;
		} else if (strcasecmp(key, "Album") == 0) {
			value_string = (char *)g_variant_get_string(value, NULL);
			DBG("Value : %s ", value_string);
			ev->attr[i].attr_id = HAL_MEDIA_ATTR_ID_ALBUM;
			g_strlcpy((gchar *)ev->attr[i].text, value_string, HAL_MAX_ATTR_STR_LEN);
			i++;
		} else if (strcasecmp(key, "Genre") == 0) {
			value_string = (char *)g_variant_get_string(value, NULL);
			DBG("Value : %s ", value_string);
			ev->attr[i].attr_id = HAL_MEDIA_ATTR_ID_GENRE;
			g_strlcpy((gchar *)ev->attr[i].text, value_string, HAL_MAX_ATTR_STR_LEN);
			i++;
		} else if (strcasecmp(key, "Duration") == 0) {
			long int val;

			val = g_variant_get_uint32(value);
			DBG("Value : %li", val);
			ev->attr[i].attr_id = HAL_MEDIA_ATTR_ID_PLAYING_TIME;
			snprintf((char*)ev->attr[i].text, HAL_MAX_ATTR_STR_LEN, "%ld", val);
			i++;
		} else if (strcasecmp(key, "NumberOfTracks") == 0) {
			long int val;

			val = g_variant_get_uint32(value);
			DBG("Value : %li", val);
			ev->attr[i].attr_id = HAL_MEDIA_ATTR_ID_NUM_TRACKS;
			snprintf((char*)ev->attr[i].text, HAL_MAX_ATTR_STR_LEN, "%ld", val);
			i++;
		} else if (strcasecmp(key, "TrackNumber") == 0) {
			long int val;

			val = g_variant_get_uint32(value);
			DBG("Value : %li", val);
			ev->attr[i].attr_id = HAL_MEDIA_ATTR_ID_TRACK_NUM;
			snprintf((char*)ev->attr[i].text, HAL_MAX_ATTR_STR_LEN, "%ld", val);
			i++;
		} else {
			DBG("%s not supported, ignoring", key);
		}

		if (i >= HAL_MAX_ATTR_NUM) {
			ERR(" Received max attribute [%d]", i);
			break;
		}
	}

	ev->num_attr = i;
	g_variant_iter_free(&iter);
	return;
}

static int __bt_media_attrval_to_val(int type, const char *value)
{
	int ret = -1;

	switch (type) {
	case HAL_PLAYER_ATTR_EQUALIZER:
		if (!strcmp(value, "off"))
			ret = 0x01;
		else
			ret = 0x02;
		break;
	case HAL_PLAYER_ATTR_REPEAT:
		if (!strcmp(value, "off"))
			ret = BTRC_PLAYER_VAL_OFF_REPEAT;
		else if (!strcmp(value, "singletrack"))
			ret = BTRC_PLAYER_VAL_SINGLE_REPEAT;
		else if (!strcmp(value, "alltracks"))
			ret = BTRC_PLAYER_VAL_ALL_REPEAT;
		else
			ret = BTRC_PLAYER_VAL_GROUP_REPEAT;
		break;
	case HAL_PLAYER_ATTR_SHUFFLE:
		if (!strcmp(value, "off"))
			ret = BTRC_PLAYER_VAL_OFF_SHUFFLE;
		else if (!strcmp(value, "alltracks"))
			ret = BTRC_PLAYER_VAL_ALL_SHUFFLE;
		else
			ret = BTRC_PLAYER_VAL_GROUP_SHUFFLE;
		break;
	case HAL_PLAYER_ATTR_SCAN:
		if (!strcmp(value, "off"))
			ret = 0x01;
		else if (!strcmp(value, "alltracks"))
			ret = 0x02;
		else
			ret = 0x03;
		break;
	default:
		ERR("Value not handled");
	}

	return ret;
}

static void __bt_hal_handle_avrcp_ctrl_events(GVariant *msg, const char *member,
		const char *path)
{
	const char *property = NULL;
	GVariant *value = NULL;
	GVariantIter iter;
	char address[BT_HAL_ADDRESS_STRING_SIZE];

	if (!msg) {
		ERR("Error returned in method call\n");
		return;
	}

	if (!avrcp_ctrl_event_cb) {
		ERR("AVRCP controller DBUS handler callback not registered");
		return;
	}

	g_variant_iter_init(&iter, msg);

	_bt_hal_convert_device_path_to_address(path, address);

	while (g_variant_iter_loop(&iter, "{sv}", &property, &value)) {
		DBG("Property = %s \n", property);
		if ((strcasecmp(property, "Equalizer") == 0) ||
				(strcasecmp(property, "Repeat") == 0) ||
				(strcasecmp(property, "Shuffle") == 0) ||
				(strcasecmp(property, "Scan") == 0)) {
			struct hal_ev_player_setting ev;
			const char *valstr;
			int val = 0;

			valstr = g_variant_get_string(value, NULL);
			DBG("Value : %s ", valstr);

			if (valstr)
				val = __bt_media_attrval_to_val(__bt_media_attr_to_type(property), valstr);

			if (val >= 0) {
				memset(&ev, 0, sizeof(ev));
				_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
				ev.num_attr = 1;
				ev.attr_ids[0] = __bt_media_attr_to_type(property);
				ev.attr_values[0] = val;

				/* Send event to application */
				avrcp_ctrl_event_cb(HAL_EV_AVRCP_CTRL_PLAYER_APP_SETTING_CHANGED, &ev, sizeof(ev));
			}
		} else if ((strcasecmp(property, "Status") == 0)) {
			struct hal_ev_play_status_changed ev;
			const char *valstr;

			valstr = g_variant_get_string(value, NULL);
			DBG("Value : %s ", valstr);

			memset(&ev, 0, sizeof(ev));
			_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
			if (valstr)
				ev.status = __bt_hal_play_status_str_to_type(valstr);

			/* Send event to application */
			avrcp_ctrl_event_cb(HAL_EV_AVRCP_CTRL_PLAY_STATUS_CHANGED, &ev, sizeof(ev));

		} else if (strcasecmp(property, "Position") == 0) {
			struct hal_ev_play_position ev;

			memset(&ev, 0, sizeof(ev));
			_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
			ev.pos = g_variant_get_uint32(value);
			DBG("Value : %d ", ev.pos);

			/* Send event to application */
			avrcp_ctrl_event_cb(HAL_EV_AVRCP_CTRL_PLAY_POSITION_CHANGED, &ev, sizeof(ev));
		} else if (strcasecmp(property, "Track") == 0) {
			struct hal_ev_track_changed ev;

			memset(&ev, 0, sizeof(ev));
			_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
			__bt_avrcp_control_parse_properties(&ev, value);

			/* Send event to application */
			avrcp_ctrl_event_cb(HAL_EV_AVRCP_CTRL_TRACK_CHANGED, &ev, sizeof(ev));
		} else {
			DBG("Property not handled");
		}
	}

	g_free((char *)property);
	g_variant_unref(value);
}

static void __bt_hal_handle_avrcp_transport_events(GVariant *msg, const char *member,
		const char *path)
{
	const char *property = NULL;
	GVariant *value = NULL;
	GVariantIter iter;
	char address[BT_HAL_ADDRESS_STRING_SIZE];
	DBG("+");

	if (!msg) {
		ERR("Error returned in method call\n");
		return;
	}

	if (!avrcp_tg_event_cb) {
		ERR("AVRCP target DBUS handler callback not registered");
		return;
	}

	g_variant_iter_init(&iter, msg);

	_bt_hal_convert_device_path_to_address(path, address);

	while (g_variant_iter_loop(&iter, "{sv}", &property, &value)) {
		DBG("Property = %s \n", property);
		if ((strcasecmp(property, "Delay") == 0)) {
			struct hal_ev_avrcp_tg_delay_changed ev;
			uint16_t val;

			memset(&ev, 0, sizeof(ev));
			_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);

			val = g_variant_get_uint16(value);
			DBG("Value : %d", val);
			ev.value = val;

			/* Send event to application */
			avrcp_tg_event_cb(HAL_EV_AVRCP_TG_DELAY_CHANGE, &ev, sizeof(ev));
		} else {
			DBG("Property not handled");
		}
	}

	DBG("-");
	g_free((char *)property);
	g_variant_unref(value);
}

/* A2DP Src Role(Remote:Sink) Events */
static void __bt_hal_send_av_connection_state_event(gboolean connected, const char *address)
{
	DBG("+");
	struct hal_ev_a2dp_conn_state ev;

	if (connected == TRUE)
		INFO("A2DP(Src) Profile Connected for address [%s]", address);
	else
		INFO("A2DP(Src) Profile DisConnected for address [%s]", address);

	ev.state = (connected == TRUE) ?
		HAL_EV_A2DP_STATE_CONNECTED :
		HAL_EV_A2DP_STATE_DISCONNECTED;

	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);

	if (!av_event_cb)
		ERR("AV event handler not registered");
	else
		av_event_cb(HAL_EV_A2DP_CONN_STATE, &ev, sizeof(ev));
}

/* A2DP Sink Role(Remote:Source) Events */
static void __bt_hal_send_a2dp_sink_connection_state_event(gboolean connected, const char *address)
{
	DBG("+");
	struct hal_ev_a2dp_conn_state ev;

	if (connected == TRUE)
		INFO("A2DP(Sink) Profile Connected for address [%s]", address);
	else
		INFO("A2DP(Sink) Profile DisConnected for address [%s]", address);

	ev.state = (connected == TRUE) ?
		HAL_EV_A2DP_STATE_CONNECTED :
		HAL_EV_A2DP_STATE_DISCONNECTED;

	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);

	if (!a2dp_sink_event_cb)
		ERR("AV event handler not registered");
	else
		a2dp_sink_event_cb(HAL_EV_A2DP_SOURCE_CONN_STATE, &ev, sizeof(ev));
}

/* HF(AG Role) Audio Events */
static void __bt_hal_send_hf_audio_connection_state_event(gboolean connected,
						const char *address)
{
	DBG("+");
	struct hal_ev_handsfree_audio_state ev;

	if (connected == TRUE)
		INFO("AG Audio Connected for address [%s]", address);
	else
		INFO("AG Audio DisConnected for address [%s]", address);

	ev.state = (connected == TRUE) ?
		HAL_EV_HANDSFREE_AUDIO_STATE_CONNECTED :
		HAL_EV_HANDSFREE_AUDIO_STATE_DISCONNECTED;

	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);

	if (!hf_event_cb)
		ERR("HF event handler not registered");
	else
		hf_event_cb(HAL_EV_HANDSFREE_AUDIO_STATE, &ev, sizeof(ev));
}

/* HF(AG Role) Profile Events */
static void __bt_hal_send_hf_connection_state_event(gboolean connected,
						const char *address)
{
	DBG("+");
	struct hal_ev_handsfree_conn_state ev;

	if (connected == TRUE)
		INFO("AG Profile Connected for address [%s]", address);
	else
		INFO("AG Profile DisConnected for address [%s]", address);

	ev.state = (connected == TRUE) ?
		HAL_EV_HANDSFREE_CONN_STATE_CONNECTED :
		HAL_EV_HANDSFREE_CONN_STATE_DISCONNECTED;

	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);

	if (!hf_event_cb)
		ERR("HF event handler not registered");
	else
		hf_event_cb(HAL_EV_HANDSFREE_CONN_STATE, &ev, sizeof(ev));
}

void _bt_hal_register_event_handler_cb(bt_hal_module_e module, handle_stack_msg cb)
{
	switch (module) {
	case HAL_HIDHOST:
		hid_event_cb = cb;
		break;
	case HAL_A2DP_SRC:
		av_event_cb = cb;
		break;
	case HAL_A2DP_SNK:
		a2dp_sink_event_cb = cb;
		break;
	case HAL_HF_AG:
		hf_event_cb = cb;
		break;
	case HAL_AVRCP_TG:
		avrcp_tg_event_cb = cb;
		break;
	case HAL_AVRCP_CTRL:
		avrcp_ctrl_event_cb = cb;
		break;
	case HAL_GATT:
		gatt_event_cb = cb;
		break;
	default:
		ERR("Unknown module: %d", module);
	}
}

void _bt_hal_unregister_event_handler_cb(bt_hal_module_e module)
{
	switch (module) {
	case HAL_HIDHOST:
		hid_event_cb = NULL;
		break;
	case HAL_A2DP_SRC:
		av_event_cb = NULL;
		break;
	case HAL_A2DP_SNK:
		a2dp_sink_event_cb = NULL;
		break;
	case HAL_HF_AG:
		hf_event_cb = NULL;
		break;
	case HAL_AVRCP_TG:
		avrcp_tg_event_cb = NULL;
		break;
	case HAL_AVRCP_CTRL:
		avrcp_ctrl_event_cb = NULL;
		break;
	case HAL_GATT:
		gatt_event_cb = NULL;
		break;
	default:
		ERR("Unknown module: %d", module);
	}
}

bool _bt_hal_get_adapter_request_state(void)
{
	return is_adapter_activating;
}

bool _bt_hal_get_le_request_state(void)
{
	return is_le_activating;
}

void _bt_hal_set_adapter_request_state(bool enable)
{
	DBG("set_adapter_request_state %d", enable);
	is_adapter_activating = enable;
}

void _bt_hal_set_le_request_state(bool enable)
{
	DBG("set_le_request_state %d", enable);
	is_le_activating = enable;
}
