/*
 * Open Adaptation Layer (OAL)
 *
 * Copyright (c) 2014-2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef _OAL_ADAPTER_MGR_H_
#define _OAL_ADAPTER_MGR_H_

#include <glib.h>
#include <sys/types.h>

#include <oal-manager.h>
#include <oal-gatt.h>

#ifdef __cplusplus
extern "C" {
#endif

#define OAL_BT_ADDRESS_STRING_SIZE 18


/**
 * @brief Enable BT chip for usage.
 *
 * @remarks Other API can only be used after successful event. \n
 *
 * @details EVENT: OAL_EVENT_ADAPTER_ENABLED/OAL_EVENT_ADAPTER_DISABLED
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre OAL API must be initialized with oal_bt_init().
 *
 * @see  adapter_disable()
 */
oal_status_t adapter_enable(void);


/**
 * @brief Disable Adapter.
 *
 * @remarks  Disables the BT stack and chip. After this, no OAL API is valid except "adapter_enable()"
 *
 * @details EVENT: OAL_EVENT_ADAPTER_DISABLED
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  adapter_enable()
 */
oal_status_t adapter_disable(void);

#ifdef TIZEN_BT_HAL

/**
 * @brief Enable LE Mode BT chip for usage.
 *
 * @remarks Other API can only be used after successful event. \n
 *
 * @details EVENT: OAL_EVENT_LE_ENABLED/OAL_EVENT_LE_DISABLED
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre OAL API must be initialized with oal_bt_init().
 *
 * @see  le_disable()
 */
oal_status_t le_enable(void);

/**
 * @brief Disable LE Adapter.
 *
 * @remarks  Disables the LE mode in the chip. After this, no OAL API is valid except "le_enable()"
 *
 * @details EVENT: OAL_EVENT_LE_DISABLED
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre LE Adapter must be enabled with le_enable() followed by OAL_EVENT_LE_ENABLED
 *
 * @see  le_enable()
 */
oal_status_t le_disable(void);


/**
 * @brief init LE Adapter.
 *
 * @remarks  Initilizes LE Advertisement slot list
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre LE Adapter must be enabled with le_enable() followed by OAL_EVENT_LE_ENABLED
 *
 * @see  le_enable()
 */
oal_status_t le_init(void);


/**
 * @brief deinit LE Adapter.
 *
 * @remarks  free LE Advertisement slot list
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre LE Adapter must be enabled with le_enable() followed by OAL_EVENT_LE_ENABLED and Initilized using le_init()
 *
 * @see  le_enable()
 */
oal_status_t le_deinit(void);

#endif

/**
 * @brief Start Custom Discovery.
 *
 * @remarks  Based on discovery type, it either starts BREDR, LE or DUAL discovery in Adapter
 *
 * @details For each device found either OAL_EVENT_ADAPTER_INQUIRY_RESULT_BREDR_ONLY
 *                or OAL_EVENT_ADAPTER_INQUIRY_RESULT_BLE will be generated.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @see  adapter_stop_inquiry()
 * @see  OAL_EVENT_ADAPTER_INQUIRY_STARTED
 * @see  OAL_EVENT_ADAPTER_INQUIRY_RESULT_BREDR_ONLY
 * @see  OAL_EVENT_ADAPTER_INQUIRY_RESULT_BLE
 * @see  OAL_EVENT_ADAPTER_INQUIRY_FINISHED
 */
oal_status_t adapter_start_custom_inquiry(discovery_type_t disc_type);


/**
 * @brief Get Adapter Powered state.
 *
 * @remarks  It is possible that stack is up but BT hardware interface is still not enabled.
 * Application can call this API to explicitly check for Powered state
 *
 * @details If API returns OAL_STATUS_FAIL, then adapter can be considered Powered Down.
 * If API returns OAL_STATUS_SUCCESS, then application needs to check powered status. It can be either
 * true or false depending on real status of the BT hardware.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 */
oal_status_t adapter_get_powered_status(gboolean *status);

/**
 * @brief Reset Adapter.
 *
 * @remarks  Disables the BT stack and chip.
 *
 * @details EVENT: OAL_EVENT_ADAPTER_DISABLED
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @see  adapter_enable()
 */
oal_status_t adapter_reset(void);

/**
 * @brief Start device discovery
 *
 * @details For each device found either OAL_EVENT_ADAPTER_INQUIRY_RESULT_BREDR_ONLY
 *                or OAL_EVENT_ADAPTER_INQUIRY_RESULT_BLE will be generated.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  adapter_stop_inquiry()
 * @see  OAL_EVENT_ADAPTER_INQUIRY_STARTED
 * @see  OAL_EVENT_ADAPTER_INQUIRY_RESULT_BREDR_ONLY
 * @see  OAL_EVENT_ADAPTER_INQUIRY_RESULT_BLE
 * @see  OAL_EVENT_ADAPTER_INQUIRY_FINISHED
 */
oal_status_t adapter_start_inquiry(unsigned short duration);

/**
 * @brief Stop device discovery
 *
 * @details OAL_EVENT_ADAPTER_INQUIRY_FINISHED will be generated by OAL
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Discovery must be in progress by calling adapter_start_inquiry()
 *
 * @see  adapter_start_inquiry()
 * @see  OAL_EVENT_ADAPTER_INQUIRY_FINISHED
 */
oal_status_t adapter_stop_inquiry(void);

/**
 * @brief Get local adapter properties
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 */
oal_status_t adapter_get_properties(void);

/**
 * @brief Get local BT chip address
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see OAL_EVENT_ADAPTER_PROPERTY_ADDRESS
 */
oal_status_t adapter_get_address(void);

/**
 * @brief Get local BT version
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see OAL_EVENT_ADAPTER_PROPERTY_VERSION
 */
oal_status_t adapter_get_version(void);

/**
 * @brief Get local BT chip name
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see OAL_EVENT_ADAPTER_PROPERTY_NAME
 */
oal_status_t adapter_get_name(void);

/**
 * @brief Set local BT chip name
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see OAL_EVENT_ADAPTER_PROPERTY_NAME
 */
oal_status_t adapter_set_name(char * name);

/**
 * @brief Sets output variable to TRUE if adapter is discoverable & connectable.
 */
oal_status_t adapter_is_discoverable(int *p_discoverable);

/**
 * @brief Sets output variable to TRUE if adapter is either discoverable or connectable.
 */
oal_status_t adapter_is_connectable(int *p_connectable);

/**
 * @brief Sets output variable to value of current discoverable timeout.
 */
oal_status_t adapter_get_discoverable_timeout(int *p_timeout);

/**
 * @brief Get List of UUIDs for services supported
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see OAL_EVENT_ADAPTER_PROPERTY_SERVICES
 */
oal_status_t adapter_get_service_uuids(void);

/**
 * @brief Get list of bonded devices
 *
 * @remarks     A list of bt_address_t is provided in the event data.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see OAL_EVENT_ADAPTER_BONDED_DEVICE_LIST
 */
oal_status_t adapter_get_bonded_devices(void);

/*
 * @brief Set connectability of adapter
 *
 * @remarks     If enabled, after success event, device can connected from remote device.
 *                      If disabled, after success event, device cannot be connected from remote device;
 *                      neither can be found in discoveries.
 *
 * @details EVENT:  OAL_EVENT_ADAPTER_MODE_CONNECTABLE/OAL_EVENT_ADAPTER_MODE_NON_CONNECTABLE
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  adapter_is_connectable()
 * @see  OAL_EVENT_ADAPTER_MODE_CONNECTABLE
 * @see  OAL_EVENT_ADAPTER_MODE_NON_CONNECTABLE
 */
oal_status_t adapter_set_connectable(int connectable);

/**
 * @brief Make adapter discoverable
 *
 * @remarks  This device appears in inquiries and can be connected by remote device.
 *
 * @details If operation successful OAL_EVENT_ADAPTER_MODE_DISCOVERABLE is generated
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  adapter_is_discoverable()
 * @see  OAL_EVENT_ADAPTER_MODE_DISCOVERABLE
 */
oal_status_t adapter_set_discoverable(void);

/**
 * @brief Set discoverable timeout
 *
 * @param timeout Timeout in seconds after which device cannot be found in discoveries
 *
 * @details Event  OAL_EVENT_ADAPTER_MODE_DISCOVERABLE_TIMEOUT will be generated.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  adapter_is_discoverable()
 * @see  OAL_EVENT_ADAPTER_MODE_DISCOVERABLE_TIMEOUT
 */
oal_status_t adapter_set_discoverable_timeout(int timeout);

/**
 * @brief Set Advertising data on specific Instance ID
 *
 * @param instance_id Multi advertising slot ID initialized by stack
 *
 * @details Event  OAL_EVENT_BLE_MULTI_ADVERTISING_SET_INST_DATA will be generated.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  gatts_multi_adv_set_inst_data()
 */
oal_status_t adapter_ble_multi_adv_set_inst_data(int instance_id, oal_ble_multi_adv_param_setup_t *adv_param_setup);

/**
 * @brief Set Advertising parameters on specific Instance ID
 *
 * @param instance_id Multi advertising slot ID initialized by stack
 *
 * @details Event  OAL_EVENT_BLE_MULTI_ADVERTISING_UPDATE event will be generated.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  gatts_multi_adv_set_inst_data()
 */
oal_status_t adapter_ble_multi_adv_update(int Ins_id, int min_intv, int max_intv,
		int adv_type, int chnl_map, int tx_power, int timeout_s);

/**
 * @brief Enable Multi Advertising on specific Instance ID
 *
 * @param instance_id Multi advertising slot ID initialized by stack
 *
 * @details Event OAL_EVENT_BLE_MULTI_ADVERTISING_ENABLE will be generated.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 */
oal_status_t adapter_ble_multi_adv_enable(int instance_id);

/**
 * @brief Disable Multi Advertising on specific Instance ID
 *
 * @param instance_id Multi advertising slot ID initialized by stack
 *
 * @details Event OAL_EVENT_BLE_MULTI_ADVERTISING_DISABLE will be generated.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 */
oal_status_t adapter_ble_multi_adv_disable(int instance_id);

/**
 * @brief Check if the advertising is enable
 *
 * @param void
 *
 * @details check from the LE adv list if the Advertising is enabled on the server created.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 */
oal_status_t is_advertising(void);
#ifdef __cplusplus
}
#endif /* __cplusplus */
#endif /*_OAL_ADAPTER_MGR_H_*/
