/*
 * Open Adaptation Layer (OAL)
 *
 * Copyright (c) 2014-2015 Samsung Electronics Co., Ltd.
 *
 * Contact: Atul Kumar Rai <a.rai@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef _OAL_HDP_H_
#define _OAL_HDP_H_

#include <glib.h>
#include <sys/types.h>

#include <oal-manager.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef enum {
	OAL_HDP_ROLE_SOURCE,
	OAL_HDP_ROLE_SINK
} oal_hdp_role_t;

typedef enum {
	OAL_CHANNEL_TYPE_RELIABLE,
	OAL_CHANNEL_TYPE_STREAMING,
	OAL_CHANNEL_TYPE_ANY
} oal_channel_type_t;

/**
 * @brief Initialize Health Device Profile
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre OAL API must be initialized with oal_bt_init().
 *
 * @see  hdp_cleanup()
 */
oal_status_t hdp_init(void);


/**
 * @brief Cleanup Health Device Profile
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre HDP should be initialized with hdp_init().
 *
 * @see  hdp_init()
 */
oal_status_t hdp_cleanup(void);

/**
 * @brief Register HDP application
 *
 * @details  Result will be notified through an OAL event
 *
 * @param[in] role: HDP application role (SOURCE/SINK)
 * @param[in] channel_type: HDP application channel type (RELIABLE/STREAMMING/ANY)
 * @param[in] name: Application name
 * @param[in] data_type: Data type supported by HDP application
 * @return Registered application app_id >= 0 on success, otherwise a negative value.
 * @remark None
 *
 * @pre HDP should be initialized with hdp_init().
 *
 * @see  OAL_EVENT_HDP_APP_REGISTERED
 */
int hdp_register_application(oal_hdp_role_t role,
	oal_channel_type_t channel_type, const char *name, int data_type);

/**
 * @brief Un-register HDP application
 *
 * @details  Result will be notified through an OAL event
 *
 * @param[in] app_id: HDP application id to be unregistered
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 * @remark None
 *
 * @pre HDP should be initialized with hdp_init().
 *
 * @see  OAL_EVENT_HDP_APP_UNREGISTERED
 */
oal_status_t hdp_unregister_application(int app_id);

/**
 * @brief Connec HDP channel
 *
 * @details  Result will be notified through an OAL event
 *
 * @pre HDP should be initialized with hdp_init().
 *
 * @param[in]   app_id HDP application id for which channel to be connected
 * @param[in]   addresss Remote HDP device address
 * @return Connected channel_id on success, otherwise a negative value.
 * @remark None
 *
 * @see  OAL_EVENT_HDP_CHANNEL_CONNECTED
 */
int hdp_connect_channel(int app_id, bt_address_t *address);

/**
 * @brief Disconnect hdp channel
 *
 * @details  Result will be notified through an OAL event
 *
 * @pre HDP should be initialized with hdp_init().
 *
 * @param[in]   channel_id HDP channel id to be disconnected
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 * @remark      None
 *
 * @see  OAL_EVENT_HDP_CHANNEL_DESTROYED
 */
oal_status_t hdp_disconnect_channel(int channel_id);

#ifdef __cplusplus
}
#endif /* __cplusplus */
#endif /* _OAL_HDP_H_ */
