/*
 * Open Adaptation Layer (OAL)
 *
 * Copyright (c) 2014-2015 Samsung Electronics Co., Ltd.
 *
 * Contact: Anupam Roy <anupam.r@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */


#ifndef _OAL_HFPC_H_
#define _OAL_HFP_H_


#include <glib.h>
#include <sys/types.h>
#include <oal-manager.h>
#include "oal-event.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief Enable Audio HFP profile or Audio Gateway Role of HFP profile
 *
 * @remarks  BT Headsets/Handsfree devices will be able to connect.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre OAL API must be initialized with oal_bt_init().
 *
 * @see  hfp_disable()
 */
oal_status_t hfp_enable(int max_hfp_clients);

/**
 * @brief Disable Audio HFP profile or Audio Gateway Role of HFP profile
 *
 * @remarks  BT Headsets/Handsfree devices will disconnect.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre AG Host should be enabled with hfp_enable().
 *
 * @see  hfp_enable()
 */
oal_status_t hfp_disable(void);

/**
 * @brief Initiate a HFP connection with a BT headset\handsfree device
 *
 * @details  Result will be notified through an OAL event
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre AG Host should be enabled with hfp_enable().
 *
 * @see  OAL_EVENT_HFP_DISCONNECTED
 */
oal_status_t hfp_connect(bt_address_t *device_address);

/**
 * @brief Terminate a HFP connection with a BT headset\handsfree device
 *
 * @details  Result will be notified through an OAL event
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre AG Host should be enabled with hfp_enable().
 *
 * @see  OAL_EVENT_HFP_CONNECTED
 */
oal_status_t hfp_disconnect(bt_address_t *device_address);

/**
 * @brief create an audio connection with a connected BT Sound device
 *
 * @details  Result will be notified through an OAL event
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre BT Audio should be connected with a BT Sound device.
 *
 * @see  OAL_EVENT_HFP_AUDIO_CONNECTED
 */
oal_status_t hfp_connect_audio(bt_address_t *device_address);

/**
 * @brief Disconnect the audio connection with a connected BT Sound device
 *
 * @details  Result will be notified through an OAL event
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre BT Audio should be connected with a BT Sound device.
 *
 * @see  OAL_EVENT_HFP_AUDIO_DISCONNECTED
 */
oal_status_t hfp_disconnect_audio(bt_address_t *device_address);

#ifdef __cplusplus
}
#endif /* __cplusplus */
#endif /*_OAL_HFP_H_*/
