/*
 * Open Adaptation Layer (OAL)
 *
 * Copyright (c) 2014-2015 Samsung Electronics Co., Ltd.
 *
 * Contact: Anupam Roy <anupam.r@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dlog.h>

#include <bluetooth.h>
#include <bt_hf.h>

#include "oal-common.h"
#include "oal-internal.h"
#include "oal-hfp.h"
#include "oal-utils.h"

#define CHECK_OAL_HF_ENABLED() \
	do { \
		if (blued_hfp_interface == NULL) { \
			BT_ERR("OAL, HFP Not Enabled"); \
			return OAL_STATUS_NOT_READY; \
		} \
	} while (0)

static void cb_hfp_connection_state(bthf_connection_state_t state, bt_bdaddr_t *bd_addr);
static void cb_hfp_audio_connection_state(bthf_audio_state_t state, bt_bdaddr_t *bd_addr);

const bt_interface_t * _bt_get_stack_interface(void);
static const bthf_interface_t *blued_hfp_interface = NULL;

static bthf_callbacks_t blued_hf_cb = {
	.size = sizeof(blued_hf_cb),
	.connection_state_cb = cb_hfp_connection_state,
	.audio_state_cb = cb_hfp_audio_connection_state,
	.vr_cmd_cb = NULL,
	.answer_call_cmd_cb = NULL,
	.hangup_call_cmd_cb = NULL,
	.volume_cmd_cb = NULL,
	.dial_call_cmd_cb = NULL,
	.dtmf_cmd_cb = NULL,
	.nrec_cmd_cb = NULL,
	.wbs_cb = NULL,
	.chld_cmd_cb = NULL,
	.cnum_cmd_cb = NULL,
	.cind_cmd_cb = NULL,
	.cops_cmd_cb = NULL,
	.clcc_cmd_cb = NULL,
	.unknown_at_cmd_cb = NULL,
	.key_pressed_cmd_cb = NULL
};

oal_status_t hfp_enable(int max_hf_clients_supported)
{
	const bt_interface_t* blued_inf;
	int ret;

	API_TRACE("HF Enable");

	if (max_hf_clients_supported > 1)
		return OAL_STATUS_NOT_SUPPORT;

	if ((blued_inf = adapter_get_stack_interface()) == NULL) {
		BT_ERR("Bluetooth module is not loaded");
		return OAL_STATUS_NOT_READY;
	}

	if (blued_hfp_interface != NULL) {
		BT_WARN("HF Interface is already initialized...");
		return OAL_STATUS_ALREADY_DONE;
	}

	if ((blued_hfp_interface = (const bthf_interface_t *)blued_inf->get_profile_interface(BT_PROFILE_HANDSFREE_ID)) == NULL) {
		BT_ERR("OAL, Failed to get Bluetooth HF Interface");
		return OAL_STATUS_INTERNAL_ERROR;
	}

	BT_DBG("Got profile interface");
	if ((ret = blued_hfp_interface->init(&blued_hf_cb, 1)) != BT_STATUS_SUCCESS) {
		BT_ERR("Failed to initialize Bluetooth HF, status: %s", status2string(ret));
		blued_hfp_interface = NULL;
		return convert_to_oal_status(ret);
	}
	BT_DBG("OAL, Bluetooth HF interface initialised");

	return OAL_STATUS_SUCCESS;
}

/* HF deinit: Resets all the HF information
 * Note: Adapter should be disabled before calling deinit
 * */
oal_status_t hfp_disable(void)
{
	API_TRACE("HF disable");
	CHECK_OAL_HF_ENABLED();

	blued_hfp_interface->cleanup();
	blued_hfp_interface = NULL;

	return OAL_STATUS_SUCCESS;
}

void hfp_cleanup(void)
{
	API_TRACE("HF Cleanup");
	blued_hfp_interface = NULL;
}

oal_status_t hfp_connect(bt_address_t *device_address)
{
	int result = OAL_STATUS_SUCCESS;
	bt_status_t status;
	bdstr_t bdstr;
	API_TRACE("HF Connect");

	CHECK_OAL_HF_ENABLED();
	OAL_CHECK_PARAMETER(device_address, return);

	BT_INFO("BT HF Address: %s", bdt_bd2str(device_address, &bdstr));
	/* Call connect function of Bluedroid*/
	status = blued_hfp_interface->connect((bt_bdaddr_t *)device_address);
	if ((status != BT_STATUS_SUCCESS) && (status != BT_STATUS_DONE)) {
		BT_ERR("Connection could not be established, err: %s", status2string(status));;
		result =  convert_to_oal_status(status);
	}
	return result;
}

oal_status_t hfp_disconnect(bt_address_t *device_address)
{
	int result = OAL_STATUS_SUCCESS;
	bdstr_t bdstr;
	bt_status_t status;
	API_TRACE("HF Disconnect");

	CHECK_OAL_HF_ENABLED();
	OAL_CHECK_PARAMETER(device_address, return);

	BT_INFO("BT HF Address: %s", bdt_bd2str(device_address, &bdstr));
	/* call Disconnect function of Bluedroid */
	status = blued_hfp_interface->disconnect((bt_bdaddr_t *)device_address);
	if ((status != BT_STATUS_SUCCESS) && (status != BT_STATUS_DONE)) {
		BT_ERR("OAL, Disconnection failed err: %s", status2string(status));
		result =  convert_to_oal_status(status);
	}
	return result;
}

oal_status_t hfp_connect_audio(bt_address_t *device_address)
{
	bt_status_t status;
	int result = OAL_STATUS_SUCCESS;
	bdstr_t bdstr;
	API_TRACE("Connect HF Audio");

	CHECK_OAL_HF_ENABLED();
	OAL_CHECK_PARAMETER(device_address, return);

	BT_INFO("BT HF Address: %s", bdt_bd2str(device_address, &bdstr));
	status = blued_hfp_interface->connect_audio((bt_bdaddr_t *)device_address);

	if (status != BT_STATUS_SUCCESS) {
		BT_ERR("Start stream failed err: %s", status2string(status));
		result = convert_to_oal_status(status);
	}
	return result;
}

oal_status_t hfp_disconnect_audio(bt_address_t *device_address)
{
	bt_status_t status;
	int result = OAL_STATUS_SUCCESS;
	bdstr_t bdstr;
	API_TRACE("DisConnect HF Audio");

	CHECK_OAL_HF_ENABLED();

	OAL_CHECK_PARAMETER(device_address, return);

	BT_INFO("BT HF Address: %s", bdt_bd2str(device_address, &bdstr));
	status = blued_hfp_interface->disconnect_audio((bt_bdaddr_t *)device_address);

	if (status != BT_STATUS_SUCCESS) {
		BT_ERR("Start stream failed err: %s", status2string(status));
		result = convert_to_oal_status(status);
	}
	return result;
}

static void cb_hfp_connection_state(bthf_connection_state_t state, bt_bdaddr_t *bd_addr)
{
	int event_type;
	bdstr_t bdstr;
	bt_address_t * event_data = NULL;
	event_data = g_new0(bt_address_t, 1);
	memcpy(event_data->addr, bd_addr->address, BT_ADDRESS_BYTES_NUM);

	if (bd_addr == NULL) {
		BT_ERR("Address is NULL");
		if (event_data)
			g_free(event_data);
		return;
	}

	BT_INFO("HFP Profile state = [%d],  BT Address = [%s]", state, bdt_bd2str((bt_address_t*)bd_addr, &bdstr));
	switch (state) {
	case BTHF_CONNECTION_STATE_DISCONNECTED: {
		 event_type = OAL_EVENT_HFP_DISCONNECTED;
		 break;
	}
	case BTHF_CONNECTION_STATE_CONNECTING: {
		event_type = OAL_EVENT_HFP_CONNECTING;
		break;
	}
	case BTHF_CONNECTION_STATE_CONNECTED: {
		event_type = OAL_EVENT_HFP_CONNECTED;
		break;
	}
	case BTHF_CONNECTION_STATE_DISCONNECTING: {
		event_type = OAL_EVENT_HFP_DISCONNECTING;
		break;
	}
	default: {
		BT_INFO("Invalid state");
		if (event_data)
			g_free(event_data);
		return;
	}
	}
	send_event_bda_trace(event_type, event_data, sizeof(bt_address_t), (bt_address_t*)bd_addr);
}

static void cb_hfp_audio_connection_state(bthf_audio_state_t state, bt_bdaddr_t *bd_addr)
{
	int event_type;
	bdstr_t bdstr;
	bt_address_t * event_data = NULL;
	event_data = g_new0(bt_address_t, 1);
	memcpy(event_data->addr, bd_addr->address, BT_ADDRESS_BYTES_NUM);

	if (bd_addr == NULL) {
		BT_ERR("Address is NULL");
		if (event_data)
			g_free(event_data);
		return;
	}
	BT_INFO("HFP Audio state = [%d],  BT Address = [%s]", state, bdt_bd2str((bt_address_t*)bd_addr, &bdstr));

	switch (state) {
	case BTHF_AUDIO_STATE_DISCONNECTED: {
		BT_INFO("HFP Audio(SCO) DisConnected");
		event_type = OAL_EVENT_HFP_AUDIO_DISCONNECTED;
		break;
	}
	case BTHF_AUDIO_STATE_CONNECTED: {
		BT_INFO("HFP Audio(SCO) Connected");
		event_type = OAL_EVENT_HFP_AUDIO_CONNECTED;
		break;
	}
	case BTHF_AUDIO_STATE_CONNECTING: {
		BT_INFO("HFP Audio(SCO) Connecting");
		event_type = OAL_EVENT_HFP_AUDIO_CONNECTING;
		break;
	}
	case BTHF_AUDIO_STATE_DISCONNECTING: {
		BT_INFO("HFP Audio(SCO) DisConnecting");
		event_type = OAL_EVENT_HFP_AUDIO_DISCONNECTING;
		break;
	}
	default: {
		BT_INFO("Invalid state");
		if (event_data)
			g_free(event_data);
		return;
	}
	}
	send_event_bda_trace(event_type, event_data, sizeof(bt_address_t), (bt_address_t*)bd_addr);
}
