/*
 * Copyright (c) 2016 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Contact: Anupam Roy <anupam.r@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <gio/gio.h>
#include <glib.h>
#include <dlog.h>
#include <string.h>
#include <vconf.h>

#include "bt-internal-types.h"
#include "bt-service-common.h"
#include "bt-service-event.h"
#include "bt-service-core-adapter.h"
#include "bt-service-core-adapter-le.h"
#include "bt-service-event-receiver.h"
#include "bt-service-gatt.h"
#include "bt-service-util.h"
#include "bt-service-core-device.h"

#include <oal-hardware.h>
#include <oal-manager.h>
#include <oal-event.h>
#include <oal-adapter-mgr.h>
#include <oal-device-mgr.h>
#include <oal-gatt.h>

#define BT_UUID_128 16
#define BT_ADV_DEFAULT_TIMEOUT 0
#define BT_ADV_DEFAULT_TX_POWER 4
#define BT_ADV_DEFAULT_CHANNEL_MAP 0

static const char BASE_UUID_CONVERTED[BT_UUID_128] = {
	0xfb, 0x34, 0x9b, 0x5f, 0x80, 0x00, 0x00, 0x80,
	0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

typedef struct {
	int adv_inst_max;
	int rpa_offloading;
	int max_filter;
	int le_2m_phy;
	int le_coded_phy;
} bt_adapter_le_feature_info_t;

/* Set Default values */
static bt_adapter_le_feature_info_t le_feature_info = {1, 0, 0, 0, 0};

typedef struct {
	int adv_handle;
	char *sender;
	int data_len;
	unsigned char data[31];
	gboolean is_adv;/* Adv or Scan Response: Only applicable if is_data_set_pending is TRUE */
	gboolean is_data_set_pending; /* Data Set pending or Adv enable request at one time */
	gboolean is_custom_adv; /* If Adv enable is custom adv enable request */
	bluetooth_advertising_params_t params; /* Adv Parameters */
} bt_pending_adv_data_set_t;

static guint timer_id = 0;

static GSList *adv_data_pending_list = NULL;
static bt_le_status_t adapter_le_state = BT_LE_DEACTIVATED;
static bt_le_discovery_state_t adapter_le_discovery_state = LE_DISCOVERY_STOPPED;

/******************************************* LE Scan *********************************************/
#define BT_LE_SCAN_INTERVAL_MIN 2.5
#define BT_LE_SCAN_INTERVAL_MAX 10240
#define BT_LE_SCAN_WINDOW_MIN 2.5
#define BT_LE_SCAN_WINDOW_MAX 10240

#define BT_ADV_INTERVAL_SPLIT 0.625

typedef struct {
	char *sender;
	GSList *filter_list;
	gboolean is_scanning;
} bt_adapter_le_scanner_t;

GSList *scanner_list = NULL;
static gboolean is_le_set_scan_parameter = FALSE;
static gboolean is_le_scanning = FALSE;

static int g_gatt_client_id = 0;

/******************************************* LE Scan *********************************************/

void _bt_adapter_set_le_status(bt_le_status_t status)
{
	BT_INFO("adapter_le_state changed [%d] -> [%d]", adapter_le_state, status);
	adapter_le_state = status;
}

bt_le_status_t _bt_adapter_get_le_status(void)
{
	return adapter_le_state;
}


/* Internal functions of core adapter service */
static void __bt_le_handle_pending_requests(int service_function, void *user_data, unsigned int size)
{
	GSList *l;
	GArray *out_param;
	invocation_info_t *req_info;
	BT_INFO("+");

	/* Get method invocation context */
	for (l = _bt_get_invocation_list(); l != NULL; l = g_slist_next(l)) {
		req_info = l->data;
		if (req_info == NULL || req_info->service_function != service_function)
			continue;

		/* Create out param */
		out_param = g_array_new(FALSE, FALSE, sizeof(gchar));

		switch (service_function) {
		case BT_ENABLE_ADAPTER_LE:
		case BT_DISABLE_ADAPTER_LE: {
			gboolean done = TRUE;
			g_array_append_vals(out_param, &done, sizeof(gboolean));
			break;
		}
		default:
			BT_ERR("Unknown service function[%d]", service_function);
		}

		_bt_service_method_return(req_info->context, out_param, req_info->result);
		g_array_free(out_param, TRUE);
		/* Now free invocation info for this request*/
		_bt_free_info_from_invocation_list(req_info);
	}
}


/* Request return handlings */
static gboolean __bt_le_post_set_enabled(gpointer user_data)
{
	BT_INFO("__bt_adapter_post_set_enabled>>");

	/* Add LE enabled post processing codes */

	return FALSE;
}


static gboolean __bt_le_post_set_disabled(gpointer user_data)
{
	BT_INFO("_bt_adapter_post_set_disabled>>");

	/* Add LE disabled post processing codes */

	return FALSE;
}

static void __bt_le_update_bt_enabled(void)
{
	int result = BLUETOOTH_ERROR_NONE;
	BT_ERR("_bt_adapter_update_bt_enabled>>");
	/* Update Bluetooth Status to notify other modules */
	if (vconf_set_int(VCONFKEY_BT_LE_STATUS, VCONFKEY_BT_LE_STATUS_ON) != 0)
		BT_ERR("Set vconf failed\n");

	/* TODO:Add timer function to handle any further post processing */
	g_idle_add((GSourceFunc)__bt_le_post_set_enabled, NULL);

	/*Return BT_ADAPTER_ENABLE Method invocation context */
	__bt_le_handle_pending_requests(BT_ENABLE_ADAPTER_LE, NULL, 0);
	/*Send BT Enabled event to application */
	_bt_send_event(BT_LE_ADAPTER_EVENT, BLUETOOTH_EVENT_LE_ENABLED,
			g_variant_new("(i)", result));
}

static void __bt_le_update_bt_disabled(void)
{
	int result = BLUETOOTH_ERROR_NONE;
	BT_INFO("_bt_adapter_update_bt_disabled>>");

	int power_off_status = 0;
	int ret;

	/* Update the vconf LE status in normal Deactivation case only */
	ret = vconf_get_int(VCONFKEY_SYSMAN_POWER_OFF_STATUS, &power_off_status);
	BT_DBG("ret : %d, power_off_status : %d", ret, power_off_status);

	/* Update LE Status to notify other modules */
	if (vconf_set_int(VCONFKEY_BT_LE_STATUS, VCONFKEY_BT_LE_STATUS_OFF) != 0)
		BT_ERR("Set vconf failed");

	/* TODO:Add timer function to handle any further post processing */
	g_idle_add((GSourceFunc)__bt_le_post_set_disabled, NULL);

	/* Return BT_ADAPTER_DISABLE Method invocation context */
	__bt_le_handle_pending_requests(BT_DISABLE_ADAPTER_LE, NULL, 0);

	/* Send BT Disabled event to application */
	_bt_send_event(BT_LE_ADAPTER_EVENT, BLUETOOTH_EVENT_LE_DISABLED,
			g_variant_new("(i)", result));
}


static void __bt_set_le_scan_status(gboolean mode)
{
	is_le_scanning = mode;
}

gboolean _bt_is_le_scanning(void)
{
	return is_le_scanning;
}

bt_adapter_le_scanner_t* __bt_find_scanner_from_list(const char *sender)
{
	GSList *l;
	bt_adapter_le_scanner_t *scanner;

	for (l = scanner_list; l != NULL; l = g_slist_next(l)) {
		scanner = l->data;
		if (g_strcmp0(scanner->sender, sender) == 0)
			return scanner;
	}

	return NULL;
}

/* Event handlers */
static void __bt_adapter_le_handle_pending_request_info(int result,
		int service_function, void *param, unsigned int size)
{
	GSList *l;
	GArray *out_param;
	invocation_info_t *req_info = NULL;
	BT_DBG("+");

	for (l = _bt_get_invocation_list(); l != NULL; ) {
		req_info = l->data;
		l = g_slist_next(l);
		if (req_info == NULL || req_info->service_function != service_function)
			continue;

		switch (service_function) {
		case BT_SET_ADVERTISING:
		case BT_SET_CUSTOM_ADVERTISING:
		case BT_SET_ADVERTISING_DATA:
		case BT_SET_SCAN_RESPONSE_DATA: {
			int *saved_handle;
			bt_pending_adv_data_set_t *data;

			BT_DBG("Service function: [%d]", service_function);
			ret_if(param == NULL);

			saved_handle = (int*)req_info->user_data;
			data = (bt_pending_adv_data_set_t*)param;
			BT_DBG("Current Sender [%s] Current Handle [%d]", data->sender, data->adv_handle);

			if (!g_strcmp0(req_info->sender, data->sender) && (*saved_handle == data->adv_handle)) {
				BT_DBG("Requester found [%s] ADV Handle [%d]", req_info->sender, *saved_handle);
				out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
				g_array_append_vals(out_param, &data->adv_handle, sizeof(int));
				_bt_service_method_return(req_info->context, out_param, result);

				/* Free data */
				g_free(req_info->user_data);
				_bt_free_info_from_invocation_list(req_info);
				g_array_free(out_param, TRUE);
			}
			break;
		}
		case BT_START_LE_DISCOVERY: {
			bt_adapter_le_scanner_t *scanner;

			BT_DBG("Request Sender: [%s]", req_info->sender);
			if (BLUETOOTH_ERROR_NONE != result) {
				scanner = __bt_find_scanner_from_list(req_info->sender);
				if (scanner && scanner->is_scanning)
					scanner->is_scanning = FALSE;
			}

			out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
			_bt_service_method_return(req_info->context, out_param, result);
			_bt_free_info_from_invocation_list(req_info);
			g_array_free(out_param, TRUE);
			break;
		}
		case BT_STOP_LE_DISCOVERY: {
			BT_DBG("Request Sender: [%s]", req_info->sender);
			out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
			_bt_service_method_return(req_info->context, out_param, result);
			_bt_free_info_from_invocation_list(req_info);
			g_array_free(out_param, TRUE);
			break;
		}
		default:
			break;
		}
	}
	BT_DBG("-");
}

static void __bt_le_handle_server_instance_registered(event_gatts_register_t *data)
{
	int adv_handle;
	GSList *l;
	char uuid_string[BLUETOOTH_UUID_STRING_MAX];
	bluetooth_advertising_data_t adv;
	bluetooth_scan_resp_data_t scan_rsp;
	int result = BLUETOOTH_ERROR_NONE;
	bt_pending_adv_data_set_t *info = NULL;

	memset(adv.data, 0, sizeof(adv.data));
	memset(scan_rsp.data, 0, sizeof(scan_rsp.data));

	_bt_uuid_to_string(&(data->server_uuid), uuid_string);
	BT_INFO("Instance ID is Intialized [%d] UUID initialized [%s]", data->server_inst, uuid_string);

	/* Get sender and adv handle for the server instance */
	_bt_get_adv_handle_from_instance(data->server_inst, &adv_handle);

		/* Check if pending Adv Data set or Scan response data set request is pending*/
		for (l = adv_data_pending_list; l != NULL; l = l->next) {
			info = l->data;

			if (info && (adv_handle == info->adv_handle)) {
				if (info->is_data_set_pending) {
					BT_INFO("LE Data set request is pending");
					if (info->is_adv) {
						memcpy(&adv.data, info->data, info->data_len);
						result = _bt_set_advertising_data(info->sender,
							info->adv_handle, &adv, info->data_len, FALSE);
						if (result != BLUETOOTH_ERROR_NONE) {
							__bt_adapter_le_handle_pending_request_info(result,
								BT_SET_ADVERTISING_DATA,
								(void*)info, sizeof(bt_pending_adv_data_set_t));
						}
						goto data_free;
					} else {
						memcpy(&scan_rsp.data, info->data, info->data_len);
						result = _bt_set_scan_response_data(info->sender,
								info->adv_handle, &scan_rsp, info->data_len, FALSE);
						if (result != BLUETOOTH_ERROR_NONE) {
							__bt_adapter_le_handle_pending_request_info(result,
								BT_SET_SCAN_RESPONSE_DATA, (void*)info,
								sizeof(bt_pending_adv_data_set_t));
						}
						goto data_free;
					}
				} else {
					BT_INFO("LE Enable Adv request is pending");
					if (info->is_custom_adv) {
						result = _bt_set_custom_advertising(info->sender, info->adv_handle,
									TRUE, &info->params, FALSE/*Reserved Slot*/);
						if (result != BLUETOOTH_ERROR_NONE) {
							__bt_adapter_le_handle_pending_request_info(result,
								BT_SET_CUSTOM_ADVERTISING,
								(void*)info, sizeof(bt_pending_adv_data_set_t));
						}
						goto data_free;
					} else {
						result = _bt_set_advertising(info->sender, info->adv_handle,
								TRUE, FALSE/*Reserved Slot*/);
						if (result != BLUETOOTH_ERROR_NONE) {
							__bt_adapter_le_handle_pending_request_info(result,
								BT_SET_ADVERTISING,
								(void*)info, sizeof(bt_pending_adv_data_set_t));
						}
						goto data_free;
				}
			}
		}
	}
	BT_DBG("-");
	return;

data_free:
	adv_data_pending_list = g_slist_remove(adv_data_pending_list, info);
	g_free(info->sender);
	g_free(info);
}

static void __bt_le_multi_advertising_enabled(event_ble_multiadv_status *event)
{
	char *sender;
	int adv_handle;
	bt_pending_adv_data_set_t *info = NULL;
	GVariant *param = NULL;
	int result = BLUETOOTH_ERROR_NONE;

	sender = _bt_get_sender_and_handle(event->server_inst, &adv_handle);
	if (sender == NULL) {
		BT_ERR("Abnormal!!");
	} else {
		if (event->status != OAL_STATUS_SUCCESS)
			result = BLUETOOTH_ERROR_INTERNAL;
		info = g_malloc0(sizeof(bt_pending_adv_data_set_t));
		info->sender = sender;
		info->adv_handle = adv_handle;
		__bt_adapter_le_handle_pending_request_info(result,
				BT_SET_CUSTOM_ADVERTISING,
				(void*)info, sizeof(bt_pending_adv_data_set_t));
		__bt_adapter_le_handle_pending_request_info(result,
				BT_SET_ADVERTISING,
				(void*)info, sizeof(bt_pending_adv_data_set_t));

		/* Send event */
		param = g_variant_new("(ii)", result, info->adv_handle);
		_bt_send_event_to_dest(info->sender, BT_ADAPTER_EVENT, BLUETOOTH_EVENT_ADVERTISING_STARTED, param);

		/* Free data */
		g_free(info->sender);
		g_free(info);
	}
}

static void __bt_le_multi_advertising_disabled(event_ble_multiadv_status *event)
{
	char *sender;
	int adv_handle;
	bt_pending_adv_data_set_t *info = NULL;
	GVariant *param = NULL;
	int result = BLUETOOTH_ERROR_NONE;
	sender = _bt_get_sender_and_handle(event->server_inst, &adv_handle);

	if (sender == NULL) {
		BT_INFO("Means application containing the adv info is already freed!!");
	} else {
		if (event->status != OAL_STATUS_SUCCESS)
			result = BLUETOOTH_ERROR_INTERNAL;
		info = g_malloc0(sizeof(bt_pending_adv_data_set_t));
		info->sender = sender;
		info->adv_handle = adv_handle;
		__bt_adapter_le_handle_pending_request_info(result,
				BT_SET_CUSTOM_ADVERTISING,
				(void*)info, sizeof(bt_pending_adv_data_set_t));
		__bt_adapter_le_handle_pending_request_info(result,
				BT_SET_ADVERTISING,
				(void*)info, sizeof(bt_pending_adv_data_set_t));
		/* Send event */
		param = g_variant_new("(ii)", result, info->adv_handle);
		_bt_send_event_to_dest(info->sender, BT_ADAPTER_EVENT, BLUETOOTH_EVENT_ADVERTISING_STOPPED, param);

		/* Free allocated slot or server instance from stack to be used for other advertisng */
		result = _bt_unregister_server_instance(sender, adv_handle);

		/* Free data */
		g_free(info->sender);
		g_free(info);
	}
}

static void __bt_le_multi_advertising_set_data(event_ble_multiadv_status *event)
{
	char *sender = NULL;
	int adv_handle;
	bt_pending_adv_data_set_t *info = NULL;
	int result = BLUETOOTH_ERROR_NONE;
	sender = _bt_get_sender_and_handle(event->server_inst, &adv_handle);

	if (sender == NULL) {
		BT_ERR("Abnormal!!");
	} else {
		BT_DBG("Sender [%s], adv handle [%d]",  sender, adv_handle);
		if (event->status != OAL_STATUS_SUCCESS)
			result = BLUETOOTH_ERROR_INTERNAL;

		info = g_malloc0(sizeof(bt_pending_adv_data_set_t));
		info->sender = sender;
		info->adv_handle = adv_handle;
		BT_DBG("Sender [%s], adv handle [%d]",  info->sender, info->adv_handle);
		__bt_adapter_le_handle_pending_request_info(result,
				BT_SET_SCAN_RESPONSE_DATA,
				(void*)info, sizeof(bt_pending_adv_data_set_t));
		__bt_adapter_le_handle_pending_request_info(result,
				BT_SET_ADVERTISING_DATA,
				(void*)info, sizeof(bt_pending_adv_data_set_t));
		g_free(info->sender);
		g_free(info);
	}
}

static int __get_advertisement_data(unsigned char *dest, unsigned char *src)
{
	int length = 0;
	int i = 0;

	while (i <= 31) {
		int len = src[i];

		if ((len == 0) || (31 < (i + len))) {
			BT_INFO("length: %d", length);
			memcpy(dest, src, length);
			break;
		}

		length += (len + 1);
		i += (len + 1);
	}

	return length;
}

static void __bt_le_handle_device_found(event_ble_scan_result_info *scan_result)
{
	int result = BLUETOOTH_ERROR_NONE;
	bt_adapter_le_scanner_t *scanner = NULL;
	char address[BT_ADDRESS_STRING_SIZE];
	int addr_type = 0x01; /* TODO: Need to get correct address type */
	unsigned char adv_ind_data[31];
	unsigned char scan_resp_data[31];
	int adv_ind_len = 0;
	int scan_resp_len = 0;
	char *tmp_str = NULL;
	GVariant *scan_data_param;
	GVariant *adv_data_param;
	GVariant *param;
	GSList *l;

	BT_DBG("+");
	ret_if(NULL == scan_result);

	_bt_convert_addr_type_to_string(address, scan_result->address.addr);

	memset(adv_ind_data, 0x00, sizeof(adv_ind_data));
	adv_ind_len = __get_advertisement_data(
			adv_ind_data, &(scan_result->adv_data[0]));

	memset(scan_resp_data, 0x00, sizeof(scan_resp_data));
	scan_resp_len = __get_advertisement_data(
			scan_resp_data, &(scan_result->adv_data[adv_ind_len]));

	BT_INFO("Address: %s, RSSI: %d, adv_ind_len: %d, scan_resp_len: %d",
			address, scan_result->rssi, adv_ind_len, scan_resp_len);

	tmp_str = _bt_service_convert_hex_to_string(scan_result->adv_data, 62);
	BT_INFO("Complete Data: [%s]", tmp_str);
	g_free(tmp_str);
	tmp_str = NULL;

	tmp_str = _bt_service_convert_hex_to_string(adv_ind_data, adv_ind_len);
	BT_INFO("Adv Ind Data: [%s]", tmp_str);
	g_free(tmp_str);
	tmp_str = NULL;

	tmp_str = _bt_service_convert_hex_to_string(scan_resp_data, scan_resp_len);
	BT_INFO("Scan resp Data: [%s]", tmp_str);
	g_free(tmp_str);
	tmp_str = NULL;

	for (l = scanner_list; l != NULL; l = g_slist_next(l)) {
		scanner = l->data;
		if (scanner->is_scanning == FALSE)
			continue;

		adv_data_param = g_variant_new_from_data((const GVariantType *)"ay",
				adv_ind_data, adv_ind_len, TRUE, NULL, NULL);
		scan_data_param = g_variant_new_from_data((const GVariantType *)"ay",
				scan_resp_data, scan_resp_len, TRUE, NULL, NULL);

		param = g_variant_new("(isnnn@ayn@ay)",
				result, address,
				addr_type, scan_result->rssi,
				adv_ind_len, adv_data_param,
				scan_resp_len, scan_data_param);

		_bt_send_event_to_dest(scanner->sender, BT_LE_ADAPTER_EVENT,
				BLUETOOTH_EVENT_REMOTE_LE_DEVICE_FOUND, param);
	}

	BT_DBG("-");
}

static void __bt_le_state_change_callback(int bt_status)
{
	BT_INFO("__bt_le_state_change_callback: status [%d]", bt_status);

	switch (bt_status) {
	case BT_LE_DEACTIVATED:
		_bt_adapter_set_le_status(bt_status);

		/* Add Adapter disabled post processing codes */
		__bt_le_update_bt_disabled();
		break;
	case BT_LE_ACTIVATED:
		_bt_adapter_set_le_status(bt_status);

		/* Add Adapter enabled post processing codes */
		if (timer_id > 0) {
			BT_DBG("g_source is removed");
			g_source_remove(timer_id);
			timer_id = 0;
		}
		__bt_le_update_bt_enabled();
		break;
	default:
		BT_ERR("Incorrect Bluetooth adapter state changed status");
		break;

	}
}

static void __bt_le_event_handler(int event_type, gpointer event_data)
{
	BT_INFO("OAL event = 0x%x, \n", event_type);

	switch (event_type) {
	case OAL_EVENT_BLE_ENABLED: {
		__bt_le_state_change_callback(BT_LE_ACTIVATED);
		break;
	}
	case OAL_EVENT_BLE_DISABLED: {
		__bt_le_state_change_callback(BT_LE_DEACTIVATED);
		break;
	}
	case OAL_EVENT_BLE_SERVER_INSTANCE_INITIALISED: {
		BT_INFO("OAL Event: Server Instance Registered");
		__bt_le_handle_server_instance_registered((event_gatts_register_t *)event_data);
		break;
	}
	case OAL_EVENT_BLE_ADVERTISING_STARTED: {
		BT_INFO("OAL Event: Legacy Advertising Enabled: Not Supported!!");
		break;
	}
	case OAL_EVENT_BLE_ADVERTISING_STOPPED: {
		BT_INFO("OAL Event: Legacy Advertising Disabled: Not Supported!!");
		break;
	}
	case OAL_EVENT_BLE_MULTI_ADVERTISING_ENABLE: {
		BT_INFO("OAL Event: Advertising Enabled");
		__bt_le_multi_advertising_enabled((event_ble_multiadv_status *)event_data);
		break;
	}
	case OAL_EVENT_BLE_MULTI_ADVERTISING_DISABLE: {
		BT_INFO("OAL Event: Advertising Disabled");
		__bt_le_multi_advertising_disabled((event_ble_multiadv_status *)event_data);
		break;
	}
	case OAL_EVENT_BLE_MULTI_ADVERTISING_SET_INST_DATA: {
		BT_INFO("OAL Event: Advertising Data set successfully");
		__bt_le_multi_advertising_set_data((event_ble_multiadv_status *)event_data);
		break;
	}
	case OAL_EVENT_BLE_MULTI_ADVERTISING_UPDATE: {
			BT_INFO("OAL Event: Advertising Params updated");
		break;
	}
	case OAL_EVENT_BLE_DISCOVERY_STARTED: {
		__bt_set_le_scan_status(TRUE);

		__bt_adapter_le_handle_pending_request_info(
				BLUETOOTH_ERROR_NONE,
				BT_START_LE_DISCOVERY, NULL, 0);
		break;
	}
	case OAL_EVENT_BLE_DISCOVERY_STOPPED: {
		if (!_bt_is_le_scanning()) {
			BT_ERR("LE discovery start failed");
			__bt_adapter_le_handle_pending_request_info(
					BLUETOOTH_ERROR_INTERNAL,
					BT_START_LE_DISCOVERY, NULL, 0);
			break;
		}

		__bt_set_le_scan_status(FALSE);
		__bt_adapter_le_handle_pending_request_info(
				BLUETOOTH_ERROR_NONE,
				BT_STOP_LE_DISCOVERY, NULL, 0);
		break;
	}
	case OAL_EVENT_BLE_REMOTE_DEVICE_FOUND: {
		event_ble_scan_result_info *scan_result = event_data;
		BT_INFO("OAL Event: OAL_EVENT_BLE_REMOTE_DEVICE_FOUND");
		__bt_le_handle_device_found(scan_result);
		break;
	}
	case OAL_EVENT_GATTC_REGISTRATION: {
		event_gattc_register_t *gattc_event = event_data;
		char uuid_str[BLUETOOTH_UUID_STRING_MAX];
		char *default_uuid_string;

		default_uuid_string = _bt_gatt_get_default_gatt_client_uuid();
		_bt_uuid_to_string((service_uuid_t*)&(gattc_event->client_uuid), uuid_str);
		BT_INFO("default UUID [%s] current registered uuid [%s]",
				default_uuid_string, uuid_str);
		if (g_strcmp0(uuid_str, default_uuid_string)) {
			BT_INFO("This is not the default GATT client that is registered");

			g_free(default_uuid_string);
			break;
		}
		BT_INFO("GATT CLient instance registered is default client: ID [%d]",
				gattc_event->client_if);
		g_free(default_uuid_string);

		g_gatt_client_id = gattc_event->client_if;
		break;
	}
	case OAL_EVENT_BLE_LOCAL_FEATURES: {
		event_adapter_le_features_t *le_features = event_data;

		le_feature_info.le_2m_phy = le_features->le_2m_phy_support;
		le_feature_info.le_coded_phy = le_features->le_coded_phy_support;

		BT_INFO("Adapter LE 2M PHY Support [%s]", le_feature_info.le_2m_phy ? "TRUE" : "FALSE");
		BT_INFO("Adapter LE CODED PHY Support [%s]", le_feature_info.le_coded_phy ? "TRUE" : "FALSE");

		break;
	}
	default:
		break;
	}
}

int _bt_le_init(void)
{
	BT_DBG("+");
	/* Register LE event handler */
	_bt_service_register_event_handler_callback(BT_ADAPTER_LE_MODULE, __bt_le_event_handler);
	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

void _bt_le_deinit(void)
{
	BT_DBG("+");
	/* Un-register LE event handler */
	_bt_service_unregister_event_handler_callback(BT_ADAPTER_LE_MODULE);
	BT_DBG("-");
}

static void __bt_le_update_discovery_status(bt_adapter_discovery_state_t status)
{
	BT_INFO("adapter_discovery_status changed [%d] -> [%d]", adapter_le_discovery_state, status);
	adapter_le_discovery_state = status;
}

static int __bt_le_state_handle_request(gboolean enable)
{
	int result = BLUETOOTH_ERROR_NONE;
	BT_DBG("");

	switch (_bt_adapter_get_le_status()) {
	case BT_LE_ACTIVATING: {
		BT_INFO("LE is currently in activating state, state [%d]",
				_bt_adapter_get_le_status());
		if (enable) {
			return BLUETOOTH_ERROR_IN_PROGRESS;
		} else {
			if (adapter_le_discovery_state == LE_DISCOVERY_STARTED ||
					adapter_le_discovery_state == LE_DISCOVERY_STARTING) {
				/*TODO Stop Discovery*/
				__bt_le_update_discovery_status(FALSE);
			}
			result = le_disable();	//change
			if (result != OAL_STATUS_SUCCESS) {
				BT_ERR("LE_enable failed: [%d]", result);
				result = BLUETOOTH_ERROR_INTERNAL;
				/*TODO: perform if anything more needs to be done to handle failure */
			} else {
				/* TODO: To be handled */
				_bt_adapter_set_le_status(BT_LE_DEACTIVATING);
				result = BLUETOOTH_ERROR_NONE;
			}
		}
		break;
	}
	case BT_LE_ACTIVATED: {
		BT_INFO("LE is currently in activated state, state [%d]",
				_bt_adapter_get_le_status());
		if (enable) {
			return BLUETOOTH_ERROR_DEVICE_ALREADY_ENABLED;
		} else {
			if (adapter_le_discovery_state == LE_DISCOVERY_STARTED ||
					adapter_le_discovery_state == LE_DISCOVERY_STARTING) {
				/*TODO Stop Discovery*/
				__bt_le_update_discovery_status(FALSE);
			}
			result = le_disable();
			if (result != OAL_STATUS_SUCCESS) {
				BT_ERR("LE_enable failed: [%d]", result);
				result = BLUETOOTH_ERROR_INTERNAL;
				/*TODO: perform if anything more needs to be done to handle failure */
			} else {
				/* TODO: To be handled */
				_bt_adapter_set_le_status(BT_LE_DEACTIVATING);
				result = BLUETOOTH_ERROR_NONE;
			}
		}
		break;
	}
	case BT_LE_DEACTIVATING: {
		BT_INFO("LE is currently in deactivating state, state [%d]",
				_bt_adapter_get_le_status());
		if (!enable) {
			return BLUETOOTH_ERROR_IN_PROGRESS;

		} else {
			result = le_enable();
			if (result != OAL_STATUS_SUCCESS && result != OAL_STATUS_PENDING) {
				BT_ERR("LE_enable failed: [%d]", result);
				le_disable();
				result = BLUETOOTH_ERROR_INTERNAL;
				/*TODO: perform if anything more needs to be done to handle failure */
			} else {
				/* TODO: To be handled */
				_bt_adapter_set_le_status(BT_LE_ACTIVATING);
				result = BLUETOOTH_ERROR_NONE;
			}
		}
		break;
	}
	case BT_LE_DEACTIVATED: {
		BT_INFO("LE is currently in deactivated state, state [%d]",
				_bt_adapter_get_le_status());
		if (!enable) {
			return BLUETOOTH_ERROR_DEVICE_NOT_ENABLED;
		} else {
			result = le_enable();
			if (result != OAL_STATUS_SUCCESS && result != OAL_STATUS_PENDING) {
				BT_ERR("LE_enable failed: [%d]", result);
				le_disable();
				result = BLUETOOTH_ERROR_INTERNAL;
				/*TODO: perform if anything more needs to be done to handle failure */
			} else {
				/* TODO: To be handled */
				_bt_adapter_set_le_status(BT_LE_ACTIVATING);
				result = BLUETOOTH_ERROR_NONE;
			}
		}
		break;
	}
	default:
		BT_ERR("Unknown state: %d", _bt_adapter_get_le_status());
		break;
	}

	if (enable && result == BLUETOOTH_ERROR_NONE) {
		/* Adapter enable request is successful, setup event handlers */
		_bt_service_register_event_handler_callback(
				BT_ADAPTER_LE_MODULE, __bt_le_event_handler);
		_bt_device_state_handle_callback_set_request();
	}
	return result;
}

int _bt_enable_adapter_le(void)
{
	return __bt_le_state_handle_request(TRUE);
}

int _bt_disable_adapter_le(void)
{
	return __bt_le_state_handle_request(FALSE);
}


static gboolean __bt_is_factory_test_mode(void)
{
	int mode = 0;

	if (vconf_get_bool(VCONFKEY_BT_DUT_MODE, &mode)) {
		BT_ERR("Get the DUT Mode fail");
		return TRUE;
	}

	if (mode != FALSE) {
		BT_INFO("DUT Test Mode !!");
		return TRUE;
	}

	return FALSE;
}

int _bt_set_advertising(const char *sender, int adv_handle, gboolean enable, gboolean use_reserved_slot)
{
	BT_CHECK_PARAMETER(sender, return);
	int server_instance;
	bt_pending_adv_data_set_t *info = NULL;
	int result = OAL_STATUS_SUCCESS;
	BT_DBG("+");

	if (__bt_is_factory_test_mode()) {
		BT_ERR("Unable to start advertising in factory binary !!");
		return BLUETOOTH_ERROR_NOT_SUPPORT;
	}

	server_instance = _bt_get_allocated_server_instance(sender, adv_handle, use_reserved_slot);

	if (server_instance == -1) {
		BT_DBG("No available slot for the current sender and handle combination");

		if (enable == FALSE) {
			BT_ERR("Advertising not even enabled on adv handle [%d] sender [%s]", adv_handle, sender);
			return BLUETOOTH_ERROR_NOT_IN_OPERATION;
		}

		server_instance = _bt_is_sender_gatt_server_with_no_adv(sender, adv_handle);

		if (server_instance == -1) {
			/* Internal Logic to register server instance if not initialized, store adv handle for future use */
			if (_bt_register_server_instance(sender, adv_handle) != BLUETOOTH_ERROR_NONE)
				return BLUETOOTH_ERROR_INTERNAL;
			else {
				/* Allocate a pending structure and mark Adv data set pending */
				info = g_malloc0(sizeof(bt_pending_adv_data_set_t));
				info->adv_handle = adv_handle;
				info->sender = g_strdup(sender);
				info->is_data_set_pending = FALSE;
				adv_data_pending_list = g_slist_append(adv_data_pending_list, info);
				return BLUETOOTH_ERROR_NONE;
			}
		}
	}

	if (enable)
		result = adapter_ble_multi_adv_enable(server_instance);
	else
		result = adapter_ble_multi_adv_disable(server_instance);
	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("OAL API adapter_ble_multi_adv_enable Fail %d", result);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	/* Update adv handle in table */
	_bt_update_adv_handle(sender, adv_handle);

	return BLUETOOTH_ERROR_NONE;
}

int _bt_set_custom_advertising(const char *sender, int adv_handle,
		gboolean enable, bluetooth_advertising_params_t *params, gboolean use_reserved_slot)
{
	BT_CHECK_PARAMETER(sender, return);
	int server_instance;
	bt_pending_adv_data_set_t *info = NULL;
	int result = OAL_STATUS_SUCCESS;
	BT_DBG("+");

	if (__bt_is_factory_test_mode()) {
		BT_ERR("Unable to start advertising in factory binary !!");
		return BLUETOOTH_ERROR_NOT_SUPPORT;
	}

	server_instance = _bt_get_allocated_server_instance(sender, adv_handle, use_reserved_slot);

	if (server_instance == -1) {
		BT_DBG("No available slot for the current sender and handle combination");

		if (enable == FALSE) {
			BT_ERR("Advertising not even enabled on adv handle [%d] sender [%s]", adv_handle, sender);
			return BLUETOOTH_ERROR_NOT_IN_OPERATION;
		}
		/* Below logic is only valid only when enabling advertising */
		server_instance = _bt_is_sender_gatt_server_with_no_adv(sender, adv_handle);

		if (server_instance == -1) {
			/* Internal Logic to register server instance if not initialized, store adv handle for future use */
			if (_bt_register_server_instance(sender, adv_handle) != BLUETOOTH_ERROR_NONE)
				return BLUETOOTH_ERROR_INTERNAL;
			else {
				/* Allocate a pending structure and mark Adv data set pending */
				info = g_malloc0(sizeof(bt_pending_adv_data_set_t));
				info->adv_handle = adv_handle;
				info->sender = g_strdup(sender);
				info->is_custom_adv = TRUE;
				info->is_data_set_pending = FALSE;
				memcpy(&info->params, params, sizeof(bluetooth_advertising_params_t));
				adv_data_pending_list = g_slist_append(adv_data_pending_list, info);
				return BLUETOOTH_ERROR_NONE;
			}
		}
	}

	if (enable) {
		/* Set Advertising parameters to Stack */
		result = adapter_ble_multi_adv_update(server_instance, params->interval_min, params->interval_max,
				params->type, BT_ADV_DEFAULT_CHANNEL_MAP, BT_ADV_DEFAULT_TX_POWER, BT_ADV_DEFAULT_TIMEOUT);
		if (result != OAL_STATUS_SUCCESS) {
			BT_ERR("OAL API adapter_ble_multi_adv_update Fail %d", result);
			return BLUETOOTH_ERROR_INTERNAL;
		}
		/* Start Advertising when Adv update event is received */
		result = adapter_ble_multi_adv_enable(server_instance);
		if (result != OAL_STATUS_SUCCESS) {
			BT_ERR("OAL API adapter_ble_multi_adv_enable Fail %d", result);
			return BLUETOOTH_ERROR_INTERNAL;
		}
		/* Update adv handle in table */
		_bt_update_adv_handle(sender, adv_handle);

		return BLUETOOTH_ERROR_NONE;
	} else
		result = adapter_ble_multi_adv_disable(server_instance);

	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("OAL API adapter_ble_multi_adv_disable Fail %d", result);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	return BLUETOOTH_ERROR_NONE;
}

int _bt_get_advertising_data(char *sender, int adv_handle, bluetooth_advertising_data_t *adv, int *length)
{
	BT_CHECK_PARAMETER(adv, return);
	BT_CHECK_PARAMETER(length, return);
	BT_CHECK_PARAMETER(sender, return);

	int server_instance;

	server_instance = _bt_get_allocated_server_instance(sender, adv_handle, FALSE);

	if (server_instance == -1) {
		BT_DBG("No available slot for the current sender and handle combination");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	_bt_get_previous_adv_data(adv, length, server_instance);
	BT_DBG("ADV Data length [%d] Server Instance [%d] Adv handle [%d]", *length, server_instance, adv_handle);

	return BLUETOOTH_ERROR_NONE;
}

int _bt_get_scan_response_data(char *sender, int adv_handle, bluetooth_scan_resp_data_t *response, int *length)
{
	BT_CHECK_PARAMETER(response, return);
	BT_CHECK_PARAMETER(length, return);
	BT_CHECK_PARAMETER(sender, return);

	int server_instance;

	server_instance = _bt_get_allocated_server_instance(sender, adv_handle, FALSE);

	if (server_instance == -1) {
		BT_DBG("No available slot for the current sender and handle combination");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	_bt_get_previous_scan_rsp_data(response, length, server_instance);
	BT_DBG("SCAN RSP Data length [%d] Server Instance [%d] Adv handle [%d]", *length, server_instance, adv_handle);

	return BLUETOOTH_ERROR_NONE;
}

static int __bt_set_multi_adv_param(oal_ble_multi_adv_param_setup_t *adv_setup,
		bluetooth_advertising_data_t *adv_data, int length)
{
	unsigned char *ptr;
	int num_uuids;

	BT_INFO("+");
	retv_if(NULL == adv_data, BLUETOOTH_ERROR_INVALID_PARAM);
	retv_if(NULL == adv_setup, BLUETOOTH_ERROR_INVALID_PARAM);

	for (ptr = adv_data->data; NULL != ptr && length > 0;) {
		int len = ptr[0];
		int type = ptr[1];

		BT_INFO("len: %d, type: %x", len, type);

		switch (type) {
		case 0xFF: /* Manufacturer Data */
			adv_setup->manufacturer_data = g_malloc0(sizeof(char) * (len - 1));
			memcpy(adv_setup->manufacturer_data, (ptr + 2), (len - 1));
			adv_setup->manufacturer_data_len = len - 1;
			break;
		case 0x15: /* 128 bit Solicit UUID */
			adv_setup->solicit_uuid = g_malloc0(sizeof(char) * (len - 1));
			memcpy((adv_setup->solicit_uuid), (ptr + 2), (len - 1));
			adv_setup->solicit_uuid_len = len;
			break;
		case 0x06: /* 128 bit Service UUID */
			adv_setup->service_uuid = g_malloc0(sizeof(char) * (len - 1));
			memcpy((adv_setup->service_uuid), (ptr + 2), (len - 1));
			adv_setup->service_uuid_len = len;
			break;
			case 0x14: {  /* 16 bit Solicit UUID */
			int c;
			int k;
			num_uuids = (len -1)/2;
			adv_setup->solicit_uuid = g_malloc0(sizeof(char) * 16 * num_uuids);
			char *tmp = adv_setup->solicit_uuid;
			adv_setup->solicit_uuid_len = 0;

			for (c = 1; c <= num_uuids; c++) {
				adv_setup->solicit_uuid_len += 16;;
				memcpy(tmp, BASE_UUID_CONVERTED, BT_UUID_128);
				memcpy(tmp+12, &ptr[c*2/* Byte Length*/], 2/* Byte Length */);

				if (c < num_uuids)
					tmp += 16;
			}
			/* DEBUG: Test*/
			for (k = 0; k < 16 * num_uuids; k++)
				BT_DBG("%x", adv_setup->solicit_uuid[k]);
			break;
		}
		case 0x02: { /* 16 bit Service UUID */
			int c;
			int k;
			num_uuids = (len -1)/2;
			adv_setup->service_uuid = g_malloc0(sizeof(char) * 16 * num_uuids);
			char *tmp = adv_setup->service_uuid;
			adv_setup->service_uuid_len = 0;

			for (c = 1; c <= num_uuids; c++) {
				adv_setup->service_uuid_len += 16;;
				memcpy(tmp, BASE_UUID_CONVERTED, BT_UUID_128);
				memcpy(tmp+12, &ptr[c*2/* Byte Length */], 2/* Byte Length */);

				if (c < num_uuids)
					tmp += 16;
			}
			/* DEBUG: Test*/
			for (k = 0; k < 16 * num_uuids; k++)
				BT_DBG("%x", adv_setup->service_uuid[k]);
			break;
		}
		case 0x16: { /* Service Data */
			adv_setup->service_data = g_malloc0(sizeof(char) * (len - 1));
			memcpy(adv_setup->service_data, (ptr + 2), (len - 1));
			adv_setup->service_data_len = len - 1;
			break;
		}
		case 0x21: {
			BT_INFO("128 Bit Service Data Not Supported!!");
			break;
		}
		case 0x0A: {
			adv_setup->include_txpower = 1;
			break;
		}
		case 0x09:
		case 0x08: {
			adv_setup->include_name = 1;
			break;
		}
		case 0x19: {
			memcpy(&adv_setup->appearance, (ptr + 2), (len - 1));
			break;
		}
		default:
			BT_ERR("Unknown type: %x", type);
			break;
	}

		length -= len + 1;
		ptr += len + 1;
	}

	BT_INFO("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_set_advertising_data(const char *sender, int adv_handle,
		bluetooth_advertising_data_t *adv, int length, gboolean use_reserved_slot)
{
	BT_CHECK_PARAMETER(adv, return);
	BT_CHECK_PARAMETER(sender, return);
	bt_pending_adv_data_set_t *info = NULL;
	int server_instance;
	bluetooth_advertising_data_t adv_old;
	int adv_data_len;
	char *old_mdata = NULL;
	int old_len = 0;
	GVariant *ad_data, *param = NULL;
	oal_ble_multi_adv_param_setup_t adv_setup;
	int result = OAL_STATUS_SUCCESS;
	BT_DBG("+");

	if (length > BLUETOOTH_ADVERTISING_DATA_LENGTH_MAX)
		return BLUETOOTH_ERROR_INVALID_PARAM;

	if (__bt_is_factory_test_mode()) {
		BT_ERR("Unable to set advertising data in factory binary !!");
		return BLUETOOTH_ERROR_NOT_SUPPORT;
	}

	server_instance = _bt_get_allocated_server_instance(sender, adv_handle, use_reserved_slot);

	if (server_instance == -1) {
		BT_DBG("No available slot for the current sender and handle combination");

		server_instance = _bt_is_sender_gatt_server_with_no_adv(sender, adv_handle);

		if (server_instance == -1) {
			/* Internal Logic to register server instance if not initialized, store adv handle for future use */
			if (_bt_register_server_instance(sender, adv_handle) != BLUETOOTH_ERROR_NONE)
				return BLUETOOTH_ERROR_INTERNAL;
			else {
				/* Allocate a pending structure and mark Adv data set pending */
				info = g_malloc0(sizeof(bt_pending_adv_data_set_t));
				info->adv_handle = adv_handle;
				info->sender = g_strdup(sender);
				info->data_len = length;
				info->is_adv = TRUE;
				info->is_data_set_pending = TRUE;
				memcpy(&info->data, &(adv->data[0]), length);
				adv_data_pending_list = g_slist_append(adv_data_pending_list, info);
				return BLUETOOTH_ERROR_NONE;
			}
		}
	}

	/* Server Instance is already allocated, set Adv data to stack */
	/* First check if adv data is already present for slot server_instance*/
	memset(&adv_old.data, 0, sizeof(adv_old.data));
	_bt_get_previous_adv_data(&adv_old, &adv_data_len, server_instance);

	/* Send Data to stack */
	memset(&adv_setup, 0, sizeof(oal_ble_multi_adv_param_setup_t));

	if (BLUETOOTH_ERROR_NONE !=
			__bt_set_multi_adv_param(&adv_setup, adv, length)) {
		if (adv_setup.manufacturer_data)
			g_free(adv_setup.manufacturer_data);
		if (adv_setup.service_uuid)
			g_free(adv_setup.service_uuid);
		if (adv_setup.service_data)
			g_free(adv_setup.service_data);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	/* Set Scan response false */
	adv_setup.set_scan_rsp = FALSE;

	/* Set Server instance */
	adv_setup.server_if = server_instance;

	/* Set Server instance[Product Requirement] */
	adv_setup.tx_power = 4;

	BT_DBG("####Service UUID len [%d], service data len [%d] Solicit UUID len [%d]",
				adv_setup.solicit_uuid_len, adv_setup.service_uuid_len, adv_setup.service_data_len);
	result = adapter_ble_multi_adv_set_inst_data(server_instance, &adv_setup);
	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("OAL API adapter_ble_multi_adv_set_inst_data Fail %d", result);
		/* Free the data */
		if (adv_setup.manufacturer_data)
			g_free(adv_setup.manufacturer_data);
		if (adv_setup.service_data)
			g_free(adv_setup.service_data);
		if (adv_setup.service_uuid)
			g_free(adv_setup.service_uuid);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	/* Data sent to Stack successfully, send manuf data changed event if applicable */
	if (adv_setup.manufacturer_data_len != 0) {
		if (adv_data_len > 0) {
			_bt_get_ad_data_by_type((char *)adv_old.data, adv_data_len, 0xff,
					&old_mdata, &old_len);
		}

		if (old_len != adv_setup.manufacturer_data_len ||
				(old_mdata && adv_setup.manufacturer_data &&
				 memcmp(old_mdata, adv_setup.manufacturer_data, adv_setup.manufacturer_data_len))) {

			ad_data = g_variant_new_from_data((const GVariantType *)"ay",
					adv_setup.manufacturer_data, adv_setup.manufacturer_data_len, TRUE, NULL, NULL);
			param = g_variant_new("(@ay)", ad_data);
			_bt_send_event(BT_ADAPTER_EVENT,
					BLUETOOTH_EVENT_ADVERTISING_MANUFACTURER_DATA_CHANGED,
					param);
		}

		g_free(old_mdata);
	}

	/* Time to update new ADV data completely in Table */
	_bt_set_new_adv_data(adv, length, server_instance);

	/* Free the data */
	if (adv_setup.manufacturer_data)
		g_free(adv_setup.manufacturer_data);
	if (adv_setup.service_data)
		g_free(adv_setup.service_data);
	if (adv_setup.service_uuid)
		g_free(adv_setup.service_uuid);

	/* Update adv handle in table */
	_bt_update_adv_handle(sender, adv_handle);

	return BLUETOOTH_ERROR_NONE;
}

int _bt_set_scan_response_data(const char *sender, int adv_handle,
				bluetooth_scan_resp_data_t *response, int length, gboolean use_reserved_slot)
{
	BT_CHECK_PARAMETER(response, return);
	BT_CHECK_PARAMETER(sender, return);
	bt_pending_adv_data_set_t *info = NULL;
	bluetooth_scan_resp_data_t scan_rsp_old;
	int scan_rsp_data_len;
	GVariant *ad_data, *param = NULL;
	oal_ble_multi_adv_param_setup_t adv_setup;
	char *old_mdata = NULL;
	int old_len = 0;
	int server_instance;
	int result = OAL_STATUS_SUCCESS;
	BT_DBG("+");

	if (length > BLUETOOTH_ADVERTISING_DATA_LENGTH_MAX)
		return BLUETOOTH_ERROR_INVALID_PARAM;

	if (__bt_is_factory_test_mode()) {
		BT_ERR("Unable to set Scan Response data in factory binary !!");
		return BLUETOOTH_ERROR_NOT_SUPPORT;
	}

	server_instance = _bt_get_allocated_server_instance(sender, adv_handle, use_reserved_slot);
	if (server_instance == -1) {
		BT_DBG("No available slot for the current sender and handle combination");

		server_instance = _bt_is_sender_gatt_server_with_no_adv(sender, adv_handle);

		if (server_instance == -1) {
			/* Internal Logic to register server instance if not initialized */
			if (_bt_register_server_instance(sender, adv_handle) != BLUETOOTH_ERROR_NONE)
				return BLUETOOTH_ERROR_INTERNAL;
			else {
				/* Allocate a pending structure and mark Adv data set pending */
				info = g_malloc0(sizeof(bt_pending_adv_data_set_t));
				info->adv_handle = adv_handle;
				info->sender = g_strdup(sender);
				info->data_len = length;
				info->is_adv = FALSE;
				info->is_data_set_pending = TRUE;
				memcpy(&info->data, &(response->data[0]), length);
				adv_data_pending_list = g_slist_append(adv_data_pending_list, info);
				return BLUETOOTH_ERROR_NONE;
			}
		}
	}

	/* Server Instance is already allocated, set Adv data to stack */
	/* First check if adv data is already present for slot server_instance*/
	memset(&scan_rsp_old.data, 0, sizeof(scan_rsp_old.data));
	_bt_get_previous_scan_rsp_data(&scan_rsp_old, &scan_rsp_data_len, server_instance);

	/* Send Data to stack */
	memset(&adv_setup, 0, sizeof(oal_ble_multi_adv_param_setup_t));

	if (BLUETOOTH_ERROR_NONE !=
			__bt_set_multi_adv_param(&adv_setup, (bluetooth_advertising_data_t*)response, length)) {
		if (adv_setup.manufacturer_data)
			g_free(adv_setup.manufacturer_data);
		if (adv_setup.service_uuid)
			g_free(adv_setup.service_uuid);
		if (adv_setup.service_data)
			g_free(adv_setup.service_data);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	/* Set Scan response to TRUE */
	adv_setup.set_scan_rsp = TRUE;

	/* Set Server instance */
	adv_setup.server_if = server_instance;

	/* Set Server instance[Product Requirement] */
	adv_setup.tx_power = 4;

	result = adapter_ble_multi_adv_set_inst_data(server_instance, &adv_setup);
	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("OAL API adapter_ble_multi_adv_set_inst_data Fail %d", result);
		/* Free the data */
		if (adv_setup.manufacturer_data)
			g_free(adv_setup.manufacturer_data);
		if (adv_setup.service_data)
			g_free(adv_setup.service_data);
		if (adv_setup.service_uuid)
			g_free(adv_setup.service_uuid);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	/* Data sent to Stack successfully, send manuf data changed event if applicable */
	if (adv_setup.manufacturer_data_len != 0) {
		if (scan_rsp_data_len > 0) {
			_bt_get_ad_data_by_type((char *)scan_rsp_old.data, scan_rsp_data_len, 0xff,
					&old_mdata, &old_len);
		}

		if (old_len != adv_setup.manufacturer_data_len ||
				(old_mdata && adv_setup.manufacturer_data &&
				 memcmp(old_mdata, adv_setup.manufacturer_data, adv_setup.manufacturer_data_len))) {

			ad_data = g_variant_new_from_data((const GVariantType *)"ay",
					adv_setup.manufacturer_data, adv_setup.manufacturer_data_len, TRUE, NULL, NULL);
			param = g_variant_new("(@ay)", ad_data);
			_bt_send_event(BT_ADAPTER_EVENT,
					BLUETOOTH_EVENT_ADVERTISING_MANUFACTURER_DATA_CHANGED,
					param);
		}

		g_free(old_mdata);
	}

	/* Time to update new Scan Response data completely in Table */
	_bt_set_new_scan_rsp_data(response, length, server_instance);

	/* Free the data */
	if (adv_setup.manufacturer_data)
		g_free(adv_setup.manufacturer_data);
	if (adv_setup.service_data)
		g_free(adv_setup.service_data);
	if (adv_setup.service_uuid)
		g_free(adv_setup.service_uuid);

	/* Update adv handle in table */
	_bt_update_adv_handle(sender, adv_handle);
	return BLUETOOTH_ERROR_NONE;
}

/*************************************** LE Scan APIs *********************************************/
int _bt_set_scan_parameters(bluetooth_le_scan_params_t *params)
{
	int itv = 0;
	int win = 0;
	int ret;

	BT_CHECK_PARAMETER(params, return);

	if (_bt_adapter_get_status() != BT_ACTIVATED &&
			_bt_adapter_get_le_status() != BT_LE_ACTIVATED) {
		return BLUETOOTH_ERROR_DEVICE_NOT_ENABLED;
	}

	if (params->interval < BT_LE_SCAN_INTERVAL_MIN ||
			params->interval > BT_LE_SCAN_INTERVAL_MAX)
		return BLUETOOTH_ERROR_INVALID_PARAM;

	if (params->window < BT_LE_SCAN_WINDOW_MIN ||
			params->window > BT_LE_SCAN_WINDOW_MAX)
		return BLUETOOTH_ERROR_INVALID_PARAM;

	if (params->window > params->interval)
		return BLUETOOTH_ERROR_INVALID_PARAM;

	itv = params->interval / BT_ADV_INTERVAL_SPLIT;
	win = params->window / BT_ADV_INTERVAL_SPLIT;

	ret = gattc_set_le_scan_param(params->type, itv, win);
	if (OAL_STATUS_SUCCESS != ret) {
		BT_ERR("gattc_set_le_scan_param failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	is_le_set_scan_parameter = TRUE;
	BT_INFO("Set scan parameters");

	return BLUETOOTH_ERROR_NONE;
}

gboolean __send_le_scan_reply(gpointer data)
{
	int value = (intptr_t)data;

	if (0 == value) {
		BT_DBG("Send reply to BT_STOP_LE_DISCOVERY");
		__bt_adapter_le_handle_pending_request_info(
				BLUETOOTH_ERROR_NONE,
				BT_STOP_LE_DISCOVERY, NULL, 0);
	} else {
		BT_DBG("Send reply to BT_START_LE_DISCOVERY");
		__bt_adapter_le_handle_pending_request_info(
				BLUETOOTH_ERROR_NONE,
				BT_START_LE_DISCOVERY, NULL, 0);
	}

	return FALSE;
}

int _bt_start_le_scan(const char *sender)
{
	bt_adapter_le_scanner_t *scanner;
	int ret;

	scanner = __bt_find_scanner_from_list(sender);
	if (!scanner) {
		scanner = g_malloc0(sizeof(bt_adapter_le_scanner_t));
		retv_if(scanner == NULL, BLUETOOTH_ERROR_INTERNAL);

		scanner->sender = g_strdup(sender);
		scanner_list = g_slist_append(scanner_list, scanner);
	}

	if (scanner->is_scanning) {
		BT_ERR("BT is already in LE scanning");
		return BLUETOOTH_ERROR_IN_PROGRESS;
	}

	if (_bt_is_le_scanning()) {
		int value = 1;

		BT_INFO("LE Full Scan is already on progress");
		g_idle_add(__send_le_scan_reply, (void *)(intptr_t)value);
		goto done;
	}

	if (is_le_set_scan_parameter == FALSE) {
		bluetooth_le_scan_params_t scan_params;

		/* Set default scan parameter same with BT_ADAPTER_LE_SCAN_MODE_LOW_ENERGY */
		scan_params.type = 0x01;
		scan_params.interval = 5000;
		scan_params.window = 500;
		_bt_set_scan_parameters(&scan_params);
	}

	BT_INFO("Start LE Full Scan");
	ret = gattc_start_le_discovery(g_gatt_client_id);

	if (OAL_STATUS_SUCCESS != ret) {
		BT_ERR("gattc_start_le_discovery failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

done:
	scanner->is_scanning = TRUE;
	return BLUETOOTH_ERROR_NONE;
}

int _bt_stop_le_scan(const char *sender)
{
	bt_adapter_le_scanner_t *scanner;
	gboolean next_scanning = FALSE;
	GSList *l;
	int ret;

	scanner = __bt_find_scanner_from_list(sender);
	if (scanner == NULL || scanner->is_scanning == FALSE)
		return BLUETOOTH_ERROR_NOT_IN_OPERATION;

	scanner->is_scanning = FALSE;

	for (l = scanner_list; l != NULL; l = g_slist_next(l)) {
		scanner = l->data;
		if (scanner->is_scanning == TRUE)
			next_scanning = TRUE;
	}

	if (next_scanning == TRUE) {
		int value = 0;

		g_idle_add(__send_le_scan_reply, (void *)(intptr_t)value);
		return BLUETOOTH_ERROR_NONE;
	} else {
		BT_INFO("Just stop LE scan");
	}

	ret = gattc_stop_le_discovery(g_gatt_client_id);

	if (OAL_STATUS_SUCCESS != ret) {
		BT_ERR("gattc_start_le_discovery failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	is_le_set_scan_parameter = FALSE;
	return BLUETOOTH_ERROR_NONE;
}

void _bt_check_le_scanner_app_termination(const char *sender)
{
	bt_adapter_le_scanner_t *scanner;

	scanner = __bt_find_scanner_from_list(sender);
	if (!scanner)
		return;

	if (scanner->is_scanning)
		_bt_stop_le_scan(sender);

	scanner_list = g_slist_remove(scanner_list, scanner);
	g_free(scanner->sender);
	g_free(scanner);
}

int _bt_service_le_init(void)
{
	le_init();

	return BLUETOOTH_ERROR_NONE;
}

void _bt_service_le_deinit(void)
{
	le_deinit();
}

int _bt_is_advertising(void)
{
	int ret = is_advertising();

	if (ret == BLUETOOTH_ERROR_NONE)
		return TRUE;
	else
		return FALSE;
}

gboolean _bt_is_le_2m_phy_supported(void)
{
	if (le_feature_info.le_2m_phy)
		return TRUE;
	else
		return FALSE;
}

gboolean _bt_is_le_coded_phy_supported(void)
{
	if (le_feature_info.le_coded_phy)
		return TRUE;
	else
		return FALSE;
}


/*************************************** LE Scan APIs *********************************************/
