/*
 * Copyright (c) 2011 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <string.h>
#include <stdio.h>
#include <glib.h>
#include <dlog.h>
#include <gio/gio.h>
#include <gio/gunixfdlist.h>
#include <cynara-client.h>
#include <cynara-creds-gdbus.h>

#include "bluetooth-api.h"
#include "bluetooth-audio-api.h"
#include "bluetooth-gatt-server-api.h"
#include "bt-request-handler.h"
#include "bt-service-common.h"
#include "bt-service-util.h"

#include "bt-service-core-adapter.h"
#include "bt-service-core-adapter-le.h"
#include "bt-service-core-device.h"

#include "bt-service-audio-common.h"
#include "bt-service-avrcp-tg.h"
#include "bt-service-avrcp-ctrl.h"
#include "bt-service-gatt.h"

#ifdef TIZEN_FEATURE_BT_DPM
#include "bt-service-dpm.h"
#endif
#include "bt-service-hidhost.h"
#include "bt-service-rfcomm.h"
#include "bt-service-hdp.h"

/*Obex*/
#include "bt-service-obex-server.h"
#include "bt-service-opp-client.h"
#include "bt-service-map-client.h"
#include "bt-service-pbap.h"
#include "bt-service-oob.h"

#ifdef TIZEN_FEATURE_BT_PAN_NAP
#include "bt-service-network.h"
#endif

#ifdef TIZEN_FEATURE_BT_IPSP
#include "bt-service-ipsp.h"
#endif
/* For maintaining Application Sync API call requests */
GSList *invocation_list = NULL;

static GDBusConnection *bt_service_conn;
static guint owner_id = 0;
static guint owner_sig_id = 0;
static gboolean is_le_intended = FALSE;


static cynara *p_cynara;
static cynara_configuration *conf;

static const gchar bt_service_introspection_xml[] =
"<node name='/org/projectx/bt_service'>"
"	<interface name='org.projectx.bt'>"
"		<method name='service_request'>"
			/* Input Parameters */
"			<arg type='i' name='service_type' direction='in' />"
"			<arg type='i' name='service_function' direction='in' />"
"			<arg type='i' name='request_type' direction='in' />"
"			<arg type='ay' name='input_param1' direction='in' />"
"			<arg type='ay' name='input_param2' direction='in' />"
"			<arg type='ay' name='input_param3' direction='in' />"
"			<arg type='ay' name='input_param4' direction='in' />"
"			<arg type='ay' name='input_param5' direction='in' />"
			/* Return Parameters */
"			<arg type='i' name='output_param1' direction='out' />"
"			<arg type='v' name='output_param2' direction='out' />"
"		</method>"
"	</interface>"
"</node>";

static gboolean name_acquired = FALSE;

static char *current_sender_playing = NULL;

GDBusNodeInfo *node_info_g = NULL;

static void __bt_service_method(GDBusConnection *connection,
		const gchar *sender,
		const gchar *object_path,
		const gchar *interface_name,
		const gchar *method_name,
		GVariant *parameters,
		GDBusMethodInvocation *invocation,
		gpointer user_data);

int __bt_bluez_request(int function_name,
		int request_type,
		int request_id,
		GDBusMethodInvocation *context,
		GVariant *in_param1,
		GVariant *in_param2,
		GVariant *in_param3,
		GVariant *in_param4,
		GArray **out_param1);
int __bt_obexd_request(int function_name,
		int request_type,
		int request_id,
		GDBusMethodInvocation *context,
		GVariant *in_param1,
		GVariant *in_param2,
		GVariant *in_param3,
		GVariant *in_param4,
		GArray **out_param1);
int __bt_agent_request(int function_name,
		int request_type,
		int request_id,
		GDBusMethodInvocation *context,
		GVariant *in_param1,
		GVariant *in_param2,
		GVariant *in_param3,
		GVariant *in_param4,
		GArray **out_param1);
int __bt_core_request(int function_name,
		int request_type,
		int request_id,
		GDBusMethodInvocation *context,
		GVariant *in_param1);

gboolean __bt_service_check_privilege(int function_name,
					int service_type,
					const char *unique_name);

/* Function definitions*/
GSList *_bt_get_invocation_list(void)
{
	return invocation_list;
}

void _bt_free_info_from_invocation_list(invocation_info_t *req_info)
{
	GSList *l;
	invocation_info_t *info;

	ret_if(NULL == req_info);
	ret_if(NULL == req_info->sender);

	for (l = invocation_list; l != NULL; l = g_slist_next(l)) {
		info = l->data;
		if (NULL == info || NULL == info->sender)
			continue;

		/* No two same sync requests from same application can exist */
		if ((strcasecmp(info->sender, req_info->sender) == 0) &&
				req_info->service_function == info->service_function) {

			invocation_list = g_slist_remove(invocation_list, req_info);
			g_free(req_info->sender);
			g_free(req_info);
			break;
		}
	}

}

void _bt_set_le_intended_status(gboolean value)
{
	is_le_intended = value;
}

static void __bt_service_get_parameters(GVariant *in_param,
		void *value, int size)
{
	void *buf = NULL;
	buf = (void *)g_variant_get_data(in_param);
	memcpy(value, buf, size);
}

static gboolean __bt_is_sync_function(int service_function)
{
	/*TODO: Keep adding sync methods with expect replies from bluetooth service */
	if (service_function == BT_ENABLE_ADAPTER
			|| service_function == BT_DISABLE_ADAPTER
			|| service_function == BT_GET_LOCAL_ADDRESS
			|| service_function == BT_GET_LOCAL_NAME
			|| service_function == BT_GET_LOCAL_VERSION
			|| service_function == BT_GET_BONDED_DEVICES
			|| service_function == BT_GET_BONDED_DEVICE
			|| service_function == BT_GET_IS_ALIAS_SET
			|| service_function == BT_GET_CONNECTED_LINK_TYPE
			|| service_function == BT_IS_SERVICE_USED
			|| service_function == BT_RFCOMM_LISTEN_AND_ACCEPT
			|| service_function == BT_RFCOMM_LISTEN
			|| service_function == BT_HDP_REGISTER_SINK_APP
			|| service_function == BT_HDP_UNREGISTER_SINK_APP
			|| service_function == BT_HDP_GET_FD
			|| service_function == BT_AVRCP_GET_TRACK_INFO
			|| service_function == BT_AVRCP_CONTROL_GET_PROPERTY
			|| service_function == BT_SET_ADVERTISING_DATA
			|| service_function == BT_SET_SCAN_RESPONSE_DATA
			|| service_function == BT_SET_ADVERTISING
			|| service_function == BT_SET_CUSTOM_ADVERTISING
			|| service_function == BT_GATT_SERVER_REGISTER
			|| service_function == BT_GATT_SERVER_ADD_SERVICE
			|| service_function == BT_GATT_SERVER_ADD_CHARACTERISTIC
			|| service_function == BT_GATT_SERVER_ADD_DESCRIPTOR
			|| service_function == BT_GATT_SERVER_START_SERVICE
			|| service_function == BT_GATT_SERVER_STOP_SERVICE
			|| service_function == BT_GATT_SERVER_DELETE_SERVICE
			|| service_function == BT_START_LE_DISCOVERY
			|| service_function == BT_STOP_LE_DISCOVERY
#ifdef TIZEN_GATT_CLIENT
			|| service_function == BT_GATT_CLIENT_REGISTER
			|| service_function == BT_GATT_GET_PRIMARY_SERVICES
			|| service_function == BT_GATT_GET_SERVICE_PROPERTIES
			|| service_function == BT_GATT_GET_INCLUDED_SERVICES
			|| service_function == BT_GATT_GET_CHARACTERISTIC_PROPERTIES
			|| service_function == BT_GATT_WATCH_CHARACTERISTIC
			|| service_function == BT_GATT_ACQUIRE_WRITE
#endif
			|| service_function == BT_AUDIO_SELECT_ROLE)
		return TRUE;
	else
		return FALSE;
}

void _bt_save_invocation_context(GDBusMethodInvocation *invocation, int result,
		char *sender, int service_function,
		gpointer invocation_data)
{
	BT_DBG("Saving the invocation context: service_function [%d]", service_function);
	invocation_info_t *info;
	info = g_malloc0(sizeof(invocation_info_t));
	info->context = invocation;
	info->result = result;
	info->sender = g_strdup(sender);
	info->service_function = service_function;
	info->user_data = invocation_data;
	invocation_list = g_slist_append(invocation_list, info);

}

static void __bt_service_method(GDBusConnection *connection,
		const gchar *sender,
		const gchar *object_path,
		const gchar *interface_name,
		const gchar *method_name,
		GVariant *parameters,
		GDBusMethodInvocation *invocation,
		gpointer user_data)
{
	FN_START;

	BT_DBG("Method[%s] Object Path[%s] Interface Name[%s]",
			method_name, object_path, interface_name);

	if (g_strcmp0(method_name, "service_request") == 0) {
		int service_type;
		int service_function;
		int request_type;
		GVariant *param1 = NULL;
		GVariant *param2 = NULL;
		GVariant *param3 = NULL;
		GVariant *param4 = NULL;
		GVariant *param5 = NULL;
		GArray *out_param1 = NULL;
		GVariant *out_var = NULL;
		GVariant *temp = NULL;
		int result = 0;
		int request_id = -1;

		g_variant_get(parameters, "(iii@ay@ay@ay@ay@ay)", &service_type,
				&service_function, &request_type,
				&param1, &param2, &param3, &param4, &param5);

		out_param1 = g_array_new(FALSE, FALSE, sizeof(gchar));

		if (service_type == BT_CORE_SERVICE) {
			BT_DBG("No need to check privilege from bt-core");
		} else if (__bt_service_check_privilege(service_function,
					service_type, (const char *)sender) == FALSE) {
			BT_ERR("Client don't have the privilege to excute this function");
			result = BLUETOOTH_ERROR_PERMISSION_DEINED;
			goto fail;
		}

		BT_DBG("SERVICE TYPE [%d] SERVICE FUNC [%d]",
				service_type, service_function);
		switch (service_type) {
		case BT_BLUEZ_SERVICE:
			result = __bt_bluez_request(service_function,
					request_type, request_id,
					invocation, param1, param2,
					param3, param4, &out_param1);
			break;
		case BT_OBEX_SERVICE:
			result = __bt_obexd_request(service_function,
					request_type, request_id,
					invocation, param1,
					param2, param3,
					param4, &out_param1);
			break;
		case BT_AGENT_SERVICE:
			result = __bt_agent_request(service_function,
					request_type, request_id,
					invocation, param1,
					param2, param3,
					param4, &out_param1);
			break;
		case BT_CORE_SERVICE:
			result = __bt_core_request(service_function,
					request_type, request_id,
					invocation, param1);
			break;
		case BT_CHECK_PRIVILEGE:
			result = BLUETOOTH_ERROR_NONE;
			break;
		default:
			BT_ERR("Unknown service type");
			result = BLUETOOTH_ERROR_INTERNAL;
			goto fail;
		}

		if (result != BLUETOOTH_ERROR_NONE)
			goto fail;

		if (service_type == BT_CHECK_PRIVILEGE)
			goto send_reply;

		if (service_function == BT_OPP_PUSH_FILES)
			goto done;

		if (request_type == BT_ASYNC_REQ ||
				__bt_is_sync_function(service_function)) {
			BT_INFO("Do not send reply to bt-api over dbus,"
				" Invocation context is already saved in service_function");
			goto done;
		}

send_reply:
		out_var = g_variant_new_from_data((const GVariantType *)"ay",
				out_param1->data, out_param1->len,
				TRUE, NULL, NULL);

		temp = g_variant_new("(iv)", result, out_var);
		g_dbus_method_invocation_return_value(invocation, temp);
		g_array_free(out_param1, TRUE);
		out_param1 = NULL;

done:
		g_variant_unref(param1);
		g_variant_unref(param2);
		g_variant_unref(param3);
		g_variant_unref(param4);
		g_variant_unref(param5);
		FN_END;
		return;
fail:
		BT_ERR_C("Request is failed [%s] [0x%x]",
				_bt_convert_error_to_string(result), result);

		out_var = g_variant_new_from_data((const GVariantType *)"ay",
				out_param1->data, out_param1->len,
				TRUE, NULL, NULL);

		temp = g_variant_new("(iv)", result, out_var);
		g_dbus_method_invocation_return_value(invocation, temp);

		g_array_free(out_param1, TRUE);
		out_param1 = NULL;

		g_variant_unref(param1);
		g_variant_unref(param2);
		g_variant_unref(param3);
		g_variant_unref(param4);
		g_variant_unref(param5);
	}

	FN_END;
	return;
}


static const GDBusInterfaceVTable method_table = {
	__bt_service_method,
	NULL,
	NULL,
	{0}
};

int __bt_bluez_request(int function_name,
		int request_type,
		int request_id,
		GDBusMethodInvocation *context,
		GVariant *in_param1,
		GVariant *in_param2,
		GVariant *in_param3,
		GVariant *in_param4,
		GArray **out_param1)
{
	int result = BLUETOOTH_ERROR_NONE;
	char *sender = NULL;

	switch (function_name) {
	case BT_ENABLE_ADAPTER: {
		result = _bt_enable_adapter();
		/* Save invocation */
		if (result == BLUETOOTH_ERROR_NONE) {
			BT_DBG("_bt_enable_adapter scheduled successfully! save invocation context");
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, NULL);
		}
		break;
	}
	case BT_DISABLE_ADAPTER: {
		result = _bt_disable_adapter();
		/* Save invocation */
		if (result == BLUETOOTH_ERROR_NONE) {
			BT_DBG("_bt_disable_adapter scheduled successfully! save invocation context");
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, NULL);
		}
		break;
	}
	case BT_RECOVER_ADAPTER:
		 result = _bt_recover_adapter();
		 break;
	case BT_RESET_ADAPTER:
		result = _bt_reset_adapter();
		break;
	case BT_CHECK_ADAPTER: {
		int enabled = BT_ADAPTER_DISABLED;
		result = _bt_check_adapter(&enabled);

		g_array_append_vals(*out_param1, &enabled,
				sizeof(int));
		break;
	}
	case BT_ENABLE_ADAPTER_LE:
		_bt_set_le_intended_status(TRUE);
		result = _bt_enable_adapter_le();
		break;
	case BT_DISABLE_ADAPTER_LE:
		_bt_set_le_intended_status(FALSE);
		result = _bt_disable_adapter_le();
		break;
	case BT_START_DISCOVERY: {
		unsigned short max_response;
		unsigned short discovery_duration;
		unsigned int classOfDeviceMask;

		__bt_service_get_parameters(in_param1,
				&max_response, sizeof(unsigned short));
		__bt_service_get_parameters(in_param2,
				&discovery_duration, sizeof(unsigned short));
		__bt_service_get_parameters(in_param3,
				&classOfDeviceMask, sizeof(unsigned int));

		result = _bt_start_discovery(max_response, discovery_duration, classOfDeviceMask);
		break;
	}
	case BT_START_CUSTOM_DISCOVERY: {
		bt_discovery_role_type_t role;

		__bt_service_get_parameters(in_param1,
				&role, sizeof(bt_discovery_role_type_t));
		result = _bt_start_custom_discovery(role);

		break;
	}
	case BT_CANCEL_DISCOVERY:
		result = _bt_cancel_discovery();
		break;
	case BT_IS_DISCOVERYING: {
		gboolean discovering = FALSE;
		discovering = _bt_is_discovering();
		g_array_append_vals(*out_param1,
				&discovering, sizeof(gboolean));
		break;
	}
	case BT_GET_LOCAL_ADDRESS: {
		result = _bt_get_local_address();

		/* Save invocation */
		if (result == BLUETOOTH_ERROR_NONE) {
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, NULL);
		}
		break;
	}
	case BT_GET_LOCAL_VERSION: {
		result = _bt_get_local_version();

		/* Save invocation */
		if (result == BLUETOOTH_ERROR_NONE) {
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, NULL);
		}
		break;
	}
	case BT_GET_LOCAL_NAME: {
		result = _bt_get_local_name();

		/* Save invocation */
		if (result == BLUETOOTH_ERROR_NONE) {
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, NULL);
		}
		break;
	}
	case BT_SET_LOCAL_NAME: {
		bluetooth_device_name_t local_name = { {0} };
		__bt_service_get_parameters(in_param1,
				&local_name, sizeof(bluetooth_device_name_t));
		result = _bt_set_local_name(local_name.name);
		break;
	}
	case BT_GET_DISCOVERABLE_MODE: {
		int discoverable_mode = BLUETOOTH_DISCOVERABLE_MODE_CONNECTABLE;
		result = _bt_get_discoverable_mode(&discoverable_mode);
		g_array_append_vals(*out_param1, &discoverable_mode, sizeof(int));
		break;
	}
	case BT_GET_DISCOVERABLE_TIME: {
		int timeout = 0;

		result = _bt_get_timeout_value(&timeout);
		g_array_append_vals(*out_param1, &timeout, sizeof(int));
		break;
	}
	case BT_SET_DISCOVERABLE_MODE: {
		int mode = BLUETOOTH_DISCOVERABLE_MODE_CONNECTABLE;
		int time = 0;

		__bt_service_get_parameters(in_param1, &mode, sizeof(int));
		__bt_service_get_parameters(in_param2, &time, sizeof(int));

		result = _bt_set_discoverable_mode(mode, time);
		break;
	}
	case BT_IS_CONNECTABLE: {
		gboolean is_connectable = FALSE;

		is_connectable = _bt_is_connectable();
		g_array_append_vals(*out_param1, &is_connectable, sizeof(gboolean));
		break;
	}
	case BT_SET_CONNECTABLE: {
		gboolean is_connectable;

		__bt_service_get_parameters(in_param1,
				&is_connectable, sizeof(gboolean));
		result = _bt_set_connectable(is_connectable);
		break;
	}
	case BT_IS_SERVICE_USED: {
		char *uuid = NULL;

		uuid = (char *)g_variant_get_data(in_param1);
		BT_INFO("UUID to be searched [%s]", uuid);
		result = _bt_is_service_used();

		/* Save invocation */
		if (result == BLUETOOTH_ERROR_NONE) {
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)uuid);
		}
		break;
	}
	case BT_GET_BONDED_DEVICES: {
		result = _bt_get_bonded_devices();
		/* Save invocation */
		if (result == BLUETOOTH_ERROR_NONE) {
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, NULL);
		}
		break;
	}
	case BT_GET_BONDED_DEVICE: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_get_bonded_device_info(&address);
		/* Save invocation */
		if (result == BLUETOOTH_ERROR_NONE) {
			char *addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
			if (!addr) {
				result = BLUETOOTH_ERROR_MEMORY_ALLOCATION;
				break;
			}

			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, addr);
		}
		break;
	}
	case BT_GET_PROFILE_CONNECTED_DEVICES: {
		char *uuid;
		uuid = (char *)g_variant_get_data(in_param1);
		result = _bt_get_profile_connected_devices(uuid, out_param1);
		break;
	}
	case BT_SET_ALIAS: {
		bluetooth_device_address_t address = { {0} };
		const char *local_name;

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));
		local_name = (const char *)g_variant_get_data(in_param2);

		result = _bt_set_alias(&address, local_name);
		break;
	}
	case BT_BOND_DEVICE: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));
		result = _bt_bond_device(&address, BLUETOOTH_DEV_CONN_DEFAULT, out_param1);

		/* Save invocation */
		if (result == BLUETOOTH_ERROR_NONE) {
			char * addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			BT_DBG("_bt_bond_device scheduled successfully! save invocation context");
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		} else {
			bluetooth_device_info_t dev_info;

			memset(&dev_info, 0x00, sizeof(bluetooth_device_info_t));
			memcpy(&dev_info.device_address, &address,
					sizeof(bluetooth_device_address_t));
			g_array_append_vals(*out_param1, &dev_info,
					sizeof(bluetooth_device_info_t));
		}
		break;
	}
	case BT_BOND_DEVICE_BY_TYPE: {
		bluetooth_device_address_t address = { {0} };
		unsigned short conn_type = 0;

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2,
				&conn_type, sizeof(unsigned short));

		result = _bt_bond_device(&address, conn_type, out_param1);
		/* Save invocation */
		if (result == BLUETOOTH_ERROR_NONE) {
			char * addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			BT_DBG("_bt_bond_device_by_type invoked successfully! save invocation context");
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		} else {
			bluetooth_device_info_t dev_info;

			memset(&dev_info, 0x00, sizeof(bluetooth_device_info_t));
			memcpy(&dev_info.device_address, &address,
					sizeof(bluetooth_device_address_t));
			g_array_append_vals(*out_param1, &dev_info,
					sizeof(bluetooth_device_info_t));
		}
		break;
	}
	case BT_UNBOND_DEVICE: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));
		result = _bt_unbond_device(&address, out_param1);

		/* Save invocation */
		if (result == BLUETOOTH_ERROR_NONE) {
			char * addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			BT_DBG("_bt_unbond_device scheduled successfully! save invocation context");
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		} else {
			g_array_append_vals(*out_param1, &address,
					sizeof(bluetooth_device_address_t));
		}
		break;
	}
	case BT_CANCEL_BONDING:{
		result = _bt_cancel_bonding();
		break;
	}
	case BT_SEARCH_SERVICE: {
		bluetooth_device_address_t address = { {0} };
		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));
		result = _bt_search_device(&address);
		/* Save invocation */
		if (result == BLUETOOTH_ERROR_NONE) {
			char * addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			BT_DBG("BT Device Service Search Request scheduled successfully! save invocation context");
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		} else {
			bt_sdp_info_t sdp_info;

			memset(&sdp_info, 0x00, sizeof(bt_sdp_info_t));
			memcpy(&sdp_info.device_addr, &address,
					sizeof(bluetooth_device_address_t));
			g_array_append_vals(*out_param1, &sdp_info, sizeof(bt_sdp_info_t));
		}
		break;
	}
	case BT_CANCEL_SEARCH_SERVICE: {
		result = _bt_cancel_search_device();
		break;
	}
	case BT_SET_PASSKEY_NOTIFICATION: {
		char *sender = NULL;
		gboolean enable;

		sender = (char *)g_dbus_method_invocation_get_sender(context);
		__bt_service_get_parameters(in_param1,
				&enable, sizeof(gboolean));

		result = _bt_set_passkey_notification(sender, enable);
		break;
	}
	case BT_SET_PIN_CODE: {
		bluetooth_device_address_t address = { {0} };
		bluetooth_device_pin_code_t pin_code = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2,
				&pin_code, sizeof(bluetooth_device_pin_code_t));

		result = _bt_set_pin_code(&address, &pin_code);
		break;
	}
	case BT_UNSET_PIN_CODE: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_unset_pin_code(&address);
		break;
	}
	case BT_PASSKEY_REPLY: {
		const char *passkey = NULL;
		gboolean authentication_reply = FALSE;
		passkey = g_variant_get_data(in_param1);
		__bt_service_get_parameters(in_param2,
				&authentication_reply, sizeof(gboolean));
		result = _bt_passkey_reply(passkey, authentication_reply);
		break;
	}
	case BT_PASSKEY_CONFIRMATION_REPLY: {
		gboolean confirmation_reply = FALSE;
		__bt_service_get_parameters(in_param1,
				&confirmation_reply, sizeof(gboolean));
		result = _bt_passkey_confirmation_reply(confirmation_reply);
		break;
	}
	case BT_GET_IS_ALIAS_SET: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_is_alias_set(&address);
		if (result == BLUETOOTH_ERROR_NONE) {
			char *addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
			if (!addr) {
				result = BLUETOOTH_ERROR_MEMORY_ALLOCATION;
				break;
			}

			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, addr);
		}
		break;
	}
	case BT_SET_AUTHORIZATION: {
		bluetooth_device_address_t address = { {0} };
		gboolean authorize;
		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2,
				&authorize, sizeof(gboolean));
		result = _bt_set_authorization(&address, authorize);
		break;
	}
	case BT_IS_DEVICE_CONNECTED: {
		bluetooth_device_address_t address = { {0} };
		gboolean connected = FALSE;
		int service_type;

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2,
				&service_type, sizeof(int));

#if defined(TIZEN_FEATURE_BT_PAN_NAP)
		if ((service_type == BLUETOOTH_NAP_SERVICE) || (service_type == BLUETOOTH_NAP_SERVER_SERVICE))
			connected = _bt_is_nap_panu_device_connected(&address, service_type);
		else
			connected = _bt_is_device_connected(&address, service_type);
#else
		connected = _bt_is_device_connected(&address, service_type);
#endif
		BT_INFO("is_connected: %d", connected);
		g_array_append_vals(*out_param1, &connected, sizeof(gboolean));
		break;
	}
	case BT_GET_CONNECTED_LINK_TYPE: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_get_connected_link(&address);
		if (result == BLUETOOTH_ERROR_NONE) {
			char *addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
			if (!addr) {
				result = BLUETOOTH_ERROR_MEMORY_ALLOCATION;
				break;
			}

			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, addr);
		}
		break;
	}
	case BT_ENABLE_RSSI: {
		bluetooth_device_address_t bd_addr;
		int link_type;
		bt_rssi_threshold_t rssi_threshold;
		int low_threshold;
		int in_range_threshold;
		int high_threshold;

		BT_DBG("Enable RSSI");

		__bt_service_get_parameters(in_param1,
				&bd_addr, sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2,
				&link_type, sizeof(int));
		__bt_service_get_parameters(in_param3,
				&rssi_threshold, sizeof(bt_rssi_threshold_t));

		low_threshold = rssi_threshold.low_threshold;
		in_range_threshold = rssi_threshold.in_range_threshold;
		high_threshold = rssi_threshold.high_threshold;

		result = _bt_enable_rssi(&bd_addr, link_type, low_threshold,
				in_range_threshold, high_threshold);
		break;
	}
	case BT_GET_RSSI: {
		int link_type;
		bluetooth_device_address_t bd_addr;

		BT_DBG("Get RSSI Strength");

		__bt_service_get_parameters(in_param1,
				&bd_addr, sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2,
				&link_type, sizeof(int));

		result = _bt_get_rssi_strength(&bd_addr, link_type);
		break;
	}
	case BT_SET_PROFILE_TRUSTED: {
		bluetooth_device_address_t bd_addr = { {0} };
		int profile;
		int trust;

		__bt_service_get_parameters(in_param1, &bd_addr,
				sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2, &profile, sizeof(int));
		__bt_service_get_parameters(in_param3, &trust, sizeof(int));

		result = _bt_set_trust_profile(&bd_addr, profile, trust);
		break;
	}
	case BT_GET_PROFILE_TRUSTED: {
		bluetooth_device_address_t bd_addr = { {0} };
		int profile;
		guint trusted_profile = 0;

		__bt_service_get_parameters(in_param1, &bd_addr,
				sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2, &profile, sizeof(int));

		result = _bt_get_trust_profile(&bd_addr, profile, &trusted_profile);
		BT_DBG("TRUST %d", trusted_profile);
		if (result == BLUETOOTH_ERROR_NONE) {
			g_array_append_vals(*out_param1, &trusted_profile,
					sizeof(guint));
		}

		break;
	}
	case BT_HID_CONNECT: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_hid_connect(&address);
		if (result != BLUETOOTH_ERROR_NONE) {
			g_array_append_vals(*out_param1, &address,
					sizeof(bluetooth_device_address_t));
		} else {
			char *addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		}
		break;
	}
	case BT_HID_DISCONNECT: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_hid_disconnect(&address);
		if (result != BLUETOOTH_ERROR_NONE) {
			g_array_append_vals(*out_param1, &address,
					sizeof(bluetooth_device_address_t));
		} else {
			char *addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		}
		break;
	}
#ifdef TIZEN_FEATURE_BT_DPM
	case BT_DPM_SET_ALLOW_BT_MODE: {
		dpm_bt_allow_t value = DPM_BT_ERROR;

		__bt_service_get_parameters(in_param1, &value,
				sizeof(int));

		result = _bt_dpm_set_allow_bluetooth_mode(value);
		break;
	}
	case BT_DPM_GET_ALLOW_BT_MODE: {
		int value = DPM_BT_ERROR;

		result = _bt_dpm_get_allow_bluetooth_mode(&value);
		g_array_append_vals(*out_param1, &value, sizeof(int));
		break;
	}
	case BT_DPM_SET_DEVICE_RESTRITION: {
		dpm_status_t value = DPM_STATUS_ERROR;

		__bt_service_get_parameters(in_param1, &value,
				sizeof(int));

		result = _bt_dpm_activate_bluetooth_device_restriction(value);
		break;
	}
	case BT_DPM_GET_DEVICE_RESTRITION: {
		int value = DPM_STATUS_ERROR;

		result = _bt_dpm_is_bluetooth_device_restriction_active(&value);
		g_array_append_vals(*out_param1, &value, sizeof(int));
		break;
	}
	case BT_DPM_SET_UUID_RESTRITION: {
		dpm_status_t value = DPM_STATUS_ERROR;

		__bt_service_get_parameters(in_param1, &value,
				sizeof(int));

		result = _bt_dpm_activate_bluetoooth_uuid_restriction(value);
		break;
	}
	case BT_DPM_GET_UUID_RESTRITION: {
		int value = DPM_STATUS_ERROR;

		result = _bt_dpm_is_bluetooth_uuid_restriction_active(&value);
		g_array_append_vals(*out_param1, &value, sizeof(int));
		break;
	}
	case BT_DPM_ADD_DEVICES_BLACKLIST: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1, &address,
			sizeof(bluetooth_device_address_t));

		result = _bt_dpm_add_bluetooth_devices_to_blacklist(&address);
		break;
	}
	case BT_DPM_ADD_DEVICES_WHITELIST: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1, &address,
			sizeof(bluetooth_device_address_t));

		result = _bt_dpm_add_bluetooth_devices_to_whitelist(&address);
		break;
	}
	case BT_DPM_ADD_UUIDS_BLACKLIST: {
		const char *uuid = NULL;

		uuid = g_variant_get_data(in_param1);

		result = _bt_dpm_add_bluetooth_uuids_to_blacklist(uuid);
		break;
	}
	case BT_DPM_ADD_UUIDS_WHITELIST: {
		const char *uuid = NULL;

		uuid = g_variant_get_data(in_param1);

		result = _bt_dpm_add_bluetooth_uuids_to_whitelist(uuid);
		break;
	}
	case BT_DPM_CLEAR_DEVICES_BLACKLIST: {
		result = _bt_dpm_clear_bluetooth_devices_from_blacklist();
		break;
	}
	case BT_DPM_CLEAR_DEVICES_WHITELIST: {
		result = _bt_dpm_clear_bluetooth_devices_from_whitelist();
		break;
	}
	case BT_DPM_CLEAR_UUIDS_BLACKLIST: {
		result = _bt_dpm_clear_bluetooth_uuids_from_blacklist();
		break;
	}
	case BT_DPM_CLEAR_UUIDS_WHITELIST: {
		result = _bt_dpm_clear_bluetooth_uuids_from_whitelist();
		break;
	}
	case BT_DPM_REMOVE_DEVICE_BLACKLIST: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1, &address,
			sizeof(bluetooth_device_address_t));

		result = _bt_dpm_remove_bluetooth_devices_from_blacklist(&address);
		break;
	}
	case BT_DPM_REMOVE_DEVICE_WHITELIST: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1, &address,
			sizeof(bluetooth_device_address_t));

		result = _bt_dpm_remove_bluetooth_devices_from_whitelist(&address);
		break;
	}
	case BT_DPM_REMOVE_UUID_BLACKLIST: {
		const char *uuid = NULL;

		uuid = g_variant_get_data(in_param1);

		result = _bt_dpm_remove_bluetooth_uuids_from_blacklist(uuid);
		break;
	}
	case BT_DPM_REMOVE_UUID_WHITELIST: {
		const char *uuid = NULL;

		uuid = g_variant_get_data(in_param1);

		result = _bt_dpm_remove_bluetooth_uuids_from_whitelist(uuid);

		break;
	}
	case BT_DPM_GET_DEVICES_BLACKLIST: {
		result = _bt_dpm_get_bluetooth_devices_from_blacklist(out_param1);
		break;
	}
	case BT_DPM_GET_DEVICES_WHITELIST: {
		result = _bt_dpm_get_bluetooth_devices_from_whitelist(out_param1);
		break;
	}
	case BT_DPM_GET_UUIDS_BLACKLIST: {
		result = _bt_dpm_get_bluetooth_uuids_from_blacklist(out_param1);
		break;
	}
	case BT_DPM_GET_UUIDS_WHITELIST: {
		result = _bt_dpm_get_bluetooth_uuids_from_whitelist(out_param1);
		break;
	}
	case BT_DPM_SET_ALLOW_OUTGOING_CALL: {
		dpm_status_t value = DPM_STATUS_ERROR;

		__bt_service_get_parameters(in_param1, &value,
				sizeof(int));

		result = _bt_dpm_set_allow_bluetooth_outgoing_call(value);

		break;
	}
	case BT_DPM_GET_ALLOW_OUTGOING_CALL: {
		int value = DPM_STATUS_ERROR;

		result = _bt_dpm_get_allow_bluetooth_outgoing_call(&value);
		g_array_append_vals(*out_param1, &value, sizeof(int));
		break;
	}
	case BT_DPM_SET_PAIRING_STATE: {
		dpm_status_t value = DPM_STATUS_ERROR;

		__bt_service_get_parameters(in_param1, &value,
				sizeof(int));

		result = _bt_dpm_set_bluetooth_pairing_state(value);

		break;
	}
	case BT_DPM_GET_PAIRING_STATE: {
		int value = DPM_STATUS_ERROR;

		result = _bt_dpm_get_bluetooth_pairing_state(&value);
		g_array_append_vals(*out_param1, &value, sizeof(int));
		break;
	}
	case BT_DPM_SET_PROFILE_STATE: {
		int value = DPM_STATUS_ERROR;
		int profile = DPM_PROFILE_NONE;

		__bt_service_get_parameters(in_param1, &profile,
				sizeof(int));
		__bt_service_get_parameters(in_param2, &value,
				sizeof(int));

		result = _bt_dpm_set_bluetooth_profile_state(profile, value);

		break;
	}
	case BT_DPM_GET_PROFILE_STATE: {
		int value = DPM_STATUS_ERROR;
		int profile = DPM_PROFILE_NONE;

		__bt_service_get_parameters(in_param1, &profile,
				sizeof(int));

		result = _bt_dpm_get_bluetooth_profile_state(profile, &value);
		g_array_append_vals(*out_param1, &value, sizeof(int));
		break;
	}
	case BT_DPM_SET_DESKROP_CONNECTIVITY_STATE: {
		int value = DPM_BT_ERROR;

		__bt_service_get_parameters(in_param1, &value,
				sizeof(int));

		result = _bt_dpm_set_bluetooth_desktop_connectivity_state(value);

		break;
	}
	case BT_DPM_GET_DESKROP_CONNECTIVITY_STATE: {
		int value = DPM_STATUS_ERROR;

		result = _bt_dpm_get_bluetooth_desktop_connectivity_state(&value);
		g_array_append_vals(*out_param1, &value, sizeof(int));
		break;
	}
	case BT_DPM_SET_DISCOVERABLE_STATE: {
		int value = DPM_STATUS_ERROR;

		__bt_service_get_parameters(in_param1, &value,
				sizeof(int));

		result = _bt_dpm_set_bluetooth_discoverable_state(value);

		break;
	}
	case BT_DPM_GET_DISCOVERABLE_STATE: {
		int value = DPM_STATUS_ERROR;

		result = _bt_dpm_get_bluetooth_discoverable_state(&value);
		g_array_append_vals(*out_param1, &value, sizeof(int));
		break;
	}
	case BT_DPM_SET_LIMITED_DISCOVERABLE_STATE: {
		int value = DPM_STATUS_ERROR;

		__bt_service_get_parameters(in_param1, &value,
				sizeof(int));

		result = _bt_dpm_set_bluetooth_limited_discoverable_state(value);

		break;
	}
	case BT_DPM_GET_LIMITED_DISCOVERABLE_STATE: {
		int value = DPM_STATUS_ERROR;

		result = _bt_dpm_get_bluetooth_limited_discoverable_state(&value);
		g_array_append_vals(*out_param1, &value, sizeof(int));
		break;
	}
	case BT_DPM_SET_DATA_TRANSFER_STATE: {
		int value = DPM_STATUS_ERROR;

		__bt_service_get_parameters(in_param1, &value,
				sizeof(int));

		result = _bt_dpm_set_bluetooth_data_transfer_state(value);

		break;
	}
	case BT_DPM_GET_DATA_TRANSFER_STATE: {
		int value = DPM_STATUS_ERROR;

		result = _bt_dpm_get_allow_bluetooth_data_transfer_state(&value);
		g_array_append_vals(*out_param1, &value, sizeof(int));
		break;
	}
#endif
	case BT_RFCOMM_CLIENT_CONNECT: {
		bluetooth_device_address_t address = { {0} };
		char *input_string;
		int connect_type;

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));
		input_string = (char *)g_variant_get_data(in_param2);
		__bt_service_get_parameters(in_param3, &connect_type, sizeof(int));

		if (connect_type == BT_RFCOMM_UUID)
			result = _bt_rfcomm_connect_using_uuid(&address, input_string);
		else
			result = _bt_rfcomm_connect_using_channel(&address, input_string);

		if (result != BLUETOOTH_ERROR_NONE) {
			bluetooth_rfcomm_connection_t conn_info;

			BT_ERR("BT_RFCOMM_CLIENT_CONNECT failed, send error");
			memset(&conn_info, 0x00, sizeof(bluetooth_rfcomm_connection_t));
			if (connect_type == BT_RFCOMM_UUID)
				g_strlcpy(conn_info.uuid, input_string, BLUETOOTH_UUID_STRING_MAX);
			else
				g_strlcpy(conn_info.uuid, "not_used", BLUETOOTH_UUID_STRING_MAX);

			conn_info.device_role = RFCOMM_ROLE_CLIENT;
			conn_info.socket_fd = -1;
			g_array_append_vals(*out_param1, &conn_info,
					sizeof(bluetooth_rfcomm_connection_t));
		} else {
			char *addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			BT_ERR("BT_RFCOMM_CLIENT_CONNECT success, save context");
			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender, function_name, addr);
		}
		break;
	}
	case BT_RFCOMM_SOCKET_DISCONNECT: {
		/*
		 * Bluetooth RFCOMM socket disconnection will be done from bt-api, call to bt-service
		 * is only used for privilege check, so return BLUETOOTH_ERROR_NONE from here.
		 */
		result = BLUETOOTH_ERROR_NONE;
		break;
	}
	case BT_RFCOMM_SOCKET_WRITE: {
		/*
		 * This call to bt-service is only used for privilege check, so return
		 * BLUETOOTH_ERROR_NONE from here.
		 */
		result = BLUETOOTH_ERROR_NONE;
		break;
	}
	case BT_RFCOMM_CREATE_SOCKET: {
		/*
		 * This call to bt-service is only used for privilege check, so return
		 * BLUETOOTH_ERROR_NONE from here.
		 */
		result = BLUETOOTH_ERROR_NONE;
		break;
	}
	case BT_RFCOMM_LISTEN_AND_ACCEPT: {
		char *uuid;
		int socket_fd = -1;

		sender = (char *)g_dbus_method_invocation_get_sender(context);
		uuid = (char *)g_variant_get_data(in_param1);

		result = _bt_rfcomm_socket_listen(sender, uuid, true);
		if (result > 0) {
			BT_ERR("BT_RFCOMM_LISTEN_AND_ACCEPT success, save context");

			result = BLUETOOTH_ERROR_NONE;
			_bt_save_invocation_context(context,
					result, sender, function_name, NULL);
		} else {
			g_array_append_vals(*out_param1, &socket_fd, sizeof(int));
		}
		break;
	}
	case BT_RFCOMM_LISTEN: {
		char *uuid;
		int socket_fd = -1;

		sender = (char *)g_dbus_method_invocation_get_sender(context);
		uuid = (char *)g_variant_get_data(in_param1);

		result = _bt_rfcomm_socket_listen(sender, uuid, false);
		if (result > 0) {
			BT_ERR("BT_RFCOMM_LISTEN success, save context");

			result = BLUETOOTH_ERROR_NONE;
			_bt_save_invocation_context(context,
					result, sender, function_name, NULL);
		} else {
			g_array_append_vals(*out_param1, &socket_fd, sizeof(int));
		}
		break;
	}
	case BT_RFCOMM_ACCEPT_CONNECTION: {
		char *address;

		address = (char *)g_variant_get_data(in_param1);
		result = _bt_rfcomm_reply_conn_authorization(address, TRUE);
		break;
	}
	case BT_RFCOMM_REJECT_CONNECTION: {
		char *address;

		address = (char *)g_variant_get_data(in_param1);
		result = _bt_rfcomm_reply_conn_authorization(address, FALSE);
		break;
	}
	case BT_RFCOMM_REMOVE_SOCKET: {
		/*
		 * This call to bt-service is only used for privilege check, so return
		 * BLUETOOTH_ERROR_NONE from here.
		 */
		result = BLUETOOTH_ERROR_NONE;
		break;
	}
	case BT_AUDIO_SELECT_ROLE: {
		bluetooth_audio_role_t role;

		__bt_service_get_parameters(in_param1,
				&role, sizeof(bluetooth_audio_role_t));

		result = _bt_audio_select_role(role);
		if (result == BLUETOOTH_ERROR_NONE) {
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, NULL);
		}
		break;
	}
	case BT_AV_CONNECT: {
		bluetooth_device_address_t address = { {0} };
		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_audio_connect(BT_AUDIO_A2DP, &address);

		if (result != BLUETOOTH_ERROR_NONE) {
			char addr[BT_ADDRESS_STRING_SIZE];
			_bt_convert_addr_type_to_string(addr, address.addr);
			g_array_append_vals(*out_param1, addr, BT_ADDRESS_STRING_SIZE);
		} else {
			char *addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		}
		break;
	}
	case BT_AUDIO_CONNECT: {
		bluetooth_device_address_t address = { {0} };
		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_audio_connect(BT_AUDIO_ALL, &address);

		if (result != BLUETOOTH_ERROR_NONE) {
			char addr[BT_ADDRESS_STRING_SIZE];
			_bt_convert_addr_type_to_string(addr, address.addr);
			g_array_append_vals(*out_param1, addr, BT_ADDRESS_STRING_SIZE);
		} else {
			char *addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		}
		break;
	}
	case BT_AUDIO_DISCONNECT: {
		bluetooth_device_address_t address = { {0} };
		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_audio_disconnect(BT_AUDIO_ALL, &address);

		if (result != BLUETOOTH_ERROR_NONE) {
			char addr[BT_ADDRESS_STRING_SIZE];
			_bt_convert_addr_type_to_string(addr, address.addr);
			g_array_append_vals(*out_param1, addr, BT_ADDRESS_STRING_SIZE);
		} else {
			char *addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		}
		break;
	}
	case BT_AV_DISCONNECT: {
		bluetooth_device_address_t address = { {0} };
		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_audio_disconnect(BT_AUDIO_A2DP, &address);

		if (result != BLUETOOTH_ERROR_NONE) {
			char addr[BT_ADDRESS_STRING_SIZE];
			_bt_convert_addr_type_to_string(addr, address.addr);
			g_array_append_vals(*out_param1, addr, BT_ADDRESS_STRING_SIZE);
		} else {
			char *addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		}
		break;
	}
	case BT_AG_CONNECT: {
		bluetooth_device_address_t address = { {0} };
		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_audio_connect(BT_AUDIO_HSP, &address);

		if (result != BLUETOOTH_ERROR_NONE) {
			char addr[BT_ADDRESS_STRING_SIZE];
			_bt_convert_addr_type_to_string(addr, address.addr);
			g_array_append_vals(*out_param1, addr, BT_ADDRESS_STRING_SIZE);
		} else {
			char *addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		}
		break;
	}
	case BT_AG_DISCONNECT: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_audio_disconnect(BT_AUDIO_HSP, &address);

		if (result != BLUETOOTH_ERROR_NONE) {
			char addr[BT_ADDRESS_STRING_SIZE];
			_bt_convert_addr_type_to_string(addr, address.addr);
			g_array_append_vals(*out_param1, addr, BT_ADDRESS_STRING_SIZE);
		} else {
			char *addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		}
		break;
	}
	case BT_AV_SOURCE_CONNECT: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_audio_connect(BT_AUDIO_A2DP_SOURCE, &address);
		if (result != BLUETOOTH_ERROR_NONE) {
			char addr[BT_ADDRESS_STRING_SIZE];
			_bt_convert_addr_type_to_string(addr, address.addr);
			g_array_append_vals(*out_param1, addr, BT_ADDRESS_STRING_SIZE);
		} else {
			char *addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		}
		break;
	}
	case BT_AV_SOURCE_DISCONNECT: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_audio_disconnect(BT_AUDIO_A2DP_SOURCE, &address);
		if (result != BLUETOOTH_ERROR_NONE) {
			char addr[BT_ADDRESS_STRING_SIZE];
			_bt_convert_addr_type_to_string(addr, address.addr);
			g_array_append_vals(*out_param1, addr, BT_ADDRESS_STRING_SIZE);
		} else {
			char *addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		}
		break;
	}
	case BT_AVRCP_TARGET_CONNECT: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_audio_connect(BT_AVRCP_TARGET, &address);
		if (result != BLUETOOTH_ERROR_NONE) {
			char addr[BT_ADDRESS_STRING_SIZE];
			_bt_convert_addr_type_to_string(addr, address.addr);
			g_array_append_vals(*out_param1, addr, BT_ADDRESS_STRING_SIZE);
		} else {
			char *addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		}
		break;
	}
	case BT_AVRCP_TARGET_DISCONNECT: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_audio_disconnect(BT_AVRCP_TARGET, &address);
		if (result != BLUETOOTH_ERROR_NONE) {
			char addr[BT_ADDRESS_STRING_SIZE];
			_bt_convert_addr_type_to_string(addr, address.addr);
			g_array_append_vals(*out_param1, addr, BT_ADDRESS_STRING_SIZE);
		} else {
			char *addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		}
		break;
	}
	case BT_AVRCP_CONTROL_CONNECT: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_audio_connect(BT_AVRCP, &address);
		if (result != BLUETOOTH_ERROR_NONE) {
			char addr[BT_ADDRESS_STRING_SIZE];
			_bt_convert_addr_type_to_string(addr, address.addr);
			g_array_append_vals(*out_param1, addr, BT_ADDRESS_STRING_SIZE);
		} else {
			char *addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		}
		break;
	}
	case BT_AVRCP_CONTROL_DISCONNECT: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_audio_disconnect(BT_AVRCP, &address);
		if (result != BLUETOOTH_ERROR_NONE) {
			char addr[BT_ADDRESS_STRING_SIZE];
			_bt_convert_addr_type_to_string(addr, address.addr);
			g_array_append_vals(*out_param1, addr, BT_ADDRESS_STRING_SIZE);
		} else {
			char *addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(addr, address.addr);
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		}
		break;
	}
	case BT_AVRCP_HANDLE_CONTROL: {
		int key_code;
		__bt_service_get_parameters(in_param1, &key_code, sizeof(int));
		result = _bt_avrcp_control_cmd(key_code);
		break;
	}
	case BT_AVRCP_HANDLE_CONTROL_TO_DEST: {
		int key_code;
		bluetooth_device_address_t address = { { 0 } };

		__bt_service_get_parameters(in_param1, &key_code, sizeof(int));
		__bt_service_get_parameters(in_param2,
						&address, sizeof(bluetooth_device_address_t));

		result = _bt_avrcp_control_cmd_to_dest(key_code, &address);
		break;
	}
	case BT_AVRCP_CONTROL_SET_PROPERTY: {
		int type;
		unsigned int value;

		__bt_service_get_parameters(in_param1,
				&type, sizeof(int));
		__bt_service_get_parameters(in_param2,
				&value, sizeof(unsigned int));

		result = _bt_avrcp_control_set_property(type, value);
		break;
	}
	case BT_AVRCP_TRANSPORT_SET_PROPERTY: {
		int type;
		unsigned int value;
		BT_DBG("+");

		__bt_service_get_parameters(in_param1,
				&type, sizeof(int));
		__bt_service_get_parameters(in_param2,
				&value, sizeof(unsigned int));

		result = _bt_avrcp_transport_set_property(type, value);
		BT_DBG("-");
		break;
	}
	case BT_AVRCP_CONTROL_GET_PROPERTY: {
		int type;

		__bt_service_get_parameters(in_param1, &type, sizeof(int));

		result = _bt_avrcp_control_get_property(type);
		/* Save invocation */
		if (result == BLUETOOTH_ERROR_NONE) {
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, g_memdup(&type, sizeof(int)));
		}
		break;
	}
	case BT_AVRCP_GET_TRACK_INFO: {
		result = _bt_avrcp_control_get_track_info();
		/* Save invocation */
		if (result == BLUETOOTH_ERROR_NONE) {
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, NULL);
		}
		break;
	}
	case BT_HDP_REGISTER_SINK_APP: {
		unsigned short data_type;
		bt_hdp_role_type_t role;
		bt_hdp_qos_type_t channel_type;
		int *app_id = g_new0(int, 1);

		__bt_service_get_parameters(in_param1,
				&data_type, sizeof(short));
		__bt_service_get_parameters(in_param2,
				&role, sizeof(bt_hdp_role_type_t));
		__bt_service_get_parameters(in_param3,
				&channel_type, sizeof(bt_hdp_qos_type_t));
		sender = (char*)g_dbus_method_invocation_get_sender(context);

		result = _bt_hdp_app_register(role,
				channel_type, data_type, sender, app_id);
		if (result != BLUETOOTH_ERROR_NONE) {
			g_array_append_vals(*out_param1, &app_id, sizeof(app_id));
			g_free(app_id);
		} else {
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)app_id);
		}
		break;
	}
	case BT_HDP_UNREGISTER_SINK_APP: {
		char *app_handle;
		int *app_id = g_new0(int, 1);

		app_handle = (char *)g_variant_get_data(in_param1);
		sscanf(app_handle, "health_app_%d", app_id);
		result = _bt_hdp_app_unregister(*app_id);
		if (result != BLUETOOTH_ERROR_NONE) {
			g_free(app_id);
		} else {
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)app_id);
		}
		break;
	}
	case BT_HDP_CONNECT: {
		int app_id = -1;
		char *app_handle;

		bt_hdp_connected_t *conn_info = NULL;

		conn_info = g_malloc0(sizeof(bt_hdp_connected_t));

		app_handle = (char *)g_variant_get_data(in_param1);
		conn_info->app_handle = app_handle;
		sscanf(app_handle, "health_app_%d", &app_id);

		__bt_service_get_parameters(in_param2,
				&(conn_info->type), sizeof(bt_hdp_qos_type_t));
		__bt_service_get_parameters(in_param3,
				&(conn_info->device_address),
				sizeof(bluetooth_device_address_t));

		result = _bt_hdp_connect(app_id, &(conn_info->device_address),
				conn_info->type, (int *)(&(conn_info->channel_id)));
		if (result != BLUETOOTH_ERROR_NONE) {
			g_array_append_vals(*out_param1, conn_info,
					sizeof(bt_hdp_connected_t));
			g_free(conn_info);
		} else {
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)conn_info);
		}
		break;
	}
	case BT_HDP_DISCONNECT: {
		bt_hdp_disconnected_t *hdp_disconn_info = NULL;

		hdp_disconn_info = g_malloc0(sizeof(bt_hdp_disconnected_t));

		__bt_service_get_parameters(in_param1,
				&(hdp_disconn_info->channel_id), sizeof(int));
		__bt_service_get_parameters(in_param2,
				&(hdp_disconn_info->device_address),
				sizeof(bluetooth_device_address_t));

		result = _bt_hdp_disconnect(hdp_disconn_info->channel_id);
		if (result != BLUETOOTH_ERROR_NONE) {
			g_array_append_vals(*out_param1, hdp_disconn_info,
					sizeof(bt_hdp_disconnected_t));
			g_free(hdp_disconn_info);
		} else {
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)hdp_disconn_info);
		}
		break;
	}
	case BT_HDP_GET_FD: {
		int *channel_id = g_new0(int, 1);

		__bt_service_get_parameters(in_param1, channel_id, sizeof(int));

		result = _bt_hdp_get_fd(*channel_id);
		if (result != BLUETOOTH_ERROR_NONE) {
			g_free(channel_id);
		} else {
			sender = (char*)g_dbus_method_invocation_get_sender(context);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)channel_id);
		}
		break;
	}
	case BT_HDP_SEND_DATA: {
		/*
		 * This call to bt-service is only used for privilege check, so return
		 * BLUETOOTH_ERROR_NONE from here.
		 */
		result = BLUETOOTH_ERROR_NONE;
		break;
	}
	case BT_OOB_READ_LOCAL_DATA: {	//208
		bt_oob_data_t local_oob_data;

		memset(&local_oob_data, 0x00, sizeof(bt_oob_data_t));
		result = _bt_oob_read_local_data(&local_oob_data);

		g_array_append_vals(*out_param1, &local_oob_data,
				sizeof(bt_oob_data_t));

		break;
	}
	case BT_OOB_ADD_REMOTE_DATA: {   //209
		bluetooth_device_address_t address = { {0} };
		unsigned short address_type;
		bt_oob_data_t remote_oob_data;

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2,
				&address_type, sizeof(unsigned short));
		__bt_service_get_parameters(in_param3,
				&remote_oob_data, sizeof(bt_oob_data_t));

		result = _bt_oob_add_remote_data(&address, address_type, &remote_oob_data);

		break;
	}
	case BT_OOB_REMOVE_REMOTE_DATA: {	//210
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_oob_remove_remote_data(&address);

		break;
	}
	case BT_AVRCP_SET_TRACK_INFO: {
		media_metadata_t data;
		media_metadata_attributes_t meta_data;

		memset(&data, 0x00, sizeof(media_metadata_t));
		memset(&meta_data, 0x00, sizeof(media_metadata_attributes_t));

		__bt_service_get_parameters(in_param1,
				&data, sizeof(media_metadata_t));

		meta_data.title = g_strdup(data.title);
		meta_data.artist = g_strdup(data.artist);
		meta_data.album = g_strdup(data.album);
		meta_data.genre = g_strdup(data.genre);
		meta_data.total_tracks = data.total_tracks;
		meta_data.number = data.number;
		meta_data.duration = (int64_t) data.duration;

		result = _bt_avrcp_set_track_info(&meta_data);

		g_free((gpointer)meta_data.title);
		g_free((gpointer)meta_data.artist);
		g_free((gpointer)meta_data.album);
		g_free((gpointer)meta_data.genre);

		break;
	}
	case BT_AVRCP_SET_PROPERTY: {
		int type;
		unsigned int value;

		sender = (char *)g_dbus_method_invocation_get_sender(context);
		BT_INFO("Sender %s", sender);

		__bt_service_get_parameters(in_param1,
				&type, sizeof(int));
		__bt_service_get_parameters(in_param2,
				&value, sizeof(unsigned int));

		if (value == STATUS_PLAYING) {
			if (current_sender_playing)
				g_free(current_sender_playing);
			current_sender_playing = g_strdup(sender);
		}
		if (g_strcmp0(sender, current_sender_playing) == 0 ||
			current_sender_playing == NULL) {
			BT_INFO("Current Player Status %d type %d", value, type);
		} else {
			BT_INFO("Current Player and this sender are different");
			result = BLUETOOTH_ERROR_NONE;
			break;
		}
		result = _bt_avrcp_set_property(type, value);

		break;
	}
	case BT_AVRCP_SET_PROPERTIES: {
		media_player_settings_t properties;

		memset(&properties, 0x00, sizeof(media_player_settings_t));
		__bt_service_get_parameters(in_param1,
				&properties, sizeof(media_player_settings_t));

		result = _bt_avrcp_set_properties(&properties);

		break;
	}
	case BT_SET_ADVERTISING_DATA: {
		char *app = NULL;
		int *adv_handle;
		bluetooth_advertising_data_t adv = { {0} };
		int length;
		gboolean use_reserved_slot = FALSE;

		app = (char *)g_dbus_method_invocation_get_sender(context);
		adv_handle = g_malloc0(sizeof(int));

		__bt_service_get_parameters(in_param1,
				adv_handle, sizeof(int));
		__bt_service_get_parameters(in_param3,
				&length, sizeof(int));
		__bt_service_get_parameters(in_param2,
				&adv, length);
		__bt_service_get_parameters(in_param4,
				&use_reserved_slot, sizeof(gboolean));
		result = _bt_set_advertising_data(app, *adv_handle,
				&adv, length, use_reserved_slot);
		if (result != BLUETOOTH_ERROR_NONE) {
			BT_ERR("Set Advertising data failed!!");
			g_free(adv_handle);
		} else {
			_bt_save_invocation_context(context, result, app,
					function_name, (gpointer)adv_handle);
		}
		break;
	}
	case BT_SET_SCAN_RESPONSE_DATA: {
		char *app = NULL;
		int *adv_handle;
		bluetooth_scan_resp_data_t rsp = { {0} };
		int length;
		gboolean use_reserved_slot = FALSE;

		app = (char *)g_dbus_method_invocation_get_sender(context);
		adv_handle = g_malloc0(sizeof(int));

		__bt_service_get_parameters(in_param1,
				adv_handle, sizeof(int));
		__bt_service_get_parameters(in_param3,
				&length, sizeof(int));
		__bt_service_get_parameters(in_param2,
				&rsp, length);
		__bt_service_get_parameters(in_param4,
				&use_reserved_slot, sizeof(gboolean));

		result = _bt_set_scan_response_data(app, *adv_handle,
				&rsp, length, use_reserved_slot);

		if (result != BLUETOOTH_ERROR_NONE) {
			BT_ERR("Set Scan Response Data failed!!");
			g_free(adv_handle);
		} else {
			_bt_save_invocation_context(context, result, app,
					function_name, (gpointer)adv_handle);
		}
		break;
	}
	case BT_SET_ADVERTISING: {
		char *app = NULL;
		int *adv_handle;
		gboolean enable = FALSE;
		gboolean use_reserved_slot = FALSE;

		adv_handle = g_malloc0(sizeof(int));
		__bt_service_get_parameters(in_param1,
				adv_handle, sizeof(int));
		__bt_service_get_parameters(in_param2,
				&enable, sizeof(gboolean));
		__bt_service_get_parameters(in_param3,
				&use_reserved_slot, sizeof(gboolean));

		app = (char *)g_dbus_method_invocation_get_sender(context);

		result = _bt_set_advertising(app, *adv_handle,
				enable, use_reserved_slot);
		if (result != BLUETOOTH_ERROR_NONE) {
			BT_ERR("Start Advertising failed!!");
			g_free(adv_handle);
		} else {
			_bt_save_invocation_context(context, result, app,
					function_name, (gpointer)adv_handle);
		}
		break;
	}
	case BT_SET_CUSTOM_ADVERTISING: {
		char *app = NULL;
		int *adv_handle;
		gboolean enable = FALSE;
		bluetooth_advertising_params_t adv_params;
		gboolean use_reserved_slot = FALSE;

		app = (char *)g_dbus_method_invocation_get_sender(context);
		adv_handle = g_malloc0(sizeof(int));

		__bt_service_get_parameters(in_param1, adv_handle,
				sizeof(int));
		__bt_service_get_parameters(in_param2, &enable,
				sizeof(gboolean));
		__bt_service_get_parameters(in_param3, &adv_params,
				sizeof(bluetooth_advertising_params_t));
		__bt_service_get_parameters(in_param4, &use_reserved_slot,
				sizeof(gboolean));

		BT_DBG("bluetooth_advertising_params_t [%f %f %d %d]",
				adv_params.interval_min, adv_params.interval_max,
				adv_params.filter_policy, adv_params.type);
		result = _bt_set_custom_advertising(app, *adv_handle,
				enable, &adv_params, use_reserved_slot);
		if (result != BLUETOOTH_ERROR_NONE) {
			BT_ERR("Start Custom Advertising failed!!");
			g_free(adv_handle);
		} else {
			_bt_save_invocation_context(context, result, app,
					function_name, (gpointer)adv_handle);
		}
		break;
	}
	case BT_GET_SCAN_RESPONSE_DATA: {
		bluetooth_scan_resp_data_t rsp = { {0} };
		char *app = NULL;
		int length = 0;
		int adv_handle;
		app = (char *)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1, &adv_handle,
				sizeof(int));

		result = _bt_get_scan_response_data(app, adv_handle, &rsp, &length);
		if (result == BLUETOOTH_ERROR_NONE)
			g_array_append_vals(*out_param1, rsp.data, length);

		break;
	}
	case BT_GET_ADVERTISING_DATA: {
		bluetooth_advertising_data_t adv = { {0} };
		char *app = NULL;
		int length = 0;
		int adv_handle;
		app = (char *)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1, &adv_handle,
				sizeof(int));

		result = _bt_get_advertising_data(app, adv_handle, &adv, &length);
		if (result == BLUETOOTH_ERROR_NONE)
			g_array_append_vals(*out_param1, adv.data, length);

		break;
	}
	case BT_GATT_REGISTER_APPLICATION: {
		BT_DBG("Register GATT application:Unhandled!!");
		break;
	}
	case BT_GATT_SERVER_REGISTER: {
		BT_DBG("GATT Server instance initialization");
		char *app;

		app = (char *)g_dbus_method_invocation_get_sender(context);
		BT_INFO("GATT Server Unique Name [%s]", app);

		/* No ADV handle: Set 0 */
		result = _bt_register_server_instance(app, 0);
		BT_INFO("GATT Server: Register Server result [%d]", result);

		if (result != BLUETOOTH_ERROR_NONE) {
			BT_ERR("GATT Server registration failed!!");
		} else {
			_bt_save_invocation_context(context, result, app,
					function_name, NULL);
		}
		break;
	}
	case BT_GATT_SERVER_DEREGISTER: {
		char *app;
		app = (char*)g_dbus_method_invocation_get_sender(context);

		result = _bt_unregister_server_instance(app, 0/* Adv Handle*/);
		break;
	}
	case BT_GATT_SERVER_ADD_SERVICE: {
		BT_DBG("GATT Server Add Service");
		int service_type;
		int num_handles;
		char *svc_uuid;
		int instance_id;
		char *app;
		int *tmp_inst_id = NULL;

		app = (char *)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1, &service_type,
				sizeof(int));
		__bt_service_get_parameters(in_param2, &num_handles,
				sizeof(int));
		svc_uuid = (char *)g_variant_get_data(in_param3);
		__bt_service_get_parameters(in_param4, &instance_id,
				sizeof(int));

		BT_INFO("GATT Server Service UUID [%s]", svc_uuid);
		result = _bt_gatt_server_add_service(app, service_type, num_handles, svc_uuid, instance_id);
		if (result != BLUETOOTH_ERROR_NONE) {
			BT_ERR("GATT Server Add Service failed!!");
		} else {
			tmp_inst_id = g_malloc0(sizeof(int));
			*tmp_inst_id = instance_id;
			BT_INFO("GATT Server: Save Instance ID in request Inst id [%d] request_id [%d]",
					*tmp_inst_id, instance_id);
			_bt_save_invocation_context(context, result, app,
					function_name, (gpointer)tmp_inst_id);
		}
		break;
	}
	case BT_GATT_SERVER_ADD_CHARACTERISTIC: {
		BT_DBG("GATT Server Add Characteristic");
		char *char_uuid;
		char *app;
		int *tmp_inst_id = NULL;
		bluetooth_gatt_server_attribute_params_t param;
		memset(&param, 0, sizeof(bluetooth_gatt_server_attribute_params_t));

		app = (char*)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1, &param,
				sizeof(bluetooth_gatt_server_attribute_params_t));
		char_uuid = (char *)g_variant_get_data(in_param2);

		BT_INFO("GATT Server Characteristic UUID [%s]", char_uuid);

		result = _bt_gatt_server_add_characteristic(app, char_uuid, &param);
		if (result != BLUETOOTH_ERROR_NONE) {
			BT_ERR("GATT Server Add Service failed!!");
		} else {
			tmp_inst_id = g_malloc0(sizeof(int));
			*tmp_inst_id = param.instance_id;
			BT_INFO("GATT Server: Save Instance ID in request Inst id [%d] request_id [%d]",
					*tmp_inst_id, param.instance_id);
			_bt_save_invocation_context(context, result, app,
					function_name, (gpointer)tmp_inst_id);
		}
		break;
	}
	case BT_GATT_SERVER_ADD_DESCRIPTOR: {
		BT_DBG("GATT Server Add Descriptor");
		char *desc_uuid;
		int instance_id;
		int service_handle;
		bt_gatt_permission_t perm;
		char *app;
		int *tmp_inst_id = NULL;
		memset(&perm, 0, sizeof(bt_gatt_permission_t));

		app = (char *)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1, &service_handle,
				sizeof(int));
		__bt_service_get_parameters(in_param2, &instance_id,
				sizeof(int));
		desc_uuid = (char *)g_variant_get_data(in_param4);
		__bt_service_get_parameters(in_param3, &perm,
				sizeof(bt_gatt_permission_t));

		BT_INFO("GATT Server Descriptor UUID [%s]", desc_uuid);

		result = _bt_gatt_server_add_descriptor(app, desc_uuid, &perm, service_handle, instance_id);
		if (result != BLUETOOTH_ERROR_NONE) {
			BT_ERR("GATT Server Add Service failed!!");
		} else {
			tmp_inst_id = g_malloc0(sizeof(int));
			*tmp_inst_id = instance_id;
			BT_INFO("GATT Server: Save Instance ID in request Inst id [%d] request_id [%d]",
					*tmp_inst_id, instance_id);
			_bt_save_invocation_context(context, result, app,
					function_name, (gpointer)tmp_inst_id);
		}
		break;
	}
	case BT_GATT_SERVER_START_SERVICE: {
		int service_handle;
		int instance_id;

		char *app;
		int *tmp_inst_id = NULL;
		app = (char*)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1, &service_handle,
				sizeof(int));
		__bt_service_get_parameters(in_param2, &instance_id,
				sizeof(int));

		result = _bt_gatt_server_start_service(app, service_handle, instance_id);

		if (BLUETOOTH_ERROR_NONE == result) {
			tmp_inst_id = g_malloc0(sizeof(int));
			*tmp_inst_id = instance_id;
			BT_INFO("GATT Server: Save Instance ID in request Inst id [%d] request_id [%d]",
					*tmp_inst_id, instance_id);
			_bt_save_invocation_context(context, result, app,
					function_name, (gpointer)tmp_inst_id);
		}
		break;
	}
	case BT_GATT_SERVER_STOP_SERVICE: {
		int service_handle;
		int instance_id;
		char *app;
		int *tmp_inst_id = NULL;
		app = (char*)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1, &service_handle,
				sizeof(int));
		__bt_service_get_parameters(in_param2, &instance_id,
				sizeof(int));

		result = _bt_gatt_server_stop_service(app, service_handle, instance_id);

		if (BLUETOOTH_ERROR_NONE == result) {
			tmp_inst_id = g_malloc0(sizeof(int));
			*tmp_inst_id = instance_id;
			BT_INFO("GATT Server: Save Instance ID in request Inst id [%d] request_id [%d]",
					*tmp_inst_id, instance_id);
			_bt_save_invocation_context(context, result, app,
					function_name, (gpointer)tmp_inst_id);
		}
		break;
	}
	case BT_GATT_SERVER_DELETE_SERVICE: {
		int service_handle;
		int instance_id;
		int *tmp_inst_id = NULL;
		char *app;
		app = (char*)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1, &service_handle,
				sizeof(int));
		__bt_service_get_parameters(in_param2, &instance_id,
				sizeof(int));

		result = _bt_gatt_server_delete_service(app, service_handle, instance_id);

		if (BLUETOOTH_ERROR_NONE == result) {
			tmp_inst_id = g_malloc0(sizeof(int));
			*tmp_inst_id = instance_id;
			BT_INFO("GATT Server: Save Instance ID in request Inst id [%d] request_id [%d]",
					*tmp_inst_id, instance_id);
			_bt_save_invocation_context(context, result, app,
					function_name, (gpointer)tmp_inst_id);
		}
		break;
	}
	case BT_GATT_SERVER_SEND_RESPONSE: {
		bluetooth_gatt_server_response_params_t param;
		bluetooth_gatt_att_data_t data;
		char *app;

		memset(&param, 0x00, sizeof(bluetooth_gatt_server_response_params_t));
		memset(&data, 0x00, sizeof(bluetooth_gatt_att_data_t));

		app = (char*)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1, &data,
				sizeof(bluetooth_gatt_att_data_t));
		__bt_service_get_parameters(in_param2, &param,
				sizeof(bluetooth_gatt_server_response_params_t));

		result = _bt_gatt_server_send_response(app, &data, &param);

		break;
	}
	case BT_GATT_SERVER_ACQURE_WRITE_RESPONSE: {

		bluetooth_gatt_server_acquire_response_params_t param;
		char *app;
		GDBusMessage *msg;
		msg = g_dbus_method_invocation_get_message(context);
		GUnixFDList *fd_list;
		int fd  = -1;;
		int *fd_list_array;
		int len;

		BT_ERR("sending acquire write  respose \n");

		fd_list = g_dbus_message_get_unix_fd_list(msg);

		memset(&param, 0x00, sizeof(bluetooth_gatt_server_acquire_response_params_t));

		app = (char*)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1, &param,
				sizeof(bluetooth_gatt_server_acquire_response_params_t));

		BT_ERR("sending acquire write  respose sent \n  ");

		fd_list_array = (int *)g_unix_fd_list_peek_fds(fd_list, &len);
		BT_INFO("Num fds in fd_list is : %d, fd_list[0]: %d", len, fd_list_array[0]);
		fd = fd_list_array[0];

		param.fd = fd;
		result = _bt_gatt_server_acquire_send_response(app, &param, fd_list);

		break;
	}
	case BT_GATT_SERVER_ACQUIRE_NOTIFY_RESPONSE: {

		bluetooth_gatt_server_acquire_response_params_t param;
		char *app;
		GDBusMessage *msg;
		msg = g_dbus_method_invocation_get_message(context);
		GUnixFDList *fd_list;
		int fd  = -1;
		int *fd_list_array;
		int len;

		BT_ERR("sending acquire write  respose \n");

		fd_list = g_dbus_message_get_unix_fd_list(msg);

		memset(&param, 0x00, sizeof(bluetooth_gatt_server_acquire_response_params_t));

		app = (char*)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1, &param,
				sizeof(bluetooth_gatt_server_acquire_response_params_t));

		BT_ERR("sending acquire write  respose sent \n  ");

		fd_list_array = (int *)g_unix_fd_list_peek_fds(fd_list, &len);
		BT_INFO("Num fds in fd_list is : %d, fd_list[0]: %d", len, fd_list_array[0]);
		fd = fd_list_array[0];

		param.fd = fd;
		result = _bt_gatt_server_acquire_send_response(app, &param, fd_list);

		break;
	}
	case BT_GATT_SERVER_SEND_INDICATION: {
		bluetooth_gatt_server_indication_params_t param;
		bluetooth_gatt_att_data_t data;
		bluetooth_device_address_t address;

		char *app;
		app = (char*)g_dbus_method_invocation_get_sender(context);

		memset(&param, 0x00, sizeof(bluetooth_gatt_server_indication_params_t));
		memset(&data, 0x00, sizeof(bluetooth_gatt_att_data_t));
		memset(&address, 0x00, sizeof(bluetooth_device_address_t));

		__bt_service_get_parameters(in_param1, &data,
				sizeof(bluetooth_gatt_att_data_t));
		__bt_service_get_parameters(in_param2, &param,
				sizeof(bluetooth_gatt_server_indication_params_t));
		__bt_service_get_parameters(in_param3, &address,
				sizeof(bluetooth_device_address_t));

		result = _bt_gatt_server_send_indication(app, &address, &data, &param);

		break;
	}
	case BT_GATT_SERVER_UPDATE_VALUE: {
		bluetooth_gatt_server_update_value_t param;
		int instance_id;
		char *app;

		memset(&param, 0x00, sizeof(bluetooth_gatt_server_update_value_t));

		app = (char*)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1, &instance_id,
				sizeof(int));
		__bt_service_get_parameters(in_param2, &param,
				sizeof(bluetooth_gatt_server_update_value_t));

		result = _bt_gatt_server_update_attribute_value(app, instance_id, &param);
		break;
	}
	case BT_GET_ATT_MTU: {
		bluetooth_device_address_t address = { {0} };
		unsigned int mtu = 0;

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));
		result = _bt_get_att_mtu(&address, &mtu);
		BT_DBG("MTU: %d", mtu);

		if (result == BLUETOOTH_ERROR_NONE) {
			g_array_append_vals(*out_param1, &mtu,
					sizeof(unsigned int));
		}
		break;
	}
#ifdef TIZEN_GATT_CLIENT
	/* Local Async */
	case BT_GATT_CLIENT_REGISTER: {
		bluetooth_device_address_t address = { {0} };
		char *addr;

		sender = (char*)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_register_gatt_client_instance(sender, &address);
		if (BLUETOOTH_ERROR_NONE == result) {
			addr = g_malloc0(BT_ADDRESS_STRING_SIZE);

			_bt_convert_addr_type_to_string(addr, address.addr);
			BT_INFO("GATT Client: Save Invocation data GATT CLient Register app[%s] address [%s]", sender, addr);

			/* Save the informations to invocation */
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		}
		break;
	}
	/* Sync */
	case BT_GATT_CLIENT_UNREGISTER: {
		int client_id;

		sender = (char*)g_dbus_method_invocation_get_sender(context);
		__bt_service_get_parameters(in_param1,
				&client_id, sizeof(int));

		result = _bt_unregister_gatt_client_instance(sender, client_id);
		break;
	}

	/* Async: Remote Interaction  */
	case BT_GATT_ACQUIRE_WRITE: {

		int fd = -1;
		int mtu = -1;
		bluetooth_gatt_client_char_prop_info_t param;

		BT_INFO("GATT Client ACQUIRE WRITE recived");

		sender = (char*)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1, &param,
					sizeof(bluetooth_gatt_client_char_prop_info_t));


		result   = _bt_gatt_acquire_write(&param , &fd, &mtu);

		g_array_append_vals(*out_param1, &mtu, sizeof(int));

		if (BLUETOOTH_ERROR_NONE == result)
			BT_INFO("GATT Client: Save Invocation data for characteristic props app[%s] fd[ %d]", sender, fd);
		else {
			BT_ERR("Gatt Client Acqure Write");
			break;
		}

		GUnixFDList *fd_list = NULL;
		GError *error = NULL;

		/* Add socket fd to unix_fd_list */
		fd_list = g_unix_fd_list_new();
		g_unix_fd_list_append(fd_list, fd, &error);
		g_assert_no_error(error);

		_bt_service_method_return_with_unix_fd_list(
				context, *out_param1, result, fd_list);

		close(fd);
		g_object_unref(fd_list);

		break;
	}
	/* Async: Remote Interaction  */
	case BT_CONNECT_LE: {
		bluetooth_device_address_t address = { {0} };
		gboolean auto_connect;
		char *addr;
		int client_id;

		sender = (char*)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2,
				&auto_connect, sizeof(gboolean));
		__bt_service_get_parameters(in_param3,
				&client_id, sizeof(int));

		BT_INFO("GATT Client: client_id[%d]", client_id);
		result = _bt_connect_le_device(&address, auto_connect, client_id);
		if (BLUETOOTH_ERROR_NONE == result) {

			addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			BT_INFO("GATT Client: Save Invocation data Connect LE app[%s]", sender);

			_bt_convert_addr_type_to_string(addr, address.addr);

			/* Save the informations to invocation */
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		} else {
			BT_INFO("GATT Client: gatt connect failed");
			g_array_append_vals(*out_param1, &address,
					sizeof(bluetooth_device_address_t));
		}
		break;
	}
	/* Async: Remote Interaction  */
	case BT_DISCONNECT_LE: {
		bluetooth_device_address_t address = { {0} };
		char *addr;
		int client_id;

		sender = (char*)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2,
				&client_id, sizeof(int));

		BT_INFO("GATT Disconnect using client Interface [%d]", client_id);
		result = _bt_disconnect_le_device(&address, client_id);
		if (BLUETOOTH_ERROR_NONE == result) {
			addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
			BT_INFO("GATT Client: Save Invocation data DisConnect LE app[%s]", sender);

			_bt_convert_addr_type_to_string(addr, address.addr);

			/* Save the informations to invocation */
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		} else {
			g_array_append_vals(*out_param1, &address,
					sizeof(bluetooth_device_address_t));
		}
		break;
	}
	case BT_GATT_GET_PRIMARY_SERVICES: {
		char *addr;

		sender = (char*)g_dbus_method_invocation_get_sender(context);
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1, &address,
				sizeof(bluetooth_device_address_t));

		addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
		_bt_convert_addr_type_to_string(addr, address.addr);

		result = _bt_gatt_get_primary_services(addr);
		if (BLUETOOTH_ERROR_NONE == result) {
			BT_INFO("GATT Client: Save Remote GATT Server address addr [%s] app[%s]",
					addr, sender);
			_bt_save_invocation_context(context, result, sender,
					function_name, (gpointer)addr);
		} else
			g_free(addr);
		break;
	}
	/* Local Async */
	case BT_GATT_GET_SERVICE_PROPERTIES: {
		bluetooth_gatt_client_svc_prop_info_t param;

		sender = (char*)g_dbus_method_invocation_get_sender(context);
		memset(&param, 0x00, sizeof(bluetooth_gatt_client_svc_prop_info_t));

		__bt_service_get_parameters(in_param1, &param,
				sizeof(bluetooth_gatt_client_svc_prop_info_t));

		result = _bt_gatt_get_all_characteristic(&param);
		if (BLUETOOTH_ERROR_NONE == result) {
			BT_INFO("GATT Client: Save Invocation data for Service props app[%s]", sender);

			/* Save the informations to invocation */
			_bt_save_invocation_context(context, result, sender,
					function_name,
					(gpointer)g_memdup(&param, sizeof(bluetooth_gatt_client_svc_prop_info_t)));
		}
		break;
	}
	case BT_GATT_GET_CHARACTERISTIC_PROPERTIES: {
		bluetooth_gatt_client_char_prop_info_t param;

		sender = (char*)g_dbus_method_invocation_get_sender(context);
		memset(&param, 0x00, sizeof(bluetooth_gatt_client_char_prop_info_t));

		__bt_service_get_parameters(in_param1, &param,
				sizeof(bluetooth_gatt_client_char_prop_info_t));

		result = _bt_gatt_get_all_characteristic_properties(&param);
		if (BLUETOOTH_ERROR_NONE == result) {
			BT_INFO("GATT Client: Save Invocation data for characteristic props app[%s]", sender);

			/* Save the informations to invocation */
			_bt_save_invocation_context(context, result, sender,
					function_name,
					(gpointer)g_memdup(&param, sizeof(bluetooth_gatt_client_char_prop_info_t)));
		}
		break;
	}
	/* Async: Remote Interaction  */
	case BT_GATT_READ_CHARACTERISTIC: {
		bluetooth_gatt_client_char_prop_info_t param;

		sender = (char*)g_dbus_method_invocation_get_sender(context);
		memset(&param, 0x00, sizeof(bluetooth_gatt_client_char_prop_info_t));

		__bt_service_get_parameters(in_param1, &param,
				sizeof(bluetooth_gatt_client_char_prop_info_t));

		result = _bt_gatt_read_characteristic_value(&param);
		if (BLUETOOTH_ERROR_NONE == result) {
			BT_INFO("GATT Client: Save Invocation data for characteristic props app[%s]", sender);

			/* Save the informations to invocation */
			_bt_save_invocation_context(context, result, sender,
					function_name,
					(gpointer)g_memdup(&param, sizeof(bluetooth_gatt_client_char_prop_info_t)));
		}
		break;
	}
	/* Async: Remote Interaction  */
	case BT_GATT_READ_DESCRIPTOR_VALUE: {
		bluetooth_gatt_client_desc_prop_info_t param;

		sender = (char*)g_dbus_method_invocation_get_sender(context);
		memset(&param, 0x00, sizeof(bluetooth_gatt_client_desc_prop_info_t));

		__bt_service_get_parameters(in_param1, &param,
				sizeof(bluetooth_gatt_client_desc_prop_info_t));

		result = _bt_gatt_read_descriptor_value(&param);
		if (BLUETOOTH_ERROR_NONE == result) {
			BT_INFO("GATT Client: Save Invocation data for Descriptor's props app[%s]", sender);

			/* Save the informations to invocation */
			_bt_save_invocation_context(context, result, sender,
					function_name,
					(gpointer)g_memdup(&param, sizeof(bluetooth_gatt_client_desc_prop_info_t)));
		}
		break;
	}
	/* Async: Remote Interaction  */
	case BT_GATT_WRITE_CHARACTERISTIC_VALUE_BY_TYPE: {
		bluetooth_gatt_client_char_prop_info_t param;
		bluetooth_gatt_att_data_t data;
		bluetooth_gatt_write_type_e write_type;

		sender = (char*)g_dbus_method_invocation_get_sender(context);

		memset(&data, 0x00, sizeof(bluetooth_gatt_att_data_t));
		memset(&param, 0x00, sizeof(bluetooth_gatt_client_char_prop_info_t));

		__bt_service_get_parameters(in_param1, &param,
				sizeof(bluetooth_gatt_client_char_prop_info_t));
		__bt_service_get_parameters(in_param2, &data,
				sizeof(bluetooth_gatt_att_data_t));
		__bt_service_get_parameters(in_param3, &write_type,
				sizeof(bluetooth_gatt_write_type_e));

		result = _bt_gatt_write_characteristic_value_by_type(&param , &data, write_type);
		if (BLUETOOTH_ERROR_NONE == result) {
			BT_INFO("GATT Client: Save Invocation data for characteristic props app[%s]", sender);

			/* Save the informations to invocation */
			_bt_save_invocation_context(context, result, sender,
					function_name,
					(gpointer)g_memdup(&param, sizeof(bluetooth_gatt_client_char_prop_info_t)));
		}
		break;
	}
	/* Async: Remote Interaction  */
	case BT_GATT_WRITE_DESCRIPTOR_VALUE: {
		bluetooth_gatt_client_desc_prop_info_t param;
		bluetooth_gatt_att_data_t data;
		bluetooth_gatt_write_type_e write_type;

		sender = (char*)g_dbus_method_invocation_get_sender(context);
		memset(&data, 0x00, sizeof(bluetooth_gatt_att_data_t));
		memset(&param, 0x00, sizeof(bluetooth_gatt_client_desc_prop_info_t));

		__bt_service_get_parameters(in_param1, &param,
				sizeof(bluetooth_gatt_client_desc_prop_info_t));
		__bt_service_get_parameters(in_param2, &data,
				sizeof(bluetooth_gatt_att_data_t));
		__bt_service_get_parameters(in_param3, &write_type,
				sizeof(bluetooth_gatt_write_type_e));

		result = _bt_gatt_write_descriptor_value_by_type(&param, &data, write_type);
		if (BLUETOOTH_ERROR_NONE == result) {
			BT_INFO("GATT Client: Save Invocation data for descriptor props app[%s]", sender);

			/* Save the informations to invocation */
			_bt_save_invocation_context(context, result, sender,
					function_name,
					(gpointer)g_memdup(&param, sizeof(bluetooth_gatt_client_desc_prop_info_t)));
		}
		break;
	}
	/* Local Async */
	case BT_GATT_WATCH_CHARACTERISTIC: {
		bluetooth_gatt_client_char_prop_info_t param;
		int client_id;
		gboolean is_notify;
		int fd = -1;
		int mtu = 0;
		gboolean is_indicate;


		memset(&param, 0x00, sizeof(bluetooth_gatt_client_char_prop_info_t));

		sender = (char*)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1, &param,
				sizeof(bluetooth_gatt_client_char_prop_info_t));
		__bt_service_get_parameters(in_param2, &client_id,
				sizeof(int));
		__bt_service_get_parameters(in_param3, &is_notify,
				sizeof(gboolean));
		__bt_service_get_parameters(in_param4, &is_indicate,
				sizeof(gboolean));

#ifdef TIZEN_FEATURE_BT_GATT_CLIENT_FD_DISABLE
		goto normal;
#endif

		if (is_indicate == false) {
			result   = _bt_gatt_acquire_notify(&param , &fd, &mtu);
			if (BLUETOOTH_ERROR_NONE == result && fd > -1) {
				BT_INFO("GATT Client: Save Invocation data for characteristic props app[%s] fd[ %d]", sender, fd);

				GUnixFDList *fd_list = NULL;
				GError *error = NULL;

				g_array_append_vals(*out_param1, &mtu, sizeof(int));
				/* Add socket fd to unix_fd_list */
				fd_list = g_unix_fd_list_new();
				g_unix_fd_list_append(fd_list, fd, &error);
				g_assert_no_error(error);
				close(fd);
				_bt_service_method_return_with_unix_fd_list(
				context, *out_param1, result, fd_list);

				g_object_unref(fd_list);
			} else
				goto normal;
		} else {

normal:
			result = _bt_gatt_watch_characteristic(&param , client_id, is_notify);
			if (BLUETOOTH_ERROR_NONE == result) {
				BT_INFO("GATT Client: Save Invocation data for characteristic props app[%s]", sender);

				/* Save the informations to invocation */
				_bt_save_invocation_context(context, result, sender, function_name,
					 (gpointer)g_memdup(&param, sizeof(bluetooth_gatt_client_char_prop_info_t)));
			}
		}
		break;
	}
	case BT_GATT_WATCH_SERVICE_CHANGED_INDICATION: {
		gboolean is_enabled;
		bluetooth_device_address_t address;

		sender = (char*)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2, &is_enabled,
				sizeof(gboolean));
		result = _bt_gatt_watch_service_changed_indication(sender, &address, is_enabled);
		BT_INFO("Service changed indication from app [%s]", sender);

		break;
	}
	case BT_LE_CONN_UPDATE: {
		char *sender = NULL;
		bluetooth_device_address_t local_address = { {0} };
		bluetooth_le_connection_param_t parameters = {0};

		sender = (char *)g_dbus_method_invocation_get_sender(context);

		__bt_service_get_parameters(in_param1, &local_address,
				sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2, &parameters,
				sizeof(bluetooth_le_connection_param_t));

		result = _bt_le_connection_update(sender,
					local_address.addr,
					parameters.interval_min,
					parameters.interval_max,
					parameters.latency,
					parameters.timeout);
		break;
	}
#endif
	case BT_START_LE_DISCOVERY: {
		sender = (char *)g_dbus_method_invocation_get_sender(context);
		result = _bt_start_le_scan(sender);
		if (result == BLUETOOTH_ERROR_NONE) {
			_bt_save_invocation_context(context, result, sender,
					function_name, NULL);
		}
		break;
	}
	case BT_STOP_LE_DISCOVERY: {
		sender = (char *)g_dbus_method_invocation_get_sender(context);
		result = _bt_stop_le_scan(sender);
		if (result == BLUETOOTH_ERROR_NONE) {
			_bt_save_invocation_context(context, result, sender,
					function_name, NULL);
		}
		break;
	}
	case BT_IS_LE_DISCOVERYING: {
		gboolean le_discovering = FALSE;

		le_discovering = _bt_is_le_scanning();
		g_array_append_vals(*out_param1,
				&le_discovering, sizeof(gboolean));
		break;
	}
	case BT_SET_SCAN_PARAMETERS: {
		bluetooth_le_scan_params_t scan_params;

		__bt_service_get_parameters(in_param1, &scan_params,
				sizeof(bluetooth_le_scan_params_t));
		BT_DBG("bluetooth_le_scan_params_t [%f %f %d]",
				scan_params.interval, scan_params.window,
				scan_params.type);
		result = _bt_set_scan_parameters(&scan_params);
		break;
	}
#ifdef TIZEN_FEATURE_BT_PAN_NAP
	case BT_NETWORK_ACTIVATE:
		BT_DBG("BT_NETWORK_ACTIVATE Request recieved Request Handler ");
		result = _bt_network_activate();
		break;
	case BT_NETWORK_DEACTIVATE:
		result = _bt_network_deactivate();
		break;
	case BT_NETWORK_CONNECT: {
		bluetooth_device_address_t address = { {0} };
		int role;

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2,
				&role, sizeof(int));

		result = _bt_network_connect(request_id, role, &address);
		if (result != BLUETOOTH_ERROR_NONE) {
			g_array_append_vals(*out_param1, &address,
					sizeof(bluetooth_device_address_t));
		}
		break;
	}
	case BT_NETWORK_DISCONNECT: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_network_disconnect(request_id, &address);
		if (result != BLUETOOTH_ERROR_NONE) {
			g_array_append_vals(*out_param1, &address,
					sizeof(bluetooth_device_address_t));
		}
		break;
	}
	case BT_NETWORK_SERVER_DISCONNECT: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1,
				&address, sizeof(bluetooth_device_address_t));

		result = _bt_network_server_disconnect(request_id, &address);
		if (result != BLUETOOTH_ERROR_NONE) {
			g_array_append_vals(*out_param1, &address,
					sizeof(bluetooth_device_address_t));
		}
		break;
	}

#endif
#ifdef TIZEN_FEATURE_BT_IPSP
	case BT_LE_IPSP_INIT:
		result = _bt_initialize_ipsp();
		break;
	case BT_LE_IPSP_DEINIT:
		result = _bt_deinitialize_ipsp();
		break;
	case BT_LE_IPSP_CONNECT: {
		bluetooth_device_address_t address = { {0} };
		__bt_service_get_parameters(in_param1, &address,
				sizeof(bluetooth_device_address_t));
		result = _bt_connect_le_ipsp_device(&address);
		break;
	}
	case BT_LE_IPSP_DISCONNECT: {
		bluetooth_device_address_t address = { {0} };
		__bt_service_get_parameters(in_param1, &address,
				sizeof(bluetooth_device_address_t));
		result = _bt_disconnect_le_ipsp_device(&address);
		break;
	}
#endif
	case BT_IS_ADVERTISING: {
		gboolean advertising = FALSE;
		advertising = _bt_is_advertising();
		g_array_append_vals(*out_param1, &advertising,
				sizeof(gboolean));
		break;
	}
	case BT_IS_LE_2M_PHY_SUPPORTED: {
		gboolean is_2m_phy_supported = FALSE;

		is_2m_phy_supported = _bt_is_le_2m_phy_supported();
		g_array_append_vals(*out_param1, &is_2m_phy_supported, sizeof(gboolean));
		break;
	}
	case BT_IS_LE_CODED_PHY_SUPPORTED: {
		gboolean is_coded_phy_supported = FALSE;

		is_coded_phy_supported = _bt_is_le_coded_phy_supported();
		g_array_append_vals(*out_param1, &is_coded_phy_supported, sizeof(gboolean));
		break;
	}
	default:
		BT_INFO("UnSupported function [%d]", function_name);
		result = BLUETOOTH_ERROR_NOT_SUPPORT;
		break;
	}

	return result;
}
#ifdef TIZEN_FEATURE_BT_OBEX
/* Function definitions*/
static void __bt_fill_garray_from_variant(GVariant *var, GArray *param)
{
	char *data;
	int size;

	size = g_variant_get_size(var);
	if (size > 0) {
		data = (char *)g_variant_get_data(var);
		if (data)
			param = g_array_append_vals(param, data, size);

	}
}
#endif

int __bt_obexd_request(int function_name,
		int request_type,
		int request_id,
		GDBusMethodInvocation *context,
		GVariant *in_param1,
		GVariant *in_param2,
		GVariant *in_param3,
		GVariant *in_param4,
		GArray **out_param1)
{
	BT_DBG("+");

	int result = BLUETOOTH_ERROR_NONE;

	BT_DBG("function_name : %x", function_name);

	switch (function_name) {
	case BT_OPP_PUSH_FILES: {
		BT_DBG("BT_OPP_PUSH_FILES");
		int i;
		bluetooth_device_address_t address = { {0} };
		bt_file_path_t *path;
		char **file_path;
		int file_count;
		GArray *param2;

		__bt_service_get_parameters(in_param1, &address,
				sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param3, &file_count,
				sizeof(int));

		file_path = g_new0(char *, file_count + 1);

		param2 = g_array_new(TRUE, TRUE, sizeof(gchar));
		__bt_fill_garray_from_variant(in_param2, param2);

		for (i = 0; i < file_count; i++) {
			path = &g_array_index(param2, bt_file_path_t, i);
			file_path[i] = g_strdup(path->path);
		}
		BT_DBG("_bt_opp_client_push_files");
		result = _bt_opp_client_push_files(request_id, context,
						&address, file_path,
						file_count);

		for (i = 0; i < file_count; i++)
			g_free(file_path[i]);

		g_free(file_path);
		g_array_free(param2, TRUE);

		break;
	}
	case BT_OPP_CANCEL_PUSH: {
		result = _bt_opp_client_cancel_push();

		break;
	}
	case BT_OPP_IS_PUSHING_FILES: {
		gboolean is_sending = FALSE;

		result = _bt_opp_client_is_sending(&is_sending);

		g_array_append_vals(*out_param1, &is_sending,
				sizeof(gboolean));
		break;
	}
	case BT_OPP_GET_TRANSFER_PROGRESS: {
		int direction;
		int transfer_id;
		guint8 progress = 0;

		__bt_service_get_parameters(in_param1, &direction,
				sizeof(int));
		__bt_service_get_parameters(in_param2, &transfer_id,
				sizeof(int));
		if (direction)
			result = _bt_opp_get_client_progress(&progress);
		else
			result = _bt_opp_get_server_progress(transfer_id, &progress);

		g_array_append_vals(*out_param1, &progress,
				sizeof(guint8));

		break;
	}

	case BT_MAP_CREATE_SESSION: {
		BT_DBG("BT_MAP_CREATE_SESSION");
		char *address = (char *)g_variant_get_data(in_param1);
		char *session_id = NULL;
		result = _bt_create_session_sync(address, &session_id);
		if (result == BLUETOOTH_ERROR_NONE)
			g_array_append_vals(*out_param1, session_id, strlen(session_id)+1);
		break;
	}

	case BT_MAP_DESTROY_SESSION: {
		BT_DBG("BT_MAP_DESTROY_SESSION");
		char* session_id = (char *)g_variant_get_data(in_param1);
		result = _bt_destroy_session_sync(session_id);
		if (result == BLUETOOTH_ERROR_NONE)
			BT_DBG("successfully destroyed session");
		break;
	}

	case BT_MAP_SET_FOLDER: {
		BT_DBG("BT_MAP_SET_FOLDER");
		char *session_id = (char *)g_variant_get_data(in_param1);
		char *name = (char *)g_variant_get_data(in_param2);
		result = _bt_map_client_set_folder(session_id, name);
		break;
	}

	case BT_MAP_LIST_FOLDERS: {
		BT_DBG("BT_MAP_LIST_FOLDERS");

		char* session_id = (char *)g_variant_get_data(in_param1);
		char* filter_serialized = (char*)g_variant_get_data(in_param2);

		result = _bt_map_client_list_folders(request_id, context, session_id, filter_serialized);
		if (result == BLUETOOTH_ERROR_NONE)
			BT_DBG("_bt_map_client_list_folders succeed");

		break;
	}

	case BT_MAP_LIST_FILTER_FIELDS: {
		BT_DBG("BT_MAP_LIST_FILTER_FIELDS");

		char* session_id = (char *)g_variant_get_data(in_param1);

		result = _bt_map_client_list_filter_fields(request_id, context, session_id);
		if (result == BLUETOOTH_ERROR_NONE)
			BT_DBG("_bt_map_client_list_filter_fields succeed");

		break;
	}

	case BT_MAP_LIST_MESSAGES: {
		BT_DBG("BT_MAP_LIST_MESSAGES");

		char* session_id = (char*)g_variant_get_data(in_param1);
		char* folder = (char*)g_variant_get_data(in_param2);
		char* filter_serialized = (char*)g_variant_get_data(in_param3);

		result = _bt_map_client_list_messages(request_id, context, session_id, folder, filter_serialized);
		if (result == BLUETOOTH_ERROR_NONE)
			BT_DBG("_bt_map_client_list_messages succeed");
		else
			BT_DBG("_bt_map_client_list_messages failed");

		break;
	}

	case BT_MAP_UPDATE_INBOX: {
		BT_DBG("BT_MAP_UPDATE_INBOX");
		char* session_id = (char *)g_variant_get_data(in_param1);
		result = _bt_map_client_update_inbox(session_id);
		break;
	}

	case BT_MAP_PUSH_MESSAGE: {
		BT_DBG("BT_MAP_PUSH_MESSAGE");

		char* session_id = (char *)g_variant_get_data(in_param1);
		char* source_file = (char *)g_variant_get_data(in_param2);
		char* folder = (char *)g_variant_get_data(in_param3);
		char* args_serialized = (char *)g_variant_get_data(in_param4);

		result = _bt_map_client_push_message(
				request_id, context, session_id, source_file, folder, args_serialized);
		if (result == BLUETOOTH_ERROR_NONE)
			BT_DBG("_bt_map_client_push_message succeed");
		else
			BT_ERR("_bt_map_client_push_message failed");

		break;
	}

	case BT_MAP_GET_MESSAGE: {
		BT_DBG("BT_MAP_GET_MESSAGE");
		// TODO session currently is not used, but should be valid
		//char* session_id = (char *)g_variant_get_data(in_param1);
		char* message_object = (char *)g_variant_get_data(in_param2);
		char* target_file = (char *)g_variant_get_data(in_param3);
		bool attachment = false;
		__bt_service_get_parameters(in_param4, &attachment, sizeof(bool));

		result = _bt_map_client_get_message(request_id, context, message_object,
				target_file, attachment);
		if (result == BLUETOOTH_ERROR_NONE)
			BT_DBG("_bt_map_client_get_message succeed");

		break;
	}

	case BT_OBEX_SERVER_ALLOCATE: {
		int app_pid;
		gboolean is_native;
		char *path;
		char *sender;

		sender = (char *)g_dbus_method_invocation_get_sender(context);

		path = (char *)g_variant_get_data(in_param1);
		__bt_service_get_parameters(in_param2, &is_native,
				sizeof(gboolean));
		__bt_service_get_parameters(in_param3, &app_pid,
				sizeof(int));

		result = _bt_obex_server_allocate(sender,
				path, app_pid, is_native);

		break;
	}
	case BT_OBEX_SERVER_DEALLOCATE: {
		int app_pid;
		gboolean is_native;

		__bt_service_get_parameters(in_param1, &is_native,
				sizeof(gboolean));
		__bt_service_get_parameters(in_param2, &app_pid,
				sizeof(int));

		result = _bt_obex_server_deallocate(app_pid, is_native);
		break;
	}
	case BT_OBEX_SERVER_IS_ACTIVATED: {
		gboolean is_activated = FALSE;

		result = _bt_obex_server_is_activated(&is_activated);

		g_array_append_vals(*out_param1, &is_activated,
				sizeof(gboolean));

		break;
	}
	case BT_OBEX_SERVER_ACCEPT_CONNECTION: {
		result = _bt_obex_server_accept_connection(request_id);

		break;
	}
	case BT_OBEX_SERVER_REJECT_CONNECTION: {
		result = _bt_obex_server_reject_connection();

		break;
	}
	case BT_OBEX_SERVER_ACCEPT_FILE: {
		char *file_name;

		file_name = (char *)g_variant_get_data(in_param1);

		result = _bt_obex_server_accept_authorize(file_name, TRUE);

		break;
	}
	case BT_OBEX_SERVER_REJECT_FILE: {
		result = _bt_obex_server_reject_authorize();

		break;
	}
	case BT_OBEX_SERVER_SET_PATH: {
		gboolean is_native;
		char *destination_path;

		destination_path = (char *)g_variant_get_data(in_param1);
		__bt_service_get_parameters(in_param2, &is_native,
				sizeof(gboolean));

		result = _bt_obex_server_set_destination_path(destination_path,
							is_native);

		break;
	}
	case BT_OBEX_SERVER_SET_ROOT: {
		char *root;

		root = (char *)g_variant_get_data(in_param1);

		result = _bt_obex_server_set_root(root);

		break;
	}
	case BT_OBEX_SERVER_CANCEL_TRANSFER: {
		int transfer_id;

		__bt_service_get_parameters(in_param1, &transfer_id,
				sizeof(int));

		result = _bt_obex_server_cancel_transfer(transfer_id);

		break;
	}
	case BT_OBEX_SERVER_CANCEL_ALL_TRANSFERS: {
		result = _bt_obex_server_cancel_all_transfers();

		break;
	}
	case BT_OBEX_SERVER_IS_RECEIVING: {
		gboolean is_receiving = FALSE;

		result = _bt_obex_server_is_receiving(&is_receiving);

		g_array_append_vals(*out_param1, &is_receiving,
				sizeof(gboolean));
		break;
	}
	case BT_PBAP_CONNECT: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1, &address,
				sizeof(bluetooth_device_address_t));

		result = _bt_pbap_connect(&address);
		break;
	}
	case BT_PBAP_DISCONNECT: {
		bluetooth_device_address_t address = { {0} };

		__bt_service_get_parameters(in_param1, &address,
				sizeof(bluetooth_device_address_t));

		result = _bt_pbap_disconnect(&address);
		break;
	}
	case BT_PBAP_GET_PHONEBOOK_SIZE: {
		bluetooth_device_address_t address = { {0} };
		bt_pbap_folder_t folder = { 0, };

		__bt_service_get_parameters(in_param1, &address,
				sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2, &folder,
				sizeof(bt_pbap_folder_t));

		result = _bt_pbap_get_phonebook_size(&address,
				folder.addressbook, folder.folder_type);
		break;
	}
	case BT_PBAP_GET_PHONEBOOK: {
		bluetooth_device_address_t address = { {0} };
		bt_pbap_folder_t folder = { 0, };
		bt_pbap_pull_parameters_t app_param = { 0, };

		__bt_service_get_parameters(in_param1, &address,
				sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2, &folder,
				sizeof(bt_pbap_folder_t));
		__bt_service_get_parameters(in_param3, &app_param,
				sizeof(bt_pbap_pull_parameters_t));

		result = _bt_pbap_get_phonebook(&address, folder.addressbook,
				folder.folder_type, &app_param);
		break;
	}
	case BT_PBAP_GET_LIST: {
		bluetooth_device_address_t address = { {0} };
		bt_pbap_folder_t folder = { 0, };
		bt_pbap_list_parameters_t app_param = { 0, };

		__bt_service_get_parameters(in_param1, &address,
				sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2, &folder,
				sizeof(bt_pbap_folder_t));
		__bt_service_get_parameters(in_param3, &app_param,
				sizeof(bt_pbap_list_parameters_t));

		result = _bt_pbap_get_list(&address, folder.addressbook,
				folder.folder_type, &app_param);
		break;
	}
	case BT_PBAP_PULL_VCARD: {
		bluetooth_device_address_t address = { {0} };
		bt_pbap_folder_t folder = { 0, };
		bt_pbap_pull_vcard_parameters_t app_param = { 0, };

		__bt_service_get_parameters(in_param1, &address,
				sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2, &folder,
				sizeof(bt_pbap_folder_t));
		__bt_service_get_parameters(in_param3, &app_param,
				sizeof(bt_pbap_pull_vcard_parameters_t));

		result = _bt_pbap_pull_vcard(&address, folder.addressbook,
				folder.folder_type, &app_param);
		break;
	}
	case BT_PBAP_PHONEBOOK_SEARCH: {
		bluetooth_device_address_t address = { {0} };
		bt_pbap_folder_t folder = { 0, };
		bt_pbap_search_parameters_t app_param = { 0, };

		__bt_service_get_parameters(in_param1, &address,
				sizeof(bluetooth_device_address_t));
		__bt_service_get_parameters(in_param2, &folder,
				sizeof(bt_pbap_folder_t));
		__bt_service_get_parameters(in_param3, &app_param,
				sizeof(bt_pbap_search_parameters_t));

		result = _bt_pbap_phonebook_search(&address, folder.addressbook,
				folder.folder_type, &app_param);
		break;
	}

	default:
		BT_ERR("Unknown function!");
		result = BLUETOOTH_ERROR_INTERNAL;
		break;
	}

	FN_END;

	return result;
}

int __bt_agent_request(int function_name,
		int request_type,
		int request_id,
		GDBusMethodInvocation *context,
		GVariant *in_param1,
		GVariant *in_param2,
		GVariant *in_param3,
		GVariant *in_param4,
		GArray **out_param1)
{
	int result;
	switch (function_name) {
	case BT_SET_AUTHORIZATION: {
		int type;
		char *uuid;
		char *path;
		int fd;

		__bt_service_get_parameters(in_param1, &type, sizeof(int));
		uuid = (char *)g_variant_get_data(in_param2);
		path = (char *)g_variant_get_data(in_param3);
		__bt_service_get_parameters(in_param4, &fd, sizeof(int));

		result = _bt_register_osp_server_in_agent(type, uuid, path, fd);
		break;
	}
	case BT_UNSET_AUTHORIZATION: {
		int type;
		char *uuid;

		__bt_service_get_parameters(in_param1, &type, sizeof(int));
		uuid = (char *)g_variant_get_data(in_param2);

		result = _bt_unregister_osp_server_in_agent(type, uuid);
		break;
	}
	default:
		BT_ERR("Unknown function!");
		result = BLUETOOTH_ERROR_INTERNAL;
		break;
	}

	return result;
}

int __bt_core_request(int function_name,
		int request_type,
		int request_id,
		GDBusMethodInvocation *context,
		GVariant *in_param1)
{
	int result;

	switch (function_name) {
	case BT_ENABLE_ADAPTER:
	{
		bt_status_t status;

		status = _bt_adapter_get_status();

		if (status == BT_ACTIVATING) {
			BT_DBG("Enabling in progress");
			result = BLUETOOTH_ERROR_IN_PROGRESS;
		} else if (status == BT_ACTIVATED) {
			BT_DBG("Already enabled");
			result = BLUETOOTH_ERROR_DEVICE_ALREADY_ENABLED;
		} else {
			_bt_adapter_set_status(BT_ACTIVATING);

#if TODO_40 /* Need to add this function */
			_bt_adapter_start_enable_timer();
#endif
			result = BLUETOOTH_ERROR_NONE;
		}

		break;
	}
	case BT_DISABLE_ADAPTER:
	{
		bt_status_t status;

		status = _bt_adapter_get_status();
		if (status == BT_DEACTIVATING) {
				BT_DBG("Disabling in progress");
				result = BLUETOOTH_ERROR_IN_PROGRESS;
		} else if (status == BT_DEACTIVATED) {
				BT_DBG("Already disabled");
				result = BLUETOOTH_ERROR_DEVICE_NOT_ENABLED;
		} else {
			_bt_adapter_set_status(BT_DEACTIVATING);
			result = BLUETOOTH_ERROR_NONE;
		}

		break;
	}
	case BT_ENABLE_ADAPTER_LE:
	{
		bt_le_status_t le_status;

		le_status = _bt_adapter_get_le_status();
		if (le_status == BT_LE_ACTIVATING) {
			BT_DBG("Enabling in progress");
			result = BLUETOOTH_ERROR_IN_PROGRESS;
		} else if (le_status == BT_LE_ACTIVATED) {
			BT_DBG("Already enabled");
			result = BLUETOOTH_ERROR_DEVICE_ALREADY_ENABLED;
		} else {
			_bt_adapter_set_le_status(BT_LE_ACTIVATING);

#if TODO_40 /* Need to add this function */
			_bt_adapter_start_le_enable_timer();
#endif
			result = BLUETOOTH_ERROR_NONE;
		}

		break;
	}
	case BT_DISABLE_ADAPTER_LE:
	{
		bt_le_status_t le_status;

		le_status = _bt_adapter_get_le_status();
		if (le_status == BT_LE_DEACTIVATING) {
				BT_DBG("Disabling in progress");
				result = BLUETOOTH_ERROR_IN_PROGRESS;
		} else if (le_status == BT_LE_DEACTIVATED) {
				BT_DBG("Already disabled");
				result = BLUETOOTH_ERROR_DEVICE_NOT_ENABLED;
		} else {
			_bt_adapter_set_le_status(BT_LE_DEACTIVATING);
			result = BLUETOOTH_ERROR_NONE;
		}

		break;
	}
	default:
		BT_ERR("Unknown function!");
		result = BLUETOOTH_ERROR_INTERNAL;
		break;
	}

	return result;
}

gboolean __bt_service_check_privilege(int function_name,
					int service_type,
					const char *unique_name)
{
	int ret_val;
	gboolean result = TRUE;
	char *client_creds = NULL;
	char *user_creds = NULL;
	char *client_session = "";
	enum cynara_client_creds client_creds_method = CLIENT_METHOD_SMACK;
	enum cynara_user_creds user_creds_method = USER_METHOD_UID;
	char err_msg[256] = {0, };

	retv_if(unique_name == NULL, FALSE);
	retv_if(bt_service_conn == NULL, FALSE);

	ret_val = cynara_creds_get_default_client_method(&client_creds_method);
	if (ret_val != CYNARA_API_SUCCESS) {
		cynara_strerror(ret_val, err_msg, sizeof(err_msg));
		BT_ERR("Fail to get default client method: %s", err_msg);
		return FALSE;
	}

	ret_val = cynara_creds_get_default_user_method(&user_creds_method);
	if (ret_val != CYNARA_API_SUCCESS) {
		cynara_strerror(ret_val, err_msg, sizeof(err_msg));
		BT_ERR("Fail to get default user method: %s", err_msg);
		return FALSE;
	}

	ret_val = cynara_creds_gdbus_get_client(bt_service_conn, unique_name, client_creds_method, &client_creds);
	if (ret_val != CYNARA_API_SUCCESS) {
		cynara_strerror(ret_val, err_msg, sizeof(err_msg));
		BT_ERR("Fail to get client credential: %s", err_msg);
		return FALSE;
	}

	ret_val = cynara_creds_gdbus_get_user(bt_service_conn, unique_name, user_creds_method, &user_creds);
	if (ret_val != CYNARA_API_SUCCESS) {
		cynara_strerror(ret_val, err_msg, sizeof(err_msg));
		BT_ERR("Fail to get user credential: %s", err_msg);
		if (client_creds)
			free(client_creds);
		return FALSE;
	}

	BT_DBG("%s, %s, %s", unique_name, client_creds, user_creds);

	switch (function_name) {
	case BT_SET_LOCAL_NAME:
	case BT_START_DISCOVERY:
	case BT_START_CUSTOM_DISCOVERY:
	case BT_CANCEL_DISCOVERY:
	case BT_OOB_ADD_REMOTE_DATA:
	case BT_OOB_REMOVE_REMOTE_DATA:
	case BT_OOB_READ_LOCAL_DATA:
	case BT_SET_ADVERTISING:
	case BT_SET_CUSTOM_ADVERTISING:
	case BT_SET_ADVERTISING_PARAMETERS:
	case BT_START_LE_DISCOVERY:
	case BT_STOP_LE_DISCOVERY:
	case BT_SET_SCAN_PARAMETERS:

	case BT_BOND_DEVICE:
	case BT_CANCEL_BONDING:
	case BT_UNBOND_DEVICE:
	case BT_SET_ALIAS:
	case BT_SET_AUTHORIZATION:
	case BT_UNSET_AUTHORIZATION:
	case BT_SEARCH_SERVICE:

	case BT_RFCOMM_CLIENT_CONNECT:
	case BT_RFCOMM_CLIENT_CANCEL_CONNECT:
	case BT_RFCOMM_SOCKET_DISCONNECT:
	case BT_RFCOMM_SOCKET_WRITE:
	case BT_RFCOMM_CREATE_SOCKET:
	case BT_RFCOMM_REMOVE_SOCKET:

	case BT_OPP_PUSH_FILES:
	case BT_OPP_CANCEL_PUSH:
	/* TODO: MAP? MAP functions, see above */

	case BT_OBEX_SERVER_ACCEPT_CONNECTION:
	case BT_OBEX_SERVER_REJECT_CONNECTION:
	case BT_OBEX_SERVER_ACCEPT_FILE:
	case BT_OBEX_SERVER_REJECT_FILE:
	case BT_OBEX_SERVER_SET_PATH:
	case BT_OBEX_SERVER_SET_ROOT:
	case BT_OBEX_SERVER_CANCEL_TRANSFER:
	case BT_OBEX_SERVER_CANCEL_ALL_TRANSFERS:

	case BT_AUDIO_SELECT_ROLE:
	case BT_AUDIO_CONNECT:
	case BT_AUDIO_DISCONNECT:
	case BT_AG_CONNECT:
	case BT_AG_DISCONNECT:
	case BT_AV_CONNECT:
	case BT_AV_DISCONNECT:
	case BT_AV_SOURCE_CONNECT:
	case BT_AV_SOURCE_DISCONNECT:
	case BT_AVRCP_TARGET_CONNECT:
	case BT_AVRCP_TARGET_DISCONNECT:
	case BT_AVRCP_CONTROL_CONNECT:
	case BT_AVRCP_CONTROL_DISCONNECT:
	case BT_AVRCP_HANDLE_CONTROL:
	case BT_AVRCP_SET_TRACK_INFO:
	case BT_AVRCP_SET_PROPERTY:
	case BT_AVRCP_SET_PROPERTIES:
	case BT_AVRCP_CONTROL_SET_PROPERTY:

	case BT_HF_CONNECT:
	case BT_HF_DISCONNECT:

	case BT_HID_CONNECT:
	case BT_HID_DISCONNECT:

	case BT_HID_DEVICE_ACTIVATE:
	case BT_HID_DEVICE_DEACTIVATE:
	case BT_HID_DEVICE_CONNECT:
	case BT_HID_DEVICE_DISCONNECT:
	case BT_HID_DEVICE_SEND_MOUSE_EVENT:
	case BT_HID_DEVICE_SEND_KEY_EVENT:
	case BT_HID_DEVICE_SEND_REPLY_TO_REPORT:
	case BT_HID_DEVICE_SEND_CUSTOM_EVENT:

	case BT_CONNECT_LE:
	case BT_DISCONNECT_LE:

	case BT_SET_ADVERTISING_DATA:
	case BT_SET_SCAN_RESPONSE_DATA:

	case BT_HDP_CONNECT:
	case BT_HDP_DISCONNECT:
	case BT_HDP_SEND_DATA:
	case BT_HDP_REGISTER_SINK_APP:
	case BT_HDP_UNREGISTER_SINK_APP:
	case BT_HDP_GET_FD:

	case BT_DPM_SET_ALLOW_BT_MODE:
	case BT_DPM_GET_ALLOW_BT_MODE:
	case BT_DPM_SET_DEVICE_RESTRITION:
	case BT_DPM_GET_DEVICE_RESTRITION:
	case BT_DPM_SET_UUID_RESTRITION:
	case BT_DPM_GET_UUID_RESTRITION:
	case BT_DPM_ADD_DEVICES_BLACKLIST:
	case BT_DPM_ADD_DEVICES_WHITELIST:
	case BT_DPM_ADD_UUIDS_BLACKLIST:
	case BT_DPM_ADD_UUIDS_WHITELIST:
	case BT_DPM_CLEAR_DEVICES_BLACKLIST:
	case BT_DPM_CLEAR_DEVICES_WHITELIST:
	case BT_DPM_CLEAR_UUIDS_BLACKLIST:
	case BT_DPM_CLEAR_UUIDS_WHITELIST:
	case BT_DPM_REMOVE_DEVICE_BLACKLIST:
	case BT_DPM_REMOVE_DEVICE_WHITELIST:
	case BT_DPM_REMOVE_UUID_BLACKLIST:
	case BT_DPM_REMOVE_UUID_WHITELIST:
	case BT_DPM_GET_DEVICES_BLACKLIST:
	case BT_DPM_GET_DEVICES_WHITELIST:
	case BT_DPM_GET_UUIDS_BLACKLIST:
	case BT_DPM_GET_UUIDS_WHITELIST:
	case BT_DPM_SET_ALLOW_OUTGOING_CALL:
	case BT_DPM_GET_ALLOW_OUTGOING_CALL:
	case BT_DPM_SET_PAIRING_STATE:
	case BT_DPM_GET_PAIRING_STATE:
	case BT_DPM_SET_PROFILE_STATE:
	case BT_DPM_GET_PROFILE_STATE:
	case BT_DPM_SET_DESKROP_CONNECTIVITY_STATE:
	case BT_DPM_GET_DESKROP_CONNECTIVITY_STATE:
	case BT_DPM_SET_DISCOVERABLE_STATE:
	case BT_DPM_GET_DISCOVERABLE_STATE:
	case BT_DPM_SET_LIMITED_DISCOVERABLE_STATE:
	case BT_DPM_GET_LIMITED_DISCOVERABLE_STATE:
	case BT_DPM_SET_DATA_TRANSFER_STATE:
	case BT_DPM_GET_DATA_TRANSFER_STATE:

	case BT_NETWORK_ACTIVATE:
	case BT_NETWORK_DEACTIVATE:
	case BT_NETWORK_CONNECT:
	case BT_NETWORK_DISCONNECT:
	case BT_NETWORK_SERVER_DISCONNECT:

	case BT_GATT_GET_PRIMARY_SERVICES:
#ifdef TIZEN_GATT_CLIENT
	case BT_GATT_GET_SERVICE_PROPERTIES:             /* GATT Client */
	case BT_GATT_GET_CHARACTERISTIC_PROPERTIES:      /* GATT Client */
	case BT_GATT_WATCH_SERVICE_CHANGED_INDICATION:   /* GATT Client */
#endif
	case BT_GATT_DISCOVER_CHARACTERISTICS:
	case BT_GATT_SET_PROPERTY_REQUEST:
	case BT_GATT_READ_CHARACTERISTIC:
#ifdef TIZEN_GATT_CLIENT
	case BT_GATT_READ_DESCRIPTOR_VALUE:              /* GATT Client */
	case BT_GATT_WRITE_CHARACTERISTIC_VALUE_BY_TYPE: /* GATT Client */
	case BT_GATT_ACQUIRE_WRITE:
	case BT_GATT_WRITE_DESCRIPTOR_VALUE:             /* GATT Client */
	case BT_GATT_WATCH_CHARACTERISTIC:               /* GATT Client */
	case BT_GATT_CLIENT_REGISTER:                    /* GATT Client */
	case BT_GATT_CLIENT_UNREGISTER:                   /* GATT Client */
#endif
	case BT_GATT_DISCOVER_CHARACTERISTICS_DESCRIPTOR:
	case BT_GATT_REGISTER_APPLICATION:
	case BT_GATT_REGISTER_SERVICE:
	case BT_GATT_SEND_RESPONSE:
#ifndef GATT_DIRECT
	case BT_GATT_SERVER_REGISTER:
	case BT_GATT_SERVER_ADD_SERVICE:
	case BT_GATT_SERVER_ADD_CHARACTERISTIC:
	case BT_GATT_SERVER_ADD_DESCRIPTOR:
	case BT_GATT_SERVER_START_SERVICE:
	case BT_GATT_SERVER_STOP_SERVICE:
	case BT_GATT_SERVER_DELETE_SERVICE:
	case BT_GATT_SERVER_SEND_RESPONSE:
	case BT_GATT_SERVER_ACQURE_WRITE_RESPONSE:
	case BT_GATT_SERVER_ACQUIRE_NOTIFY_RESPONSE:
	case BT_GATT_SERVER_SEND_INDICATION:
	case BT_GATT_SERVER_UPDATE_VALUE:
	case BT_GATT_SERVER_DEREGISTER:
#endif
	case BT_REQ_ATT_MTU:
	case BT_PBAP_CONNECT:
	case BT_PBAP_DISCONNECT:
	case BT_PBAP_GET_PHONEBOOK_SIZE:
	case BT_PBAP_GET_PHONEBOOK:
	case BT_PBAP_GET_LIST:
	case BT_PBAP_PULL_VCARD:
	case BT_PBAP_PHONEBOOK_SEARCH:

	case BT_RFCOMM_LISTEN_AND_ACCEPT:

	ret_val = cynara_check(p_cynara, client_creds, client_session, user_creds,
						BT_PRIVILEGE_PUBLIC);

	if (ret_val != CYNARA_API_ACCESS_ALLOWED) {
		BT_ERR("Fail to access: %s", BT_PRIVILEGE_PUBLIC);
		result = FALSE;
	}

	/* Need to check mediastorage privilege */
	if (function_name == BT_PBAP_GET_PHONEBOOK ||
		function_name == BT_PBAP_PULL_VCARD) {
		ret_val = cynara_check(p_cynara, client_creds, client_session, user_creds,
							MEDIASTORAGE_PRIVILEGE);

		if (ret_val != CYNARA_API_ACCESS_ALLOWED) {
		BT_ERR("Fail to access: %s", MEDIASTORAGE_PRIVILEGE);
		result = FALSE;
		}
	}
	break;

	case BT_ENABLE_ADAPTER:
	case BT_DISABLE_ADAPTER:
	case BT_RESET_ADAPTER:
	case BT_RECOVER_ADAPTER:
	case BT_ENABLE_ADAPTER_LE:
	case BT_DISABLE_ADAPTER_LE:
	case BT_SET_CONNECTABLE:
	case BT_SET_DISCOVERABLE_MODE:
	case BT_ADD_WHITE_LIST:
	case BT_REMOVE_WHITE_LIST:
	case BT_CLEAR_WHITE_LIST:
	case BT_SET_MANUFACTURER_DATA:

	case BT_CANCEL_SEARCH_SERVICE:
	case BT_ENABLE_RSSI:

	case BT_RFCOMM_ACCEPT_CONNECTION:
	case BT_RFCOMM_REJECT_CONNECTION:
	case BT_RFCOMM_LISTEN:

	case BT_HID_ENABLE_BARCODE_FEATURE:

	case BT_AVRCP_CONTROL_GET_PROPERTY:
	case BT_AVRCP_GET_TRACK_INFO:
	case BT_AVRCP_TRANSPORT_SET_PROPERTY:
	case BT_AVRCP_HANDLE_CONTROL_TO_DEST:

	case BT_SET_CONTENT_PROTECT:
	case BT_BOND_DEVICE_BY_TYPE:
	case BT_SET_LE_PRIVACY:
	case BT_SET_LE_STATIC_RANDOM_ADDRESS:
	case BT_LE_CONN_UPDATE:
	case BT_UPDATE_LE_CONNECTION_MODE:
	case BT_GET_DEVICE_IDA:
	case BT_LE_READ_MAXIMUM_DATA_LENGTH:
	case BT_LE_WRITE_HOST_SUGGESTED_DATA_LENGTH:
	case BT_LE_READ_HOST_SUGGESTED_DATA_LENGTH:
	case BT_LE_SET_DATA_LENGTH:

	case BT_LE_IPSP_INIT:
	case BT_LE_IPSP_DEINIT:
	case BT_LE_IPSP_CONNECT:
	case BT_LE_IPSP_DISCONNECT:
	case BT_PXP_MONITOR_SET_PROPERTY:
	case BT_PXP_MONITOR_GET_PROPERTY:
	case BT_PXP_MONITOR_GET_SUPPORTED_SERIVCES:
	case BT_PXP_REPORTER_REGISTER:
	case BT_PXP_REPORTER_UNREGISTER:
	case BT_PXP_REPORTER_GET_PROPERTY:

	/* TDS */
	case BT_TDS_PROVIDER_REGISTER:
	case BT_TDS_PROVIDER_UNREGISTER:
	case BT_TDS_PROVIDER_SET_MANUF_DATA:
	case BT_TDS_PROVIDER_CREATE:
	case BT_TDS_PROVIDER_DESTROY:
	case BT_TDS_PROVIDER_SET_TRANSPORT_DATA:
	case BT_TDS_SEND_ACTIVATION_RESPONSE:
	case BT_TDS_READ_TRANSPORT_DATA:
	case BT_TDS_ENABLE_CONTROL_POINT:
	case BT_TDS_ACTIVATE_CONTROL_POINT:

	/* OTP Server */
	case BT_OTP_SERVER_INIT:
	case BT_OTP_SERVER_DEINIT:
	case BT_OTP_READ_VALUE:
	case BT_OTP_ENABLE_NOTIFICATION:
	case BT_OTP_WRITE_VALUE:
	case BT_LE_OTC_CONNECT:
	case BT_LE_OTC_DISCONNECT:

	case BT_MAP_CREATE_SESSION:
	case BT_MAP_DESTROY_SESSION:
	case BT_MAP_SET_FOLDER:
	case BT_MAP_LIST_FOLDERS:
	case BT_MAP_LIST_FILTER_FIELDS:
	case BT_MAP_LIST_MESSAGES:
	case BT_MAP_UPDATE_INBOX:
	case BT_MAP_PUSH_MESSAGE:
	case BT_MAP_GET_MESSAGE:

	ret_val = cynara_check(p_cynara, client_creds, client_session, user_creds,
				BT_PRIVILEGE_PLATFORM);

	if (ret_val != CYNARA_API_ACCESS_ALLOWED) {
		BT_ERR("Fail to access: %s", BT_PRIVILEGE_PLATFORM);
		result = FALSE;
	}
	break;

	case BT_CHECK_ADAPTER:
	case BT_GET_RSSI:

	case BT_GET_LOCAL_NAME:
	case BT_GET_LOCAL_ADDRESS:
	case BT_GET_LOCAL_VERSION:
	case BT_IS_SERVICE_USED:
	case BT_GET_DISCOVERABLE_MODE:
	case BT_GET_DISCOVERABLE_TIME:
	case BT_IS_DISCOVERYING:
	case BT_IS_LE_DISCOVERYING:
	case BT_IS_CONNECTABLE:
	case BT_GET_BONDED_DEVICES:
	case BT_GET_PROFILE_CONNECTED_DEVICES:
	case BT_GET_BONDED_DEVICE:
	case BT_PASSKEY_REPLY:
	case BT_PASSKEY_CONFIRMATION_REPLY:
	case BT_GET_IS_ALIAS_SET:
	case BT_IS_DEVICE_CONNECTED:
	case BT_GET_CONNECTED_LINK_TYPE:
	case BT_SET_PROFILE_TRUSTED:
	case BT_GET_PROFILE_TRUSTED:
	case BT_GET_SPEAKER_GAIN:
	case BT_SET_SPEAKER_GAIN:
	case BT_RFCOMM_CLIENT_IS_CONNECTED:
	case BT_RFCOMM_IS_UUID_AVAILABLE:
	case BT_GET_ADVERTISING_DATA:
	case BT_GET_SCAN_RESPONSE_DATA:
	case BT_IS_ADVERTISING:
	case BT_IS_LE_2M_PHY_SUPPORTED:
	case BT_IS_LE_CODED_PHY_SUPPORTED:
	case BT_REGISTER_SCAN_FILTER:
	case BT_UNREGISTER_SCAN_FILTER:
	case BT_UNREGISTER_ALL_SCAN_FILTERS:
	case BT_IS_SCAN_FILTER_SUPPORTED:
	case BT_GET_ATT_MTU:

	case BT_OBEX_SERVER_ALLOCATE:
	case BT_OBEX_SERVER_DEALLOCATE:
	case BT_OBEX_SERVER_IS_ACTIVATED:
	case BT_OPP_GET_TRANSFER_PROGRESS:
		/* Non-privilege control */
		break;
	default:
		BT_ERR("Unknown function!");
		result = FALSE;
		break;
	}

	if (client_creds)
		free(client_creds);

	if (user_creds)
		free(user_creds);

	return result;
}

GDBusNodeInfo *__bt_service_create_method_node_info
					(const gchar *introspection_data)
{
	GError *err = NULL;
	GDBusNodeInfo *node_info = NULL;

	if (introspection_data == NULL) {
		BT_ERR("Introspection XML not present");
		return NULL;
	}

	node_info = g_dbus_node_info_new_for_xml(introspection_data, &err);

	if (err) {
		BT_ERR("Unable to create node: %s", err->message);
		g_clear_error(&err);
	}
	return node_info;
}

int __bt_service_register_object(GDBusConnection *conn,
		GDBusNodeInfo *node_info, gboolean reg)
{
	static guint service_id = 0;
	GError *error = NULL;

	if (reg) {
		if (node_info == NULL)
			return -1;

		service_id = g_dbus_connection_register_object(conn,
				BT_SERVICE_PATH,
				node_info->interfaces[0],
				&method_table,
				NULL, NULL, &error);
		if (service_id == 0)
			return -1;
	} else {
		if (service_id > 0) {
			g_dbus_connection_unregister_object(conn,
					service_id);
			service_id = 0;
		}
	}

	return 0;
}

static void __name_owner_changed(GDBusConnection *connection,
		const gchar *sender_name,
		const gchar *object_path,
		const gchar *interface_name,
		const gchar *signal_name,
		GVariant *parameters,
		gpointer user_data)
{
	const char *name = NULL;
	const char *old_owner = NULL;
	const char *new_owner = NULL;

	g_variant_get(parameters, "(&s&s&s)", &name, &old_owner, &new_owner);
	if (*new_owner != '\0')
		return;

	_bt_check_hdp_app_termination(name);

	/* Advertising App Termination */
	_bt_check_adv_app_termination(name);

	/* Check if le_scanning app is terminated */
	_bt_check_le_scanner_app_termination(name);
}

static void __bt_service_bus_acquired_handler(GDBusConnection *connection,
		const gchar *name, gpointer user_data)
{
	GDBusNodeInfo *node_info = NULL;

	BT_INFO("bus acquired");

	ret_if(connection == NULL);

	node_info = __bt_service_create_method_node_info(
			bt_service_introspection_xml);
	ret_if(node_info == NULL);

	__bt_service_register_object(connection, node_info, TRUE);
	g_dbus_node_info_unref(node_info);

	bt_service_conn = connection;
}

static void __bt_service_name_acquired_handler(GDBusConnection *connection,
		const gchar *name, gpointer user_data)
{
	BT_INFO("name acquired");
	name_acquired = TRUE;
}

static void __bt_service_name_lost_handler(GDBusConnection *connection,
		const gchar *name, gpointer user_data)
{
	BT_INFO("name lost");
	name_acquired = FALSE;
}

gboolean _is_name_acquired(void)
{
	return name_acquired;
}

int _bt_service_register(void)
{
	GDBusConnection *conn;
	GError *err = NULL;

	conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &err);
	retv_if(conn == NULL, BLUETOOTH_ERROR_INTERNAL);
	bt_service_conn = conn;

	owner_id = g_bus_own_name(G_BUS_TYPE_SYSTEM,
				BT_SERVICE_NAME,
				G_BUS_NAME_OWNER_FLAGS_NONE,
				__bt_service_bus_acquired_handler,
				__bt_service_name_acquired_handler,
				__bt_service_name_lost_handler,
				NULL, NULL);
	BT_DBG("owner_id is [%d]", owner_id);
	if (owner_id == 0)
		goto fail;

	/* Subscribe for name owner changed signal */
	owner_sig_id = g_dbus_connection_signal_subscribe(conn,
			BT_FREEDESKTOP_SERVICE, BT_FREEDESKTOP_INTERFACE,
			BT_NAME_OWNER_CHANGED, NULL, NULL, G_DBUS_SIGNAL_FLAGS_NONE,
			__name_owner_changed, NULL, NULL);
	BT_INFO("owner_sig_id: %d", owner_sig_id);

	return BLUETOOTH_ERROR_NONE;

fail:
	if (bt_service_conn) {
		g_object_unref(bt_service_conn);
		bt_service_conn = NULL;
	}

	return BLUETOOTH_ERROR_INTERNAL;
}

void _bt_service_unregister(void)
{
	if (bt_service_conn) {
		if (owner_sig_id > 0) {
			g_dbus_connection_signal_unsubscribe(
					bt_service_conn, owner_sig_id);
			owner_sig_id = 0;
		}

		__bt_service_register_object(bt_service_conn, NULL, FALSE);
		if (bt_service_conn) {
			g_object_unref(bt_service_conn);
			bt_service_conn = NULL;
		}
		if (owner_id > 0) {
			g_bus_unown_name(owner_id);
			owner_id = 0;
		}
	}
}

int _bt_service_cynara_init(void)
{
	int result;
	char err_msg[256] = {0, };

	retv_if(p_cynara != NULL, BLUETOOTH_ERROR_ALREADY_INITIALIZED);

	result = cynara_initialize(&p_cynara, conf);

	if (result != CYNARA_API_SUCCESS) {
		cynara_strerror(result, err_msg, sizeof(err_msg));
		BT_ERR("Fail to initialize cynara: [%s]", err_msg);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	return BLUETOOTH_ERROR_NONE;
}

void _bt_service_cynara_deinit(void)
{
	int result;
	char err_msg[256] = {0, };

	ret_if(p_cynara == NULL);

	result = cynara_finish(p_cynara);

	if (result != CYNARA_API_SUCCESS) {
		cynara_strerror(result, err_msg, sizeof(err_msg));
		BT_ERR("Fail to finish cynara: [%s]", err_msg);
		return;
	}

	p_cynara = NULL;
	conf = NULL;
}

void _bt_service_method_return(GDBusMethodInvocation *invocation,
		GArray *out_param, int result)
{
	GVariant *out_var;
	BT_DBG("+");
	out_var = g_variant_new_from_data((const GVariantType *)"ay",
			out_param->data, out_param->len, TRUE, NULL, NULL);

	g_dbus_method_invocation_return_value(invocation,
			g_variant_new("(iv)", result, out_var));
	BT_DBG("-");
}

void _bt_service_method_return_with_unix_fd_list(GDBusMethodInvocation *invocation,
		GArray *out_param, int result, GUnixFDList *fd_list)
{
	GVariant *out_var;
	BT_DBG("+");
	out_var = g_variant_new_from_data((const GVariantType *)"ay",
			out_param->data, out_param->len, TRUE, NULL, NULL);

	g_dbus_method_invocation_return_value_with_unix_fd_list(invocation,
			g_variant_new("(iv)", result, out_var), fd_list);
	BT_DBG("-");
}
