/*
 * Copyright (c) 2015 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Contact: Anupam Roy <anupam.r@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <gio/gio.h>
#include <glib.h>
#include <dlog.h>
#include <string.h>
#include <vconf.h>
#include <vconf-internal-keys.h>
#include <bundle.h>
#include <bundle_internal.h>
#include <eventsystem.h>

/*bt-service headers */
#include "bt-internal-types.h"
#include "bt-service-common.h"
#include "bt-service-util.h"
#include "bt-service-main.h"
#include "bt-service-core-device.h"
#include "bt-service-core-adapter.h"
#include "bt-service-event-receiver.h"
#include "bt-request-handler.h"
#include "bt-service-event.h"
#include "bt-service-agent-util.h"
#include "bt-service-a2dp-src.h"
#include "bt-service-a2dp-sink.h"
#ifdef TIZEN_FEATURE_BT_OBEX
#include "bt-service-obex-server.h"
#endif
#include "bt-service-device-internal.h"

#ifdef TIZEN_GATT_CLIENT
#include "bt-service-gatt.h"
#endif

/* OAL headers */
#include <oal-event.h>
#include <oal-manager.h>
#include <oal-adapter-mgr.h>
#include <oal-device-mgr.h>

#define MAX_BOND_RETRY_COUNT 3
#define BT_PASSKEY_MAX_LENGTH 4

/* Bonding Info structure */
typedef struct {
	int result;
	char *addr;
	gboolean is_autopair;
	unsigned short conn_type;
	gboolean is_cancelled_by_user;
	gboolean is_device_creating;
	bluetooth_device_address_t *dev_addr;
	bt_remote_dev_info_t *dev_info;
} bt_bond_data_t;

/* Searching Info structure */
typedef struct {
	int result;
	char *addr;
	gboolean is_cancelled_by_user;
	bluetooth_device_address_t *dev_addr;
	bt_remote_dev_info_t *dev_info;
} bt_service_search_info_data_t;

/* Pairing Info structure */
typedef struct {
	char *addr;
	gboolean is_autopair;
	int is_ssp;
} bt_pairing_data_t;

typedef struct {
	char addr[BT_ADDRESS_STRING_SIZE];
	bt_remote_dev_info_t *dev_info;
} bt_incoming_bond_data_t;

/* Bonding and Pairing Informations */
bt_bond_data_t *trigger_bond_info;
bt_bond_data_t *trigger_unbond_info;
bt_pairing_data_t *trigger_pairing_info;
bt_service_search_info_data_t *service_search_info;

bt_incoming_bond_data_t *incoming_bond_info;

typedef enum {
	BT_DEVICE_BOND_STATE_NONE,
	BT_DEVICE_BOND_STATE_CANCEL_DISCOVERY,
	BT_DEVICE_BOND_STATE_DISCOVERY_CANCELLED,
	BT_DEVICE_BOND_STATE_REMOVE_BONDING,
	BT_DEVICE_BOND_STATE_REMOVED_BONDING,
	BT_DEVICE_BOND_STATE_STARTED,
	BT_DEVICE_BOND_STATE_WAIT_PROP,
	BT_DEVICE_BOND_STATE_WAIT_DID
} bt_bond_state_e;

typedef enum {
	BT_DEVICE_BOND_INFO,
	BT_DEVICE_INCOMING_BOND_INFO,
	BT_DEVICE_UNBOND_INFO
} bt_bond_info_e;

/* BT device bond state variable */
static bt_bond_state_e bt_device_bond_state;
static int bond_retry_count;

static char *passkey_watcher;
static GSList *pin_info_list = NULL;

#ifdef TIZEN_GATT_CLIENT
typedef struct {
	char *address;
	float interval_min;
	float interval_max;
	GSList *senders;
} bt_connected_le_dev_t;

typedef struct {
	char *sender;
	float interval_min;
	float interval_max;
	guint16 latency;
	guint16 time_out;
	float key;
} bt_le_conn_param_t;

static GSList *le_connected_dev_list = NULL;

#define BT_LE_CONN_INTERVAL_MIN 7.5 /* msec */
#define BT_LE_CONN_INTERVAL_MAX 4000 /* msec */
#define BT_LE_CONN_SUPER_TO_MIN 100 /* msec */
#define BT_LE_CONN_SUPER_TO_MAX 32000 /* msec */
#define BT_LE_CONN_SLAVE_LATENCY_MAX 499
#define BT_LE_CONN_TO_SPLIT 10 /* msec */
#define BT_LE_CONN_INTERVAL_SPLIT 1.25 /* msec */

static void _bt_handle_le_connected_dev_info(const char *address, gboolean connected);
#endif

/* Forward declaration */
static void __bt_device_event_handler(int event_type, gpointer event_data);
static void __bt_device_remote_device_found_callback(gpointer event_data, gboolean is_ble);


static int __bt_device_handle_bond_state(void);
static void __bt_free_bond_info(uint8_t type);
static void __bt_free_service_search_info(bt_service_search_info_data_t **p_info);
static void __bt_device_handle_bond_completion_event(bt_address_t *bd_addr);
static void __bt_device_handle_bond_removal_event(bt_address_t *bd_addr);
static void __bt_device_handle_bond_failed_event(event_dev_bond_failed_t* bond_fail_event);
static void __bt_handle_ongoing_bond(bt_remote_dev_info_t *remote_dev_info, gboolean incoming_bond);
static void __bt_device_acl_state_changed_callback(event_dev_conn_status_t *acl_event,
		gboolean connected, unsigned char type);
static void __bt_free_pairing_info(bt_pairing_data_t **p_info);

static void __bt_device_ssp_consent_callback(remote_device_t* dev_info);
static void __bt_device_pin_request_callback(remote_device_t* pin_req_event);
static void __bt_device_ssp_passkey_display_callback(event_dev_passkey_t *dev_info);
static void __bt_device_ssp_passkey_confirmation_callback(event_dev_passkey_t *dev_info);
static void __bt_device_ssp_passkey_entry_callback(remote_device_t* dev_info);
static void __bt_device_authorization_request_callback(event_dev_authorize_req_t* auth_event);

static void __bt_device_services_callback(event_dev_services_t* uuid_list);
static void __bt_handle_ongoing_device_service_search(bt_remote_dev_info_t *remote_dev_info);

static void __bt_device_trusted_callback(gboolean trusted, event_dev_trust_t* info);

static int __bt_get_device_pin_code(const char *address, char *pin_code);

gboolean _bt_is_device_creating(void)
{
	if (!trigger_bond_info)
		return FALSE;
	return trigger_bond_info->is_device_creating;
}

void _bt_device_state_handle_callback_set_request(void)
{
	_bt_service_register_event_handler_callback(
			BT_DEVICE_MODULE, __bt_device_event_handler);
}

void __bt_device_handle_pending_requests(int result, int service_function,
		void *user_data, unsigned int size)
{
	GSList *l;
	GArray *out_param;
	invocation_info_t *req_info = NULL;

	BT_DBG("+");

	/* Get method invocation context */
	for (l = _bt_get_invocation_list(); l != NULL; l = g_slist_next(l)) {
		req_info = l->data;
		if (req_info == NULL || req_info->service_function != service_function)
			continue;

		switch (service_function) {
		case BT_SEARCH_SERVICE: {
			char *address = (char *)user_data;
			if (strncmp((char*)req_info->user_data, address, BT_ADDRESS_STRING_SIZE)) {
				BT_ERR("Unexpected: Info request pending for a different address!!");
				return;
			} else {
				BT_INFO("Found info request addr [%s]", (char*)req_info->user_data);
				bt_sdp_info_t sdp_info;

				memset(&sdp_info, 0x00, sizeof(bt_sdp_info_t));
				_bt_convert_addr_string_to_type(sdp_info.device_addr.addr, address);

				out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
				g_array_append_vals(out_param, &sdp_info, sizeof(bt_sdp_info_t));
				_bt_service_method_return(req_info->context, out_param, result);

				g_free(req_info->user_data);
				_bt_free_info_from_invocation_list(req_info);
				g_array_free(out_param, TRUE);
			}
			break;
		}
		case BT_BOND_DEVICE: {
			char *address = (char *)user_data;
			if (strncmp((char*)req_info->user_data, address, BT_ADDRESS_STRING_SIZE)) {
				BT_ERR("Unexpected: Info request pending for a different address!!");
				return;
			} else {
				BT_INFO("Found info request addr [%s]", (char*)req_info->user_data);
				bluetooth_device_info_t dev_info;
				memset(&dev_info, 0x00, sizeof(bluetooth_device_info_t));
				_bt_convert_addr_string_to_type(dev_info.device_address.addr,
						address);
				out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
				g_array_append_vals(out_param, &dev_info,
						sizeof(bluetooth_device_info_t));
				_bt_service_method_return(req_info->context, out_param, result);

				g_free(req_info->user_data);
				_bt_free_info_from_invocation_list(req_info);
				g_array_free(out_param, TRUE);
			}
			break;
		}
		case BT_UNBOND_DEVICE: {
			char *address = (char *)user_data;
			if (strncmp((char*)req_info->user_data, address, BT_ADDRESS_STRING_SIZE)) {
				BT_ERR("Unexpected: Info request pending for a different address!!");
				return;
			} else {
				BT_INFO("Found info request addr [%s]", (char*)req_info->user_data);
				bluetooth_device_address_t dev_addr;
				_bt_convert_addr_string_to_type(dev_addr.addr, address);
				out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
				g_array_append_vals(out_param, &dev_addr,
						sizeof(bluetooth_device_address_t));
				_bt_service_method_return(req_info->context, out_param, result);

				g_free(req_info->user_data);
				_bt_free_info_from_invocation_list(req_info);
				g_array_free(out_param, TRUE);
			}
			break;
		}
		default:
			BT_ERR("Unhandled case");
			break;
		}
	}
	BT_INFO("-");
}

/*
 * Remote device properties are received on all following conditions
 * a. When Bonding in on-going
 * b. When device properties are updated\changed for a connected device
 *    (due to SDP or any other reason)
 */
static void __bt_device_remote_properties_callback(event_dev_properties_t *oal_dev_props)
{
	bt_remote_dev_info_t *rem_info = NULL;

	BT_DBG("+");
	rem_info = g_malloc0(sizeof(bt_remote_dev_info_t));
	_bt_copy_remote_dev(rem_info, &(oal_dev_props->device_info));

	if (oal_dev_props->adv_len > 0) {
		int k;

		rem_info->manufacturer_data_len = oal_dev_props->adv_len;
		rem_info->manufacturer_data =
			g_memdup(oal_dev_props->adv_data,
					oal_dev_props->adv_len);
		BT_DBG("----Advertising Data Length: %d",
				rem_info->manufacturer_data_len);

		for (k = 0; k < rem_info->manufacturer_data_len; k++) {
			BT_INFO("Check data[%d] = [[0x%x]",
					k, oal_dev_props->adv_data[k]);
		}
	} else {
		rem_info->manufacturer_data = NULL;
		rem_info->manufacturer_data_len = 0;
	}

	/* a. Check if bonding is on-going, if yes, we MUST update the bonding device properties */
	if (trigger_bond_info  && !strcmp(trigger_bond_info->addr, rem_info->address)) {
		BT_INFO("Bonding is ongoing, try update properties");
		if (!trigger_bond_info->dev_info ||
				(!trigger_bond_info->dev_info->name &&
				 !trigger_bond_info->dev_info->alias) ||
					!trigger_bond_info->dev_info->address ||
						trigger_bond_info->dev_info->uuid_count == 0) {
			BT_INFO("Complete data is not present, Assigning rem_info");
			if (!trigger_bond_info->dev_info)
				trigger_bond_info->dev_info = g_malloc0(sizeof(bt_remote_dev_info_t));
			_bt_copy_remote_dev_info(trigger_bond_info->dev_info, rem_info);
		}

		BT_DBG("Bonding dev addr has matched with remote dev properties address [%s]", rem_info->address);
		__bt_handle_ongoing_bond(trigger_bond_info->dev_info, FALSE);
	} else if (incoming_bond_info && !g_strcmp0(incoming_bond_info->addr, rem_info->address)) {
		BT_INFO("Incoming Bond is ongoing, try update properties");
		if (!incoming_bond_info->dev_info ||
				(!incoming_bond_info->dev_info->name &&
				 !incoming_bond_info->dev_info->alias) ||
					!incoming_bond_info->dev_info->address ||
						incoming_bond_info->dev_info->uuid_count == 0) {
			BT_INFO("Complete data is not present, Assigning rem_info");
			if (!incoming_bond_info->dev_info)
				incoming_bond_info->dev_info = g_malloc0(sizeof(bt_remote_dev_info_t));
			_bt_copy_remote_dev_info(incoming_bond_info->dev_info, rem_info);
		}

		BT_DBG("Incoming Bond addr matches with remote dev properties address [%s]", rem_info->address);
		__bt_handle_ongoing_bond(incoming_bond_info->dev_info, TRUE);
	}

	/* Add device to bonded list */
	_bt_service_add_device_to_bonded_list(rem_info);

	/* Handle SDP Device properties update */
	if (service_search_info && service_search_info->dev_info) {
		if (!strcmp(service_search_info->addr, rem_info->address)) {
			BT_DBG("Properties received and SDP request pending, fill device properties and send event");
			service_search_info->dev_info->class = rem_info->class;
			service_search_info->dev_info->paired = rem_info->paired;
			service_search_info->dev_info->connected = rem_info->connected;
			service_search_info->dev_info->rssi = rem_info->rssi;
			service_search_info->dev_info->addr_type = rem_info->addr_type;
			service_search_info->dev_info->trust = rem_info->trust;

			/* TODO*/
			service_search_info->dev_info->manufacturer_data = NULL;
			service_search_info->dev_info->manufacturer_data_len = 0;

			__bt_handle_ongoing_device_service_search(service_search_info->dev_info);
		}
	}

	_bt_free_remote_dev(rem_info);
	BT_DBG("-");
}

static int __get_oal_trusted_profile(bluetooth_trusted_profile_t profile)
{
	switch (profile) {
	case TRUSTED_PROFILE_PBAP:
		return OAL_TRUSTED_PROFILE_PBAP;
	case TRUSTED_PROFILE_MAP:
		return OAL_TRUSTED_PROFILE_MAP;
	case TRUSTED_PROFILE_SAP:
		return OAL_TRUSTED_PROFILE_SAP;
	case TRUSTED_PROFILE_HFP_HF:
		return OAL_TRUSTED_PROFILE_HFP_HF;
	case TRUSTED_PROFILE_A2DP:
		return OAL_TRUSTED_PROFILE_A2DP;
	case TRUSTED_PROFILE_ALL:
		return OAL_TRUSTED_PROFILE_ALL;
	default:
		return 0;
	}
}

int _bt_set_trust_profile(bluetooth_device_address_t *addr,
		bluetooth_trusted_profile_t profile, gboolean trust)
{
	int result;
	bt_address_t bd_addr;
	oal_trusted_profile_e oal_profile;

	BT_DBG("+");

	retv_if(!addr, BLUETOOTH_ERROR_INVALID_PARAM);

	memcpy(bd_addr.addr, addr, BLUETOOTH_ADDRESS_LENGTH);
	oal_profile = __get_oal_trusted_profile(profile);
	retv_if(0 == oal_profile, BLUETOOTH_ERROR_NOT_SUPPORT);

	result = device_set_trust_profile(&bd_addr, oal_profile, trust);
	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("device_set_trust_profile error: [%d]", result);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_get_trust_profile(bluetooth_device_address_t *addr,
		bluetooth_trusted_profile_t profile, guint *trust)
{
	int result;
	bt_address_t bd_addr;
	oal_trusted_profile_e oal_profile;

	BT_DBG("+");

	retv_if(!addr, BLUETOOTH_ERROR_INVALID_PARAM);

	memcpy(bd_addr.addr, addr, BLUETOOTH_ADDRESS_LENGTH);
	oal_profile = __get_oal_trusted_profile(profile);
	retv_if(0 == oal_profile, BLUETOOTH_ERROR_NOT_SUPPORT);

	result = device_get_trust_profile(&bd_addr, oal_profile, trust);
	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("device_set_trust_profile error: [%d]", result);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

static void __bt_handle_ongoing_device_service_search(bt_remote_dev_info_t *remote_dev_info)
{
	GVariant *param = NULL;
	GVariant *uuids = NULL;
	GVariantBuilder *builder = NULL;
	GVariant *manufacturer_data;
	unsigned int i = 0;
	char *name = NULL;

	BT_INFO("Send Service Search request event");

	if (remote_dev_info->alias)
		name = remote_dev_info->alias;
	else
		name = remote_dev_info->name;

	builder = g_variant_builder_new(G_VARIANT_TYPE("as"));
	for (i = 0; i < remote_dev_info->uuid_count; i++) {
		g_variant_builder_add(builder, "s",
				remote_dev_info->uuids[i]);
	}
	uuids = g_variant_new("as", builder);
	g_variant_builder_unref(builder);
	manufacturer_data = g_variant_new_from_data((const GVariantType *)"ay",
			remote_dev_info->manufacturer_data, remote_dev_info->manufacturer_data_len,
			TRUE, NULL, NULL);

	param = g_variant_new("(isunsbub@asn@ay)",
			BLUETOOTH_ERROR_NONE,
			remote_dev_info->address,
			remote_dev_info->class,
			remote_dev_info->rssi,
			name,
			remote_dev_info->paired,
			remote_dev_info->connected,
			remote_dev_info->trust,
			uuids,
			remote_dev_info->manufacturer_data_len,
			manufacturer_data);
	/* Send the event to application */
	_bt_send_event(BT_ADAPTER_EVENT,
			BLUETOOTH_EVENT_SERVICE_SEARCHED,
			param);

	__bt_free_service_search_info(&service_search_info);
	BT_DBG("-");
}

static void __bt_device_services_callback(event_dev_services_t* uuid_list)
{
	bt_remote_dev_info_t *rem_info = NULL;
	unsigned int i;
	BT_DBG("+");

	if (trigger_bond_info && _bt_compare_adddress(trigger_bond_info->dev_addr,
				(bluetooth_device_address_t *)&uuid_list->address) == TRUE) {
		bluetooth_device_address_t *dev_addr = trigger_bond_info->dev_addr;

		BT_DBG("Bonding dev addr has matched");
		/* Bonding ongoing, Query device properties again */
		if (BLUETOOTH_ERROR_NONE ==
			_bt_device_get_bonded_device_info(dev_addr))
			BT_DBG("_bt_device_get_bonded_device_info success");
		else
			BT_ERR("_bt_device_get_bonded_device_info failed");
	}

	if (service_search_info == NULL) {
		/* Send reply */
		BT_DBG("searching_info == NULL");
		return;
	}

	if (_bt_compare_adddress(service_search_info->dev_addr,
				(bluetooth_device_address_t *)&uuid_list->address) == FALSE) {
		BT_DBG("This device is not queried");
		return;
	}

	rem_info = g_malloc0(sizeof(bt_remote_dev_info_t));
	memset(rem_info, 0x00, sizeof(bt_remote_dev_info_t));

	rem_info->address = g_new0(char, BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(rem_info->address, uuid_list->address.addr);

	rem_info->uuid_count = uuid_list->num;

	BT_INFO("Address [%s]", rem_info->address);
	BT_INFO("Number of UUID's [%d]", rem_info->uuid_count);
	if (rem_info->uuid_count > 0)
		rem_info->uuids = g_new0(char *, rem_info->uuid_count);

	/* Fill Remote Device Service List list */
	for (i = 0; i < rem_info->uuid_count; i++) {
		rem_info->uuids[i] = g_malloc0(BLUETOOTH_UUID_STRING_MAX);
		_bt_uuid_to_string((service_uuid_t *)&uuid_list->service_list[i].uuid, rem_info->uuids[i]);
		BT_DBG("UUID value=%s", rem_info->uuids[i]);
	}

	/* Update local cache */
	_bt_update_remote_dev_property(rem_info->address, DEV_PROP_SERVICES, (void *)rem_info);

	BT_DBG("DBUS return");
	__bt_device_handle_pending_requests(BLUETOOTH_ERROR_NONE, BT_SEARCH_SERVICE,
			service_search_info->addr, BT_ADDRESS_STRING_SIZE);

	/* Save UUID List of remote devices */
	if (service_search_info->dev_info)
		_bt_free_remote_dev(service_search_info->dev_info);
	service_search_info->dev_info = rem_info;

	/* Query Other device properties */
	if (_bt_device_get_bonded_device_info(service_search_info->dev_addr) == BLUETOOTH_ERROR_NONE) {
		BT_DBG("Bonded device info query posted to stack successfully");
	} else {
		BT_DBG("Possibly internal stack error in bonded device info query, perform cleanup");
		__bt_free_service_search_info(&service_search_info);
	}
	BT_DBG("-");
}

static void __bt_device_authorization_request_callback(event_dev_authorize_req_t* auth_event)
{
	oal_service_t service_d = auth_event->service_id;
	gchar address[BT_ADDRESS_STR_LEN];
	int res;

	_bt_convert_addr_type_to_string(address, auth_event->address.addr);

	BT_INFO("service_d: %d", service_d);

	switch (service_d) {
	case HID_SERVICE_ID:
		BT_INFO("Incoming HID Profile conn Req from device addr [%s]", address);
		break;
	case A2DP_SERVICE_ID:
		BT_INFO("Incoming A2DP(Remote Sink) profile conn Req from device addr [%s]", address);
		_bt_a2dp_src_handle_incoming_authorization(address, service_d);
		return;
	case A2DP_SRC_SERVICE_ID:
		BT_INFO("Incoming A2DP(Remote Source) Profile conn Req from device addr [%s]", address);
		_bt_a2dp_sink_handle_incoming_authorization(address, service_d);
		break;
	case AVRCP_SERVICE_ID:
		BT_INFO("Incoming AVRCP (Remote) Profile conn Req from device addr [%s]", address);
		break;
	case AVRCP_CT_SERVICE_ID:
		BT_INFO("Incoming AVRCP (Controller) Profile conn Req from device addr [%s]", address);
		break;
#ifdef TIZEN_FEATURE_BT_OBEX
	case OPP_SERVICE_ID: {
		GVariant *param = NULL;
		char *name = NULL;

		if (_bt_obex_server_is_custom() == false) {
			/* Allow the connection for native OPP server */
			break;
		}

		name = g_strdup(address);

		BT_INFO("Incoming OPP conn Req from device addr [%s]", address);
		_bt_obex_server_set_pending_conn_auth_device_addr(address);
		param = g_variant_new("(iss)", BLUETOOTH_ERROR_NONE, address, name);
		_bt_send_event(BT_OPP_SERVER_EVENT,
				BLUETOOTH_EVENT_OBEX_SERVER_CONNECTION_AUTHORIZE, param);
		g_free(name);
		return;
	}
#endif
	case HSP_SERVICE_ID:
		BT_INFO("Incoming HSP_SERVICE_ID conn Req from device addr [%s]", address);
		break;
	case HFP_SERVICE_ID:
		BT_INFO("Incoming HFP_SERVICE_ID conn Req from device addr [%s]", address);
		break;
	case SAP_SERVICE_ID:
		BT_INFO("Incoming SAP_SERVICE_ID conn Req from device addr [%s]", address);
		break;
	case HSP_HS_SERVICE_ID:
		BT_INFO("Incoming HSP_HS_SERVICE_ID conn Req from device addr [%s]", address);
		break;
	case HFP_HS_SERVICE_ID:
		BT_INFO("Incoming HFP_HS_SERVICE_ID conn Req from device addr [%s]", address);
		break;
#ifdef TIZEN_BT_HAL
	case IOTIVITY_SERVICE_ID:
		BT_INFO("Incoming IOTIVITY_SERVICE_ID conn Req from device addr [%s]", address);
		break;
#endif
	default:
		/* For now, reject authorization for any service apart from above switch cases */
		BT_INFO("Incoming Profile conn req with service ID [%d] from device addr [%s]", service_d, address);
		res = device_reply_auth_request((bt_address_t*)&auth_event->address, service_d, FALSE, FALSE);
		if (res != OAL_STATUS_SUCCESS)
			BT_ERR("authorize_response: %d", res);
		return;
	}

	/* Auto accept authorization request for HID, A2DP and AVRCP profiles */
	BT_INFO("Auto Accept authorization");
	res = device_reply_auth_request((bt_address_t*)&auth_event->address, service_d, TRUE, FALSE);
	if (res != OAL_STATUS_SUCCESS)
		BT_ERR("authorize_response: %d", res);
	BT_INFO("-");
}

static void __bt_handle_ongoing_bond(bt_remote_dev_info_t *remote_dev_info, gboolean incoming_bond)
{
	GVariant *param = NULL;
	BT_DBG("+");

	if ((remote_dev_info->name || remote_dev_info->alias)
			&& remote_dev_info->address
			&& remote_dev_info->uuids) {
		BT_INFO("All properties updated,  time to send bonding finished event");
		GVariant *uuids = NULL;
		GVariantBuilder *builder = NULL;
		GVariant *manufacturer_data;
		unsigned int i = 0;
		char *name = NULL;

		if (remote_dev_info->alias)
			name = remote_dev_info->alias;
		else
			name = remote_dev_info->name;

		builder = g_variant_builder_new(G_VARIANT_TYPE("as"));
		for (i = 0; i < remote_dev_info->uuid_count; i++) {
			g_variant_builder_add(builder, "s",
					remote_dev_info->uuids[i]);
		}
		uuids = g_variant_new("as", builder);
		g_variant_builder_unref(builder);
		manufacturer_data = g_variant_new_from_data((const GVariantType *)"ay",
				remote_dev_info->manufacturer_data, remote_dev_info->manufacturer_data_len,
				TRUE, NULL, NULL);

		param = g_variant_new("(isunsbub@asn@ay)",
				BLUETOOTH_ERROR_NONE,
				remote_dev_info->address,
				remote_dev_info->class,
				remote_dev_info->rssi,
				name,
				remote_dev_info->paired,
				remote_dev_info->connected,
				remote_dev_info->trust,
				uuids,
				remote_dev_info->manufacturer_data_len,
				manufacturer_data);
		/* Send the event to application */
		_bt_send_event(BT_ADAPTER_EVENT,
				BLUETOOTH_EVENT_BONDING_FINISHED,
				param);
		if (incoming_bond) {
			__bt_free_bond_info(BT_DEVICE_INCOMING_BOND_INFO);
		} else {
			__bt_free_bond_info(BT_DEVICE_BOND_INFO);
			__bt_free_pairing_info(&trigger_pairing_info);
		}
	} else {
		BT_INFO("Lets wait for more remote device properties");
	}
}

static void __handle_incoming_bond_created_event(bt_address_t *bd_addr)
{
	char address[BT_ADDRESS_STRING_SIZE];
	bluetooth_device_address_t dev_addr;

	BT_INFO("+");

	/*
	 * BlueZ sends paired signal for each paired device, during activation,
	 * We should ignore this, otherwise application thinks that a new device
	 * got paired
	 */
	if (_bt_adapter_get_status() != BT_ACTIVATED) {
		BT_DBG("BT is not activated, so ignore this");
		return;
	}

	_bt_convert_addr_type_to_string(address, bd_addr->addr);
	if (!incoming_bond_info) {
		incoming_bond_info = g_malloc0(sizeof(bt_incoming_bond_data_t));
	} else {
		if (g_strcmp0(incoming_bond_info->addr, address)) {
			BT_DBG("Previous Bond address: [%s] differs from new address: [%s]",
					address, incoming_bond_info->addr);
			__bt_free_bond_info(BT_DEVICE_INCOMING_BOND_INFO);
			incoming_bond_info = g_malloc0(sizeof(bt_incoming_bond_data_t));
		}
	}

	BT_INFO("Incoming bond successfully completed");
	g_strlcpy(incoming_bond_info->addr, address, BT_ADDRESS_STRING_SIZE);
	incoming_bond_info->dev_info = NULL;

	_bt_convert_addr_string_to_type(dev_addr.addr, incoming_bond_info->addr);
	if (BLUETOOTH_ERROR_NONE == _bt_device_get_bonded_device_info(&dev_addr)) {
		BT_DBG("Bonded device info query posted to stack successfully");
	} else {
		BT_ERR("Bonded device info query failed");
		__bt_free_bond_info(BT_DEVICE_INCOMING_BOND_INFO);
	}

	BT_INFO("-");
}

static void __bt_device_handle_bond_completion_event(bt_address_t *bd_addr)
{
	gchar address[BT_ADDRESS_STR_LEN];
	bluetooth_device_address_t dev_addr;

	BT_INFO("+");

	if (trigger_bond_info == NULL) {
		/* Send reply */
		BT_DBG("trigger_bond_info == NULL, Handle incomming bond event");
		__handle_incoming_bond_created_event(bd_addr);
		return;
	}

	_bt_convert_addr_type_to_string(address, bd_addr->addr);
	if (g_strcmp0(trigger_bond_info->addr, address)) {
		BT_DBG("Bonding address= [%s] is different from requested address =[%s]",
				address, trigger_bond_info->addr);
		__handle_incoming_bond_created_event(bd_addr);
		return;
	}

	BT_INFO("Bonding successfully completed");
	/* Bonding state will be cleaned up & BONDING FINISHED EVENT
	   will be sent only when Properties are fetched from stack
	   Till that time lets not free trigger_bond_info.
	   However it is possible that while fetching device properties, internal
	   stack error can occur which can lead to no valid properties or
	   no properties at all. So in such cases, we must not wait for properties,
	   otherwise, it can lead to infinite wait  */
	_bt_convert_addr_string_to_type(dev_addr.addr,
			trigger_bond_info->addr);

	if (_bt_device_get_bonded_device_info(&dev_addr) == BLUETOOTH_ERROR_NONE) {
		BT_DBG("BOnded device info query posted to stack successfully");
		__bt_device_handle_pending_requests(BLUETOOTH_ERROR_NONE, BT_BOND_DEVICE,
				trigger_bond_info->addr, BT_ADDRESS_STRING_SIZE);
	} else {
		BT_DBG("Possibly internal stack error in bonded device info query, perform cleanup");
		__bt_device_handle_pending_requests(BLUETOOTH_ERROR_INTERNAL, BT_BOND_DEVICE,
				trigger_bond_info->addr, BT_ADDRESS_STRING_SIZE);
		/* Destroy if at all device got bonded at stack level */
		device_destroy_bond((bt_address_t *)trigger_bond_info->dev_addr);

		__bt_free_bond_info(BT_DEVICE_BOND_INFO);
		__bt_free_pairing_info(&trigger_pairing_info);
	}

	BT_INFO("-");
}

/**********************************************************************************************
*  Bond removal event can be triggered for following reasons -
*  a. If Bonding procedure if failed (for Auth failed, Page timeout, cancelled by user etc)
*  b. If Application requests for explicitly removing the bond
*  c. When application attempt to create bond,bond is removed first which triggers this event
*     c. is in-line with Bluedroid bond create\emoval architecture
*********************************************************************************************/
static void __bt_device_handle_bond_removal_event(bt_address_t *bd_addr)
{
	char address[BT_ADDRESS_STRING_SIZE];

	BT_INFO("+");

	_bt_convert_addr_type_to_string(address, bd_addr->addr);
	_bt_service_remove_device_from_bonded_list(address);

	if (trigger_unbond_info) {
		BT_INFO("Bond removal request successfully handled, return DBUS and send event");
		GVariant *param = NULL;
		__bt_device_handle_pending_requests(BLUETOOTH_ERROR_NONE, BT_UNBOND_DEVICE,
				trigger_unbond_info->addr, BT_ADDRESS_STRING_SIZE);
		param = g_variant_new("(is)", BLUETOOTH_ERROR_NONE, trigger_unbond_info->addr);
		_bt_send_event(BT_ADAPTER_EVENT,
				BLUETOOTH_EVENT_BONDED_DEVICE_REMOVED,
				param);
		__bt_free_bond_info(BT_DEVICE_UNBOND_INFO);
		__bt_free_pairing_info(&trigger_pairing_info);
	} else if (trigger_bond_info) {
		BT_ERR("Bonding was removed");
		__bt_device_handle_bond_state();
	}
	BT_INFO("-");
}

static void __bt_device_handle_bond_failed_event(event_dev_bond_failed_t* bond_fail_event)
{
	BT_INFO("+");
	oal_status_t status = bond_fail_event->status;
	BT_INFO("Bonding failed, reason: %d", status);

	switch (status) {
	case OAL_STATUS_RMT_DEVICE_DOWN:
	{
		if (trigger_bond_info) {
			BT_INFO("OAL_STATUS_RMT_DEVICE_DOWN:Lets retry bonding!! retry count [%d]",
					bond_retry_count);
			int ret = OAL_STATUS_SUCCESS;
			if (bond_retry_count < MAX_BOND_RETRY_COUNT) {
				ret = device_create_bond((bt_address_t *)trigger_bond_info->dev_addr, trigger_bond_info->conn_type);
				if (ret != OAL_STATUS_SUCCESS) {
					BT_ERR("Create Bond procedure could not suceed");
					__bt_device_handle_pending_requests(BLUETOOTH_ERROR_INTERNAL, BT_BOND_DEVICE,
							trigger_bond_info->addr, BT_ADDRESS_STRING_SIZE);
					__bt_free_bond_info(BT_DEVICE_BOND_INFO);
					__bt_free_pairing_info(&trigger_pairing_info);
					bond_retry_count = 0;
				} else
					bond_retry_count++;
			} else {
				BT_ERR("Create Bond failed MAX_BOND_RETRY_COUNT TIMES!!");
				__bt_device_handle_pending_requests(BLUETOOTH_ERROR_INTERNAL, BT_BOND_DEVICE,
						trigger_bond_info->addr, BT_ADDRESS_STRING_SIZE);
				__bt_free_bond_info(BT_DEVICE_BOND_INFO);
				__bt_free_pairing_info(&trigger_pairing_info);
				bond_retry_count = 0;
			}
		}
		break;
	}
	case OAL_STATUS_AUTH_FAILED:
	{
		/*TODO Auto pairing status set & ignore auto pairing logics can be done at this point.
		  To be considered later*/
		int result = BLUETOOTH_ERROR_INTERNAL;
		BT_INFO("BT_OPERATION_STATUS_AUTH_FAILED");
		if (trigger_bond_info) {
			BT_ERR("Create Bond procedure could not suceed, check if cancelled by User");
			if (trigger_bond_info->is_cancelled_by_user) {
				BT_ERR("Bonding is cancelled by user");
				result = BLUETOOTH_ERROR_CANCEL_BY_USER;
			}
			__bt_device_handle_pending_requests(result, BT_BOND_DEVICE,
					trigger_bond_info->addr, BT_ADDRESS_STRING_SIZE);
			__bt_free_bond_info(BT_DEVICE_BOND_INFO);
		}

		__bt_free_pairing_info(&trigger_pairing_info);
		break;
	}
	case OAL_STATUS_INTERNAL_ERROR:
	{
		BT_INFO("OAL_STATUS_INTERNAL_ERROR");
		if (trigger_unbond_info) {
			BT_INFO("Bond removal request failed, return DBUS and send event");
			GVariant *param = NULL;
			__bt_device_handle_pending_requests(BLUETOOTH_ERROR_INTERNAL, BT_UNBOND_DEVICE,
					trigger_unbond_info->addr, BT_ADDRESS_STRING_SIZE);
			param = g_variant_new("(is)", BLUETOOTH_ERROR_INTERNAL, trigger_unbond_info->addr);
			_bt_send_event(BT_ADAPTER_EVENT,
					BLUETOOTH_EVENT_BONDED_DEVICE_REMOVED,
					param);
			__bt_free_bond_info(BT_DEVICE_UNBOND_INFO);
		} else if (trigger_bond_info) {
			if (__bt_device_handle_bond_state() != BLUETOOTH_ERROR_NONE) {
				if (trigger_bond_info) {
					__bt_device_handle_pending_requests(BLUETOOTH_ERROR_INTERNAL, BT_BOND_DEVICE,
							trigger_bond_info->addr, BT_ADDRESS_STRING_SIZE);
					__bt_free_bond_info(BT_DEVICE_BOND_INFO);
				}
			}
		}

		__bt_free_pairing_info(&trigger_pairing_info);
		break;
	}
	default:
	{
		BT_ERR("Unknown status of Bond failed event status [%d]", status);
		break;
	}

	}
	BT_INFO("-");
}

static void __bt_device_event_handler(int event_type, gpointer event_data)
{
	int eventcheck = OAL_EVENT_DEVICE_PROPERTIES;
	BT_INFO("event [%d] Event check = [%d]", event_type, eventcheck);

	switch (event_type) {
	case OAL_EVENT_ADAPTER_INQUIRY_RESULT_BREDR_ONLY: {
		BT_INFO("BREDR Device Found");
		__bt_device_remote_device_found_callback(event_data, FALSE);
		break;
	}
	case OAL_EVENT_ADAPTER_INQUIRY_RESULT_BLE: {
		BT_INFO("Dual Device Found");
		__bt_device_remote_device_found_callback(event_data, FALSE);
		break;
	}
	case OAL_EVENT_DEVICE_PROPERTIES: {
		BT_INFO("Remote Device properties Received");
		__bt_device_remote_properties_callback((event_dev_properties_t *)event_data);
		break;
	}
	case OAL_EVENT_DEVICE_BONDING_SUCCESS: {
		BT_INFO("Bonding Success event Received");
		__bt_device_handle_bond_completion_event((bt_address_t *)event_data);
		break;
	}
	case OAL_EVENT_DEVICE_BONDING_REMOVED: {
		BT_INFO("Bonding Removed event Received");
		__bt_device_handle_bond_removal_event((bt_address_t *)event_data);
		break;
	}
	case OAL_EVENT_DEVICE_BONDING_FAILED: {
		BT_INFO("Bonding Failed event Received");
		__bt_device_handle_bond_failed_event((event_dev_bond_failed_t*) event_data);
		break;
	}
	case OAL_EVENT_DEVICE_ACL_CONNECTED: {
		BT_INFO("ACL Connected event Received");
		event_dev_conn_status_t* param = event_data;
		__bt_device_acl_state_changed_callback(param, TRUE, 0);
		break;
	}
	case OAL_EVENT_DEVICE_ACL_DISCONNECTED: {
		BT_INFO("ACL Disconnected event Received");
		__bt_device_acl_state_changed_callback((event_dev_conn_status_t *)event_data, FALSE, 0);
		break;
	}
	case OAL_EVENT_DEVICE_LE_CONNECTED: {
		BT_INFO("LE Connected event Received");
		event_dev_conn_status_t* param = event_data;
		__bt_device_acl_state_changed_callback(param, TRUE, 1);
		break;
	}
	case OAL_EVENT_DEVICE_LE_DISCONNECTED: {
		BT_INFO("LE Disconnected event Received");
		__bt_device_acl_state_changed_callback((event_dev_conn_status_t *)event_data, FALSE, 1);
		break;
	}
	case OAL_EVENT_DEVICE_PIN_REQUEST: {
		BT_INFO("PIN Request Received");
		__bt_device_pin_request_callback((remote_device_t*)event_data);
		break;
	}
	case OAL_EVENT_DEVICE_PASSKEY_ENTRY_REQUEST: {
		BT_INFO("Passkey Entry request Received");
		__bt_device_ssp_passkey_entry_callback((remote_device_t*)event_data);
		break;
	}
	case OAL_EVENT_DEVICE_PASSKEY_CONFIRMATION_REQUEST:{
		BT_INFO("Passkey Confirmation Request Received");
		__bt_device_ssp_passkey_confirmation_callback((event_dev_passkey_t *)event_data);
		break;
	}
	case OAL_EVENT_DEVICE_PASSKEY_DISPLAY: {
		BT_INFO("Passkey Display Request Received");
		__bt_device_ssp_passkey_display_callback((event_dev_passkey_t *)event_data);
		break;
	}
	case OAL_EVENT_DEVICE_SSP_CONSENT_REQUEST: {
		BT_INFO("SSP Consent Request Received");
		 __bt_device_ssp_consent_callback((remote_device_t*)event_data);
		break;
	}
	case OAL_EVENT_DEVICE_SERVICES: {
		BT_INFO("Remote Device Services Received");
		__bt_device_services_callback((event_dev_services_t*)event_data);
		break;
	}
	case OAL_EVENT_DEVICE_AUTHORIZE_REQUEST: {
		BT_INFO("Remote Device Authorization Request");
		__bt_device_authorization_request_callback((event_dev_authorize_req_t*)event_data);
		break;
	}
	case OAL_EVENT_DEVICE_TRUSTED: {
		BT_INFO("Remote Device Trusted");
		__bt_device_trusted_callback(TRUE, (event_dev_trust_t*)event_data);
		break;
	}
	case OAL_EVENT_DEVICE_UNTRUSTED: {
		BT_INFO("Remote Device UnTrusted");
		__bt_device_trusted_callback(FALSE, (event_dev_trust_t*)event_data);
		break;
	}
	case OAL_EVENT_DEVICE_NAME: {
		remote_device_t *rem_dev = event_data;
		gchar address[BT_ADDRESS_STR_LEN];

		_bt_convert_addr_type_to_string(address, rem_dev->address.addr);
		BT_INFO("Remote Device name Received");
		BT_INFO("Name: %s, Address: %s", rem_dev->name, address);

		/* Update local cache */
		_bt_update_remote_dev_property(address, DEV_PROP_NAME, (void *)rem_dev->name);
		break;
	}
	case OAL_EVENT_DEVICE_TRUSTED_PROFILES_CHANGED: {
		event_device_trusted_profiles_t *ev = event_data;
		char address[BT_ADDRESS_STRING_SIZE];

		ret_if(NULL == ev);

		_bt_convert_addr_type_to_string(address, ev->address.addr);
		_bt_send_event(BT_DEVICE_EVENT,
				BLUETOOTH_EVENT_SUPPORTED_PROFILE_TRUSTED,
				g_variant_new("(isi)", BLUETOOTH_ERROR_NONE,
					address, ev->trust_val));
		break;
	}
	case OAL_EVENT_RSSI_MONITORING_ENABLED: {
		event_dev_rssi_info_t *ev = event_data;
		char address[BT_ADDRESS_STRING_SIZE];
		GVariant *param;

		ret_if(NULL == ev);

		_bt_convert_addr_type_to_string(address, ev->address.addr);
		param = g_variant_new("(isib)", BLUETOOTH_ERROR_NONE,
				address, ev->link_type, TRUE);
		_bt_send_event(BT_DEVICE_EVENT, BLUETOOTH_EVENT_RSSI_ENABLED, param);
		break;
	}
	case OAL_EVENT_RSSI_MONITORING_DISABLED: {
		event_dev_rssi_info_t *ev = event_data;
		char address[BT_ADDRESS_STRING_SIZE];
		GVariant *param;

		ret_if(NULL == ev);

		_bt_convert_addr_type_to_string(address, ev->address.addr);
		param = g_variant_new("(isib)", BLUETOOTH_ERROR_NONE,
				address, ev->link_type, FALSE);
		_bt_send_event(BT_DEVICE_EVENT, BLUETOOTH_EVENT_RSSI_ENABLED, param);
		break;
	}
	case OAL_EVENT_RSSI_ALERT_RECEIVED: {
		event_dev_rssi_info_t *ev = event_data;
		char address[BT_ADDRESS_STRING_SIZE];
		GVariant *param;

		ret_if(NULL == ev);

		_bt_convert_addr_type_to_string(address, ev->address.addr);
		param = g_variant_new("(isiii)", BLUETOOTH_ERROR_NONE,
				address, ev->link_type, ev->alert_type, ev->rssi);
		_bt_send_event(BT_DEVICE_EVENT, BLUETOOTH_EVENT_RSSI_ALERT, param);
		break;
	}
	case OAL_EVENT_RAW_RSSI_RECEIVED: {
		event_dev_rssi_info_t *ev = event_data;
		char address[BT_ADDRESS_STRING_SIZE];
		GVariant *param;

		ret_if(NULL == ev);

		_bt_convert_addr_type_to_string(address, ev->address.addr);
		param = g_variant_new("(isii)", BLUETOOTH_ERROR_NONE,
				address, ev->link_type, ev->rssi);
		_bt_send_event(BT_DEVICE_EVENT, BLUETOOTH_EVENT_RAW_RSSI, param);
		break;
	}
	default:
		BT_INFO("Unhandled event..");
	}
}

/* Legacy Pairing event handler */
static void __bt_device_pin_request_callback(remote_device_t* pin_req_event)
{
	GVariant *param;
	char address[BT_ADDRESS_STRING_SIZE];
	char pin_code[BLUETOOTH_PIN_CODE_MAX_LENGTH + 1];
	BT_DBG("+");

	_bt_convert_addr_type_to_string(address, pin_req_event->address.addr);

	BT_INFO("Address[%s]", address);
	BT_INFO("Name[%s]", pin_req_event->name);
	BT_INFO("COD[%d]", pin_req_event->cod);

	if (trigger_pairing_info) {
		/* BTAPI support only one pairing at a time */
		BT_ERR("Already Pairing address [%s]", trigger_pairing_info->addr);
		BT_ERR("New PIN request address [%s]", address);
		device_reject_pin_request(&pin_req_event->address);
		return;
	}

	/* If user initiated bonding and auto response is possible, just reply with default 0000*/
	if (_bt_is_bonding_device_address(address) == TRUE &&
			_bt_agent_is_auto_response(pin_req_event->cod, address, pin_req_event->name)) {
		/* Note: Currently even if SYSPOPUP is supported, we use Fixed PIN "0000" for basic pairing
		   as BT SYSPOPUP is currently not working for PIN KEY entry in Tizen platform. This needs
		   to be checked and fixed apropriately */
		_bt_set_autopair_status_in_bonding_info(TRUE);
		device_accept_pin_request(&pin_req_event->address, "0000");
	} else if (_bt_agent_is_hid_keyboard(pin_req_event->cod)) {
		BT_DBG("Remote Device is HID keyboard Type..");
		char str_passkey[BT_PASSKEY_MAX_LENGTH + 1] = { 0 };

		if (_bt_agent_generate_passkey(str_passkey,
					BT_PASSKEY_MAX_LENGTH) != 0) {
			device_reject_pin_request(&pin_req_event->address);
			goto done;
		}
		device_accept_pin_request(&pin_req_event->address, str_passkey);

		BT_DBG("Send BLUETOOTH_EVENT_KEYBOARD_PASSKEY_DISPLAY");
		param = g_variant_new("(isss)", BLUETOOTH_ERROR_NONE, address, pin_req_event->name, str_passkey);
		_bt_send_event(BT_ADAPTER_EVENT,
				BLUETOOTH_EVENT_KEYBOARD_PASSKEY_DISPLAY, param);
		BT_DBG("Key board pairing in process");
	} else if (BLUETOOTH_ERROR_NONE == __bt_get_device_pin_code(address, pin_code)) {
		BT_DBG("Use stored PIN code [%s]", pin_code);
		device_accept_pin_request(&pin_req_event->address, pin_code);
	} else {
		if (_bt_is_bonding_device_address(address) == TRUE) {
			BT_DBG("Show Pin entry");
			trigger_pairing_info = g_malloc0(sizeof(bt_pairing_data_t));
			trigger_pairing_info->addr = g_strdup(address);
			trigger_pairing_info->is_ssp = FALSE;

			BT_DBG("Send BLUETOOTH_EVENT_PIN_REQUEST");
			param = g_variant_new("(iss)", BLUETOOTH_ERROR_NONE, address, pin_req_event->name);
			_bt_send_event(BT_ADAPTER_EVENT,
					BLUETOOTH_EVENT_PIN_REQUEST, param);
		}
	}

done:
	_bt_agent_release_memory();
	BT_DBG("-");
}

/* SSP Pairing event handler */
static void __bt_device_ssp_passkey_entry_callback(remote_device_t* dev_info)
{
	GVariant *param;
	gchar address[BT_ADDRESS_STR_LEN];
	char *p_addr;
	gchar *name;
	int result = BLUETOOTH_ERROR_NONE;
	BT_DBG("+");

	_bt_convert_addr_type_to_string(address, dev_info->address.addr);
	p_addr = address;
	name = dev_info->name;

	BT_INFO("Address[%s]", address);
	BT_INFO("Name[%s]", name);
	BT_INFO("COD[%d]", dev_info->cod);

	if (trigger_pairing_info) {
		/* BTAPI support only one pairing at a time */
		BT_ERR("Already Pairing address [%s]", trigger_pairing_info->addr);
		BT_ERR("New PIN request address [%s]", address);
		device_reject_pin_request(&dev_info->address);
		BT_DBG("-");
		return;
	}

	/* Set pairing data */
	trigger_pairing_info = g_malloc0(sizeof(bt_pairing_data_t));
	trigger_pairing_info->addr = g_strdup(address);
	trigger_pairing_info->is_ssp = TRUE;

	param = g_variant_new("(iss)", result, p_addr, name);
	_bt_send_event(BT_ADAPTER_EVENT,
			BLUETOOTH_EVENT_PASSKEY_REQUEST, param);
	BT_DBG("-");
}

static void __bt_device_ssp_passkey_confirmation_callback(event_dev_passkey_t *dev_info)
{
	GVariant *param;
	gchar address[BT_ADDRESS_STR_LEN];
	char *p_addr;
	gchar *name;
	char str_passkey[7];
	int result = BLUETOOTH_ERROR_NONE;
	BT_DBG("+");

	_bt_convert_addr_type_to_string(address, dev_info->device_info.address.addr);
	p_addr = address;
	name = dev_info->device_info.name;

	BT_INFO("Address[%s]", address);
	BT_INFO("Name[%s]", name);
	BT_INFO("COD[%d]", dev_info->device_info.cod);

	if (trigger_pairing_info) {
		/* BTAPI support only one pairing at a time */
		BT_ERR("Already Pairing address [%s]", trigger_pairing_info->addr);
		BT_ERR("New PIN request address [%s]", address);
		device_reject_pin_request(&dev_info->device_info.address);
		BT_DBG("-");
		return;
	}

	/* Set pairing data */
	trigger_pairing_info = g_malloc0(sizeof(bt_pairing_data_t));
	trigger_pairing_info->addr = g_strdup(address);
	trigger_pairing_info->is_ssp = TRUE;

	BT_DBG("Send BLUETOOTH_EVENT_PASSKEY_CONFIRMATION");
	snprintf(str_passkey, sizeof(str_passkey), "%.6d", dev_info->pass_key);

	param = g_variant_new("(isss)", result, p_addr, name, str_passkey);
	_bt_send_event(BT_ADAPTER_EVENT,
			BLUETOOTH_EVENT_PASSKEY_CONFIRM_REQUEST, param);
	BT_DBG("-");
}

static void __bt_device_ssp_passkey_display_callback(event_dev_passkey_t *dev_info)
{
	GVariant *param;
	gchar address[BT_ADDRESS_STR_LEN];
	char *p_addr;
	gchar *name;
	char str_passkey[7];
	int result = BLUETOOTH_ERROR_NONE;
	BT_DBG("+");

	_bt_convert_addr_type_to_string(address, dev_info->device_info.address.addr);
	p_addr = address;
	name = dev_info->device_info.name;

	BT_INFO("Address[%s]", address);
	BT_INFO("Name[%s]", name);
	BT_INFO("COD[%d]", dev_info->device_info.cod);

	if (trigger_pairing_info) {
		/* BTAPI support only one pairing at a time */
		BT_ERR("Already Pairing address [%s]", trigger_pairing_info->addr);
		BT_ERR("New PIN request address [%s]", address);
		device_reject_pin_request(&dev_info->device_info.address);
		BT_DBG("-");
		return;
	}

	/* Set pairing data */
	trigger_pairing_info = g_malloc0(sizeof(bt_pairing_data_t));
	trigger_pairing_info->addr = g_strdup(address);
	trigger_pairing_info->is_ssp = TRUE;

	BT_DBG("Send BLUETOOTH_EVENT_KEYBOARD_PASSKEY_DISPLAY");
	snprintf(str_passkey, sizeof(str_passkey), "%06d", dev_info->pass_key);

	param = g_variant_new("(isss)", result, p_addr, name, str_passkey);
	if (passkey_watcher) {
		BT_INFO("Send passkey to %s", passkey_watcher);
		_bt_send_event_to_dest(passkey_watcher, BT_ADAPTER_EVENT,
				BLUETOOTH_EVENT_PASSKEY_NOTIFICATION, param);
	} else {
		_bt_send_event(BT_ADAPTER_EVENT,
				BLUETOOTH_EVENT_KEYBOARD_PASSKEY_DISPLAY, param);
	}
	BT_DBG("-");
}

static void __bt_device_ssp_consent_callback(remote_device_t* dev_info)
{
	gchar address[BT_ADDRESS_STR_LEN];
	gchar *name;
	int local_major;
	int local_minor;
	int cod;
	BT_DBG("+");

	_bt_convert_addr_type_to_string(address, dev_info->address.addr);
	name = dev_info->name;
	cod = dev_info->cod;

	BT_INFO("Address[%s]", address);
	BT_INFO("Name[%s]", name);
	BT_INFO("COD[%d]", cod);

	if (trigger_pairing_info) {
		/* BTAPI support only one pairing at a time */
		BT_ERR("Already Pairing address [%s]", trigger_pairing_info->addr);
		BT_ERR("New PIN request address [%s]", address);
		device_reject_pin_request(&dev_info->address);
		BT_DBG("-");
		return;
	}

	/* Set pairing data */
	trigger_pairing_info = g_malloc0(sizeof(bt_pairing_data_t));
	trigger_pairing_info->addr = g_strdup(address);
	trigger_pairing_info->is_ssp = TRUE;

	local_major = ((cod >> 8) & 0x001f);
	local_minor = (cod & 0x00fc);
	BT_DBG("SSP_CONSENT: Major type=[0x%x] and Minor type=[0x%x]", local_major, local_minor);

	/*TODO: BLUETOOTH_EVENT_SSP_CONSENT_REQUEST to be handled in Tizen */
	BT_DBG("-");
}

static int __bt_oal_status_to_bt_error(int oal_status)
{
        int ret = 0;

	switch (oal_status) {
	case OAL_STATUS_SUCCESS:
		ret = BLUETOOTH_ERROR_NONE;
		break;
	case OAL_STATUS_CONN_TIMEOUT:
	case OAL_STATUS_LINK_LOSS:
		BT_INFO("Connection Timeout");
		ret = BLUETOOTH_ERROR_CONNECTION_TIMEOUT;
		break;
#ifdef TIZEN_BT_HAL
	case OAL_STATUS_CONN_TERM_LOCAL_HOST:
		ret = BLUETOOTH_ERROR_LOCAL_HOST_TERM;
		break;
	case OAL_STATUS_CONN_TERM_RMT_HOST:
		ret = BLUETOOTH_ERROR_REMOTE_USER_TERM;
		break;
#endif
	case OAL_STATUS_INTERNAL_ERROR:
		ret = BLUETOOTH_ERROR_INTERNAL;
		break;
	default:
		ret = BLUETOOTH_ERROR_INTERNAL;
		break;
	}
	return ret;
}

static void __bt_device_acl_state_changed_callback(event_dev_conn_status_t *acl_event,
		gboolean connected, unsigned char type)
{
	gchar address[BT_ADDRESS_STR_LEN];
	int result = BLUETOOTH_ERROR_NONE;
	GVariant *param = NULL;
	bt_device_conn_info_t conn_info;

	BT_DBG("+");
	_bt_convert_addr_type_to_string(address, acl_event->address.addr);

	_bt_logging_connection(connected, type);

	result = __bt_oal_status_to_bt_error(acl_event->status);
	BT_INFO("Result [0x%x]", result);

	if (connected) {
		param = g_variant_new("(isy)", result, address, type);
		_bt_send_event(BT_DEVICE_EVENT,
				BLUETOOTH_EVENT_DEVICE_CONNECTED,
				param);
	} else {
		param = g_variant_new("(isy)", result, address, type);
		_bt_send_event(BT_DEVICE_EVENT,
				BLUETOOTH_EVENT_DEVICE_DISCONNECTED,
				param);
	}

	conn_info.connected = connected;
	conn_info.type = type;
	/* Update local cache */
	_bt_update_remote_dev_property(address, DEV_PROP_CONNECTED, (void *)&conn_info);

#ifdef TIZEN_GATT_CLIENT
	/*handle LE connected device info*/
	if (type) {
		BT_DBG("handle LE connected device info");
		_bt_handle_le_connected_dev_info(address, connected);
	}
#endif

	BT_DBG("-");
}

static void __bt_device_remote_device_found_callback(gpointer event_data, gboolean is_ble)
{
	bt_remote_dev_info_t *dev_info = NULL;
	int result = BLUETOOTH_ERROR_NONE;
	GVariant *param = NULL;
	GVariant *uuids = NULL;
	GVariant *manufacturer_data = NULL;
	GVariantBuilder *builder = NULL;
	unsigned int i;

	BT_INFO("+");
	ret_if(_bt_is_discovering() == FALSE);
	ret_if(event_data == NULL);

	dev_info = g_malloc0(sizeof(bt_remote_dev_info_t));

	if (is_ble) {
		event_ble_dev_found_t * oal_ble_dev = event_data;
		BT_INFO("Device type [%d]", oal_ble_dev->device_info.type);

		_bt_copy_remote_dev(dev_info, &oal_ble_dev->device_info);

		dev_info->manufacturer_data_len = oal_ble_dev->adv_len;
		if (dev_info->manufacturer_data_len)
			dev_info->manufacturer_data = g_memdup(oal_ble_dev->adv_data,
					dev_info->manufacturer_data_len);
		else
			dev_info->manufacturer_data = NULL;
		BT_DBG("----Advertising Data Length: %d", dev_info->manufacturer_data_len);
	} else {
		event_dev_found_t * oal_dev = event_data;
		_bt_copy_remote_dev(dev_info, &oal_dev->device_info);
	}

	/* If Remote device name is NULL or still RNR is not done then display address as name. */
	if (dev_info->name == NULL)
		dev_info->name = g_strdup(dev_info->address);
	BT_DBG("Name %s", dev_info->name);

	builder = g_variant_builder_new(G_VARIANT_TYPE("as"));
	for (i = 0; i < dev_info->uuid_count; i++)
		g_variant_builder_add(builder, "s", dev_info->uuids[i]);

	uuids = g_variant_new("as", builder);
	g_variant_builder_unref(builder);

	manufacturer_data = g_variant_new_from_data(G_VARIANT_TYPE_BYTESTRING,
			dev_info->manufacturer_data,
			dev_info->manufacturer_data_len,
			TRUE,
			NULL, NULL);

	param = g_variant_new("(isunsbub@asn@ay)", result,
			dev_info->address,
			dev_info->class,
			dev_info->rssi,
			dev_info->name,
			dev_info->paired,
			dev_info->connected,
			dev_info->trust,
			uuids,
			dev_info->manufacturer_data_len,
			manufacturer_data);

	_bt_send_event(BT_ADAPTER_EVENT,
			BLUETOOTH_EVENT_REMOTE_DEVICE_FOUND,
			param);

	_bt_free_remote_dev(dev_info);
	BT_DBG("-");
}

static void __bt_device_trusted_callback(gboolean trusted, event_dev_trust_t* info)
{
	gchar address[BT_ADDRESS_STR_LEN];
	int result = BLUETOOTH_ERROR_NONE;
	GVariant *param = NULL;
	int event;
	BT_DBG("+");

	_bt_convert_addr_type_to_string(address, info->address.addr);

	/* Update local cache */
	_bt_update_remote_dev_property(address, DEV_PROP_TRUST, (void *)&trusted);

	param = g_variant_new("(is)", result, address);
	event = trusted ? BLUETOOTH_EVENT_DEVICE_AUTHORIZED :
		BLUETOOTH_EVENT_DEVICE_UNAUTHORIZED;
	/* Send event to application */
	_bt_send_event(BT_DEVICE_EVENT,
			event,
			param);

	BT_DBG("-");
}

static void __bt_free_pairing_info(bt_pairing_data_t **p_info)
{
	BT_DBG("+");
	bt_pairing_data_t * info = *p_info;
	if (info) {
		if (info->addr)
			g_free(info->addr);

		g_free(info);
	}
	*p_info = NULL;
	BT_DBG("-");
}

static void __bt_free_bond_info(uint8_t type)
{
	BT_INFO("+");

	switch (type) {
	case BT_DEVICE_BOND_INFO:
		if (!trigger_bond_info)
			break;

		if (trigger_bond_info->addr)
			g_free(trigger_bond_info->addr);
		if (trigger_bond_info->dev_addr)
			g_free(trigger_bond_info->dev_addr);
		if (trigger_bond_info->dev_info)
			_bt_free_remote_dev(trigger_bond_info->dev_info);
		g_free(trigger_bond_info);
		trigger_bond_info = NULL;
		break;
	case BT_DEVICE_INCOMING_BOND_INFO:
		if (!incoming_bond_info)
			break;

		if (incoming_bond_info->dev_info)
			_bt_free_remote_dev(incoming_bond_info->dev_info);
		g_free(incoming_bond_info);
		incoming_bond_info = NULL;
		break;
	case BT_DEVICE_UNBOND_INFO:
		if (!trigger_unbond_info)
			break;

		if (trigger_unbond_info->addr)
			g_free(trigger_unbond_info->addr);
		if (trigger_unbond_info->dev_addr)
			g_free(trigger_unbond_info->dev_addr);
		if (trigger_unbond_info->dev_info)
			_bt_free_remote_dev(trigger_unbond_info->dev_info);
		g_free(trigger_unbond_info);
		trigger_unbond_info = NULL;
		break;
	}
	BT_INFO("-");
}

static void __bt_free_service_search_info(bt_service_search_info_data_t **p_info)
{
	bt_service_search_info_data_t * info = *p_info;
	if (info) {
		if (info->addr) {
			g_free(info->addr);
			info->addr = NULL;
		}

		if (info->dev_addr) {
			g_free(info->dev_addr);
			info->dev_addr = NULL;
		}

		if (info->dev_info) {
			_bt_free_remote_dev(info->dev_info);
			info->dev_info = NULL;
		}

		g_free(info);
	}
	*p_info = NULL;
}

static int __bt_device_handle_bond_state(void)
{
	BT_INFO("Current Bond state: %d", bt_device_bond_state);
	int ret = OAL_STATUS_INTERNAL_ERROR;

	switch (bt_device_bond_state) {
	case BT_DEVICE_BOND_STATE_CANCEL_DISCOVERY:
		/*TODO:Bonding during discovery: Unhandled!!*/
		BT_INFO("Bonding during discovery: Unhandled!!");
		break;
	case BT_DEVICE_BOND_STATE_DISCOVERY_CANCELLED:
		/*TODO:Bonding during discovery: Unhandled!!*/
		BT_INFO("Bonding during discovery: Unhandled!!");
		break;
	case BT_DEVICE_BOND_STATE_REMOVE_BONDING:
		bt_device_bond_state = BT_DEVICE_BOND_STATE_REMOVED_BONDING;
		ret = device_destroy_bond((bt_address_t *)trigger_bond_info->dev_addr);
		if (ret != OAL_STATUS_SUCCESS)
			ret = __bt_device_handle_bond_state();
		break;
	case BT_DEVICE_BOND_STATE_REMOVED_BONDING:
		bt_device_bond_state = BT_DEVICE_BOND_STATE_NONE;
		ret = device_create_bond((bt_address_t *)trigger_bond_info->dev_addr, trigger_bond_info->conn_type);
		/* Bonding procedure was started but unfortunately could not complete.
		   Basically removed bonding was success, but create bond request could not proceed
		   So lets cleanup the context */
		if (ret != OAL_STATUS_SUCCESS) {
			BT_ERR("Create Bond procedure could not suceed");
			__bt_device_handle_pending_requests(BLUETOOTH_ERROR_INTERNAL, BT_BOND_DEVICE,
					trigger_bond_info->addr, BT_ADDRESS_STRING_SIZE);
			__bt_free_bond_info(BT_DEVICE_BOND_INFO);
			__bt_free_pairing_info(&trigger_pairing_info);
		}
		break;
	case BT_DEVICE_BOND_STATE_NONE:
		BT_INFO("Create Bond failed!!");
		if (trigger_bond_info) {
			__bt_device_handle_pending_requests(BLUETOOTH_ERROR_INTERNAL, BT_BOND_DEVICE,
					trigger_bond_info->addr, BT_ADDRESS_STRING_SIZE);
			__bt_free_bond_info(BT_DEVICE_BOND_INFO);
			__bt_free_pairing_info(&trigger_pairing_info);
		}
		break;
	default:
		break;
	}

	if (ret != OAL_STATUS_SUCCESS)
		return BLUETOOTH_ERROR_INTERNAL;
	else
		return BLUETOOTH_ERROR_NONE;
}

int _bt_device_get_bonded_device_info(bluetooth_device_address_t *addr)
{
	int result;
	bt_address_t bd_addr;

	BT_DBG("+");

	retv_if(!addr, BLUETOOTH_ERROR_INVALID_PARAM);

	memcpy(bd_addr.addr, addr, BLUETOOTH_ADDRESS_LENGTH);
	result = device_query_attributes(&bd_addr);
	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("device_query_attributes error: [%d]", result);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_set_alias(bluetooth_device_address_t *device_address, const char *alias)
{
	int ret;
	char address[BT_ADDRESS_STRING_SIZE];

	BT_DBG("+");
	BT_CHECK_PARAMETER(alias, return);

	ret = device_set_alias((bt_address_t *)device_address, (char *)alias);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("device_set_alias: %d", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	/* Update local cache */
	_bt_convert_addr_type_to_string(address, device_address->addr);
	_bt_update_remote_dev_property(address, DEV_PROP_ALIAS, (void *)alias);

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_bond_device(bluetooth_device_address_t *device_address,
		unsigned short conn_type, GArray **out_param1)
{
	int result = BLUETOOTH_ERROR_NONE;
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };
	bluetooth_device_info_t dev_info;
	BT_DBG("+");

	retv_if(device_address == NULL, BLUETOOTH_ERROR_INVALID_PARAM);

	/* If bonding or discovery already going on */
	if (trigger_bond_info || _bt_is_discovering()) {
		BT_ERR("Device is buzy, bonding can not proceed now..");
		result = BLUETOOTH_ERROR_DEVICE_BUSY;
		goto fail;
	}

	/*TODO: If unbonding with same device going on */
	_bt_convert_addr_type_to_string(address, device_address->addr);

	trigger_bond_info = g_malloc0(sizeof(bt_bond_data_t));
	trigger_bond_info->addr = g_strdup(address);
	trigger_bond_info->conn_type = conn_type;
	trigger_bond_info->is_device_creating = TRUE;
	trigger_bond_info->dev_addr = g_memdup(device_address, sizeof(bluetooth_device_address_t));
	trigger_bond_info->dev_info = NULL;

	/* Ready to initiate bonding */

	/* In Tizen, we will first remove bond and then attempt to create bond to keep
	   consistency with bluedroid. Even if remove bond fails due to device not already
	   bonded, then straight away create bond is triggered. This is because, remove bond
	   is handled differently in bluedroid and bluez. In Bluez, if device is
	   already removed, remove bond call fails.
	   However in bluedroid, remove bond on already removed device returns success. So we will
	   handle the cases transparently*/
	bt_device_bond_state = BT_DEVICE_BOND_STATE_REMOVE_BONDING;
	bond_retry_count = 0;
	result = __bt_device_handle_bond_state();

	if (result != BLUETOOTH_ERROR_NONE)
		goto fail;

	BT_DBG("-");
	return result;

fail:
	memset(&dev_info, 0x00, sizeof(bluetooth_device_info_t));
	memcpy(dev_info.device_address.addr, device_address->addr,
			BLUETOOTH_ADDRESS_LENGTH);

	g_array_append_vals(*out_param1, &dev_info,
			sizeof(bluetooth_device_info_t));
	__bt_free_bond_info(BT_DEVICE_BOND_INFO);

	BT_DBG("-");
	return result;
}

int _bt_unbond_device(bluetooth_device_address_t *device_address,
		GArray **out_param1)
{
	int result = OAL_STATUS_SUCCESS;
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };
	bluetooth_device_info_t dev_info;
	BT_INFO("+");

	retv_if(device_address == NULL, BLUETOOTH_ERROR_INVALID_PARAM);

	_bt_convert_addr_type_to_string(address, device_address->addr);

	trigger_unbond_info = g_malloc0(sizeof(bt_bond_data_t));
	trigger_unbond_info->addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
	trigger_unbond_info->addr = g_strdup(address);
	trigger_unbond_info->dev_addr = g_memdup(device_address, sizeof(bluetooth_device_address_t));

	/* Check if Bonding is already going on, we should not abruptly remove bonding*/
	if (trigger_bond_info && strncmp(trigger_bond_info->addr, trigger_unbond_info->addr, BT_ADDRESS_STRING_SIZE) == 0) {
		BT_ERR("Bonding with same device already ongoing");
		result = BLUETOOTH_ERROR_PERMISSION_DEINED;
		goto fail;
	}

	result = device_destroy_bond((bt_address_t *)device_address);
	if (result != OAL_STATUS_SUCCESS)
		goto fail;

	return result;

fail:
	memset(&dev_info, 0x00, sizeof(bluetooth_device_info_t));
	_bt_convert_addr_string_to_type(dev_info.device_address.addr,
			trigger_unbond_info->addr);

	g_array_append_vals(*out_param1, &dev_info,
			sizeof(bluetooth_device_info_t));
	__bt_free_bond_info(BT_DEVICE_UNBOND_INFO);

	return result;
}

int _bt_cancel_bonding(void)
{
	int result = OAL_STATUS_SUCCESS;
	BT_INFO("+");

	retv_if(trigger_bond_info == NULL, BLUETOOTH_ERROR_NOT_IN_OPERATION);

	result = device_stop_bond((bt_address_t *)trigger_bond_info->dev_addr);

	if (result == OAL_STATUS_SUCCESS)
		trigger_bond_info->is_cancelled_by_user = TRUE;

	return result;
}

int _bt_passkey_reply(const char *passkey, gboolean cnfm_reply)
{
	bluetooth_device_address_t device_address;
	int ret = OAL_STATUS_SUCCESS;
	BT_INFO("reply: %d", cnfm_reply);

	retv_if(trigger_pairing_info == NULL, BLUETOOTH_ERROR_INTERNAL);
	retv_if(trigger_pairing_info->addr == NULL, BLUETOOTH_ERROR_INTERNAL);

	_bt_convert_addr_string_to_type(device_address.addr, trigger_pairing_info->addr);

	if (trigger_pairing_info->is_ssp) {
		if (cnfm_reply)
			ret = device_accept_passkey_entry((bt_address_t *)&device_address, atoi(passkey));
		else
			ret = device_reject_passkey_entry((bt_address_t *)&device_address);
		trigger_pairing_info->is_ssp = FALSE;
	} else {
		if (cnfm_reply)
			ret = device_accept_pin_request((bt_address_t *)&device_address, passkey);
		else
			ret = device_reject_pin_request((bt_address_t *)&device_address);
	}

	__bt_free_pairing_info(&trigger_pairing_info);

	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("_bt_device_handle_passkey_reply: err [%d]", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_INFO("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_passkey_confirmation_reply(gboolean cnfm_reply)
{
	BT_INFO("BT_PASSKEY_CONFIRMATION_REPLY");
	bluetooth_device_address_t device_address;
	int ret = OAL_STATUS_SUCCESS;
	BT_INFO("reply: %d", cnfm_reply);

	retv_if(trigger_pairing_info == NULL, BLUETOOTH_ERROR_INTERNAL);
	retv_if(trigger_pairing_info->addr == NULL, BLUETOOTH_ERROR_INTERNAL);

	_bt_convert_addr_string_to_type(device_address.addr, trigger_pairing_info->addr);

	ret = device_reply_passkey_confirmation((bt_address_t *)&device_address, cnfm_reply);

	__bt_free_pairing_info(&trigger_pairing_info);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("_bt_device_handle_passkey_confirmation_reply: err [%d]", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_INFO("-");
	return BLUETOOTH_ERROR_NONE;
}

gboolean _bt_device_is_pairing(void)
{
	return (trigger_pairing_info) ? TRUE : FALSE;
}

gboolean _bt_device_is_bonding(void)
{
	return (trigger_bond_info) ? TRUE : FALSE;
}

gboolean _bt_is_bonding_device_address(const char *address)
{
	if (trigger_bond_info == NULL || trigger_bond_info->addr == NULL)
		return FALSE;

	if (g_strcmp0(trigger_bond_info->addr, address) == 0) {
		BT_DBG("[%s]  is bonding device", address);
		return TRUE;
	}

	BT_DBG("[%s]  is NOT bonding device", address);
	return FALSE;
}

void _bt_set_autopair_status_in_bonding_info(gboolean is_autopair)
{
	ret_if(trigger_bond_info == NULL);
	trigger_bond_info->is_autopair = is_autopair;
}

int _bt_search_device(bluetooth_device_address_t *device_address)
{
	int result = OAL_STATUS_SUCCESS;
	BT_DBG("+");

	BT_CHECK_PARAMETER(device_address, return);

	if (trigger_bond_info) {
		BT_ERR("Bonding in progress");
		return BLUETOOTH_ERROR_DEVICE_BUSY;
	}

	if (service_search_info) {
		BT_ERR("Service searching in progress");
		return BLUETOOTH_ERROR_DEVICE_BUSY;
	}

	/* allocate user data so that it can be retrieved in callback */
	service_search_info = g_malloc0(sizeof(bt_service_search_info_data_t));
	service_search_info->addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
	service_search_info->dev_addr = g_memdup(device_address, sizeof(bluetooth_device_address_t));

	_bt_convert_addr_type_to_string(service_search_info->addr,
			device_address->addr);

	result = device_query_services((bt_address_t *)device_address);

	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("Device Service Search Failed..: %d", result);
		__bt_free_service_search_info(&service_search_info);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	return BLUETOOTH_ERROR_NONE;
}

int _bt_cancel_search_device(void)
{
	int ret = OAL_STATUS_SUCCESS;
	retv_if(service_search_info == NULL, BLUETOOTH_ERROR_NOT_IN_OPERATION);

	ret = device_stop_query_sevices((bt_address_t *)service_search_info->dev_addr);

	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("SDP Cancel request failed [%d]", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	__bt_device_handle_pending_requests(BLUETOOTH_ERROR_CANCEL_BY_USER, BT_SEARCH_SERVICE,
			service_search_info->addr, BT_ADDRESS_STRING_SIZE);

	__bt_free_service_search_info(&service_search_info);

	return BLUETOOTH_ERROR_NONE;
	BT_DBG("-");
}

int _bt_set_authorization(bluetooth_device_address_t *device_address,
		gboolean authorize)
{
	int ret = OAL_STATUS_SUCCESS;
	BT_DBG("+");

	BT_CHECK_PARAMETER(device_address, return);
	BT_INFO("Device to be Trusted? [%d]", authorize);

	ret = device_set_authorized((bt_address_t*)device_address, authorize);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("device_set_authorized: %d", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	return BLUETOOTH_ERROR_NONE;
}

gboolean _bt_is_device_connected(bluetooth_device_address_t *device_address, int svc_type)
{
	gboolean is_connected;
	oal_service_t svc_id;

	retv_if(device_address == NULL, BLUETOOTH_ERROR_INVALID_PARAM);

	/*
	 * TODO: While adding support for new profiles, need to add more
	 * <svc_type, svc_id> mapping here.
	 */
	switch (svc_type) {
	case BT_PROFILE_CONN_HID:
		svc_id = HID_SERVICE_ID;
		break;
	case BT_PROFILE_CONN_A2DP:
		svc_id = A2DP_SERVICE_ID; /* Remote is A2DP Sink */
		break;
	case BT_PROFILE_CONN_A2DP_SINK:
		svc_id = A2DP_SRC_SERVICE_ID; /* Remote is A2DP Source*/
		break;
	case BT_PROFILE_CONN_HSP:
		svc_id = HFP_HS_SERVICE_ID; /* Remote is HFP HF Unit */
		break;
#ifdef TIZEN_GATT_CLIENT
	case BT_PROFILE_CONN_GATT:
		return _bt_is_remote_gatt_device_connected(device_address); /* Remote is GATT client or Server */
#endif
	default:
		BT_DBG("Unknown svc_type: %d", svc_type);
		return FALSE;
	}

	is_connected = device_get_svc_conn_state((bt_address_t*)device_address, svc_id);

	BT_DBG("svc_type: %d, is_connected: %s",
			svc_type, is_connected ? "TRUE" : "FALSE");

	return is_connected;
}

int _bt_rfcomm_reply_conn_authorization(char *address, gboolean reply)
{
	bt_address_t bd_addr;
	int res;

	BT_DBG("+");

	retv_if(NULL == address, BLUETOOTH_ERROR_INVALID_PARAM);
	_bt_convert_addr_string_to_type(bd_addr.addr, address);
	res = device_reply_auth_request(&bd_addr, 0, reply, FALSE);
	if (res != OAL_STATUS_SUCCESS) {
		BT_ERR("authorize_response: %d", res);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_enable_rssi(bluetooth_device_address_t *addr, int link_type,
		int low_threshold, int in_range_threshold, int high_threshold)
{
	int result;
	bt_address_t bd_addr;

	BT_DBG("+");

	retv_if(!addr, BLUETOOTH_ERROR_INVALID_PARAM);

	memcpy(bd_addr.addr, addr, BLUETOOTH_ADDRESS_LENGTH);
	result = device_enable_rssi_monitoring(&bd_addr, link_type,
			low_threshold, in_range_threshold, high_threshold);
	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("device_get_connected_link_rssi_strength error: [%d]", result);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_get_rssi_strength(bluetooth_device_address_t *addr, int link_type)
{
	int result;
	bt_address_t bd_addr;

	BT_DBG("+");

	retv_if(!addr, BLUETOOTH_ERROR_INVALID_PARAM);

	memcpy(bd_addr.addr, addr, BLUETOOTH_ADDRESS_LENGTH);
	result = device_get_connected_link_rssi_strength(&bd_addr, link_type);
	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("device_get_connected_link_rssi_strength error: [%d]", result);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_set_passkey_notification(const char *sender, gboolean enable)
{
	int result;

	BT_INFO("Set passkey notification(sender:%s, %s)",
			sender, enable ? "Enable" : "Disable");

	result = device_enable_gap_auth_notifications(OAL_PASSKEY_DISPLAY, enable);
	if (OAL_STATUS_SUCCESS != result) {
		BT_ERR("device_enable_gap_auth_notifications error: [%d]", result);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	g_free(passkey_watcher);
	if (enable == TRUE)
		passkey_watcher = g_strdup(sender);
	else
		passkey_watcher = NULL;

	return BLUETOOTH_ERROR_NONE;
}

static int __bt_get_device_pin_code(const char *address, char *pin_code)
{
	GSList *l = NULL;

	BT_CHECK_PARAMETER(address, return);
	BT_CHECK_PARAMETER(pin_code, return);

	for (l = pin_info_list; l != NULL; l = l->next) {
		bt_pin_code_info_t *pin_info = l->data;

		if (!pin_info || !pin_info->address)
			continue;

		if (g_strcmp0(pin_info->address, address) == 0) {
			g_strlcpy(pin_code, pin_info->pin_code,
					BLUETOOTH_PIN_CODE_MAX_LENGTH + 1);
			return BLUETOOTH_ERROR_NONE;
		}
	}

	return BLUETOOTH_ERROR_NOT_FOUND;
}

int _bt_set_pin_code(bluetooth_device_address_t *device_address,
		bluetooth_device_pin_code_t *pin_code)
{
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };
	bt_pin_code_info_t *pin_info = NULL;
	GSList *l = NULL;

	BT_CHECK_PARAMETER(device_address, return);
	BT_CHECK_PARAMETER(pin_code, return);
	retv_if(g_slist_length(pin_info_list) >= BT_DEVICE_PIN_CODE_SLOT_MAX,
			BLUETOOTH_ERROR_NO_RESOURCES);

	_bt_convert_addr_type_to_string(address, device_address->addr);

	for (l = pin_info_list; l != NULL; l = l->next) {
		pin_info = l->data;

		if (!pin_info || !pin_info->address)
			continue;

		if (g_strcmp0(pin_info->address, address) == 0) {
			g_free(pin_info->pin_code);
			pin_info->pin_code = g_strdup(pin_code->pin_code);
			return BLUETOOTH_ERROR_NONE;
		}
	}

	pin_info = g_malloc0(sizeof(bt_pin_code_info_t));
	pin_info->address = g_strdup(address);
	pin_info->pin_code = g_strdup(pin_code->pin_code);
	pin_info_list = g_slist_append(pin_info_list, pin_info);

	return BLUETOOTH_ERROR_NONE;
}

int _bt_unset_pin_code(bluetooth_device_address_t *device_address)
{
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };
	bt_pin_code_info_t *pin_info = NULL;
	GSList *l = NULL;

	BT_DBG("+");

	BT_CHECK_PARAMETER(device_address, return);

	_bt_convert_addr_type_to_string(address, device_address->addr);

	for (l = pin_info_list; l != NULL; l = l->next) {
		pin_info = l->data;

		if (!pin_info || !pin_info->address)
			continue;

		if (g_strcmp0(pin_info->address, address) == 0) {
			pin_info_list = g_slist_remove(pin_info_list, pin_info);
			g_free(pin_info->address);
			g_free(pin_info->pin_code);
			g_free(pin_info);
			break;
		}
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

#ifdef TIZEN_GATT_CLIENT
static bt_connected_le_dev_t *__bt_get_le_connected_dev_info(const char *address)
{
	GSList *l = NULL;
	bt_connected_le_dev_t *dev;

	if (!address)
		return NULL;

	for (l = le_connected_dev_list; l; l = g_slist_next(l)) {
		dev = l->data;

		if (g_strcmp0(dev->address, address) == 0)
			return dev;
	}
	return NULL;
}

static void __bt_le_conn_param_free(void *data)
{
	bt_le_conn_param_t *param = (bt_le_conn_param_t *)data;

	BT_DBG("%s", param->sender);
	g_free(param->sender);
	g_free(param);
}

static void _bt_add_le_connected_dev_info(const char *address)
{
	bt_connected_le_dev_t *dev = NULL;

	if (!address)
		return;

	dev = g_malloc0(sizeof(bt_connected_le_dev_t));
	dev->address = g_strdup(address);

	le_connected_dev_list = g_slist_append(le_connected_dev_list, dev);

	return;
}

static void _bt_remove_le_connected_dev_info(const char *address)
{
	bt_connected_le_dev_t *dev = NULL;

	if (!address)
		return;

	dev = __bt_get_le_connected_dev_info(address);
	if (!dev)
		return;

	g_slist_free_full(dev->senders, __bt_le_conn_param_free);
	le_connected_dev_list = g_slist_remove(le_connected_dev_list, dev);
	g_free(dev->address);
	g_free(dev);

	return;
}


static void _bt_handle_le_connected_dev_info(const char *address, gboolean connected)
{
	BT_DBG("+");

	if (connected)
		_bt_add_le_connected_dev_info(address);
	else
		_bt_remove_le_connected_dev_info(address);
}

static bt_le_conn_param_t *__bt_get_le_conn_param_info(bt_connected_le_dev_t *dev, const char *sender)
{
	GSList *l = NULL;
	bt_le_conn_param_t *param = NULL;

	if (!dev || !sender)
		return NULL;

	for (l = dev->senders; l; l = g_slist_next(l)) {
		param = l->data;
		if (g_strcmp0(param->sender, sender) == 0)
			return param;
	}

	return NULL;
}

static gint __bt_compare_le_conn_param_key(gpointer *a, gpointer *b)
{
	bt_le_conn_param_t *parama = (bt_le_conn_param_t *)a;
	bt_le_conn_param_t *paramb = (bt_le_conn_param_t *)b;

	return parama->key > paramb->key;
}


int _bt_add_le_conn_param_info(const char *address, const char *sender,
		float interval_min, float interval_max, guint16 latency, guint16 time_out)
{
	bt_connected_le_dev_t *dev = NULL;
	bt_le_conn_param_t *param = NULL;
	bt_le_conn_param_t *data = NULL;

	BT_DBG("+");

	if (!address || !sender)
		return BLUETOOTH_ERROR_INVALID_PARAM;

	dev = __bt_get_le_connected_dev_info(address);
	if (!dev)
		return BLUETOOTH_ERROR_INTERNAL;

	param = __bt_get_le_conn_param_info(dev, sender);

	data = g_malloc0(sizeof(bt_le_conn_param_t));
	data->sender = g_strdup(sender);
	data->interval_min = interval_min;
	data->interval_max = interval_max;
	data->latency = latency;
	data->time_out = time_out;
	data->key = interval_min + (interval_max - interval_min)/2;

	if (param == NULL) {
		BT_DBG("Add param %s %s %f %f", address, sender, interval_min, interval_max);
		dev->senders = g_slist_append(dev->senders, data);
	} else {
		BT_DBG("Update param %s %s %f %f", address, sender, interval_min, interval_max);
		dev->senders = g_slist_remove(dev->senders, param);
		g_free(param->sender);
		g_free(param);
		dev->senders = g_slist_append(dev->senders, data);
	}

	/* Sorting. First element have the minimum interval */
	dev->senders = g_slist_sort(dev->senders,
			(GCompareFunc)__bt_compare_le_conn_param_key);

	return BLUETOOTH_ERROR_NONE;
}


static int __bt_le_set_conn_parameter(const char *address,
		float interval_min, float interval_max,
		guint16 latency, guint16 time_out)
{
	bt_address_t dev_addr = { {0} };
	guint32 min, max, to;

	BT_INFO("Min interval: %f, Max interval: %f, Latency: %u, Supervision timeout: %u",
			interval_min, interval_max, latency, time_out);

	min = interval_min / BT_LE_CONN_INTERVAL_SPLIT;
	max = interval_max / BT_LE_CONN_INTERVAL_SPLIT;
	to = time_out / BT_LE_CONN_TO_SPLIT;

	BT_INFO("updating: Min interval: %d, Max interval: %d, Latency: %d, Supervision timeout: %d",
			min, max, latency, to);

	_bt_convert_addr_string_to_type(dev_addr.addr, address);

	return gattc_conn_param_update(&dev_addr, min, max, latency, to);
}

int _bt_remove_le_conn_param_info(const char *address, const char *sender)
{
	bt_connected_le_dev_t *dev = NULL;
	bt_le_conn_param_t *param = NULL;

	if (!address || !sender)
		return BLUETOOTH_ERROR_INVALID_PARAM;

	dev = __bt_get_le_connected_dev_info(address);
	if (!dev)
		return BLUETOOTH_ERROR_INTERNAL;

	param = __bt_get_le_conn_param_info(dev, sender);
	if (param) {
		BT_DBG("Remove param %s %s ", address, sender);
		dev->senders = g_slist_remove(dev->senders, param);
		g_free(param->sender);
		g_free(param);
	}

	return BLUETOOTH_ERROR_NONE;
}


int _bt_le_connection_update(const char *sender,
		unsigned char *device_address,
		float interval_min, float interval_max,
		guint16 latency, guint16 time_out)
{
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };
	guint32 min_supervision_to;
	bt_connected_le_dev_t *dev = NULL;
	bt_le_conn_param_t *param = NULL;
	int ret = BLUETOOTH_ERROR_NONE;

	BT_CHECK_PARAMETER(device_address, return);

	BT_INFO("Sender %s, Min interval: %f, Max interval: %f, Latency: %u, Supervision timeout: %u",
			sender, interval_min, interval_max, latency, time_out);

	if (interval_min > interval_max ||
			interval_min < BT_LE_CONN_INTERVAL_MIN ||
			interval_max > BT_LE_CONN_INTERVAL_MAX) {
		ret = BLUETOOTH_ERROR_INVALID_PARAM;
		goto fail;
	}

	if (time_out < BT_LE_CONN_SUPER_TO_MIN ||
			time_out > BT_LE_CONN_SUPER_TO_MAX) {
		ret = BLUETOOTH_ERROR_INVALID_PARAM;
		goto fail;
	}

	if (latency > BT_LE_CONN_SLAVE_LATENCY_MAX) {
		ret = BLUETOOTH_ERROR_INVALID_PARAM;
		goto fail;
	}

	/*
	 * The Supervision_Timeout in milliseconds shall be larger than
	 * (1 + Conn_Latency) * Conn_Interval_Max * 2,
	 * where Conn_Interval_Max is given in milliseconds.
	 */

	min_supervision_to = (1 + latency) * interval_max * 2;
	if (time_out <= min_supervision_to) {
		ret = BLUETOOTH_ERROR_INVALID_PARAM;
		goto fail;
	}

	_bt_convert_addr_type_to_string(address, device_address);
	BT_DBG("Remote device address: %s", address);

	_bt_add_le_conn_param_info(address, sender, interval_min, interval_max, 0, 2000);

	dev = __bt_get_le_connected_dev_info(address);
	if (dev == NULL) {
		BT_DBG("device not found in the list");
		ret = BLUETOOTH_ERROR_NOT_CONNECTED;
		goto fail;
	}

	if (g_slist_length(dev->senders) == 1)
		goto update;
	else {
		param = dev->senders->data;

		BT_DBG("dev %f, param %f, input %f", dev->interval_min, param->interval_min, interval_min);

		if (dev->interval_min == param->interval_min && dev->interval_max == param->interval_max) {
			BT_DBG("Skip due to same interval");
			return ret;
		}

		interval_min = param->interval_min;
		interval_max = param->interval_max;
	}

update:
	ret = __bt_le_set_conn_parameter(address, interval_min, interval_max, latency, time_out);

	if (ret != OAL_STATUS_SUCCESS) {
		_bt_remove_le_conn_param_info(address, sender);
		BT_DBG("fail to update the LE connection parameter");
		ret = BLUETOOTH_ERROR_INTERNAL;
		goto fail;
	}

	BT_DBG("updated LE connection parameter");
	dev->interval_min = interval_min;
	dev->interval_max = interval_max;

	return BLUETOOTH_ERROR_NONE;
fail:
	return ret;
}

#endif
