/*
 * Bluetooth-frwk
 *
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Author: Atul Rai <a.rai@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <gio/gunixfdlist.h>
#include <glib.h>
#include <dlog.h>

/* OAL headers */
#include <oal-event.h>
#include <oal-hdp.h>

/* bt-service headers */
#include "bt-internal-types.h"
#include "bt-service-common.h"
#include "bt-service-util.h"
#include "bt-service-event-receiver.h"
#include "bt-service-event.h"
#include "bt-service-hdp.h"

typedef struct {
	int fd;
	void *data;
} bt_hdp_conn_info_t;

typedef struct {
	int app_id;
	char *owner;
} bt_hdp_app_info_t;

static GSList *app_list;
static GSList *pending_conn_list;

static bt_hdp_app_info_t *__find_app_by_id(int app_id)
{
	GSList *l;

	for (l = app_list; NULL != l; l = g_slist_next(l)) {
		bt_hdp_app_info_t *app = l->data;
		if (app && (app->app_id == app_id))
			return app;
	}

	return NULL;
}

static void __bt_hdp_handle_pending_request_info(int result,
		int service_function, void *data, unsigned int size)
{
	GSList *l;
	GArray *out_param;
	invocation_info_t *req_info = NULL;

	ret_if(data == NULL);

	BT_DBG("+");

	for (l = _bt_get_invocation_list(); l != NULL; l = g_slist_next(l)) {
		req_info = l->data;
		if (req_info == NULL || req_info->service_function != service_function)
			continue;

		switch (service_function) {
		case BT_HDP_REGISTER_SINK_APP:
		case BT_HDP_UNREGISTER_SINK_APP: {
			int app_id = *((int *)data);
			char *app_handle;

			if (*((int *)req_info->user_data) != app_id)
				break;

			app_handle = g_strdup_printf("health_app_%d", app_id);
			BT_DBG("app_handle: %s", app_handle);

			out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
			g_array_append_vals(out_param, app_handle, strlen(app_handle));
			_bt_service_method_return(req_info->context, out_param, result);
			g_free(req_info->user_data);
			_bt_free_info_from_invocation_list(req_info);
			g_array_free(out_param, TRUE);
			g_free(app_handle);
			break;
		}
		case BT_HDP_CONNECT: {
			bt_hdp_conn_info_t *info = data;
			bt_hdp_connected_t *conn_data = info->data;
			bt_hdp_connected_t *req_conn_data = req_info->user_data;

			if (!conn_data)
				break;

			if (req_conn_data->channel_id != conn_data->channel_id)
				break;

			out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
			g_array_append_vals(out_param, conn_data, sizeof(bt_hdp_connected_t));

			_bt_service_method_return(req_info->context, out_param, result);
			g_free(req_info->user_data);
			_bt_free_info_from_invocation_list(req_info);
			g_array_free(out_param, TRUE);
			break;
		}
		case BT_HDP_DISCONNECT: {
			bt_hdp_disconnected_t *disconn_data = data;
			bt_hdp_disconnected_t *req_disconn_data = req_info->user_data;

			if (req_disconn_data->channel_id != disconn_data->channel_id)
				break;

			out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
			g_array_append_vals(out_param, disconn_data, sizeof(bt_hdp_disconnected_t));
			_bt_service_method_return(req_info->context, out_param, result);
			g_free(req_info->user_data);
			_bt_free_info_from_invocation_list(req_info);
			g_array_free(out_param, TRUE);
			break;
		}
		case BT_HDP_GET_FD: {
			GUnixFDList *fd_list = NULL;

			out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
			if (BLUETOOTH_ERROR_NONE == result) {
				GError *error = NULL;
				bt_hdp_conn_info_t *info = data;
				bt_hdp_connected_t *conn_data = info->data;

				if (*((int *)req_info->user_data) != (int)conn_data->channel_id)
					break;

				BT_DBG("fd: %d", info->fd);

				fd_list = g_unix_fd_list_new();
				g_unix_fd_list_append(fd_list, info->fd, &error);
				g_assert_no_error(error);
				close(info->fd);
			}

			_bt_service_method_return_with_unix_fd_list(
					req_info->context, out_param, result, fd_list);
			if (fd_list)
				g_object_unref(fd_list);
			g_free(req_info->user_data);
			_bt_free_info_from_invocation_list(req_info);
			g_array_free(out_param, TRUE);
			return;
		}
		default:
			BT_ERR("Unknown service function");
			break;
		}
	}

	BT_DBG("-");
}

static bt_hdp_qos_type_t __convert_oal_ch_type_to_hdp_qos_type(int channel_type)
{
	switch (channel_type) {
	case OAL_CHANNEL_TYPE_RELIABLE:
		return HDP_QOS_RELIABLE;
	case OAL_CHANNEL_TYPE_STREAMING:
		return HDP_QOS_STREAMING;
	default:
		return HDP_QOS_ANY;
	}
}

static void __handle_hdp_channel_connected(event_hdp_channel_conn_t *event)
{
	GVariant *param;
	bt_hdp_app_info_t *app;
	bt_hdp_conn_info_t *hdp_conn;
	bt_hdp_connected_t *conn_info;
	char address[BT_ADDRESS_STRING_SIZE];
	char *app_handle;

	BT_DBG("+");

	app = __find_app_by_id(event->app_id);
	ret_if(NULL == app);

	app_handle = g_strdup_printf("health_app_%d", event->app_id);

	conn_info = g_new0(bt_hdp_connected_t, 1);
	conn_info->type = __convert_oal_ch_type_to_hdp_qos_type(event->ch_type);
	conn_info->channel_id = event->channel_id;
	conn_info->app_handle = app_handle;
	memcpy(conn_info->device_address.addr, event->address.addr, BLUETOOTH_ADDRESS_LENGTH);

	hdp_conn = g_new0(bt_hdp_conn_info_t, 1);
	hdp_conn->fd = event->fd;
	hdp_conn->data = conn_info;
	pending_conn_list = g_slist_append(pending_conn_list, hdp_conn);

	__bt_hdp_handle_pending_request_info(BLUETOOTH_ERROR_NONE,
			BT_HDP_CONNECT, (void *)hdp_conn, sizeof(bt_hdp_conn_info_t));

	_bt_convert_addr_type_to_string(address, event->address.addr);
	BT_DBG("HDP connected to %s, app_handle: %s", address, conn_info->app_handle);

	param = g_variant_new("(issui)",
			BLUETOOTH_ERROR_NONE,
			address, conn_info->app_handle,
			conn_info->channel_id, conn_info->type);
	_bt_send_event_to_dest(app->owner,
		BT_HDP_EVENT, BLUETOOTH_EVENT_HDP_CONNECTED, param);

	g_free(app_handle);

	BT_DBG("-");
}

static void __handle_hdp_channel_disconnected(event_hdp_channel_conn_t *event)
{
	bt_hdp_app_info_t *app;


	BT_DBG("+");

	app = __find_app_by_id(event->app_id);
	ret_if(NULL == app);

	/* reply to pending requests if any */
	if (0 > event->fd) {
		bt_hdp_conn_info_t hdp_conn;
		bt_hdp_connected_t conn_info;
		char *app_handle;

		BT_DBG("Connect request failed for channel_id: %d", event->channel_id);
		memset(&conn_info, 0x00, sizeof(bt_hdp_connected_t));
		conn_info.type = __convert_oal_ch_type_to_hdp_qos_type(event->ch_type);
		conn_info.channel_id = event->channel_id;

		app_handle = g_strdup_printf("health_app_%d", event->app_id);
		conn_info.app_handle = app_handle;

		memcpy(conn_info.device_address.addr, event->address.addr, BLUETOOTH_ADDRESS_LENGTH);

		hdp_conn.fd = event->fd;
		hdp_conn.data = &conn_info;

		__bt_hdp_handle_pending_request_info(BLUETOOTH_ERROR_INTERNAL,
				BT_HDP_CONNECT, (void *)&hdp_conn, sizeof(bt_hdp_conn_info_t));

		g_free(app_handle);
	} else {
		bt_hdp_disconnected_t disconn_info;
		char address[BT_ADDRESS_STRING_SIZE];
		char *app_handle;
		GVariant *param;

		memset(&disconn_info, 0x00, sizeof(bt_hdp_disconnected_t));
		disconn_info.channel_id = event->channel_id;
		memcpy(disconn_info.device_address.addr, event->address.addr, BLUETOOTH_ADDRESS_LENGTH);

		__bt_hdp_handle_pending_request_info(BLUETOOTH_ERROR_NONE,
				BT_HDP_DISCONNECT, (void *)&disconn_info, sizeof(bt_hdp_disconnected_t));

		/* Send event to application */
		_bt_convert_addr_type_to_string(address, event->address.addr);
		app_handle = g_strdup_printf("health_app_%d", event->app_id);

		BT_DBG("HDP disconnected from %s, app_handle: %s", address, app_handle);
		param = g_variant_new("(isu)",
				BLUETOOTH_ERROR_NONE,
				address, event->channel_id);
		_bt_send_event_to_dest(app->owner,
				BT_HDP_EVENT, BLUETOOTH_EVENT_HDP_DISCONNECTED, param);

		g_free(app_handle);
	}

	BT_DBG("-");
}

static void __bt_hdp_event_handler(int event_type, gpointer event_data)
{
	BT_INFO("OAL event = 0x%x, \n", event_type);

	switch (event_type) {
	case OAL_EVENT_HDP_APP_REGISTERED: {
		event_hdp_app_reg_t *app_data = event_data;
		int result;
		int app_id;

		ret_if(NULL == event_data);
		app_id = app_data->app_id;
		if (OAL_STATUS_SUCCESS != app_data->status) {
			bt_hdp_app_info_t *app = __find_app_by_id(app_id);
			if (app) {
				app_list = g_slist_remove(app_list, app);
				g_free(app->owner);
				g_free(app);
			}
			result = BLUETOOTH_ERROR_INTERNAL;
		} else {
			result = BLUETOOTH_ERROR_NONE;
		}

		BT_DBG("app_id: %d, resut: %d", app_id, result);
		__bt_hdp_handle_pending_request_info(result,
			BT_HDP_REGISTER_SINK_APP, (void *)&app_id, sizeof(app_id));
		break;
	}
	case OAL_EVENT_HDP_APP_UNREGISTERED: {
		event_hdp_app_reg_t *app_data = event_data;
		int result;
		int app_id;

		ret_if(NULL == event_data);
		app_id = app_data->app_id;
		if (OAL_STATUS_SUCCESS != app_data->status) {
			result = BLUETOOTH_ERROR_INTERNAL;
		} else {
			bt_hdp_app_info_t *app = __find_app_by_id(app_id);
			if (app) {
				app_list = g_slist_remove(app_list, app);
				g_free(app->owner);
				g_free(app);
			}
			result = BLUETOOTH_ERROR_NONE;
		}

		BT_DBG("app_id: %d, resut: %d", app_id, result);
		__bt_hdp_handle_pending_request_info(result,
			BT_HDP_UNREGISTER_SINK_APP, (void *)&app_id, sizeof(app_id));
		break;
	}
	case OAL_EVENT_HDP_CHANNEL_CONNECTED: {
		ret_if(NULL == event_data);
		__handle_hdp_channel_connected((event_hdp_channel_conn_t *)event_data);
		break;
	}
	case OAL_EVENT_HDP_CHANNEL_DESTROYED: {
		ret_if(NULL == event_data);
		__handle_hdp_channel_disconnected((event_hdp_channel_conn_t *)event_data);
		break;
	}
	default:
		BT_ERR("Invalid event:%d\n", event_type);
		break;
	}
}

int _bt_hdp_init(void)
{
	BT_DBG("+");

	if (OAL_STATUS_SUCCESS != hdp_init()) {
		BT_ERR("hdp_init failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	/* Register SOCKET event handler */
	_bt_service_register_event_handler_callback(BT_HEALTH_MODULE, __bt_hdp_event_handler);
	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

void _bt_hdp_deinit(void)
{
	BT_DBG("+");

	if (OAL_STATUS_SUCCESS != hdp_cleanup())
		BT_ERR("hdp_cleanup failed");

	/* Un-register SOCKET event handler */
	_bt_service_unregister_event_handler_callback(BT_HEALTH_MODULE);
	BT_DBG("-");
}

int _bt_hdp_app_register(bt_hdp_role_type_t role, bt_hdp_qos_type_t channel_type,
		unsigned short data_type, char *sender, int *app_id)
{
	oal_channel_type_t ch_type;
	oal_hdp_role_t dev_role;
	bt_hdp_app_info_t *app;

	BT_DBG("+");

	retv_if(NULL == app_id, BLUETOOTH_ERROR_INVALID_PARAM);

	switch (role) {
	case HDP_ROLE_SOURCE:
		dev_role = OAL_HDP_ROLE_SOURCE;
		break;
	case HDP_ROLE_SINK:
		dev_role = OAL_HDP_ROLE_SINK;
		break;
	default:
		BT_ERR("Invalid HDP role");
		return BLUETOOTH_ERROR_INVALID_PARAM;
	}

	switch (channel_type) {
	case HDP_QOS_RELIABLE:
		ch_type = OAL_CHANNEL_TYPE_RELIABLE;
		break;
	case HDP_QOS_STREAMING:
		ch_type = OAL_CHANNEL_TYPE_STREAMING;
		break;
	case HDP_QOS_ANY:
		ch_type = OAL_CHANNEL_TYPE_ANY;
		break;
	default:
		BT_ERR("Invalid channel type");
		return BLUETOOTH_ERROR_INVALID_PARAM;
	}

	*app_id = hdp_register_application(dev_role, ch_type, sender, data_type);
	if (0 > *app_id) {
		BT_ERR("hdp_register_application failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	/* Add application to the list */
	app = g_new0(bt_hdp_app_info_t, 1);
	app->app_id = *app_id;
	app->owner = g_strdup(sender);
	app_list = g_slist_append(app_list, app);

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_hdp_app_unregister(int app_id)
{
	int ret;

	BT_DBG("+");

	ret = hdp_unregister_application(app_id);
	if (OAL_STATUS_SUCCESS != ret) {
		BT_ERR("hdp_unregister_application failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_hdp_connect(int app_id, bluetooth_device_address_t *address,
		bt_hdp_qos_type_t channel_type, int *channel_id)
{
	bt_address_t bd_addr;

	BT_DBG("+");

	retv_if(NULL == address, BLUETOOTH_ERROR_INVALID_PARAM);
	retv_if(NULL == channel_id, BLUETOOTH_ERROR_INVALID_PARAM);

	memset(&bd_addr, 0x00, sizeof(bt_address_t));
	memcpy(bd_addr.addr, address->addr, BT_ADDRESS_BYTES_NUM);

	*channel_id = hdp_connect_channel(app_id, &bd_addr);
	if (0 > *channel_id) {
		BT_ERR("hdp_connect_channel failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_hdp_disconnect(int channel_id)
{
	int ret;

	BT_DBG("+");

	ret = hdp_disconnect_channel(channel_id);
	if (OAL_STATUS_SUCCESS != ret) {
		BT_ERR("hdp_disconnect_channel failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

gboolean __send_fd(gpointer data)
{
	bt_hdp_conn_info_t *info = data;

	BT_DBG("+");

	if (info) {
		pending_conn_list = g_slist_remove(pending_conn_list, info);
		__bt_hdp_handle_pending_request_info(BLUETOOTH_ERROR_NONE,
				BT_HDP_GET_FD, (void *)info, sizeof(bt_hdp_conn_info_t));
		g_free(info->data);
		g_free(info);
	} else {
		__bt_hdp_handle_pending_request_info(
				BLUETOOTH_ERROR_INTERNAL, BT_HDP_GET_FD, NULL, 0);
	}

	BT_DBG("-");
	return FALSE;
}

int _bt_hdp_get_fd(int channel_id)
{
	GSList *l;
	bt_hdp_conn_info_t *info = NULL;
	bt_hdp_connected_t *conn_info = NULL;

	BT_DBG("+");

	BT_DBG("channel_id: %d, pending_conn_list: %p", channel_id, pending_conn_list);
	for (l = pending_conn_list; NULL != l; l = g_slist_next(l)) {
		info = l->data;

		if (!info)
			continue;

		conn_info = (bt_hdp_connected_t *)(info->data);
		if (conn_info && ((int)conn_info->channel_id == channel_id)) {
			BT_DBG("Match found");
			break;
		}

		info = NULL;
		conn_info = NULL;
	}

	g_idle_add(__send_fd, info);
	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

void _bt_check_hdp_app_termination(const char *name)
{
	GSList *l;
	bt_hdp_app_info_t *app = NULL;

	ret_if(NULL == name);

	for (l = app_list; NULL != l; l = g_slist_next(l)) {
		app = l->data;

		if (!app)
			continue;

		if (app->owner && !strncasecmp(app->owner, name, strlen(name))) {
			BT_DBG("Match found, name: %s", name);
			_bt_hdp_app_unregister(app->app_id);
		}
	}
}
