/*
 * Copyright (c) 2011 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <gio/gio.h>
#include <glib.h>
#include <dlog.h>
#include <string.h>
#include <vconf.h>
#include <syspopup_caller.h>

#include "bt-internal-types.h"
#include "bt-service-common.h"
#include "bt-service-event.h"
#include "bt-service-core-adapter.h"
#include "bt-service-core-adapter-le.h"
#include "bt-service-util.h"

int _bt_connect_le_ipsp_device(const bluetooth_device_address_t *bd_addr)
{
	char device_address[BT_ADDRESS_STRING_SIZE] = { 0 };
	gchar *device_path = NULL;
	GError *error = NULL;
	GDBusProxy *device_proxy = NULL;
	GDBusProxy *adapter_proxy;
	GDBusConnection *conn;
	int ret = BLUETOOTH_ERROR_NONE;

	BT_CHECK_PARAMETER(bd_addr, return);

	_bt_convert_addr_type_to_string(device_address,
			(unsigned char *)bd_addr->addr);

	conn = _bt_gdbus_get_system_gconn();
	retv_if(conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	adapter_proxy = _bt_get_adapter_proxy();
	retv_if(adapter_proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	device_path = _bt_get_device_object_path(device_address);
	if (device_path == NULL) {
		BT_DBG("device_path NULL");
		ret = BLUETOOTH_ERROR_INTERNAL;
		return ret;
	}

	retv_if(device_path == NULL, BLUETOOTH_ERROR_INTERNAL);

	device_proxy = g_dbus_proxy_new_sync(conn, G_DBUS_PROXY_FLAGS_NONE,
							NULL, BT_BLUEZ_NAME,
							device_path, BT_DEVICE_INTERFACE,  NULL, NULL);
	g_free(device_path);
	retv_if(device_proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	g_dbus_proxy_call_sync(device_proxy, "ConnectIpsp",
				NULL,
				G_DBUS_CALL_FLAGS_NONE,
				-1,
				NULL,
				&error);
	if (error) {
		BT_ERR("ConnectIpsp Call Error %s[%s]", error->message, device_address);
		g_error_free(error);
		g_object_unref(device_proxy);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	g_object_unref(device_proxy);

	/* IPSP daemon launch */
	GDBusProxy *ipsp_proxy;

	ipsp_proxy = _bt_get_ipsp_proxy();
	retv_if(ipsp_proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	g_dbus_proxy_call(ipsp_proxy, "EnableIpsp",
			NULL, G_DBUS_CALL_FLAGS_NONE,
			-1, NULL, NULL, NULL);

	return ret;
}

int _bt_disconnect_le_ipsp_device(const bluetooth_device_address_t *bd_addr)
{
	char device_address[BT_ADDRESS_STRING_SIZE] = { 0 };
	gchar *device_path = NULL;
	GError *error = NULL;
	GDBusProxy *device_proxy = NULL;
	GDBusProxy *adapter_proxy;
	GDBusConnection *conn;
	int ret = BLUETOOTH_ERROR_NONE;

	BT_CHECK_PARAMETER(bd_addr, return);

	_bt_convert_addr_type_to_string(device_address,
			(unsigned char *)bd_addr->addr);

	conn = _bt_gdbus_get_system_gconn();
	retv_if(conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	adapter_proxy = _bt_get_adapter_proxy();
	retv_if(adapter_proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	device_path = _bt_get_device_object_path(device_address);
	if (device_path == NULL) {
		BT_DBG("device_path NULL");
		ret = BLUETOOTH_ERROR_INTERNAL;
		return ret;
	}

	retv_if(device_path == NULL, BLUETOOTH_ERROR_INTERNAL);

	device_proxy = g_dbus_proxy_new_sync(conn, G_DBUS_PROXY_FLAGS_NONE,
							NULL, BT_BLUEZ_NAME,
							device_path, BT_DEVICE_INTERFACE,  NULL, NULL);
	g_free(device_path);
	retv_if(device_proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	g_dbus_proxy_call_sync(device_proxy, "DisconnectIpsp",
				NULL,
				G_DBUS_CALL_FLAGS_NONE,
				-1,
				NULL,
				&error);
	if (error) {
		BT_ERR("DisconnectIpsp Call Error %s[%s]", error->message, device_address);
		g_error_free(error);
		g_object_unref(device_proxy);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	g_object_unref(device_proxy);

	return ret;
}

int _bt_initialize_ipsp(void)
{
	BT_DBG("+");
	GDBusProxy *proxy;
	GError *error = NULL;
	GVariant *ret;

	if (_bt_adapter_get_status() != BT_ACTIVATED &&
		_bt_adapter_get_le_status() != BT_LE_ACTIVATED) {
		return BLUETOOTH_ERROR_DEVICE_NOT_ENABLED;
	}

	proxy = _bt_get_adapter_proxy();
	retv_if(proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	ret = g_dbus_proxy_call_sync(proxy, "InitializeIpsp",
				NULL, G_DBUS_CALL_FLAGS_NONE,
				-1, NULL, &error);
	if (error) {
		BT_ERR("Initialize IPSP Failed :[%s]", error->message);
		g_clear_error(&error);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	if (ret)
		g_variant_unref(ret);

	BT_INFO("IPSP initialization called successfully");

	return BLUETOOTH_ERROR_NONE;
}

int _bt_deinitialize_ipsp(void)
{
	BT_DBG("+");
	GDBusProxy *proxy;
	GError *error = NULL;
	GVariant *ret;

	if (_bt_adapter_get_status() != BT_ACTIVATED &&
		_bt_adapter_get_le_status() != BT_LE_ACTIVATED) {
		return BLUETOOTH_ERROR_DEVICE_NOT_ENABLED;
	}

	proxy = _bt_get_adapter_proxy();
	retv_if(proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	ret = g_dbus_proxy_call_sync(proxy, "DeinitializeIpsp",
				NULL, G_DBUS_CALL_FLAGS_NONE,
				-1, NULL, &error);
	if (error) {
		BT_ERR("De-Initialize IPSP Failed :[%s]", error->message);
		g_clear_error(&error);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	if (ret)
		g_variant_unref(ret);

	BT_INFO("IPSP De-initialization called successfully");

	return BLUETOOTH_ERROR_NONE;
}




