/*
 * Copyright (c) 2011 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <glib.h>
#include <string.h>
#include <dlog.h>
#include <vconf.h>
#include <vconf-internal-bt-keys.h>

#include "bluetooth-api.h"
#include "bt-internal-types.h"

#include "bt-service-common.h"
#include "bt-service-event.h"
#include "bt-service-main.h"
#include "bt-service-core-device.h"
#include "bt-service-obex-server.h"
#include "bt-service-obex-agent.h"
#include "bt-service-pbap.h"
#include "bt-service-opp-client.h"
#include "bt-service-map-client.h"
#include "bt-service-core-adapter.h"

#ifdef TIZEN_FEATURE_BT_DPM
#include "bt-service-dpm.h"
#endif

#define DBUS_TIMEOUT 20 * 1000 /* 20 Sec */
static GDBusConnection *manager_conn;
static GDBusConnection *obexd_conn;
static GDBusConnection *opc_obexd_conn;
static GDBusConnection *map_obexd_conn;

static guint event_id;

static guint session_reinit_timer;

/**
 * obexd connection type
 */
typedef enum {
	OBEX_OPP = (1 << 1),
	OBEX_FTP = (1 << 2),
	OBEX_BIP = (1 << 3),
	OBEX_PBAP = (1 << 4),
	OBEX_IRMC = (1 << 5),
	OBEX_PCSUITE = (1 << 6),
	OBEX_SYNCEVOLUTION = (1 << 7),
	OBEX_MAS = (1 << 8),
	OBEX_MAP = (1 << 9),
} bluetooth_obex_connection_type_t;


void _bt_handle_property_changed_event(GVariant *msg, const char *object_path);
void _bt_opc_property_changed_event(GVariant *msg, char *path);
void _bt_map_property_changed_event(GVariant *msg, const char *path);
int _bt_register_service_event(GDBusConnection *g_conn, int event_type);
void _bt_unregister_service_event(GDBusConnection *g_conn, int event_type);
void _bt_opp_client_event_deinit(void);
void _bt_map_client_event_deinit(void);
void _bt_map_on_transfer_finished(const char *transfer_object_path, const int error);

/* Temp Adapter changes required to make OBEX work for handling device events here ,
   This code will be removed and moved to OAL Event Handling part .
*/

#if 0
static void __bt_get_uuids(GVariant *value, bt_remote_dev_info_t *info)
{
	ret_if(value == NULL);
	ret_if(info == NULL);

	gsize uuid_count = 0;

	info->uuids = g_variant_dup_strv(value, &uuid_count);
	info->uuid_count = (unsigned int)uuid_count;

	BT_DBG("uuid count : %d", uuid_count);
}
#endif

void _bt_convert_addr_string_to_secure_string(char *addr,
					const char *address)
{
	int len;

	ret_if(address == NULL);
	ret_if(addr == NULL);

	len = strlen(address);
	ret_if(len != BT_ADDRESS_STRING_SIZE - 1);

	g_strlcpy(addr, address, BT_ADDRESS_STRING_SIZE);

	addr[len-1] = 'X';
	addr[len-2] = 'X';

	return;
}

bt_status_t _bt_adapter_get_status_for_Obex(void)
{
	int value = VCONFKEY_BT_STATUS_OFF;

	/* check VCONFKEY_BT_STATUS */
	if (vconf_get_int(VCONFKEY_BT_STATUS, &value) != 0) {
		BT_ERR("fail to get vconf key!");
		return BLUETOOTH_ADAPTER_DISABLED;
	}

	return value;
}

static int __bt_get_owner_info(GVariant *msg, char **name,
				char **previous, char **current)
{
	g_variant_get(msg, "(sss)", name, previous, current);
	return BLUETOOTH_ERROR_NONE;
}

static int __bt_get_agent_signal_info(GVariant *msg, char **address,
				char **name, char **uuid)
{
	g_variant_get(msg, "(sss)", address, name, uuid);
	return BLUETOOTH_ERROR_NONE;
}

gboolean __bt_handle_is_flight_mode_enabled(void)
{
	if (TIZEN_FEATURE_FLIGHTMODE_ENABLED) {
		int is_flight_mode = 0;
		int ret = -1;
		ret = vconf_get_bool(VCONFKEY_TELEPHONY_FLIGHT_MODE, &is_flight_mode);
		if (ret != 0)
			BT_ERR("vconf_get_bool failed");

		return (is_flight_mode == 0) ? FALSE : TRUE;
	} else {
		return FALSE;
	}
}

void _bt_handle_adapter_event(GVariant *msg, const char *member)
{
	BT_DBG("+");
	ret_if(member == NULL);

	if (strcasecmp(member, "DeviceCreated") == 0) {
		char *object_path = NULL;
		char *address;
		ret_if(_bt_is_device_creating() == FALSE);

		/* Bonding from remote device */
		address = g_malloc0(BT_ADDRESS_STRING_SIZE);

		g_variant_get(msg, "(&o)", &object_path);
		_bt_convert_device_path_to_address((const char*)object_path, address);

		g_free(address);
	} else if (strcasecmp(member, BT_HARDWARE_ERROR) == 0) {
		BT_ERR_C("### Hardware error received from BLUEZ");
	} else if (strcasecmp(member, BT_TX_TIMEOUT_ERROR) == 0) {
		BT_ERR_C("### Tx timeout error received from BLUEZ");
	}
	BT_DBG("-");
}

static void __bt_obex_property_changed_event(GVariant *msg, const char *path)
{
	BT_DBG("+");

	GVariantIter value_iter;
	GVariant *child = NULL, *val = NULL;
	char *property = NULL;
	g_variant_iter_init(&value_iter, msg);
	while ((child = g_variant_iter_next_value(&value_iter))) {
		g_variant_get(child, "{sv}", &property, &val);

		ret_if(property == NULL);

		BT_DBG("property :%s", property);

		if (strcasecmp(property, "Status") == 0) {
			char  *status;
			g_variant_get(val, "s", &status);

			if (strcasecmp(status, "active") == 0) {
				_bt_obex_transfer_started(path);
			} else if (strcasecmp(status, "complete") == 0) {
				_bt_obex_transfer_completed(path, TRUE);
				_bt_pbap_obex_transfer_completed(path, TRUE);
			} else if (strcasecmp(status, "error") == 0) {
				_bt_obex_transfer_completed(path, FALSE);
				_bt_pbap_obex_transfer_completed(path, FALSE);
			}
			g_free(status);
		} else if (strcasecmp(property, "Transferred") == 0) {
			guint64 transferred  = 0;
			/* As Transferred is expected guint64 so change int to guint64 and
			 * eariler transferred is static because of it can overwrite data
			 * on present on opc_obex_conn or obexd_conn as these are
			 * static memory are in sequential */
			g_variant_get(val, "t", &transferred);

			_bt_obex_transfer_progress(path, transferred);
		}
		/* TODO: MAP, "Complete"? see above */
		g_free(property);
		g_variant_unref(val);
		g_variant_unref(child);
	}
	BT_DBG("-");
}

void _bt_handle_property_changed_event(GVariant *msg, const char *object_path)
{
	char *interface_name = NULL;
	GVariant *val = NULL;

	g_variant_get(msg, "(&s@a{sv}@as)", &interface_name, &val, NULL);
	BT_DBG("_bt_handle_property_changed_event");

	if (strcasecmp(interface_name, BT_ADAPTER_INTERFACE) == 0) {
		gboolean ret = FALSE;
		gboolean powered = FALSE;

		/* Check if Adapter is disabled, cancle queued transfers if any */
		ret = g_variant_lookup(msg, "Powered", "b", &powered);
		if (ret && !powered)
			_bt_cancel_queued_transfers();
	} else if (strcasecmp(interface_name, BT_OBEX_TRANSFER_INTERFACE) == 0) {
		BT_DBG("BT_OBEX_TRANSFER_INTERFACE");
		__bt_obex_property_changed_event(val,
					object_path);
	}
	g_variant_unref(val);
}

void __bt_opc_property_changed_event(GVariant *msg,
						const char *path)
{
	GVariantIter value_iter;
	char *property = NULL;
	GVariant *val = NULL;
	GVariant *child = NULL;

	g_variant_iter_init(&value_iter, msg);
	if ((child = g_variant_iter_next_value(&value_iter))) {
		g_variant_get(child, "{sv}", &property, &val);
		ret_if(property == NULL);

		if (strcasecmp(property, "Status") == 0) {
			char *status = NULL;
			g_variant_get(val, "s", &status);
			BT_DBG("Status is %s", status);

			if (strcasecmp(status, "active") == 0)
				_bt_obex_client_started(path);
			else if (strcasecmp(status, "complete") == 0)
				_bt_obex_client_completed(path, TRUE);
			else if (strcasecmp(status, "error") == 0)
				_bt_obex_client_completed(path, FALSE);

			g_free(status);
		} else if (strcasecmp(property, "Transferred") == 0) {
			guint64 transferred  = 0;
			g_variant_get(val, "t", &transferred);

			_bt_obex_client_progress(path, transferred);
		} else {
			BT_DBG("property : [%s]", property);
		}
		g_free(property);
		g_variant_unref(child);
		g_variant_unref(val);
	}
}

void _bt_opc_property_changed_event(GVariant *msg, char *path)
{
	char *interface_name = NULL;
	GVariant *value = NULL;
	g_variant_get(msg, "(&s@a{sv}@as)", &interface_name, &value, NULL);
	BT_INFO("interface_name = %s", interface_name);
	if (strcasecmp(interface_name, BT_OBEX_TRANSFER_INTERFACE) == 0) {
		__bt_opc_property_changed_event(value,
					path);
	} else {
		BT_DBG("interface_name : [%s]", interface_name);
	}
	g_variant_unref(value);
}


void __bt_map_property_changed_event(GVariant *msg,
						const char *path)
{
	BT_DBG("Entered");
	GVariantIter value_iter;
	char *property = NULL;
	GVariant *val = NULL;
	GVariant *child = NULL;

	g_variant_iter_init(&value_iter, msg);
	while ((child = g_variant_iter_next_value(&value_iter))) {
		g_variant_get(child, "{sv}", &property, &val);
		ret_if(property == NULL);

		if (strcasecmp(property, "Status") == 0) {
			char *status = NULL;
			g_variant_get(val, "s", &status);
			BT_DBG("Status is %s", status);

			if (strcasecmp(status, "active") == 0) {
				BT_DBG("EVENT : STARTED");
				// currently doing nothing
			} else if (strcasecmp(status, "complete") == 0) {
				BT_DBG("EVENT : COMPLETED");
				_bt_map_on_transfer_finished(path, BLUETOOTH_ERROR_NONE);
			} else if (strcasecmp(status, "error") == 0) {
				BT_DBG("EVENT : FAILED");
				_bt_map_on_transfer_finished(path, BLUETOOTH_ERROR_INTERNAL);
			}
			g_free(status);
		} else if (strcasecmp(property, "Transferred") == 0) {
			guint64 transferred  = 0;
			g_variant_get(val, "t", &transferred);

			BT_DBG("EVENT : PROGRESS CALLBACK");
			// currently doing nothing - progress callback type is not used
		} else {
			BT_DBG("OTHER EVENT : property : [%s]", property);
		}
		g_free(property);
		g_variant_unref(child);
		g_variant_unref(val);
	}
}

void _bt_map_property_changed_event(GVariant *msg, const char *path)
{
	BT_DBG("Entered _bt_map_property_changed_event");
	char *interface_name = NULL;
	GVariant *value = NULL;
	g_variant_get(msg, "(&s@a{sv}@as)", &interface_name, &value, NULL);
	BT_INFO("interface_name = %s", interface_name);
	if (strcasecmp(interface_name, BT_OBEX_TRANSFER_INTERFACE) == 0) {
		__bt_map_property_changed_event(value,
				path);
	} else {
		BT_DBG("interface_name : [%s]", interface_name);
	}
	g_variant_unref(value);
}


void _bt_handle_agent_event(GVariant *msg, const char *member)
{
	int result = BLUETOOTH_ERROR_NONE;
	char *address = NULL;
	char *name = NULL;
	char *uuid = NULL;
	GVariant *param = NULL;
	ret_if(member == NULL);

	if (strcasecmp(member, "ObexAuthorize") == 0) {
		__bt_get_agent_signal_info(msg, &address, &name, &uuid);
		param = g_variant_new("(iss)", result, address, name);
		_bt_send_event(BT_OPP_SERVER_EVENT,
			BLUETOOTH_EVENT_OBEX_SERVER_CONNECTION_AUTHORIZE,
			param);
		/* TODO: MAP? see above */
		g_free(address);
		g_free(name);
	}
}

static int __bt_get_object_path(GVariant *msg, char **path)
{
	g_variant_get(msg, "(o*)", path, NULL);
	if (*path == NULL)
		return BLUETOOTH_ERROR_INTERNAL;

	return BLUETOOTH_ERROR_NONE;
}

char *_bt_get_bonded_device_name(char *address)
{
	bt_remote_dev_info_t *dev_info;

	retv_if(address == NULL, strdup(""));

	dev_info = _bt_service_get_remote_dev_info(address);
	retv_if(dev_info == NULL, strdup(""));
	retv_if(dev_info->name == NULL, strdup(""));

	return g_strdup(dev_info->name);
}
/* Temp Adapter Util changes to make OBEX work.
*/

void _bt_handle_device_event(GVariant *msg, const char *member, const char *path)
{
	char *address;
	char secure_address[BT_ADDRESS_STRING_SIZE] = { 0 };
	ret_if(path == NULL);

	if (strcasecmp(member, "Disconnected") == 0) {
		unsigned char disc_reason = 0;
		unsigned char addr_type = 0;
		char *dev_name = NULL;
		gboolean sending = FALSE;

		g_variant_get(msg, "(yys)", &addr_type, &disc_reason, &dev_name);

		address = g_malloc0(BT_ADDRESS_STRING_SIZE);

		_bt_convert_device_path_to_address(path, address);

		/* 0x00 BDADDR_BRDER
		      0x01 BDADDR_LE_PUBLIC
		      0x02 BDADDR_LE_RANDOM */
		_bt_convert_addr_string_to_secure_string(secure_address, address);
		BT_INFO("Address : %s Type : %d", secure_address, addr_type);
		BT_ERR_C("### Disconnected [%s] [%d : %s] [%s]", !addr_type ? "BREDR" : "LE",
				disc_reason, _bt_convert_disc_reason_to_string(disc_reason),
				!addr_type ? dev_name : secure_address);
		g_free(dev_name);

		if (!addr_type) {
			/*Check for any OPP transfer on the device and cancel
			 * the transfer
			 */
			_bt_obex_check_pending_transfer(address);
			_bt_opp_client_is_sending(&sending);
			if (sending == TRUE)
				_bt_opp_client_check_pending_transfer(address);
			/* TODO: MAP? see above */
		}
		g_free(address);
	} else if (strcasecmp(member, "ProfileStateChanged") == 0) {
		int state = 0;
		char *profile_uuid = NULL;
		bluetooth_device_address_t bd_addr;

		g_variant_get(msg, "(si)", &profile_uuid, &state);

		address = g_malloc0(BT_ADDRESS_STRING_SIZE);

		_bt_convert_device_path_to_address(path, address);
		_bt_convert_addr_string_to_type(bd_addr.addr, address);

		_bt_convert_addr_string_to_secure_string(secure_address, address);
		BT_DBG("Address: %s", secure_address);
		BT_DBG("Profile UUID: %s", profile_uuid);
		BT_DBG("State: %d", state);
		g_free(address);
	}
}

static  void __bt_manager_event_filter(GDBusConnection *connection,
					const gchar *sender_name,
					const gchar *object_path,
					const gchar *interface_name,
					const gchar *signal_name,
					GVariant *parameters,
					gpointer user_data)
{

	GVariant *value;
	char *obj_path = NULL;

	if (signal_name == NULL)
		return;

	if (strcasecmp(signal_name, "InterfacesAdded") == 0) {
		g_variant_get(parameters, "(&o@a{sa{sv}})", &obj_path, &value);

		if (strcasecmp(obj_path, BT_BLUEZ_HCI_PATH) == 0)
			if (!TIZEN_FEATURE_BT_USB_DONGLE) {
				if (_bt_register_obex_server() != BLUETOOTH_ERROR_NONE)
					BT_ERR("Fail to init obex server");
			}
		g_variant_unref(value);
	} else if (strcasecmp(signal_name, "InterfacesRemoved") == 0) {
			if (TIZEN_FEATURE_BT_USB_DONGLE) {
				if (_bt_unregister_obex_server() != BLUETOOTH_ERROR_NONE)
					BT_ERR("Fail to unregister obex server");
			}
	} else if (strcasecmp(signal_name, "NameOwnerChanged") == 0) {
		gboolean value;
		char *name = NULL;
		char *previous = NULL;
		char *current = NULL;

		if (__bt_get_owner_info(parameters, &name, &previous, &current)) {
			BT_ERR("Fail to get the owner info");
			return;
		}

		if (*current != '\0') {
			g_free(current);
			if (name)
				g_free(name);
			if (previous)
				g_free(previous);
			return;
		}

		if (strcasecmp(name, BT_BLUEZ_NAME) == 0) {
			BT_INFO_C("### Bluetoothd is terminated");

			if (_bt_unregister_obex_server() != BLUETOOTH_ERROR_NONE)
				BT_ERR("Fail to unregister obex server");

		}

		_bt_obex_server_check_allocation(&value);

		if (value == TRUE) {
			/* Check if the obex server was terminated abnormally */
			_bt_obex_server_check_termination(name);
		}

		g_free(name);
		g_free(previous);
		g_free(current);
	} else if (g_strcmp0(interface_name, BT_AGENT_INTERFACE) == 0) {
		_bt_handle_agent_event(parameters, signal_name);
	} else if (g_strcmp0(interface_name, BT_DEVICE_INTERFACE) == 0) {
		_bt_handle_device_event(parameters, signal_name, object_path);
	}

	return;
}

static gboolean __bt_is_obexd_event(GVariant *msg, const char *interface)
{

	if (g_strcmp0(interface, BT_PROPERTIES_INTERFACE) == 0) {
		char *interface_name = NULL;

		g_variant_get(msg, "(&s@a{sv}@as)", &interface_name, NULL, NULL);
		retv_if(interface_name == NULL, FALSE);

		if (strcasecmp(interface_name, BT_OBEX_TRANSFER_INTERFACE) == 0) {
			BT_DBG("BT_OBEX_TRANSFER_INTERFACE");
			return TRUE;
		}
	}

	return FALSE;
}

static  void __bt_obexd_event_filter(GDBusConnection *connection,
					const gchar *sender_name,
					const gchar *object_path,
					const gchar *interface_name,
					const gchar *signal_name,
					GVariant *parameters,
					gpointer user_data)
{
	const char *member = signal_name;
	char *obj_path = NULL;
	ret_if(member == NULL);

	if (strcasecmp(member, "InterfacesAdded") == 0) {
		if (__bt_get_object_path(parameters, &obj_path)) {
			BT_ERR("Fail to get the path");
			return;
		}
		BT_INFO("object_path = [%s]", obj_path);

		/*Handle OPP_SERVER_CONNECTED_EVENT here */
		/* TODO: MAP? see above */
		if (strncmp(obj_path, BT_SESSION_BASEPATH_SERVER,
				strlen(BT_SESSION_BASEPATH_SERVER)) != 0) {
			g_free(obj_path);
			return;
		}

		if (g_strrstr(obj_path, "session") && g_strrstr(obj_path, "transfer")) {
			BT_DBG("Obex_Server_Session_Transfer connected");
			_bt_obex_transfer_connected(obj_path);
		}
		g_free(obj_path);
	} else if (strcasecmp(member, "InterfacesRemoved") == 0) {
		/*Handle OPP_SERVER_DISCONNECTED_EVENT here */
		/* TODO: MAP? see above */
		if (__bt_get_object_path(parameters, &obj_path)) {
			BT_ERR("Fail to get the path");
			return;
		}
		BT_INFO("object_path = [%s]", obj_path);

		if (strncmp(obj_path, BT_SESSION_BASEPATH_CLIENT,
				strlen(BT_SESSION_BASEPATH_CLIENT)) == 0) {
			BT_DBG("Call PBAP Disconnected");
			_bt_obex_pbap_client_disconnect(obj_path);
		}

		if (strncmp(obj_path, BT_SESSION_BASEPATH_SERVER,
				strlen(BT_SESSION_BASEPATH_SERVER)) != 0) {
			g_free(obj_path);
			return;
		}

		if (g_strrstr(obj_path, "session") && g_strrstr(obj_path, "transfer")) {
			BT_DBG("Obex_Server_Session_Transfer disconnected %s",
								obj_path);

			_bt_obex_transfer_disconnected(obj_path);
		}
		g_free(obj_path);
	} else if (__bt_is_obexd_event(parameters, interface_name) == TRUE) {
		const char *path = object_path;

		if (strncmp(path, BT_SESSION_BASEPATH_SERVER,
				strlen(BT_SESSION_BASEPATH_SERVER)) != 0 &&
			strncmp(path, BT_SESSION_BASEPATH_CLIENT,
				strlen(BT_SESSION_BASEPATH_CLIENT)) != 0) {
			BT_DBG("DBUS_HANDLER_RESULT_NOT_YET_HANDLED");
			return;
		}

		_bt_handle_property_changed_event(parameters, path);
	}
	return;
}

static gboolean __bt_is_obexd_client_event(GVariant *msg, const char *interface)
{
	BT_DBG("+");

	if (g_strcmp0(interface, BT_PROPERTIES_INTERFACE) == 0) {
		char *interface_name = NULL;

		g_variant_get(msg, "(&s@a{sv}@as)", &interface_name, NULL, NULL);

		retv_if(interface_name == NULL, FALSE);

		if (strcasecmp(interface_name,
					BT_OBEX_TRANSFER_INTERFACE) == 0) {
			BT_DBG("-");
			return TRUE;
		}
	}

	BT_DBG("-");

	return FALSE;
}

static  void __bt_opc_event_filter(GDBusConnection *connection,
					const gchar *sender_name,
					const gchar *object_path,
					const gchar *interface_name,
					const gchar *signal_name,
					GVariant *parameters,
					gpointer user_data)
{
	const char *member = signal_name;
	char *obj_path = NULL;
	if (strcasecmp(member, "InterfacesAdded") == 0) {
		BT_DBG("InterfacesAdded");
	} else if (strcasecmp(member, "InterfacesRemoved") == 0) {

		if (__bt_get_object_path(parameters, &obj_path)) {
			BT_ERR("Fail to get the path");
			return;
		}

		BT_DBG("object_path = %s", obj_path);

		if (strncmp(obj_path, BT_SESSION_BASEPATH_CLIENT,
				strlen(BT_SESSION_BASEPATH_CLIENT)) != 0
				|| strstr(obj_path, "transfer") == NULL) {
			g_free(obj_path);
			return;
		} else if (strncmp(obj_path, BT_SESSION_BASEPATH_CLIENT,
				strlen(BT_SESSION_BASEPATH_CLIENT)) == 0) {
			BT_DBG("Going to call opc disconnected");
			_bt_opc_disconnected(obj_path);
		}

		_bt_sending_files();
		g_free(obj_path);
	} else if (__bt_is_obexd_client_event(parameters, interface_name) == TRUE) {
		char *path = (char *)object_path;
		BT_INFO("object_path %s", path);
		if (strncmp(path, BT_SESSION_BASEPATH_CLIENT,
			strlen(BT_SESSION_BASEPATH_CLIENT)) != 0) {
			BT_DBG("NOT BT_SESSION_BASEPATH_CLIENT");
			return;
		}

		_bt_opc_property_changed_event(parameters, path);
	}

	return;
}

int _bt_opp_client_event_init(void)
{
	GError *error = NULL;

	if (opc_obexd_conn == NULL) {
		opc_obexd_conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);

		if (!opc_obexd_conn) {
			if (error) {
				BT_ERR("Unable to connect to dbus: %s", error->message);
				g_clear_error(&error);
			}
		return BLUETOOTH_ERROR_INTERNAL;
		}
	}

	if (_bt_register_service_event(opc_obexd_conn,
			BT_OPP_CLIENT_EVENT) != BLUETOOTH_ERROR_NONE) {
			g_object_unref(opc_obexd_conn);
			opc_obexd_conn = NULL;
			return BLUETOOTH_ERROR_INTERNAL;
	}

	return BLUETOOTH_ERROR_NONE;
}

void _bt_opp_client_event_deinit(void)
{
	if (opc_obexd_conn) {
		_bt_unregister_service_event(opc_obexd_conn,
						BT_OPP_CLIENT_EVENT);
		 g_object_unref(opc_obexd_conn);
		 opc_obexd_conn = NULL;
	}
}

static  void __bt_map_event_filter(GDBusConnection *connection,
					const gchar *sender_name,
					const gchar *object_path,
					const gchar *interface_name,
					const gchar *signal_name,
					GVariant *parameters,
					gpointer user_data)
{
	BT_DBG("Entered __bt_map_event_filter");
	const char *member = signal_name;

	if (strcasecmp(member, "InterfacesAdded") == 0) {
		BT_DBG("------------------------------------ADDED------------------------------------");
		// currently doing nothing
	} else if (strcasecmp(member, "InterfacesRemoved") == 0) {
		BT_DBG("------------------------------------REMOVED------------------------------------");
		// TODO check if something should be called here?
		//_bt_map_on_transfer_finished(object_path, error);
	} else if (__bt_is_obexd_client_event(parameters, interface_name) == TRUE) {
		BT_DBG("------------------------------------CLIENT EVENT------------------------------------");
		_bt_map_property_changed_event(parameters, object_path);
	}

	return;
}

int _bt_map_client_event_init(void)
{
	GError *error = NULL;

	if (map_obexd_conn == NULL) {
		map_obexd_conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);

		if (!map_obexd_conn) {
			if (error) {
				BT_ERR("Unable to connect to dbus: %s", error->message);
				g_clear_error(&error);
			}
		return BLUETOOTH_ERROR_INTERNAL;
		}
	}

	if (_bt_register_service_event(map_obexd_conn,
			BT_MAP_CLIENT_EVENT) != BLUETOOTH_ERROR_NONE) {
			g_object_unref(map_obexd_conn);
			map_obexd_conn = NULL;
			return BLUETOOTH_ERROR_INTERNAL;
	}

	return BLUETOOTH_ERROR_NONE;
}

void _bt_map_client_event_deinit(void)
{
	if (map_obexd_conn) {
		_bt_unregister_service_event(map_obexd_conn,
						BT_MAP_CLIENT_EVENT);
		 g_object_unref(map_obexd_conn);
		 map_obexd_conn = NULL;
	}
}

int _bt_register_manager_subscribe_signal(GDBusConnection *conn,
		int subscribe)
{
	if (conn == NULL)
		return -1;

	static int subs_interface_added_id = -1;
	static int subs_interface_removed_id = -1;
	static int subs_name_owner_changed = -1;

	if (subscribe) {
		if (subs_interface_added_id == -1) {
			subs_interface_added_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_MANAGER_INTERFACE,
				BT_INTERFACES_ADDED, NULL, NULL, 0,
				__bt_manager_event_filter,
				NULL, NULL);
		}
		if (subs_interface_removed_id == -1) {
			subs_interface_removed_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_MANAGER_INTERFACE,
				BT_INTERFACES_REMOVED, NULL, NULL, 0,
				__bt_manager_event_filter,
				NULL, NULL);
		}
		if (subs_name_owner_changed == -1) {
			subs_name_owner_changed = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_EVENT_FREEDESKTOP,
				BT_NAME_OWNER_CHANGED, NULL, NULL, 0,
				__bt_manager_event_filter, NULL, NULL);
		}

	} else {
		if (subs_interface_added_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_interface_added_id);
			subs_interface_added_id = -1;
		}
		if (subs_interface_removed_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_interface_removed_id);
			subs_interface_removed_id = -1;
		}
		if (subs_name_owner_changed != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_name_owner_changed);
			subs_name_owner_changed = -1;
		}

	}
	return 0;
}

int _bt_register_device_subscribe_signal(GDBusConnection *conn,
		int subscribe)
{
	if (conn == NULL)
		return -1;

	static int subs_device_id = -1;

	if (subscribe) {
		if (subs_device_id == -1) {
			subs_device_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_DEVICE_INTERFACE,
				NULL, NULL, NULL, 0,
				__bt_manager_event_filter,
				NULL, NULL);
		}
	} else {
		if (subs_device_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_device_id);
			subs_device_id = -1;
		}
	}
	return 0;
}

int _bt_register_input_subscribe_signal(GDBusConnection *conn,
		int subscribe)
{
	if (conn == NULL)
		return -1;

	static int subs_input_id = -1;

	if (subscribe) {
		if (subs_input_id == -1) {
			subs_input_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_INPUT_INTERFACE,
				NULL, NULL, NULL, 0,
				__bt_manager_event_filter,
				NULL, NULL);
		}
	} else {
		if (subs_input_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_input_id);
			subs_input_id = -1;
		}
	}
	return 0;
}



int _bt_register_opp_server_subscribe_signal(GDBusConnection *conn,
		int subscribe)
{
	if (conn == NULL)
		return -1;

	static int subs_opp_server_interface_added_id = -1;
	static int subs_opp_server_interface_removed_id = -1;
	static int subs_opp_server_property_id = -1;


	if (subscribe) {
		if (subs_opp_server_interface_added_id == -1) {
			subs_opp_server_interface_added_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_MANAGER_INTERFACE,
				BT_INTERFACES_ADDED, NULL, NULL, 0,
				__bt_obexd_event_filter,
				NULL, NULL);
		}
		if (subs_opp_server_interface_removed_id == -1) {
			subs_opp_server_interface_removed_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_MANAGER_INTERFACE,
				BT_INTERFACES_REMOVED, NULL, NULL, 0,
				__bt_obexd_event_filter,
				NULL, NULL);
		}
		if (subs_opp_server_property_id == -1) {
			subs_opp_server_property_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_PROPERTIES_INTERFACE,
				BT_PROPERTIES_CHANGED, NULL, NULL, 0,
				__bt_obexd_event_filter,
				NULL, NULL);
		}
	} else {
		if (subs_opp_server_interface_added_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_opp_server_interface_added_id);
			subs_opp_server_interface_added_id = -1;
		}
		if (subs_opp_server_interface_removed_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_opp_server_interface_removed_id);
			subs_opp_server_interface_removed_id = -1;
		}
		if (subs_opp_server_property_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_opp_server_property_id);
			subs_opp_server_property_id = -1;
		}
	}
	return 0;
}

int _bt_register_opp_client_subscribe_signal(GDBusConnection *conn,
		int subscribe)
{
	if (conn == NULL)
		return -1;

	static int subs_opp_client_interface_added_id = -1;
	static int subs_opp_client_interface_removed_id = -1;
	static int subs_opp_client_property_id = -1;


	if (subscribe) {
		if (subs_opp_client_interface_added_id == -1) {
			subs_opp_client_interface_added_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_MANAGER_INTERFACE,
				BT_INTERFACES_ADDED, NULL, NULL, 0,
				__bt_opc_event_filter,
				NULL, NULL);
		}
		if (subs_opp_client_interface_removed_id == -1) {
			subs_opp_client_interface_removed_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_MANAGER_INTERFACE,
				BT_INTERFACES_REMOVED, NULL, NULL, 0,
				__bt_opc_event_filter,
				NULL, NULL);
		}
		if (subs_opp_client_property_id == -1) {
			subs_opp_client_property_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_PROPERTIES_INTERFACE,
				BT_PROPERTIES_CHANGED, NULL, NULL, 0,
				__bt_opc_event_filter,
				NULL, NULL);
		}
	} else {
		if (subs_opp_client_interface_added_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_opp_client_interface_added_id);
			subs_opp_client_interface_added_id = -1;
		}
		if (subs_opp_client_interface_removed_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_opp_client_interface_removed_id);
			subs_opp_client_interface_removed_id = -1;
		}
		if (subs_opp_client_property_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_opp_client_property_id);
			subs_opp_client_property_id = -1;
		}
	}
	return 0;
}

int _bt_register_map_client_subscribe_signal(GDBusConnection *conn,
		int subscribe)
{
	if (conn == NULL)
		return -1;

	static int subs_map_client_interface_added_id = -1;
	static int subs_map_client_interface_removed_id = -1;
	static int subs_map_client_property_id = -1;


	if (subscribe) {
		if (subs_map_client_interface_added_id == -1) {
			subs_map_client_interface_added_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_MANAGER_INTERFACE,
				BT_INTERFACES_ADDED, NULL, NULL, 0,
				__bt_map_event_filter,
				NULL, NULL);
		}
		if (subs_map_client_interface_removed_id == -1) {
			subs_map_client_interface_removed_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_MANAGER_INTERFACE,
				BT_INTERFACES_REMOVED, NULL, NULL, 0,
				__bt_map_event_filter,
				NULL, NULL);
		}
		if (subs_map_client_property_id == -1) {
			subs_map_client_property_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_PROPERTIES_INTERFACE,
				BT_PROPERTIES_CHANGED, NULL, NULL, 0,
				__bt_map_event_filter,
				NULL, NULL);
		}
	} else {
		if (subs_map_client_interface_added_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_map_client_interface_added_id);
			subs_map_client_interface_added_id = -1;
		}
		if (subs_map_client_interface_removed_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_map_client_interface_removed_id);
			subs_map_client_interface_removed_id = -1;
		}
		if (subs_map_client_property_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_map_client_property_id);
			subs_map_client_property_id = -1;
		}
	}
	return 0;
}


int _bt_register_service_event(GDBusConnection *g_conn, int event_type)
{
	BT_DBG("+");

	retv_if(g_conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	switch (event_type) {
	case BT_MANAGER_EVENT:
		BT_DBG("BT_MANAGER_EVENT: register service event");
		_bt_register_manager_subscribe_signal(g_conn, TRUE);
		break;
	case BT_DEVICE_EVENT:
		BT_DBG("BT_DEVICE_EVENT: register service event");
		_bt_register_device_subscribe_signal(g_conn, TRUE);
		break;
	case BT_OPP_SERVER_EVENT:
		BT_DBG("BT_OPP_SERVER_EVENT: register service event");
		_bt_register_opp_server_subscribe_signal(g_conn, TRUE);
		break;
	case BT_OPP_CLIENT_EVENT:
		BT_DBG("BT_OPP_CLIENT_EVENT: register service event");
		_bt_register_opp_client_subscribe_signal(g_conn, TRUE);
		break;
	case BT_MAP_CLIENT_EVENT:
		BT_DBG("BT_MAP_CLIENT_EVENT: register service event");
		_bt_register_map_client_subscribe_signal(g_conn, TRUE);
		break;
	default:
		BT_ERR("Unknown event");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	return BLUETOOTH_ERROR_NONE;
}

void _bt_unregister_service_event(GDBusConnection *g_conn, int event_type)
{
	BT_DBG("+");

	ret_if(g_conn == NULL);

	switch (event_type) {
	case BT_MANAGER_EVENT:
		_bt_register_manager_subscribe_signal(g_conn, FALSE);
		_bt_register_device_subscribe_signal(g_conn, FALSE);
		_bt_register_input_subscribe_signal(g_conn, FALSE);
		break;
	case BT_OPP_SERVER_EVENT:
		_bt_register_opp_server_subscribe_signal(g_conn, FALSE);
		break;
	case BT_OPP_CLIENT_EVENT:
		_bt_register_opp_client_subscribe_signal(g_conn, FALSE);
		break;
	case BT_MAP_CLIENT_EVENT:
		_bt_register_map_client_subscribe_signal(g_conn, FALSE);
		break;
	default:
		BT_ERR("Unknown event");
		return;
	}

	BT_DBG("-");
}

static int __bt_init_manager_receiver(void)
{
	BT_DBG("+");

	GError *error = NULL;

	if (manager_conn == NULL) {
		manager_conn =  g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);
		if (error != NULL) {
			BT_ERR("ERROR: Can't get on system bus [%s]", error->message);
			g_clear_error(&error);
		}
		retv_if(manager_conn == NULL, BLUETOOTH_ERROR_INTERNAL);
	}

	if (_bt_register_service_event(manager_conn,
				BT_MANAGER_EVENT) != BLUETOOTH_ERROR_NONE)
		goto fail;

	if (_bt_register_service_event(manager_conn,
				BT_DEVICE_EVENT) != BLUETOOTH_ERROR_NONE)
		goto fail;

	return BLUETOOTH_ERROR_NONE;
fail:
	if (manager_conn) {
		g_object_unref(manager_conn);
		manager_conn = NULL;
	}

	BT_DBG("-");

	return BLUETOOTH_ERROR_INTERNAL;
}

static int __bt_init_obexd_receiver(void)
{
	BT_DBG("+");
	if (!TIZEN_PROFILE_TV) { /* TODO: obexd doesn't work in TV profile. It should be resolved later. */
		GError *error = NULL;

		if (obexd_conn == NULL) {
			obexd_conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);
			if (error != NULL) {
				BT_ERR("ERROR: Can't get on session bus [%s]", error->message);
				g_clear_error(&error);
			}
			retv_if(obexd_conn == NULL, BLUETOOTH_ERROR_INTERNAL);
		}

		if (_bt_register_service_event(obexd_conn,
					BT_OPP_SERVER_EVENT) != BLUETOOTH_ERROR_NONE) {
			BT_ERR("Error while registering service event");
			g_object_unref(obexd_conn);
			obexd_conn = NULL;
			return BLUETOOTH_ERROR_INTERNAL;
		}
	}

	BT_DBG("-");

	return BLUETOOTH_ERROR_NONE;
}

gboolean __bt_reinit_obexd_receiver(gpointer user_data)
{
	static int retry_cnt = 0;
	int result = BLUETOOTH_ERROR_NONE;

	BT_DBG("+");

	result = __bt_init_obexd_receiver();
	if (result != BLUETOOTH_ERROR_NONE) {
		/* 20 ms * 50 = 10 seconds. During 10 seconds fail to initialize,
		   then it is not the timing issue. Just can't use the session bus connection */
		if (retry_cnt > 100) {
			BT_ERR("Fail to init obexd receiver by 50 times.");
			retry_cnt = 0;
			session_reinit_timer = 0;
			return FALSE;
		}
		retry_cnt++;
		BT_DBG("Retry to initialize the obexd receiver");
		return TRUE;
	}

	retry_cnt = 0;
	session_reinit_timer = 0;

	BT_DBG("-");

	return FALSE;
}

/* To receive the event from bluez */
int _bt_init_obex_event_receiver(void)
{
	BT_DBG("+");

	int result;

	result = __bt_init_manager_receiver();
	retv_if(result != BLUETOOTH_ERROR_NONE, result);

	result = __bt_init_obexd_receiver();
	if (result != BLUETOOTH_ERROR_NONE) {
		BT_ERR("Fail to init obexd receiver");

		/* Try to re-initialize obexd receiver in the timer */
		if (session_reinit_timer > 0)
			g_source_remove(session_reinit_timer);

		session_reinit_timer = g_timeout_add(200,
							(GSourceFunc)__bt_reinit_obexd_receiver, NULL);
	}

	BT_DBG("-");

	return BLUETOOTH_ERROR_NONE;
}

void _bt_deinit_obex_event_receiver(void)
{
	BT_DBG("+");

	_bt_unregister_service_event(manager_conn, BT_MANAGER_EVENT);

	_bt_unregister_service_event(obexd_conn, BT_OPP_SERVER_EVENT);

	if (manager_conn) {
		g_object_unref(manager_conn);
		manager_conn = NULL;
	}

	if (obexd_conn) {
		g_object_unref(obexd_conn);
		obexd_conn = NULL;
	}

	if (event_id > 0)
		g_source_remove(event_id);
	event_id = 0;

	BT_DBG("-");
}

