/*
 * Bluetooth-frwk
 *
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Author:  Atul Kumar Rai <a.rai@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <glib.h>
#include <gio/gio.h>
#include <gio/gunixfdlist.h>
#include <dlog.h>
#include <string.h>

#include "bluetooth-api.h"
#include "bt-internal-types.h"
#include "bt-request-handler.h"
#include "bt-service-util.h"
#include "bt-service-event.h"
#include "bt-service-common.h"
#include "bt-service-rfcomm.h"
#include "bt-service-socket.h"

static void __bt_rfcomm_reply_pending_request(int result,
		int service_function, void *user_data, unsigned int size)
{
	GSList *l;
	GArray *out_param;
	invocation_info_t *req_info;

	/* Get method invocation context */
	for (l = _bt_get_invocation_list(); l != NULL; l = g_slist_next(l)) {
		req_info = l->data;
		if (req_info == NULL || req_info->service_function != service_function)
			continue;

		/* Create out param */
		out_param = g_array_new(FALSE, FALSE, sizeof(gchar));

		switch (service_function) {
		case BT_RFCOMM_CLIENT_CONNECT: {
			GUnixFDList *fd_list = NULL;
			GError *error = NULL;
			char address[BT_ADDRESS_STRING_SIZE];
			bluetooth_rfcomm_connection_t *ptr = user_data;

			/* Check if connect address matched with requested address */
			_bt_convert_addr_type_to_string(address, ptr->device_addr.addr);
			if (strncasecmp(address, (char *)req_info->user_data,
						BT_ADDRESS_STRING_SIZE)) {
				BT_INFO("RFCOMM connected addr: [%s], requested addr: [%s]",
						address, (char *)req_info->user_data);
				break;
			}

			g_array_append_vals(out_param, user_data, size);

			if (BLUETOOTH_ERROR_NONE == result) {
				fd_list = g_unix_fd_list_new();
				g_unix_fd_list_append(fd_list, ptr->socket_fd, &error);
				g_assert_no_error(error);
				close(ptr->socket_fd);
			}

			_bt_service_method_return_with_unix_fd_list(
					req_info->context, out_param, result, fd_list);
			if (fd_list)
				g_object_unref(fd_list);

			g_free(req_info->user_data);
			_bt_free_info_from_invocation_list(req_info);
			g_array_free(out_param, TRUE);
			break;
		}
		case BT_RFCOMM_LISTEN_AND_ACCEPT:
		case BT_RFCOMM_LISTEN: {
			GUnixFDList *fd_list = NULL;
			GError *error = NULL;
			int *socket_fd = user_data;

			BT_INFO("Server socket fd: %d", *socket_fd);

			g_array_append_vals(out_param, user_data, size);

			/* Add socket fd to unix_fd_list */
			fd_list = g_unix_fd_list_new();
			g_unix_fd_list_append(fd_list, *socket_fd, &error);
			g_assert_no_error(error);

			_bt_service_method_return_with_unix_fd_list(
					req_info->context, out_param, result, fd_list);

			close(*socket_fd);
			g_object_unref(fd_list);

			_bt_free_info_from_invocation_list(req_info);
			g_array_free(out_param, TRUE);
			break;
		}
		default:
			BT_ERR("Unknown Service function");
		}
	}

	return;
}

static void __bt_rfcomm_socket_conn_cb(int result, int sock_fd, char *address, char *uuid, int chan)
{
	bluetooth_rfcomm_connection_t conn_info;

	ret_if(NULL == address);
	ret_if(NULL == uuid);

	BT_DBG("+");

	BT_INFO("result: %d, socket_fd: %d, address: %s, uuid: %s, chan: %d",
			result, sock_fd, address, uuid, chan);

	/* Fill RFCOMM connection structure and send reply to pending request */
	memset(&conn_info, 0x00, sizeof(bluetooth_rfcomm_connection_t));
	conn_info.socket_fd = sock_fd;
	conn_info.device_role = RFCOMM_ROLE_CLIENT;
	g_strlcpy(conn_info.uuid, uuid, BLUETOOTH_UUID_STRING_MAX);
	_bt_convert_addr_string_to_type(conn_info.device_addr.addr, address);

	__bt_rfcomm_reply_pending_request(
			result, BT_RFCOMM_CLIENT_CONNECT,
			(void *)&conn_info, sizeof(bluetooth_rfcomm_connection_t));

	BT_DBG("-");
}

int _bt_rfcomm_connect_using_uuid(bluetooth_device_address_t *device_address, char *remote_uuid)
{
	int result;
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };

	BT_DBG("+");

	retv_if(NULL == device_address, BLUETOOTH_ERROR_INVALID_PARAM);
	retv_if(NULL == remote_uuid, BLUETOOTH_ERROR_INVALID_PARAM);

	_bt_convert_addr_type_to_string(address, device_address->addr);
	BT_INFO("RFCOMM connect called for [%s], uuid: [%s]", address, remote_uuid);

	result = _bt_socket_client_connect(SOCK_TYPE_RFCOMM,
			address, remote_uuid, -1, __bt_rfcomm_socket_conn_cb);
	if (BLUETOOTH_ERROR_NONE != result) {
		BT_ERR("_bt_socket_client_connect failed");
		return result;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

/* Range of the Channel : 0 <= channel <= 30 */
int _bt_rfcomm_connect_using_channel(bluetooth_device_address_t *device_address, char *chan_str)
{
	int channel;
	int result = BLUETOOTH_ERROR_NONE;
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };

	BT_DBG("+");

	retv_if(NULL == device_address, BLUETOOTH_ERROR_INVALID_PARAM);
	retv_if(NULL == chan_str, BLUETOOTH_ERROR_INVALID_PARAM);

	_bt_convert_addr_type_to_string(address, device_address->addr);
	channel = atoi(chan_str);
	BT_INFO("RFCOMM connect called for [%s], channel: %d", address, channel);

	result = _bt_socket_client_connect(SOCK_TYPE_RFCOMM,
			address, NULL, channel, __bt_rfcomm_socket_conn_cb);
	if (BLUETOOTH_ERROR_NONE != result) {
		BT_ERR("_bt_socket_client_connect failed");
		return result;
	}

	BT_DBG("-");
	return result;
}

gboolean __bt_send_rfcomm_server_fd(gpointer user_data)
{
	BT_DBG("+");

	__bt_rfcomm_reply_pending_request(BLUETOOTH_ERROR_NONE,
			BT_RFCOMM_LISTEN, user_data, sizeof(int));

	g_free(user_data);
	BT_DBG("-");
	return FALSE;
}

gboolean __bt_send_rfcomm_server_fd_with_accept(gpointer user_data)
{
	BT_DBG("+");

	__bt_rfcomm_reply_pending_request(BLUETOOTH_ERROR_NONE,
			BT_RFCOMM_LISTEN_AND_ACCEPT, user_data, sizeof(int));

	g_free(user_data);
	BT_DBG("-");
	return FALSE;
}

int _bt_rfcomm_socket_listen(char *svc_name, char *uuid, bool accept)
{
	int channel = 0;
	int sock_fd;

	BT_DBG("+");

	retv_if(NULL == svc_name, BLUETOOTH_ERROR_INVALID_PARAM);
	retv_if(NULL == uuid, BLUETOOTH_ERROR_INVALID_PARAM);

	BT_INFO("RFCOMM create socket called with svc name: [%s], uuid: [%s]", svc_name, uuid);
	sock_fd = _bt_socket_listen(SOCK_TYPE_RFCOMM, svc_name, uuid, channel);
	if (sock_fd < 0) {
		BT_ERR("_bt_socket_listen failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	if (accept == false)
		g_idle_add(__bt_send_rfcomm_server_fd, g_memdup(&sock_fd, sizeof(int)));
	else
		g_idle_add(__bt_send_rfcomm_server_fd_with_accept, g_memdup(&sock_fd, sizeof(int)));

	BT_DBG("-");
	return sock_fd;
}
