/*
 * Copyright (c) 2011 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <dlog.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <net_connection.h>
#include <dbus/dbus.h>
#include <glib.h>
#include <dlog.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <net_connection.h>
#include <bundle.h>
#include <eventsystem.h>

#include "bluetooth-api.h"
#include "bt-service-common.h"
#include "bt-service-agent.h"

static GDBusConnection *system_conn;
static GDBusConnection *session_conn;
static GDBusProxy *manager_proxy;
static GDBusProxy *adapter_proxy;
static void *net_conn;

static GDBusProxy *adapter_properties_proxy;

GDBusConnection *_bt_gdbus_init_system_gconn(void)
{
	GError *error = NULL;

	dbus_threads_init_default();

	if (system_conn != NULL)
		return system_conn;

	system_conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);

	if (!system_conn) {
		BT_ERR("Unable to connect to dbus: %s", error->message);
		g_clear_error(&error);
	}

	return system_conn;
}

GDBusConnection *_bt_gdbus_get_system_gconn(void)
{
	GDBusConnection *local_system_gconn = NULL;
	GError *error = NULL;

	if (system_conn == NULL) {
		system_conn = _bt_gdbus_init_system_gconn();
	} else if (g_dbus_connection_is_closed(system_conn)) {
		local_system_gconn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);

		if (!local_system_gconn) {
			BT_ERR("Unable to connect to dbus: %s", error->message);
			g_clear_error(&error);
		}

		system_conn = local_system_gconn;
	}

	return system_conn;
}

GDBusConnection *_bt_gdbus_init_session_gconn(void)
{
	GError *error = NULL;

	if (!g_thread_supported())
		g_thread_init(NULL);

	dbus_threads_init_default();

	if (session_conn != NULL)
		return session_conn;

	session_conn = g_bus_get_sync(G_BUS_TYPE_SESSION, NULL, &error);

	if (!session_conn) {
		BT_ERR("Unable to connect to dbus: %s", error->message);
		g_clear_error(&error);
	}

	return session_conn;
}

GDBusConnection *_bt_gdbus_get_session_gconn(void)
{
	GDBusConnection *local_session_gconn = NULL;
	GError *error = NULL;

	if (session_conn == NULL) {
		session_conn = _bt_gdbus_init_session_gconn();
	} else if (g_dbus_connection_is_closed(session_conn)) {
		local_session_gconn = g_bus_get_sync(G_BUS_TYPE_SESSION, NULL, &error);

		if (!local_session_gconn) {
			BT_ERR("Unable to connect to dbus: %s", error->message);
			g_clear_error(&error);
		}

		session_conn = local_session_gconn;
	}

	return session_conn;
}

static GDBusProxy *__bt_init_manager_proxy(void)
{
	GDBusConnection *g_conn;
	GDBusProxy *proxy;

	dbus_threads_init_default();

	g_conn = _bt_gdbus_get_system_gconn();
	retv_if(g_conn == NULL, NULL);

	proxy = g_dbus_proxy_new_sync(g_conn, G_DBUS_PROXY_FLAGS_NONE,
								NULL, BT_BLUEZ_NAME,
								BT_MANAGER_PATH, BT_MANAGER_INTERFACE,  NULL, NULL);

	if (!proxy) {
		BT_ERR("Unable to get proxy");
		return NULL;
	}

	manager_proxy = proxy;

	return proxy;
}

static GDBusProxy *__bt_init_adapter_proxy(void)
{
	GDBusConnection *g_conn;
	GDBusProxy *manager_proxy;
	GDBusProxy *proxy;
	char *adapter_path = NULL;

	dbus_threads_init_default();

	g_conn = _bt_gdbus_get_system_gconn();
	retv_if(g_conn == NULL, NULL);

	manager_proxy = _bt_get_manager_proxy();
	retv_if(manager_proxy == NULL, NULL);

	adapter_path = _bt_get_adapter_path();
	retv_if(adapter_path == NULL, NULL);

	proxy = g_dbus_proxy_new_sync(g_conn, G_DBUS_PROXY_FLAGS_NONE,
								NULL, BT_BLUEZ_NAME,
								adapter_path, BT_ADAPTER_INTERFACE,  NULL, NULL);

	g_free(adapter_path);

	retv_if(proxy == NULL, NULL);

	adapter_proxy = proxy;

	return proxy;
}

static GDBusProxy *__bt_init_adapter_properties_proxy(void)
{
	GDBusConnection *g_conn;
	GDBusProxy *manager_proxy;
	GDBusProxy *proxy;
	char *adapter_path = NULL;

	dbus_threads_init_default();

	g_conn = _bt_gdbus_get_system_gconn();
	retv_if(g_conn == NULL, NULL);

	manager_proxy = _bt_get_manager_proxy();
	retv_if(manager_proxy == NULL, NULL);

	adapter_path = _bt_get_adapter_path();
	retv_if(adapter_path == NULL, NULL);

	proxy = g_dbus_proxy_new_sync(g_conn, G_DBUS_PROXY_FLAGS_NONE,
									NULL, BT_BLUEZ_NAME,
									adapter_path, BT_PROPERTIES_INTERFACE,  NULL, NULL);

	g_free(adapter_path);

	retv_if(proxy == NULL, NULL);

	adapter_properties_proxy = proxy;

	return proxy;
}

GDBusProxy *_bt_get_manager_proxy(void)
{
	if (manager_proxy) {
		const gchar *path =  g_dbus_proxy_get_object_path(manager_proxy);
		if (path == NULL) {
			BT_ERR("Already proxy released hence creating new proxy");
			return  __bt_init_manager_proxy();
		}
		return manager_proxy;
	}
	return  __bt_init_manager_proxy();
}

static void *__bt_init_net_conn(void)
{
	int result;
	connection_h connection = NULL;

	if (net_conn == NULL) {
		result = connection_create(&connection);

	if (result != CONNECTION_ERROR_NONE ||
					connection == NULL) {
		BT_DBG("connection_create() failed: %d", result);
		net_conn = NULL;
		return NULL;
	}
		net_conn = connection;
	}
	return net_conn;
}

void *_bt_get_net_conn(void)
{
	return (net_conn) ? net_conn : __bt_init_net_conn();
}

GDBusProxy *_bt_get_adapter_proxy(void)
{
	if (adapter_proxy) {
		const char *path =  g_dbus_proxy_get_object_path(adapter_proxy);
		if (path == NULL) {
			BT_ERR("Already proxy released hence creating new proxy");
			return  __bt_init_adapter_proxy();
		}

		return adapter_proxy;
	}
	return  __bt_init_adapter_proxy();

}

GDBusProxy *_bt_get_adapter_properties_proxy(void)
{
	return (adapter_properties_proxy) ? adapter_properties_proxy :
					__bt_init_adapter_properties_proxy();
}

static char *__bt_extract_adapter_path(GVariantIter *iter)
{
	char *object_path = NULL;
	GVariantIter *interface_iter;
	GVariantIter *svc_iter;
	char *interface_str = NULL;

	/* Parse the signature: oa{sa{sv}}} */
	while (g_variant_iter_loop(iter, "{&oa{sa{sv}}}", &object_path,
			&interface_iter)) {

		if (object_path == NULL)
			continue;

		while (g_variant_iter_loop(interface_iter, "{&sa{sv}}",
				&interface_str, &svc_iter)) {
			if (g_strcmp0(interface_str, "org.bluez.Adapter1") != 0)
				continue;

			BT_DBG("Object Path: %s", object_path);
			g_variant_iter_free(svc_iter);
			g_variant_iter_free(interface_iter);
			return g_strdup(object_path);
		}
	}
	return NULL;
}

char *_bt_get_adapter_path(void)
{
	GDBusConnection *conn;
	GDBusProxy *manager_proxy;
	GVariant *result = NULL;
	GVariantIter *iter = NULL;
	char *adapter_path = NULL;

	conn = _bt_gdbus_get_system_gconn();
	retv_if(conn == NULL, NULL);

	manager_proxy = _bt_get_manager_proxy();
	retv_if(manager_proxy == NULL, NULL);

	result = g_dbus_proxy_call_sync(manager_proxy, "GetManagedObjects",
				NULL,
				G_DBUS_CALL_FLAGS_NONE,
				-1,
				NULL,
				NULL);
	if (!result) {
		BT_ERR("Can't get managed objects");
		return NULL;
	}

	/* signature of GetManagedObjects:  a{oa{sa{sv}}} */
	g_variant_get(result, "(a{oa{sa{sv}}})", &iter);

	adapter_path = __bt_extract_adapter_path(iter);
	g_variant_iter_free(iter);
	g_variant_unref(result);
	return adapter_path;
}

void _bt_deinit_bluez_proxy(void)
{
	if (manager_proxy) {
		g_object_unref(manager_proxy);
		manager_proxy = NULL;
	}

	if (adapter_proxy) {
		g_object_unref(adapter_proxy);
		adapter_proxy = NULL;
	}
	if (adapter_properties_proxy) {
		g_object_unref(adapter_properties_proxy);
		adapter_properties_proxy = NULL;
	}
}

void _bt_deinit_proxys(void)
{
	int ret;
	_bt_deinit_bluez_proxy();

	if (system_conn) {
		g_object_unref(system_conn);
		system_conn = NULL;
	}

	if (session_conn) {
		g_object_unref(session_conn);
		session_conn = NULL;
	}

	if (net_conn) {
		ret = connection_destroy(net_conn);
		net_conn = NULL;
		if (ret != 0)
			BT_ERR("connection_destroy failed : %d", ret);
	}
}

void _bt_convert_device_path_to_address(const char *device_path,
						char *device_address)
{
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };
	char *dev_addr;

	ret_if(device_path == NULL);
	ret_if(device_address == NULL);

	dev_addr = strstr(device_path, "dev_");
	if (dev_addr != NULL) {
		char *pos = NULL;
		dev_addr += 4;
		g_strlcpy(address, dev_addr, sizeof(address));

		while ((pos = strchr(address, '_')) != NULL)
			*pos = ':';

		g_strlcpy(device_address, address, BT_ADDRESS_STRING_SIZE);
	}
}


void _bt_convert_addr_string_to_type(unsigned char *addr,
					const char *address)
{
	int i;
	char *ptr = NULL;

	ret_if(address == NULL);
	ret_if(addr == NULL);

	for (i = 0; i < BT_ADDRESS_LENGTH_MAX; i++) {
		addr[i] = strtol(address, &ptr, 16);
		if (ptr[0] != '\0') {
			if (ptr[0] != ':')
				return;

		address = ptr + 1;
		}
	}
}

void _bt_convert_addr_string_to_secure_string(char *addr,
					const char *address)
{
	int len;

	ret_if(address == NULL);
	ret_if(addr == NULL);

	len = strlen(address);
	ret_if(len != BT_ADDRESS_STRING_SIZE - 1);

	strncpy(addr, address, len);
	addr[len] = '\0';

	addr[len-7] = 'X';
	addr[len-8] = 'X';
	addr[len-10] = 'X';
	addr[len-11] = 'X';
	addr[len-13] = 'X';
	addr[len-14] = 'X';
	addr[len-16] = 'X';
	addr[len-17] = 'X';

	return;
}

void _bt_convert_addr_type_to_string(char *address,
				unsigned char *addr)
{
	ret_if(address == NULL);
	ret_if(addr == NULL);

	snprintf(address, BT_ADDRESS_STRING_SIZE,
			"%2.2X:%2.2X:%2.2X:%2.2X:%2.2X:%2.2X",
			addr[0], addr[1], addr[2],
			addr[3], addr[4], addr[5]);
}

void _bt_print_device_address_t(const bluetooth_device_address_t *addr)
{
	BT_DBG("%2.2X:%2.2X:%2.2X:%2.2X:%2.2X:%2.2X\n", addr->addr[0], addr->addr[1], addr->addr[2],
				addr->addr[3], addr->addr[4], addr->addr[5]);
}

void _bt_divide_device_class(bluetooth_device_class_t *device_class,
				unsigned int cod)
{
	ret_if(device_class == NULL);

	device_class->major_class = (unsigned short)(cod & 0x00001F00) >> 8;
	device_class->minor_class = (unsigned short)((cod & 0x000000FC));
	device_class->service_class = (unsigned long)((cod & 0x00FF0000));

	if (cod & 0x002000) {
		device_class->service_class |=
		BLUETOOTH_DEVICE_SERVICE_CLASS_LIMITED_DISCOVERABLE_MODE;
	}
}

void _bt_free_device_info(bt_remote_dev_info_t *dev_info)
{
	int i;

	ret_if(dev_info == NULL);

	g_free(dev_info->address);
	g_free(dev_info->name);
	g_free(dev_info->manufacturer_data);

	if (dev_info->uuids) {
		for (i = 0; i < dev_info->uuid_count && dev_info->uuids[i]; i++)
			g_free(dev_info->uuids[i]);

		g_free(dev_info->uuids);
	}

	g_free(dev_info);
}

void _bt_free_le_device_info(bt_remote_le_dev_info_t *le_dev_info)
{
	ret_if(le_dev_info == NULL);

	g_free(le_dev_info->adv_data);
	g_free(le_dev_info);
}

int _bt_copy_utf8_string(char *dest, const char *src, unsigned int length)
{
	int i;
	const char *p = src;
	char *next;
	int count;

	if (dest == NULL || src == NULL)
		return BLUETOOTH_ERROR_INVALID_PARAM;

	BT_DBG("+src : %s", src);
	BT_DBG("+dest : %s", dest);

	i = 0;
	while (*p != '\0' && i < length) {
		next = g_utf8_next_char(p);
		count = next - p;

		while (count > 0 && ((i + count) < length)) {
			dest[i++] = *p;
			p++;
			count--;
		}
		p = next;
	}
	return BLUETOOTH_ERROR_NONE;
}

gboolean _bt_utf8_validate(char *name)
{
	BT_DBG("+");
	gunichar2 *u16;
	glong items_written = 0;

	if (FALSE == g_utf8_validate(name, -1, NULL))
		return FALSE;

	u16 = g_utf8_to_utf16(name, -1, NULL, &items_written, NULL);
	if (u16 == NULL)
		return FALSE;

	g_free(u16);

	if (items_written != g_utf8_strlen(name, -1))
		return FALSE;

	BT_DBG("-");
	return TRUE;
}

int _bt_register_osp_server_in_agent(int type, char *uuid, char *path, int fd)
{
	BT_DBG("+");
	if (!_bt_agent_register_osp_server(type, uuid, path, fd))
		return BLUETOOTH_ERROR_INTERNAL;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_unregister_osp_server_in_agent(int type, char *uuid)
{
	BT_DBG("+");
	if (!_bt_agent_unregister_osp_server(type, uuid))
		return BLUETOOTH_ERROR_INTERNAL;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_set_socket_non_blocking(int socket_fd)
{
	/* Set Nonblocking */
	long arg;

	arg = fcntl(socket_fd, F_GETFL);

	if (arg < 0)
		return -errno;

	if (arg & O_NONBLOCK)
		BT_ERR("Already Non-blocking \n");

	arg |= O_NONBLOCK;

	if (fcntl(socket_fd, F_SETFL, arg) < 0)
		return -errno;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_set_non_blocking_tty(int sk)
{
	struct termios ti = {0,};
	int err;

	err = _bt_set_socket_non_blocking(sk);

	if (err < 0) {
		BT_ERR("Error in set non blocking!\n");
		return err;
	}

	tcflush(sk, TCIOFLUSH);

	/* Switch tty to RAW mode */
	cfmakeraw(&ti);
	tcsetattr(sk, TCSANOW, &ti);

	return BLUETOOTH_ERROR_NONE;
}

static char *__bt_extract_device_path(GVariantIter *iter, char *address)
{
	char *object_path = NULL;
	char device_address[BT_ADDRESS_STRING_SIZE] = { 0 };

	/* Parse the signature: oa{sa{sv}}} */
	while (g_variant_iter_loop(iter, "{&oa{sa{sv}}}", &object_path,
			NULL)) {
		if (!object_path) {
			BT_ERR("Unable to get object path");
			return NULL;
		}
		_bt_convert_device_path_to_address(object_path, device_address);
		if (g_strcmp0(address, device_address) == 0)
			return g_strdup(object_path);

	}

	BT_ERR("Unable to get object path");
	return NULL;
}

char *_bt_get_device_object_path(char *address)
{
	char *object_path = NULL;
	GDBusConnection *conn;
	GDBusProxy *manager_proxy;
	GVariant *result = NULL;
	GVariantIter *iter = NULL;

	conn = _bt_gdbus_get_system_gconn();
	retv_if(conn == NULL, NULL);

	manager_proxy = _bt_get_manager_proxy();
	retv_if(manager_proxy == NULL, NULL);

	result = g_dbus_proxy_call_sync(manager_proxy, "GetManagedObjects",
				NULL,
				G_DBUS_CALL_FLAGS_NONE,
				-1,
				NULL,
				NULL);
	if (!result) {
		BT_ERR("Can't get managed objects");
		return NULL;
	}

	/* signature of GetManagedObjects:  a{oa{sa{sv}}} */
	g_variant_get(result, "(a{oa{sa{sv}}})", &iter);
	object_path = __bt_extract_device_path(iter, address);
	g_variant_iter_free(iter);
	g_variant_unref(result);
	return object_path;
}

char *_bt_get_profile_uuid128(bt_profile_type_t profile_type)
{
	switch (profile_type) {
	case BT_PROFILE_CONN_RFCOMM:
		return strdup(RFCOMM_UUID_STR);
	case BT_PROFILE_CONN_A2DP:
		return strdup(A2DP_SINK_UUID);
	case BT_PROFILE_CONN_A2DP_SINK:
		return strdup(A2DP_SOURCE_UUID);
	case BT_PROFILE_CONN_HSP:
		return strdup(HFP_HS_UUID);
	case BT_PROFILE_CONN_HID:
		return strdup(HID_UUID);
	case BT_PROFILE_CONN_NAP:
		return strdup(NAP_UUID);
	case BT_PROFILE_CONN_HFG:
		return strdup(HFP_AG_UUID);
	case BT_PROFILE_CONN_PBAP:
		return strdup(PBAP_UUID);
	case BT_PROFILE_CONN_GATT:
	case BT_PROFILE_CONN_ALL: /* NULL UUID will connect to both the audio profiles*/
	default:
		return NULL;
	};
}

const char *_bt_convert_uuid_to_string(const char *uuid)
{
#define SHORT_UUID_COUNT	162
#define LONG_UUID_COUNT		17

	if (!uuid)
		return NULL;

	int offset = 0;
	int uuid_len = 4;
	typedef struct {
		const char *uuid;
		const char *specification_name;
	} uuid_name_s;
	static uuid_name_s short_uuid_name[SHORT_UUID_COUNT] = {
		// List should be sorted by UUID
		/* BT Classic Services */
		{"1101", "Serial Port Service"},
		{"1102", "LAN Access Using PPP Service"},
		{"1103", "Dialup Networking Service"},
		{"1104", "IrMCSync Service"},
		{"1105", "OBEX Object Push Service"},
		{"1106", "OBEX File Transfer Service"},
		{"1107", "IrMC Sync Command Service"},
		{"1108", "Headset Service"},
		{"1109", "Cordless Telephony Service"},
		{"110A", "Audio Source Service"},
		{"110B", "Audio Sink Service"},
		{"110C", "AV Remote Control Target Service"},
		{"110D", "Advanced Audio Distribution Profile"},
		{"110E", "AV Remote Control Service"},
		{"110F", "Video Conferencing Service"},
		{"1110", "Intercom Service"},
		{"1111", "Fax Service"},
		{"1112", "Headset Audio Gateway Service"},
		{"1113", "WAP Service"},
		{"1114", "WAP Client Service"},
		{"1115", "PANU Service"},
		{"1116", "NAP Service"},
		{"1117", "GN Service"},
		{"1118", "Direct Printing Service"},
		{"1119", "Reference Printing Service"},
		{"111A", "Basic Imaging Profile"},
		{"111B", "Imaging Responder Service"},
		{"111C", "Imaging Automatic Archive Service"},
		{"111D", "Imaging Reference Objects Service"},
		{"111E", "Handsfree Service"},
		{"111F", "Handsfree Audio Gateway Service"},
		{"1120", "Direct Printing Reference Objects Service"},
		{"1121", "Reflected UI Service"},
		{"1122", "Basic Printing Profile"},
		{"1123", "Printing Status Service"},
		{"1124", "Human Interface Device Service"},
		{"1125", "Hardcopy Cable Replacement Profile"},
		{"1126", "HCR Print Service"},
		{"1127", "HCR Scan Service"},
		{"112D", "SIM Access Service"},
		{"112E", "Phonebook Access PCE Service"},
		{"112F", "Phonebook Access PSE Service"},
		{"1130", "Phonebook Access Profile"},
		{"1132", "Message Access Server Service"},
		{"1133", "Message Notification Server Service"},
		{"1134", "Message Access Profile"},
		{"1200", "PnP Information Service"},
		{"1201", "Generic Networking Service"},
		{"1202", "Generic File Transfer Service"},
		{"1203", "Generic Audio Service"},
		{"1204", "Generic Telephony Service"},
		{"1205", "UPnP Service"},
		{"1206", "UPnP Ip Service"},
		{"1303", "Video Source Service"},
		{"1304", "Video Sink Service"},
		{"1305", "Video Distribution Profile"},
		{"1400", "Health Device Profile"},
		{"1401", "HDP Source Service"},
		{"1402", "HDP Sink Service"},

		/* GATT Services */
		{"1800", "Generic Access"},
		{"1801", "Generic Attribute"},
		{"1802", "Immediate Alert"},
		{"1803", "Link Loss"},
		{"1804", "Tx Power"},
		{"1805", "Current Time Service"},
		{"1806", "Reference Time Update Service"},
		{"1807", "Next DST Change Service"},
		{"1808", "Glucose"},
		{"1809", "Health Thermometer"},
		{"180A", "Device Information"},
		{"180D", "Heart Rate"},
		{"180F", "Battery Service"},
		{"1810", "Blood Pressure"},
		{"1811", "Alert Notification Service"},
		{"1812", "Human Interface Device"},
		{"1813", "Scan Parameters"},
		{"1814", "Running Speed and Cadence"},
		{"1815", "Automation IO"},
		{"1816", "Cycling Speed and Cadence"},
		{"1818", "Cycling Power"},
		{"1819", "Location and Navigation"},
		{"181A", "Environmental Sensing"},
		{"181B", "Body Composition"},
		{"181C", "User Data"},
		{"181D", "Weight Scale"},
		{"181E", "Bond Management"},
		{"181F", "Continuous Glucose Monitoring"},

		/* GATT Declarations */
		{"2800", "Primary Service Declaration"},
		{"2801", "Secondary Service Declaration"},
		{"2802", "Include Declaration"},
		{"2803", "Characteristic Declaration"},

		/* GATT Descriptors */
		{"2900", "Characteristic Extended Properties"},
		{"2901", "Characteristic User Description"},
		{"2902", "Client Characteristic Configuration"},
		{"2903", "Server Characteristic Configuration"},
		{"2904", "Characteristic Format"},
		{"2905", "Characteristic Aggregate Formate"},
		{"2906", "Valid Range"},
		{"2907", "External Report Reference"},
		{"2908", "Report Reference"},

		/* GATT Characteristics */
		{"2A00", "Device Name"},
		{"2A01", "Appearance"},
		{"2A02", "Peripheral Privacy Flag"},
		{"2A03", "Reconnection Address"},
		{"2A04", "Peripheral Preferred Connection Parameters"},
		{"2A05", "Service Changed"},
		{"2A06", "Alert Level"},
		{"2A07", "Tx Power Level"},
		{"2A08", "Date Time"},
		{"2A09", "Day of Week"},
		{"2A0A", "Day Date Time"},
		{"2A11", "Time with DST"},
		{"2A12", "Time Accuracy"},
		{"2A13", "Time Source"},
		{"2A14", "Reference Time Information"},
		{"2A16", "Time Update Control Point"},
		{"2A17", "Time Update State"},
		{"2A18", "Glucose Measurement"},
		{"2A19", "Battery Level"},
		{"2A1C", "Temperature Measurement"},
		{"2A1D", "Temperature Type"},
		{"2A1E", "Intermediate Temperature"},
		{"2A21", "Measurement Interval"},
		{"2A23", "System ID"},
		{"2A24", "Model Number String"},
		{"2A25", "Serial Number String"},
		{"2A26", "Firmware Revision String"},
		{"2A27", "Hardware Revision String"},
		{"2A28", "Software Revision String"},
		{"2A29", "Manufacturer Name String"},
		{"2A2A", "IEEE 11073-20601 Regulatory Certification Data List"},
		{"2A2B", "Current Time"},
		{"2A34", "Glucose Measurement Context"},
		{"2A35", "Blood Pressure Measurement"},
		{"2A37", "Heart Rate Measurement"},
		{"2A38", "Body Sensor Location"},
		{"2A39", "Heart Rate Control Point"},
		{"2A3F", "Alert Status"},
		{"2A46", "New Alert"},
		{"2A49", "Blood Pressure Feature"},
		{"2A4A", "HID Information"},
		{"2A4C", "HID Control Point"},
		{"2A50", "PnP ID"},
		{"2A51", "Glucose Feature"},
		{"2A52", "Record Access Control Point"},
		{"2A53", "RSC Measurement"},
		{"2A54", "RSC Feature"},
		{"2A55", "SC Control Point"},
		{"2A56", "Digital"},
		{"2A58", "Analog"},
		{"2A5A", "Aggregate"},
		{"2A5B", "CSC Measurement"},
		{"2A5C", "CSC Feature"},
		{"2A5D", "Sensor Location"},
		{"2A63", "Cycling Power Measurement"},
		{"2A64", "Cycling Power Vector"},
		{"2A65", "Cycling Power Feature"},
		{"2A66", "Cycling Power Control Point"},
		{"2A67", "Location and Speed"},
		{"2A68", "Navigation"},
		{"2A6D", "Pressure"},
		{"2A6E", "Temperature"},
	};
	static uuid_name_s long_uuid_name[LONG_UUID_COUNT] = {
		// List should be sorted by UUID
		/* Custom uuids */
		{"1AB7C24D-185A-45B9-90D4-F7AB1A71949A", "Samsung Health Service"},
		{"22EAC6E9-24D6-4BB5-BE44-B36ACE7C7BFB", "Data Source"},
		{"2F7CABCE-808D-411F-9A0C-BB92BA96C102", "Entity Update"},
		{"32D1955A-E5AA-4A96-9A49-08538DA8B8F6", "Samsung Gear Fit Manager Service"},
		{"69D1D8F3-45E1-49A8-9821-9BBDFDAAD9D9", "Control Point"},
		{"7905F431-B5CE-4E99-A40F-4B1E122D00D0", "Apple Notification Center Service"},
		{"89D3502B-0F36-433A-8EF4-C502AD55F8DC", "Apple Media Service"},
		{"9A3F68E0-86CE-11E5-A309-0002A5D5C51B", "Samsung Gear Manager Service"},
		{"9B3C81D8-57B1-4A8A-B8DF-0E56F7CA51C2", "Remote Command"},
		{"9FBF120D-6301-42D9-8C58-25E699A21DBD", "Notifications Source"},
		{"A49EB41E-CB06-495C-9F4F-BB80A90CDF00", "Samsung Gear Manager Service"},
		{"ADE3D529-C784-4F63-A987-EB69F70EE816", "IoT OIC Service"},
		{"C2051EE0-804D-4D50-A12C-15E243852100", "Notifications Source"},
		{"C2F2CC0F-C085-4DD4-BE5A-ACA3074BBC72", "Control Point"},
		{"C6B2F38C-23AB-46D8-A6AB-A3A870BBD5D7", "Entity Attribute"},
		{"CECE518B-28D7-4171-92D5-76A1E249A3B9", "Notifications Source"},
		{"FE53FF98-B259-4337-B56A-0EC9F82C6BAD", "Control Point"},
	};
	const uuid_name_s *uuid_name = short_uuid_name;
	static const char *unknown_name = "Unknown";
	int start = 0;
	int end = SHORT_UUID_COUNT - 1;
	int p;
	int ret;

	if (strlen(uuid) == 36) {
		if (!g_ascii_strncasecmp(uuid + 9, "0000-1000-8000-00805F9B34FB", 27))
			offset = 4;
		else {
			offset = 0;
			uuid_len = 36;
			end = LONG_UUID_COUNT - 1;
			uuid_name = long_uuid_name;
		}
	} else if (strlen(uuid) >= 8)
		offset = 4;

	while (start <= end) {
		p = start + (end - start) / 2;
		ret = g_ascii_strncasecmp(uuid + offset, uuid_name[p].uuid, uuid_len);
		if (ret == 0)
			return uuid_name[p].specification_name;
		else if (ret < 0)
			end = p - 1;
		else
			start = p + 1;
	}

	BT_INFO("Unknown uuid : %s", uuid);
	return unknown_name;
}

const char *_bt_convert_error_to_string(int error)
{
	switch (error) {
	case BLUETOOTH_ERROR_CANCEL:
		return "CANCELLED";
	case BLUETOOTH_ERROR_INVALID_PARAM:
		return "INVALID_PARAMETER";
	case BLUETOOTH_ERROR_INVALID_DATA:
		return "INVALID DATA";
	case BLUETOOTH_ERROR_MEMORY_ALLOCATION:
	case BLUETOOTH_ERROR_OUT_OF_MEMORY:
		return "OUT_OF_MEMORY";
	case BLUETOOTH_ERROR_TIMEOUT:
		return "TIMEOUT";
	case BLUETOOTH_ERROR_NO_RESOURCES:
		return "NO_RESOURCES";
	case BLUETOOTH_ERROR_INTERNAL:
		return "INTERNAL";
	case BLUETOOTH_ERROR_NOT_SUPPORT:
		return "NOT_SUPPORT";
	case BLUETOOTH_ERROR_DEVICE_NOT_ENABLED:
		return "NOT_ENABLED";
	case BLUETOOTH_ERROR_DEVICE_ALREADY_ENABLED:
		return "ALREADY_ENABLED";
	case BLUETOOTH_ERROR_DEVICE_BUSY:
		return "DEVICE_BUSY";
	case BLUETOOTH_ERROR_ACCESS_DENIED:
		return "ACCESS_DENIED";
	case BLUETOOTH_ERROR_MAX_CLIENT:
		return "MAX_CLIENT";
	case BLUETOOTH_ERROR_NOT_FOUND:
		return "NOT_FOUND";
	case BLUETOOTH_ERROR_SERVICE_SEARCH_ERROR:
		return "SERVICE_SEARCH_ERROR";
	case BLUETOOTH_ERROR_PARING_FAILED:
		return "PARING_FAILED";
	case BLUETOOTH_ERROR_NOT_PAIRED:
		return "NOT_PAIRED";
	case BLUETOOTH_ERROR_SERVICE_NOT_FOUND:
		return "SERVICE_NOT_FOUND";
	case BLUETOOTH_ERROR_NOT_CONNECTED:
		return "NOT_CONNECTED";
	case BLUETOOTH_ERROR_ALREADY_CONNECT:
		return "ALREADY_CONNECT";
	case BLUETOOTH_ERROR_CONNECTION_BUSY:
		return "CONNECTION_BUSY";
	case BLUETOOTH_ERROR_CONNECTION_ERROR:
		return "CONNECTION_ERROR";
	case BLUETOOTH_ERROR_MAX_CONNECTION:
		return "MAX_CONNECTION";
	case BLUETOOTH_ERROR_NOT_IN_OPERATION:
		return "NOT_IN_OPERATION";
	case BLUETOOTH_ERROR_CANCEL_BY_USER:
		return "CANCEL_BY_USER";
	case BLUETOOTH_ERROR_REGISTRATION_FAILED:
		return "REGISTRATION_FAILED";
	case BLUETOOTH_ERROR_IN_PROGRESS:
		return "IN_PROGRESS";
	case BLUETOOTH_ERROR_AUTHENTICATION_FAILED:
		return "AUTHENTICATION_FAILED";
	case BLUETOOTH_ERROR_HOST_DOWN:
		return "HOST_DOWN";
	case BLUETOOTH_ERROR_END_OF_DEVICE_LIST:
		return "END_OF_DEVICE_LIST";
	case BLUETOOTH_ERROR_AGENT_ALREADY_EXIST:
		return "AGENT_ALREADY_EXIST";
	case BLUETOOTH_ERROR_AGENT_DOES_NOT_EXIST:
		return "AGENT_DOES_NOT_EXIST";
	case BLUETOOTH_ERROR_ALREADY_INITIALIZED:
		return "ALREADY_INITIALIZED";
	case BLUETOOTH_ERROR_PERMISSION_DEINED:
		return "PERMISSION_DEINED";
	case BLUETOOTH_ERROR_ALREADY_DEACTIVATED:
		return "ALREADY_DEACTIVATED";
	case BLUETOOTH_ERROR_NOT_INITIALIZED:
		return "NOT_INITIALIZED";
	case BLUETOOTH_ERROR_AUTHENTICATION_REJECTED:
		return "AUTHENTICATION REJECTED";
	case BLUETOOTH_ERROR_AUTHORIZATION_REJECTED:
		return "AUTHORIZATION REJECTED";
	default:
		return "UNKNOWN";
	}
}

const char *_bt_convert_disc_reason_to_string(int reason)
{
	switch (reason) {
	case (int)BLUETOOTH_ERROR_PAGE_TIMEOUT:
		return "Page timeout";
	case (int)BLUETOOTH_ERROR_CONNECTION_TIMEOUT:
		return "Link loss";
	case (int)BLUETOOTH_ERROR_PIN_OR_KEY_MISSING:
		return "PIN or Key missing";
	case (int)BLUETOOTH_ERROR_LOCAL_HOST_TERM:
		return "Connection terminated by local host";
	case (int)BLUETOOTH_ERROR_REMOTE_USER_TERM:
	case (int)BLUETOOTH_ERROR_REMOTE_LOW_RESOURCES:
	case (int)BLUETOOTH_ERROR_REMOTE_POWER_OFF:
		return "Remote user terminated connection";
	case (int)BLUETOOTH_ERROR_AUTH_FAILURE:
		return "Authentication Failure";
	case (int)BLUETOOTH_ERROR_REPEATED_ATTEMPTS:
		return "Repeated attempts";
	case (int)BLUETOOTH_ERROR_LMP_RESPONSE_TIMEOUT:
		return "LMP response timeout";
	case (int)BLUETOOTH_ERROR_LMP_TRANSACTION_COLLISION:
		return "LMP transaction collision";
	case (int)BLUETOOTH_ERROR_INSTANT_PASSED:
		return "Instant passed";
	case (int)BLUETOOTH_ERROR_INSUFFICIENT_SECURITY:
		return "Insufficient security";
	case (int)BLUETOOTH_ERROR_CONNECTION_TERMINATED_DUE_TO_MIC_FAILURE:
		return "Connection terminated due to MIC failure";
	case (int)BLUETOOTH_ERROR_CONNECTION_FAILED_TO_BE_ESTABLISHED:
		return "Connection failed to be established";
	default:
		return "Unknown";
	}
}

const char *_bt_convert_profile_state_to_string(bt_profile_state_t state)
{
	switch (state) {
	case BT_PROFILE_STATE_UNAVAILABLE:
		return "UNAVAILABLE";
	case BT_PROFILE_STATE_DISCONNECTED:
		return "DISCONNECTED";
	case BT_PROFILE_STATE_CONNECTING:
		return "CONNECTING";
	case BT_PROFILE_STATE_CONNECTED:
		return "CONNECTED";
	case BT_PROFILE_STATE_DISCONNECTING:
		return "DISCONNECTING";
	}

	return NULL;
}

const char *_bt_convert_service_function_to_string(int function)
{
	int i;

	typedef struct {
		int function;
		const char *function_name;
	} bt_function_name_t;

	const bt_function_name_t bt_functions[] = {
		{BT_CHECK_ADAPTER, "BT_CHECK_ADAPTER"},
		{BT_ENABLE_ADAPTER, "BT_ENABLE_ADAPTER"},
		{BT_DISABLE_ADAPTER, "BT_DISABLE_ADAPTER"},
		{BT_RECOVER_ADAPTER, "BT_RECOVER_ADAPTER"},
		{BT_SET_DISCOVERABLE_TIME, "BT_SET_DISCOVERABLE_TIME"},
		{BT_GET_DISCOVERABLE_TIME, "BT_GET_DISCOVERABLE_TIME"},
		{BT_IGNORE_AUTO_PAIRING, "BT_IGNORE_AUTO_PAIRING"},
		{BT_GET_LOCAL_ADDRESS, "BT_GET_LOCAL_ADDRESS"},
		{BT_GET_LOCAL_VERSION, "BT_GET_LOCAL_VERSION"},
		{BT_GET_LOCAL_NAME, "BT_GET_LOCAL_NAME"},
		{BT_SET_LOCAL_NAME, "BT_SET_LOCAL_NAME"},
		{BT_IS_SERVICE_USED, "BT_IS_SERVICE_USED"},
		{BT_GET_DISCOVERABLE_MODE, "BT_GET_DISCOVERABLE_MODE"},
		{BT_SET_DISCOVERABLE_MODE, "BT_SET_DISCOVERABLE_MODE"},
		{BT_START_DISCOVERY, "BT_START_DISCOVERY"},
		{BT_START_CUSTOM_DISCOVERY, "BT_START_CUSTOM_DISCOVERY"},
		{BT_CANCEL_DISCOVERY, "BT_CANCEL_DISCOVERY"},
		{BT_START_LE_DISCOVERY, "BT_START_LE_DISCOVERY"},
		{BT_STOP_LE_DISCOVERY, "BT_STOP_LE_DISCOVERY"},
		{BT_IS_DISCOVERYING, "BT_IS_DISCOVERYING"},
		{BT_IS_LE_DISCOVERYING, "BT_IS_LE_DISCOVERYING"},
		{BT_ENABLE_RSSI, "BT_ENABLE_RSSI"},
		{BT_GET_RSSI, "BT_GET_RSSI"},
		{BT_IS_CONNECTABLE, "BT_IS_CONNECTABLE"},
		{BT_SET_CONNECTABLE, "BT_SET_CONNECTABLE"},
		{BT_GET_BONDED_DEVICES, "BT_GET_BONDED_DEVICES"},
		{BT_RESET_ADAPTER, "BT_RESET_ADAPTER"},
		{BT_SET_ADVERTISING, "BT_SET_ADVERTISING"},
		{BT_SET_CUSTOM_ADVERTISING, "BT_SET_CUSTOM_ADVERTISING"},
		{BT_SET_ADVERTISING_PARAMETERS, "BT_SET_ADVERTISING_PARAMETERS"},
		{BT_GET_ADVERTISING_DATA, "BT_GET_ADVERTISING_DATA"},
		{BT_SET_ADVERTISING_DATA, "BT_SET_ADVERTISING_DATA"},
		{BT_SET_SCAN_PARAMETERS, "BT_SET_SCAN_PARAMETERS"},
		{BT_GET_SCAN_RESPONSE_DATA, "BT_GET_SCAN_RESPONSE_DATA"},
		{BT_SET_SCAN_RESPONSE_DATA, "BT_SET_SCAN_RESPONSE_DATA"},
		{BT_IS_ADVERTISING, "BT_IS_ADVERTISING"},
		{BT_SET_MANUFACTURER_DATA, "BT_SET_MANUFACTURER_DATA"},
		{BT_LE_CONN_UPDATE, "BT_LE_CONN_UPDATE"},
		{BT_LE_READ_MAXIMUM_DATA_LENGTH, "BT_LE_READ_MAXIMUM_DATA_LENGTH"},
		{BT_LE_WRITE_HOST_SUGGESTED_DATA_LENGTH, "BT_LE_WRITE_HOST_SUGGESTED_DATA_LENGTH"},
		{BT_LE_READ_HOST_SUGGESTED_DATA_LENGTH, "BT_LE_READ_HOST_SUGGESTED_DATA_LENGTH"},
		{BT_LE_SET_DATA_LENGTH, "BT_LE_SET_DATA_LENGTH"},
		{BT_ADD_WHITE_LIST, "BT_ADD_WHITE_LIST"},
		{BT_REMOVE_WHITE_LIST, "BT_REMOVE_WHITE_LIST"},
		{BT_CLEAR_WHITE_LIST, "BT_CLEAR_WHITE_LIST"},
		{BT_REGISTER_SCAN_FILTER, "BT_REGISTER_SCAN_FILTER"},
		{BT_UNREGISTER_SCAN_FILTER, "BT_UNREGISTER_SCAN_FILTER"},
		{BT_UNREGISTER_ALL_SCAN_FILTERS, "BT_UNREGISTER_ALL_SCAN_FILTERS"},
		{BT_IS_SCAN_FILTER_SUPPORTED, "BT_IS_SCAN_FILTER_SUPPORTED"},
		{BT_GET_PROFILE_CONNECTED_DEVICES, "BT_GET_PROFILE_CONNECTED_DEVICES"},
		{BT_ENABLE_FORCE_HCI_DUMP, "BT_ENABLE_FORCE_HCI_DUMP"},
		{BT_SET_PASSKEY_NOTIFICATION, "BT_SET_PASSKEY_NOTIFICATION"},
		{BT_DISCONNECT_DEVICE, "BT_DISCONNECT_DEVICE"},
		{BT_BOND_DEVICE, "BT_BOND_DEVICE"},
		{BT_BOND_DEVICE_BY_TYPE, "BT_BOND_DEVICE_BY_TYPE"},
		{BT_CANCEL_BONDING, "BT_CANCEL_BONDING"},
		{BT_PASSKEY_REPLY, "BT_PASSKEY_REPLY"},
		{BT_PASSKEY_CONFIRMATION_REPLY, "BT_PASSKEY_CONFIRMATION_REPLY"},
		{BT_UNBOND_DEVICE, "BT_UNBOND_DEVICE"},
		{BT_SEARCH_SERVICE, "BT_SEARCH_SERVICE"},
		{BT_CANCEL_SEARCH_SERVICE, "BT_CANCEL_SEARCH_SERVICE"},
		{BT_GET_BONDED_DEVICE, "BT_GET_BONDED_DEVICE"},
		{BT_GET_IS_ALIAS_SET, "BT_GET_IS_ALIAS_SET"},
		{BT_SET_ALIAS, "BT_SET_ALIAS"},
		{BT_SET_AUTHORIZATION, "BT_SET_AUTHORIZATION"},
		{BT_UNSET_AUTHORIZATION, "BT_UNSET_AUTHORIZATION"},
		{BT_IS_DEVICE_CONNECTED, "BT_IS_DEVICE_CONNECTED"},
		{BT_GET_CONNECTED_LINK_TYPE, "BT_GET_CONNECTED_LINK_TYPE"},
		{BT_SET_PIN_CODE, "BT_SET_PIN_CODE"},
		{BT_UNSET_PIN_CODE, "BT_UNSET_PIN_CODE"},
		{BT_UPDATE_LE_CONNECTION_MODE, "BT_UPDATE_LE_CONNECTION_MODE"},
		{BT_SET_PROFILE_TRUSTED, "BT_SET_PROFILE_TRUSTED"},
		{BT_GET_PROFILE_TRUSTED, "BT_GET_PROFILE_TRUSTED"},
		{BT_HID_CONNECT, "BT_HID_CONNECT"},
		{BT_HID_DISCONNECT, "BT_HID_DISCONNECT"},
		{BT_HID_DEVICE_ACTIVATE, "BT_HID_DEVICE_ACTIVATE"},
		{BT_HID_DEVICE_DEACTIVATE, "BT_HID_DEVICE_DEACTIVATE"},
		{BT_HID_DEVICE_CONNECT, "BT_HID_DEVICE_CONNECT"},
		{BT_HID_DEVICE_DISCONNECT, "BT_HID_DEVICE_DISCONNECT"},
		{BT_HID_DEVICE_SEND_MOUSE_EVENT, "BT_HID_DEVICE_SEND_MOUSE_EVENT"},
		{BT_HID_DEVICE_SEND_KEY_EVENT, "BT_HID_DEVICE_SEND_KEY_EVENT"},
		{BT_HID_DEVICE_SEND_CUSTOM_EVENT, "BT_HID_DEVICE_SEND_CUSTOM_EVENT"},
		{BT_HID_DEVICE_SEND_REPLY_TO_REPORT, "BT_HID_DEVICE_SEND_REPLY_TO_REPORT"},
		{BT_HID_ENABLE_BARCODE_FEATURE, "BT_HID_ENABLE_BARCODE_FEATURE"},
		{BT_NETWORK_ACTIVATE, "BT_NETWORK_ACTIVATE"},
		{BT_NETWORK_DEACTIVATE, "BT_NETWORK_DEACTIVATE"},
		{BT_NETWORK_CONNECT, "BT_NETWORK_CONNECT"},
		{BT_NETWORK_DISCONNECT, "BT_NETWORK_DISCONNECT"},
		{BT_NETWORK_SERVER_DISCONNECT, "BT_NETWORK_SERVER_DISCONNECT"},
		{BT_AUDIO_CONNECT, "BT_AUDIO_CONNECT"},
		{BT_AUDIO_DISCONNECT, "BT_AUDIO_DISCONNECT"},
		{BT_AG_CONNECT, "BT_AG_CONNECT"},
		{BT_AG_DISCONNECT, "BT_AG_DISCONNECT"},
		{BT_AV_CONNECT, "BT_AV_CONNECT"},
		{BT_AV_DISCONNECT, "BT_AV_DISCONNECT"},
		{BT_AV_SOURCE_CONNECT, "BT_AV_SOURCE_CONNECT"},
		{BT_AV_SOURCE_DISCONNECT, "BT_AV_SOURCE_DISCONNECT"},
		{BT_HF_CONNECT, "BT_HF_CONNECT"},
		{BT_HF_DISCONNECT, "BT_HF_DISCONNECT"},
		{BT_GET_SPEAKER_GAIN, "BT_GET_SPEAKER_GAIN"},
		{BT_SET_SPEAKER_GAIN, "BT_SET_SPEAKER_GAIN"},
		{BT_SET_CONTENT_PROTECT, "BT_SET_CONTENT_PROTECT"},
		{BT_OOB_READ_LOCAL_DATA, "BT_OOB_READ_LOCAL_DATA"},
		{BT_OOB_ADD_REMOTE_DATA, "BT_OOB_ADD_REMOTE_DATA"},
		{BT_OOB_REMOVE_REMOTE_DATA, "BT_OOB_REMOVE_REMOTE_DATA"},
		{BT_AVRCP_SET_TRACK_INFO, "BT_AVRCP_SET_TRACK_INFO"},
		{BT_AVRCP_SET_PROPERTY, "BT_AVRCP_SET_PROPERTY"},
		{BT_AVRCP_SET_PROPERTIES, "BT_AVRCP_SET_PROPERTIES"},
		{BT_AVRCP_CONTROL_CONNECT, "BT_AVRCP_CONTROL_CONNECT"},
		{BT_AVRCP_CONTROL_DISCONNECT, "BT_AVRCP_CONTROL_DISCONNECT"},
		{BT_AVRCP_TARGET_CONNECT, "BT_AVRCP_TARGET_CONNECT"},
		{BT_AVRCP_TARGET_DISCONNECT, "BT_AVRCP_TARGET_DISCONNECT"},
		{BT_AVRCP_HANDLE_CONTROL, "BT_AVRCP_HANDLE_CONTROL"},
		{BT_AVRCP_CONTROL_SET_PROPERTY, "BT_AVRCP_CONTROL_SET_PROPERTY"},
		{BT_AVRCP_CONTROL_GET_PROPERTY, "BT_AVRCP_CONTROL_GET_PROPERTY"},
		{BT_AVRCP_GET_TRACK_INFO, "BT_AVRCP_GET_TRACK_INFO"},
		{BT_OPP_PUSH_FILES, "BT_OPP_PUSH_FILES"},
		{BT_OPP_CANCEL_PUSH, "BT_OBT_OPP_IS_PUSHING_FILESPP_CANCEL_PUSH"},
		{BT_OPP_IS_PUSHING_FILES, "BT_OPP_IS_PUSHING_FILES"},
		{BT_OPP_GET_TRANSFER_PROGRESS, "BT_OPP_GET_TRANSFER_PROGRESS"},
		{BT_MAP_CREATE_SESSION, "BT_MAP_CREATE_SESSION"},
		{BT_MAP_DESTROY_SESSION, "BT_MAP_DESTROY_SESSION"},
		{BT_MAP_SET_FOLDER, "BT_MAP_SET_FOLDER"},
		{BT_MAP_LIST_FOLDERS, "BT_MAP_LIST_FOLDERS"},
		{BT_MAP_LIST_FILTER_FIELDS, "BT_MAP_LIST_FILTER_FIELDS"},
		{BT_MAP_LIST_MESSAGES, "BT_MAP_LIST_MESSAGES"},
		{BT_MAP_UPDATE_INBOX, "BT_MAP_UPDATE_INBOX"},
		{BT_MAP_PUSH_MESSAGE, "BT_MAP_PUSH_MESSAGE"},
		{BT_MAP_GET_MESSAGE, "BT_MAP_GET_MESSAGE"},
		{BT_OBEX_SERVER_ALLOCATE, "BT_OBEX_SERVER_ALLOCATE"},
		{BT_OBEX_SERVER_DEALLOCATE, "BT_OBEX_SERVER_DEALLOCATE"},
		{BT_OBEX_SERVER_IS_ACTIVATED, "BT_OBEX_SERVER_IS_ACTIVATED"},
		{BT_OBEX_SERVER_ACCEPT_CONNECTION, "BT_OBEX_SERVER_ACCEPT_CONNECTION"},
		{BT_OBEX_SERVER_REJECT_CONNECTION, "BT_OBEX_SERVER_REJECT_CONNECTION"},
		{BT_OBEX_SERVER_ACCEPT_FILE, "BT_OBEX_SERVER_ACCEPT_FILE"},
		{BT_OBEX_SERVER_REJECT_FILE, "BT_OBEX_SERVER_REJECT_FILE"},
		{BT_OBEX_SERVER_SET_PATH, "BT_OBEX_SERVER_SET_PATH"},
		{BT_OBEX_SERVER_SET_ROOT, "BT_OBEX_SERVER_SET_ROOT"},
		{BT_OBEX_SERVER_CANCEL_TRANSFER, "BT_OBEX_SERVER_CANCEL_TRANSFER"},
		{BT_OBEX_SERVER_CANCEL_ALL_TRANSFERS, "BT_OBEX_SERVER_CANCEL_ALL_TRANSFERS"},
		{BT_OBEX_SERVER_IS_RECEIVING, "BT_OBEX_SERVER_IS_RECEIVING"},
		{BT_RFCOMM_CLIENT_CONNECT, "BT_RFCOMM_CLIENT_CONNECT"},
		{BT_RFCOMM_CLIENT_CANCEL_CONNECT, "BT_RFCOMM_CLIENT_CANCEL_CONNECT"},
		{BT_RFCOMM_CLIENT_IS_CONNECTED, "BT_RFCOMM_CLIENT_IS_CONNECTED"},
		{BT_RFCOMM_SOCKET_DISCONNECT, "BT_RFCOMM_SOCKET_DISCONNECT"},
		{BT_RFCOMM_SOCKET_WRITE, "BT_RFCOMM_SOCKET_WRITE"},
		{BT_RFCOMM_CREATE_SOCKET, "BT_RFCOMM_CREATE_SOCKET"},
		{BT_RFCOMM_REMOVE_SOCKET, "BT_RFCOMM_REMOVE_SOCKET"},
		{BT_RFCOMM_LISTEN, "BT_RFCOMM_LISTEN"},
		{BT_RFCOMM_IS_UUID_AVAILABLE, "BT_RFCOMM_IS_UUID_AVAILABLE"},
		{BT_RFCOMM_ACCEPT_CONNECTION, "BT_RFCOMM_ACCEPT_CONNECTION"},
		{BT_RFCOMM_REJECT_CONNECTION, "BT_RFCOMM_REJECT_CONNECTION"},
		{BT_RFCOMM_CREATE_SOCKET_EX, "BT_RFCOMM_CREATE_SOCKET_EX"},
		{BT_RFCOMM_REMOVE_SOCKET_EX, "BT_RFCOMM_REMOVE_SOCKET_EX"},
		{BT_PBAP_CONNECT, "BT_PBAP_CONNECT"},
		{BT_PBAP_DISCONNECT, "BT_PBAP_DISCONNECT"},
		{BT_PBAP_GET_PHONEBOOK_SIZE, "BT_PBAP_GET_PHONEBOOK_SIZE"},
		{BT_PBAP_GET_PHONEBOOK, "BT_PBAP_GET_PHONEBOOK"},
		{BT_PBAP_GET_LIST, "BT_PBAP_GET_LIST"},
		{BT_PBAP_PULL_VCARD, "BT_PBAP_PULL_VCARD"},
		{BT_PBAP_PHONEBOOK_SEARCH, "BT_PBAP_PHONEBOOK_SEARCH"},
		{BT_ENABLE_ADAPTER_LE, "BT_ENABLE_ADAPTER_LE"},
		{BT_DISABLE_ADAPTER_LE, "BT_DISABLE_ADAPTER_LE"},
		{BT_CONNECT_LE, "BT_CONNECT_LE"},
		{BT_DISCONNECT_LE, "BT_DISCONNECT_LE"},
		{BT_SET_LE_PRIVACY, "BT_SET_LE_PRIVACY"},
		{BT_REQ_ATT_MTU, "BT_REQ_ATT_MTU"},
		{BT_GET_ATT_MTU, "BT_GET_ATT_MTU"},
		{BT_GET_DEVICE_IDA, "BT_GET_DEVICE_IDA"},
		{BT_SET_LE_STATIC_RANDOM_ADDRESS, "BT_SET_LE_STATIC_RANDOM_ADDRESS"},
		{BT_HDP_CONNECT, "BT_HDP_CONNECT"},
		{BT_HDP_DISCONNECT, "BT_HDP_DISCONNECT"},
		{BT_HDP_SEND_DATA, "BT_HDP_SEND_DATA"},
		{BT_HDP_REGISTER_SINK_APP, "BT_HDP_REGISTER_SINK_APP"},
		{BT_HDP_UNREGISTER_SINK_APP, "BT_HDP_UNREGISTER_SINK_APP"},
		{BT_GATT_GET_PRIMARY_SERVICES, "BT_GATT_GET_PRIMARY_SERVICES"},
		{BT_GATT_DISCOVER_CHARACTERISTICS, "BT_GATT_DISCOVER_CHARACTERISTICS"},
		{BT_GATT_SET_PROPERTY_REQUEST, "BT_GATT_SET_PROPERTY_REQUEST"},
		{BT_GATT_READ_CHARACTERISTIC, "BT_GATT_READ_CHARACTERISTIC"},
		{BT_GATT_DISCOVER_CHARACTERISTICS_DESCRIPTOR, "BT_GATT_DISCOVER_CHARACTERISTICS_DESCRIPTOR"},
		{BT_GATT_REGISTER_APPLICATION, "BT_GATT_REGISTER_APPLICATION"},
		{BT_GATT_REGISTER_SERVICE, "BT_GATT_REGISTER_SERVICE"},
		{BT_GATT_SEND_RESPONSE, "BT_GATT_SEND_RESPONSE"},
		{BT_LE_IPSP_INIT, "BT_LE_IPSP_INIT"},
		{BT_LE_IPSP_DEINIT, "BT_LE_IPSP_DEINIT"},
		{BT_LE_IPSP_CONNECT, "BT_LE_IPSP_CONNECT"},
		{BT_LE_IPSP_DISCONNECT, "BT_LE_IPSP_DISCONNECT"},
		{BT_DPM_SET_ALLOW_BT_MODE, "BT_DPM_SET_ALLOW_BT_MODE"},
		{BT_DPM_GET_ALLOW_BT_MODE, "BT_DPM_GET_ALLOW_BT_MODE"},
		{BT_DPM_SET_DEVICE_RESTRITION, "BT_DPM_SET_DEVICE_RESTRITION"},
		{BT_DPM_GET_DEVICE_RESTRITION, "BT_DPM_GET_DEVICE_RESTRITION"},
		{BT_DPM_SET_UUID_RESTRITION, "BT_DPM_SET_UUID_RESTRITION"},
		{BT_DPM_GET_UUID_RESTRITION, "BT_DPM_GET_UUID_RESTRITION"},
		{BT_DPM_ADD_DEVICES_BLACKLIST, "BT_DPM_ADD_DEVICES_BLACKLIST"},
		{BT_DPM_ADD_DEVICES_WHITELIST, "BT_DPM_ADD_DEVICES_WHITELIST"},
		{BT_DPM_ADD_UUIDS_BLACKLIST, "BT_DPM_ADD_UUIDS_BLACKLIST"},
		{BT_DPM_ADD_UUIDS_WHITELIST, "BT_DPM_ADD_UUIDS_WHITELIST"},
		{BT_DPM_CLEAR_DEVICES_BLACKLIST, "BT_DPM_CLEAR_DEVICES_BLACKLIST"},
		{BT_DPM_CLEAR_DEVICES_WHITELIST, "BT_DPM_CLEAR_DEVICES_WHITELIST"},
		{BT_DPM_CLEAR_UUIDS_BLACKLIST, "BT_DPM_CLEAR_UUIDS_BLACKLIST"},
		{BT_DPM_CLEAR_UUIDS_WHITELIST, "BT_DPM_CLEAR_UUIDS_WHITELIST"},
		{BT_DPM_REMOVE_DEVICE_BLACKLIST, "BT_DPM_REMOVE_DEVICE_BLACKLIST"},
		{BT_DPM_REMOVE_DEVICE_WHITELIST, "BT_DPM_REMOVE_DEVICE_WHITELIST"},
		{BT_DPM_REMOVE_UUID_BLACKLIST, "BT_DPM_REMOVE_UUID_BLACKLIST"},
		{BT_DPM_REMOVE_UUID_WHITELIST, "BT_DPM_REMOVE_UUID_WHITELIST"},
		{BT_DPM_GET_DEVICES_BLACKLIST, "BT_DPM_GET_DEVICES_BLACKLIST"},
		{BT_DPM_GET_DEVICES_WHITELIST, "BT_DPM_GET_DEVICES_WHITELIST"},
		{BT_DPM_GET_UUIDS_BLACKLIST, "BT_DPM_GET_UUIDS_BLACKLIST"},
		{BT_DPM_GET_UUIDS_WHITELIST, "BT_DPM_GET_UUIDS_WHITELIST"},
		{BT_DPM_SET_ALLOW_OUTGOING_CALL, "BT_DPM_SET_ALLOW_OUTGOING_CALL"},
		{BT_DPM_GET_ALLOW_OUTGOING_CALL, "BT_DPM_GET_ALLOW_OUTGOING_CALL"},
		{BT_DPM_SET_PAIRING_STATE, "BT_DPM_SET_PAIRING_STATE"},
		{BT_DPM_GET_PAIRING_STATE, "BT_DPM_GET_PAIRING_STATE"},
		{BT_DPM_SET_PROFILE_STATE, "BT_DPM_SET_PROFILE_STATE"},
		{BT_DPM_GET_PROFILE_STATE, "BT_DPM_GET_PROFILE_STATE"},
		{BT_DPM_SET_DESKROP_CONNECTIVITY_STATE, "BT_DPM_SET_DESKROP_CONNECTIVITY_STATE"},
		{BT_DPM_GET_DESKROP_CONNECTIVITY_STATE, "BT_DPM_GET_DESKROP_CONNECTIVITY_STATE"},
		{BT_DPM_SET_DISCOVERABLE_STATE, "BT_DPM_SET_DISCOVERABLE_STATE"},
		{BT_DPM_GET_DISCOVERABLE_STATE, "BT_DPM_GET_DISCOVERABLE_STATE"},
		{BT_DPM_SET_LIMITED_DISCOVERABLE_STATE, "BT_DPM_SET_LIMITED_DISCOVERABLE_STATE"},
		{BT_DPM_GET_LIMITED_DISCOVERABLE_STATE, "BT_DPM_GET_LIMITED_DISCOVERABLE_STATE"},
		{BT_DPM_SET_DATA_TRANSFER_STATE, "BT_DPM_SET_DATA_TRANSFER_STATE"},
		{BT_DPM_GET_DATA_TRANSFER_STATE, "BT_DPM_GET_DATA_TRANSFER_STATE"},
		{BT_PXP_MONITOR_SET_PROPERTY, "BT_PXP_MONITOR_SET_PROPERTY"},
		{BT_PXP_MONITOR_GET_PROPERTY, "BT_PXP_MONITOR_GET_PROPERTY"},
		{BT_PXP_MONITOR_GET_SUPPORTED_SERIVCES, "BT_PXP_MONITOR_GET_SUPPORTED_SERIVCES"},
		{BT_PXP_REPORTER_REGISTER, "BT_PXP_REPORTER_REGISTER"},
		{BT_PXP_REPORTER_UNREGISTER, "BT_PXP_REPORTER_UNREGISTER"},
		{BT_PXP_REPORTER_GET_PROPERTY, "BT_PXP_REPORTER_GET_PROPERTY"},
		{BT_TDS_PROVIDER_REGISTER, "BT_TDS_PROVIDER_REGISTER"},
		{BT_TDS_PROVIDER_UNREGISTER, "BT_TDS_PROVIDER_UNREGISTER"},
		{BT_TDS_PROVIDER_SET_MANUF_DATA, "BT_TDS_PROVIDER_SET_MANUF_DATA"},
		{BT_TDS_PROVIDER_CREATE, "BT_TDS_PROVIDER_CREATE"},
		{BT_TDS_PROVIDER_DESTROY, "BT_TDS_PROVIDER_DESTROY"},
		{BT_TDS_PROVIDER_SET_TRANSPORT_DATA, "BT_TDS_PROVIDER_SET_TRANSPORT_DATA"},
		{BT_TDS_SEND_ACTIVATION_RESPONSE, "BT_TDS_SEND_ACTIVATION_RESPONSE"},
		{BT_TDS_READ_TRANSPORT_DATA, "BT_TDS_READ_TRANSPORT_DATA"},
		{BT_TDS_ENABLE_CONTROL_POINT, "BT_TDS_ENABLE_CONTROL_POINT"},
		{BT_TDS_ACTIVATE_CONTROL_POINT, "BT_TDS_ACTIVATE_CONTROL_POINT"},
		{-1, ""},
	};

	for (i = 0; bt_functions[i].function != -1; i++) {
		if (bt_functions[i].function == function)
			return bt_functions[i].function_name;
	}

	return NULL;
}

void _bt_logging_connection(gboolean connect, int addr_type)
{
	static int le_conn = 0;
	static int le_disc = 0;
	static int edr_conn = 0;
	static int edr_disc = 0;

	if (connect) {
		if (addr_type)
			le_conn++;
		else
			edr_conn++;
	} else {
		if (addr_type)
			le_disc++;
		else
			edr_disc++;
	}

	BT_INFO("[PM] Number of LE conn: %d disc: %d, Number of BR/EDR conn: %d disc: %d",
			le_conn, le_disc, edr_conn, edr_disc);
}

void _bt_swap_byte_ordering(char *data, int data_len)
{
	char temp;
	int i, j;
	int half = data_len / 2;

	ret_if(data == NULL);
	/* Swap to opposite endian */
	for (i = 0, j = data_len - 1; i < half; i++, j--) {
		temp = data[i];
		data[i] = data[j];
		data[j] = temp;
	}
}

int _bt_byte_arr_cmp(const char *data1, const char *data2, int data_len)
{
	int i;

	retv_if(data1 == NULL, -1);
	retv_if(data2 == NULL, -1);
	for (i = 0; i < data_len; i++) {
		if (data1[i] != data2[i])
			return data1[i] - data2[i];
		}
	return 0;
}
int _bt_byte_arr_cmp_with_mask(const char *data1, const char *data2,
	const char *mask, int data_len)
{
	int i;
	char a, b;

	retv_if(data1 == NULL, -1);
	retv_if(data2 == NULL, -1);
	retv_if(mask == NULL, -1);
	for (i = 0; i < data_len; i++) {
		a = data1[i] & mask[i];
		b = data2[i] & mask[i];
		if (a != b)
			return (int)(a - b);
		}
	return 0;
}

int _bt_eventsystem_set_value(const char *event, const char *key, const char *value)
{
	int ret = ES_R_OK;
/* Send event system event in bt-core process because bt-service's permission is not system in now */

#if 0
	bundle *b = NULL;

	b = bundle_create();

	bundle_add_str(b, key, value);

	ret = eventsystem_send_system_event(event, b);

	BT_DBG("eventsystem_send_system_event result: %d", ret);

	bundle_free(b);
#endif
	return ret;
}

void __bt_get_auth_info(GVariant *reply, char *auth_info)
{
	int cursor;
	GVariant *tmp_value;
	char *manufacturer_data = NULL;
	int manufacturer_data_len;
	gboolean is_alias_set;
	GVariantIter *value_iter;
	guint8 m_value;
	int i = 0;

	tmp_value = g_variant_lookup_value(reply, "IsAliasSet",
								G_VARIANT_TYPE_BOOLEAN);
	if (tmp_value) {
		is_alias_set = g_variant_get_boolean(tmp_value);
		g_variant_unref(tmp_value);
	} else {
		is_alias_set = FALSE;
	}
	if (is_alias_set == FALSE) {
		tmp_value = g_variant_lookup_value(reply, "LegacyManufacturerDataLen",
								G_VARIANT_TYPE_UINT16);
		if (tmp_value) {
			manufacturer_data_len = g_variant_get_uint16(tmp_value);
			if (manufacturer_data_len >
					BLUETOOTH_MANUFACTURER_DATA_LENGTH_MAX) {
				BT_ERR("manufacturer_data_len is too long");
				manufacturer_data_len = BLUETOOTH_MANUFACTURER_DATA_LENGTH_MAX;
			}
			g_variant_unref(tmp_value);
		} else
			manufacturer_data_len = 0;

		tmp_value = g_variant_lookup_value(reply, "LegacyManufacturerData",
								G_VARIANT_TYPE_ARRAY);
		if (tmp_value) {
			if ((manufacturer_data_len == 0) ||
					manufacturer_data_len != g_variant_get_size(tmp_value)) {
				BT_ERR("manufacturer data length doesn't match");
				manufacturer_data_len = 0;
				manufacturer_data = NULL;
			} else {
				manufacturer_data = g_malloc0(manufacturer_data_len);
				g_variant_get(tmp_value, "ay", &value_iter);
				while (g_variant_iter_loop(value_iter, "y", &m_value))
					manufacturer_data[i++] = m_value;
			}
			g_variant_unref(tmp_value);
		} else {
			BT_INFO("manufacture data is not a G_VARIANT_TYPE_ARRAY ");
			manufacturer_data_len = 0;
			manufacturer_data = NULL;
		}
		/*minimum Size of the samsung specific manufacturer data is greater than 30 */
		if (manufacturer_data_len < 30) {
			g_free(manufacturer_data);
			return;
		}
		if (manufacturer_data[0] != 0x00 || manufacturer_data[1] != 0x75) {
			BT_DBG("This is not a samsung specific manufaturer data");
			g_free(manufacturer_data);
			return;
		}

		/* 2  samsung (0x00 0x75) + 1 (control and version) + 1 (service ID) +
		1 (discovery version) + 1 (associated service ID)
		2 (Proxamity and locality) + 2 (Device type and icon) */

		cursor = 10;

		memcpy(auth_info, &(manufacturer_data[cursor]), 5);
	}
	 g_free(manufacturer_data);
}

int _bt_convert_gerror(GError *g_error)
{
	int ret = BLUETOOTH_ERROR_INTERNAL;
	gchar *str;

	if (!g_dbus_error_strip_remote_error(g_error))
		return ret;

	str = g_error->message;

	if (g_strcmp0(str, "Connection refused") == 0)
		ret = BLUETOOTH_ERROR_AUTHENTICATION_REJECTED;
	else if (g_strcmp0(str, "Connection timed out") == 0)
		ret = BLUETOOTH_ERROR_TIMEOUT;
	else if (g_strcmp0(str, "In Progress") == 0)
		ret = BLUETOOTH_ERROR_IN_PROGRESS;

	return ret;
}
