/*
 * Copyright (c) 2011 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <glib.h>
#include <gio/gio.h>
#include <dlog.h>
#include <string.h>
#include <syspopup_caller.h>
#include <vconf.h>
#include <bundle_internal.h>

#include "bluetooth-api.h"
#include "bt-internal-types.h"

#include "bt-service-common.h"
#include "bt-service-device.h"
#include "bt-service-proximity.h"

char *_bt_convert_alert_level_to_string(int value)
{
	if (value == BT_PXP_ALERT_MILD)
		return g_strdup("mild");
	else if (value == BT_PXP_ALERT_HIGH)
		return g_strdup("high");
	else
		return g_strdup("none");
}

int _bt_convert_string_to_alert_level(const char *str)
{
	if (g_strcmp0("high", str) == 0)
		return BT_PXP_ALERT_HIGH;
	else if (g_strcmp0("mild", str) == 0)
		return BT_PXP_ALERT_MILD;

	return BT_PXP_ALERT_NONE;
}

char *_bt_convert_property_to_string(int value)
{
	if (value == BT_PXP_PROPERTY_LLS)
		return g_strdup("LinkLossAlertLevel");
	else if (value == BT_PXP_PROPERTY_IAS)
		return g_strdup("ImmediateAlertLevel");
	else if (value == BT_PXP_PROPERTY_TX_POWER)
		return g_strdup("SignalLevel");

	return NULL;
}

int bt_set_proximity_property(bluetooth_device_address_t *device_address,
		unsigned int property, int alert_level)
{
	GDBusProxy *proxy;
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };

	GDBusConnection *conn;
	char *device_path = NULL;
	GError *error = NULL;
	GVariant *ret = NULL;
	char *value_str = NULL;
	char *property_str = NULL;

	BT_CHECK_PARAMETER(device_address, return);

	conn = _bt_gdbus_get_system_gconn();
	retv_if(conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	_bt_convert_addr_type_to_string(address, device_address->addr);

	device_path = _bt_get_device_object_path(address);

	if (device_path == NULL)
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	else
		BT_INFO("device_path is created[%s]", device_path);

	value_str = _bt_convert_alert_level_to_string(alert_level);
	property_str = _bt_convert_property_to_string(property);

	if (value_str == NULL || property_str == NULL) {
		g_free(property_str);
		g_free(value_str);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	proxy = g_dbus_proxy_new_sync(conn, G_DBUS_PROXY_FLAGS_NONE,
							NULL, BT_BLUEZ_NAME,
							device_path, BT_PROPERTIES_INTERFACE,  NULL, NULL);

	g_free(device_path);
	retv_if(proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	ret = g_dbus_proxy_call_sync(proxy, "Set",
				g_variant_new("(ssv)", BT_PROXIMITY_MONITOR_INTERFACE,  property_str, g_variant_new("s", value_str)),
				G_DBUS_CALL_FLAGS_NONE,
				-1,
				NULL,
				&error);
	if (ret)
		g_variant_unref(ret);
	g_object_unref(proxy);
	g_free(property_str);
	g_free(value_str);

	if (error) {
		 BT_ERR("SetProperty error: [%s]", error->message);
		 g_error_free(error);
		 return BLUETOOTH_ERROR_INTERNAL;
	}

	return BLUETOOTH_ERROR_NONE;
}

int bt_get_proximity_property(bluetooth_device_address_t *device_address,
		unsigned int property, int *alert_level)
{
	GDBusProxy *proxy;
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };
	GDBusConnection *conn;
	char *device_path = NULL;
	GError *error = NULL;
	GVariant *result = NULL;
	GVariant *tmp_value;
	GVariant *value;
	char *value_str = NULL;
	char *property_str = NULL;

	BT_CHECK_PARAMETER(device_address, return);

	conn = _bt_gdbus_get_system_gconn();
	retv_if(conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	_bt_convert_addr_type_to_string(address, device_address->addr);

	device_path = _bt_get_device_object_path(address);

	if (device_path == NULL)
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	else
		BT_INFO("device_path is created[%s]", device_path);

	property_str = _bt_convert_property_to_string(property);

	if (property_str == NULL)
		return BLUETOOTH_ERROR_INTERNAL;

	proxy = g_dbus_proxy_new_sync(conn, G_DBUS_PROXY_FLAGS_NONE,
							NULL, BT_BLUEZ_NAME,
							device_path, BT_PROPERTIES_INTERFACE,  NULL, NULL);

	g_free(device_path);
	retv_if(proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	result = g_dbus_proxy_call_sync(proxy, "GetAll",
			g_variant_new("(s)", BT_PROXIMITY_MONITOR_INTERFACE),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);
	if (result == NULL) {
		if (error != NULL) {
			BT_ERR("Error occured in Proxy call [%s]\n", error->message);
			g_error_free(error);
		}
		g_object_unref(proxy);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	g_variant_get(result , "(@a{sv})", &value);
	g_variant_unref(result);

	tmp_value = g_variant_lookup_value(value, property_str, G_VARIANT_TYPE_STRING);
	if (tmp_value == NULL) {
		g_object_unref(proxy);
		g_variant_unref(value);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	value_str = (char *)g_variant_get_string(tmp_value, NULL);
	if (value_str)
		*alert_level = _bt_convert_string_to_alert_level(value_str);

	g_variant_unref(tmp_value);
	g_variant_unref(value);
	g_object_unref(proxy);
	g_free(property_str);

	return BLUETOOTH_ERROR_NONE;
}

int bt_get_proximity_supported_services(bluetooth_device_address_t *device_address,
		unsigned int *supported_services)
{
	GDBusProxy *proxy;
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };

	GDBusConnection *conn;
	char *device_path = NULL;
	GError *error = NULL;
	GVariant *result = NULL;
	GVariant *tmp_value;
	GVariant *value;

	BT_CHECK_PARAMETER(device_address, return);

	conn = _bt_gdbus_get_system_gconn();
	retv_if(conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	_bt_convert_addr_type_to_string(address, device_address->addr);

	device_path = _bt_get_device_object_path(address);

	if (device_path == NULL)
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	else
		BT_INFO("device_path is created[%s]", device_path);

	proxy = g_dbus_proxy_new_sync(conn, G_DBUS_PROXY_FLAGS_NONE,
							NULL, BT_BLUEZ_NAME,
							device_path, BT_PROPERTIES_INTERFACE,  NULL, NULL);

	g_free(device_path);
	retv_if(proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	result = g_dbus_proxy_call_sync(proxy, "GetAll",
			g_variant_new("(s)", BT_PROXIMITY_MONITOR_INTERFACE),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);
	if (result == NULL) {
		if (error != NULL) {
			BT_ERR("Error occured in Proxy call [%s]\n", error->message);
			g_error_free(error);
		}
		g_object_unref(proxy);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	g_variant_get(result , "(@a{sv})", &value);
	g_variant_unref(result);

	*supported_services = 0;
	tmp_value = g_variant_lookup_value(value, "LinkLossAlertLevel", G_VARIANT_TYPE_STRING);
	if (tmp_value == NULL) {
		g_object_unref(proxy);
		g_variant_unref(value);
		return BLUETOOTH_ERROR_INTERNAL;
	} else {
		*supported_services |= BT_PXP_PROPERTY_LLS;
		g_variant_unref(tmp_value);
	}

	tmp_value = g_variant_lookup_value(value, "ImmediateAlertLevel", G_VARIANT_TYPE_STRING);
	if (tmp_value == NULL) {
		if (*supported_services == 0) {
			g_object_unref(proxy);
			g_variant_unref(value);
			return BLUETOOTH_ERROR_INTERNAL;
		}
	} else {
		*supported_services |= BT_PXP_PROPERTY_IAS;
		g_variant_unref(tmp_value);
	}

	tmp_value = g_variant_lookup_value(value, "SignalLevel", G_VARIANT_TYPE_STRING);
	if (tmp_value == NULL) {
		if (*supported_services == 0) {
			g_object_unref(proxy);
			g_variant_unref(value);
			return BLUETOOTH_ERROR_INTERNAL;
		}
	} else {
		*supported_services |= BT_PXP_PROPERTY_TX_POWER;
		g_variant_unref(tmp_value);
	}

	g_variant_unref(value);
	g_object_unref(proxy);

	return BLUETOOTH_ERROR_NONE;
}

int bt_register_proximity_reporter()
{
	GDBusProxy *proxy;

	GDBusConnection *conn;
	char *adapter_path = NULL;
	GError *error = NULL;
	GVariant *result = NULL;

	conn = _bt_gdbus_get_system_gconn();
	retv_if(conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	adapter_path = _bt_get_adapter_path();
	if (adapter_path == NULL) {
		BT_ERR("Could not get adapter path\n");
		return BLUETOOTH_ERROR_DEVICE_NOT_ENABLED;
	}

	BT_INFO("Adapter path %s", adapter_path);

	proxy = g_dbus_proxy_new_sync(conn, G_DBUS_PROXY_FLAGS_NONE,
					NULL, BT_BLUEZ_NAME, adapter_path,
					BT_PROXIMITY_REPORTER_INTERFACE, NULL, NULL);

	g_free(adapter_path);
	retv_if(proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	result = g_dbus_proxy_call_sync(proxy, "RegisterProximity",
			NULL, G_DBUS_CALL_FLAGS_NONE, -1,
			NULL, &error);
	if (result == NULL) {
		if (error != NULL) {
			BT_ERR("Error occured in Proxy call [%s]\n", error->message);
			g_error_free(error);
		}
		g_object_unref(proxy);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	g_object_unref(proxy);

	return BLUETOOTH_ERROR_NONE;
}

int bt_unregister_proximity_reporter()
{
	GDBusProxy *proxy;

	GDBusConnection *conn;
	char *adapter_path = NULL;
	GError *error = NULL;
	GVariant *result = NULL;

	conn = _bt_gdbus_get_system_gconn();
	retv_if(conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	adapter_path = _bt_get_adapter_path();
	if (adapter_path == NULL) {
		BT_ERR("Could not get adapter path\n");
		return BLUETOOTH_ERROR_DEVICE_NOT_ENABLED;
	}

	BT_INFO("Adapter path %s", adapter_path);

	proxy = g_dbus_proxy_new_sync(conn, G_DBUS_PROXY_FLAGS_NONE,
					NULL, BT_BLUEZ_NAME, adapter_path,
					BT_PROXIMITY_REPORTER_INTERFACE, NULL, NULL);

	g_free(adapter_path);
	retv_if(proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	result = g_dbus_proxy_call_sync(proxy, "UnregisterProximity",
			NULL, G_DBUS_CALL_FLAGS_NONE, -1,
			NULL, &error);
	if (result == NULL) {
		if (error != NULL) {
			BT_ERR("Error occured in Proxy call [%s]\n", error->message);
			g_error_free(error);
		}
		g_object_unref(proxy);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	g_object_unref(proxy);

	return BLUETOOTH_ERROR_NONE;
}
