/*
 * Bluetooth-frwk
 *
 * Contact: Anupam Roy (anupam.r@samsung.com)
 *
 * Copyright (c) 2012-2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef __BLUETOOTH_GATT_CLIENT_API_H
#define __BLUETOOTH_GATT_CLIENT_API_H

#include <stdint.h>
#include <glib.h>
#include <dlog.h>

#include <bluetooth-api.h>
#include <bluetooth-gatt-server-api.h>

#define BLUETOOTH_GATT_ATT_DATA_LENGTH_MAX      610 /**< GATT ATT value max len */
#define BLUETOOTH_UUID_HEX_MAX_LEN      16         /**< UUID complete len in hex */

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
	int event;
	int result;
	void *param_data;
	void *user_data;
} gatt_client_event_param_t;

typedef struct {
	bluetooth_device_address_t device_addr;
	int count;
	char uuids[BLUETOOTH_MAX_SERVICES_FOR_DEVICE][BLUETOOTH_UUID_STRING_MAX];
	int inst_id[BLUETOOTH_MAX_SERVICES_FOR_DEVICE];
	int primary[BLUETOOTH_MAX_SERVICES_FOR_DEVICE];
} bt_services_browse_info_t;


typedef enum {
	BLUETOOTH_GATT_TYPE_WRITE_NO_RESPONSE = 0x01,      /* Write Without Response */
	BLUETOOTH_GATT_TYPE_WRITE = 0x2,                           /* Write */
} bluetooth_gatt_write_type_e;


/* Structure used for browsing a particular service's properties(chars and Incldued svcs) */
typedef struct {
	bluetooth_device_address_t device_address;      /**< device address */
	bt_gatt_handle_property_t svc;
} bluetooth_gatt_client_svc_prop_info_t;

typedef struct {
	bluetooth_device_address_t device_addr;
	int count;
	unsigned char svc_uuid[BLUETOOTH_UUID_HEX_MAX_LEN];
	int svc_inst_id;
	char uuids[BLUETOOTH_MAX_SERVICES_FOR_DEVICE][BLUETOOTH_UUID_STRING_MAX];
	int inst_id[BLUETOOTH_MAX_SERVICES_FOR_DEVICE];
	int props[BLUETOOTH_MAX_SERVICES_FOR_DEVICE];
} bt_char_browse_info_t;

typedef struct {
	bluetooth_device_address_t device_addr;
	int count;
	unsigned char svc_uuid[BLUETOOTH_UUID_HEX_MAX_LEN];
	int svc_inst_id;
	char uuids[BLUETOOTH_MAX_SERVICES_FOR_DEVICE][BLUETOOTH_UUID_STRING_MAX];
	int inst_id[BLUETOOTH_MAX_SERVICES_FOR_DEVICE];
} bt_included_services_browse_info_t;

typedef struct {
	bluetooth_device_address_t device_addr;
	int count;
	unsigned char svc_uuid[BLUETOOTH_UUID_HEX_MAX_LEN];
	int svc_inst_id;
	unsigned char char_uuid[BLUETOOTH_UUID_HEX_MAX_LEN];
	int char_inst_id;
	int char_props_map;
	char uuids[BLUETOOTH_MAX_SERVICES_FOR_DEVICE][BLUETOOTH_UUID_STRING_MAX];
	int inst_id[BLUETOOTH_MAX_SERVICES_FOR_DEVICE];
} bt_descriptor_browse_info_t;

/* Structure used for browsing a particular charactertic's properties(descriptors)  or reading char's value */
typedef struct {
	bluetooth_device_address_t device_address;      /**< device address */
	bt_gatt_handle_property_t svc;
	bt_gatt_handle_property_t characteristic;
} bluetooth_gatt_client_char_prop_info_t;

/* Structure used for reading Descriptor's value */
typedef struct {
	bluetooth_device_address_t device_address;      /**< device address */
	bt_gatt_handle_property_t svc;
	bt_gatt_handle_property_t characteristic;
	bt_gatt_handle_property_t descriptor;
} bluetooth_gatt_client_desc_prop_info_t;


typedef void (*gatt_client_cb_func_ptr)(int, gatt_client_event_param_t *, void *);

/**
 * @fn int bluetooth_gatt_client_init(int *client_id, gatt_client_cb_func_ptr callback_ptr, void *user_data)
 * @brief Initialize GATT Client instance with BT stack and register the callback
 *
 * This function is a synchronous call.
 *
 * @remark      None
 *
 */
int bluetooth_gatt_client_init(int *client_id,
		const bluetooth_device_address_t *address,
		gatt_client_cb_func_ptr callback_ptr);

int bluetooth_gatt_client_deinit(int client_id);

int bluetooth_gatt_client_get_primary_services(const bluetooth_device_address_t *address,
		bt_gatt_handle_info_t *prim_svc);

int bluetooth_gatt_client_get_service_property(const char *address,
		bt_gatt_handle_property_t *service_handle,
		bt_gatt_service_property_t *service);

int bluetooth_gatt_client_get_characteristics_property(const char *address,
		bt_gatt_handle_property_t *service_handle,
		bt_gatt_handle_property_t *char_handle,
		bt_gatt_char_property_t *char_prop);

int bluetooth_gatt_client_get_char_descriptor_property(const char *address,
		bt_gatt_handle_property_t *service_handle,
		bt_gatt_handle_property_t *char_handle,
		bt_gatt_handle_property_t *descriptor_handle,
		bt_gatt_char_descriptor_property_t *desc_prop);

int bluetooth_gatt_client_read_characteristic_value(const char *address,
		bt_gatt_handle_property_t *service_handle,
		bt_gatt_handle_property_t *char_handle);

int bluetooth_gatt_client_read_descriptor_value(const char *address,
		bt_gatt_handle_property_t *service_handle,
		bt_gatt_handle_property_t *char_handle,
		bt_gatt_handle_property_t *descriptor_handle);


int bluetooth_gatt_client_write_characteristic_value_by_type(
		const char *address,
		bt_gatt_handle_property_t *service_handle,
		bt_gatt_handle_property_t *char_handle,
		bluetooth_gatt_att_data_t *data,
		bluetooth_gatt_write_type_e write_type);

int bluetooth_gatt_client_write_descriptor_value(
		const char *address,
		bt_gatt_handle_property_t *service_handle,
		bt_gatt_handle_property_t *char_handle,
		bt_gatt_handle_property_t *descriptor_handle,
		bluetooth_gatt_att_data_t *data,
		bluetooth_gatt_write_type_e write_type);

int bluetooth_gatt_client_set_service_change_watcher(
		const bluetooth_device_address_t *address,
		gboolean enable);

int bluetooth_gatt_client_watch_characteristics(
			const char *address,
			bt_gatt_handle_property_t *service_handle,
			bt_gatt_handle_property_t *char_handle,
			int client_id,
			gboolean is_notify,
			gboolean is_indicate);

#ifdef __cplusplus
}
#endif
#endif /* __BLUETOOTH_GATT_CLIENT_API_H */
