/*
 * Bluetooth-frwk
 *
 * Contact: Anupam Roy (anupam.r@samsung.com)
 *
 * Copyright (c) 2012-2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef __BLUETOOTH_GATT_SERVER_API_H
#define __BLUETOOTH_GATT_SERVER_API_H

#include <stdint.h>
#include <glib.h>
#include <dlog.h>

#include <bluetooth-api.h>

#define BLUETOOTH_GATT_ATT_DATA_LENGTH_MAX      610 /**< GATT ATT value max len */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * GATT Server Request type
 */
typedef enum {
	BLUETOOTH_GATT_REQUEST_TYPE_READ = 0x00,       /* Read Requested */
	BLUETOOTH_GATT_REQUEST_TYPE_WRITE = 0x01,      /* Write Requested */
	BLUETOOTH_GATT_REQUEST_TYPE_EXEC_WRITE = 0x02, /* Exec Write Requested */
	BLUETOOTH_GATT_REQUEST_TYPE_ACQUIRE_WRITE = 0x03,
	BLUETOOTH_GATT_REQUEST_TYPE_ACQUIRE_NOTIFY = 0x04
} bluetooth_gatt_att_request_type_e;

/**
 * GATT Server Response/Indication data
 */
typedef struct {
	guint8 data[BLUETOOTH_GATT_ATT_DATA_LENGTH_MAX];
	int length;
	int offset;
} bluetooth_gatt_att_data_t;

/**
 * GATT Server service add data
 */
typedef struct {
	int service_type;
	int num_handles;
	int instance_id;
} bluetooth_gatt_server_service_params_t;

/**
 * GATT Server attributes(characteristic/descriptors) add data
 */
typedef struct {
	bt_gatt_permission_t permissions;
	bt_gatt_characteristic_property_t properties;
	int service_handle;
	int instance_id;
} bluetooth_gatt_server_attribute_params_t;

/**
 * GATT Server Indication data
 */
typedef struct {
	int atrribute_handle;
	bluetooth_device_address_t device_address;      /**< device address */
	gboolean need_confirmation;
	int instance_id;
} bluetooth_gatt_server_indication_params_t;

/**
 * GATT Server connection/disconnection info
 */
typedef struct {
	bluetooth_device_address_t device_address;      /** < Device address */
	int server_instance;				/** < GATT Server Instance Id */
} bluetooth_gatt_server_connection_info_t;

/**
* GATT Server Service Added info
*/
typedef struct {
	int server_instance;       /** < GATT Server Instance Id */
	int service_handle;        /** < GATT Server Service handle */
	int is_primary;            /** < GATT Server Is Primary Service */
	char uuid[BLUETOOTH_UUID_STRING_MAX];
} bluetooth_gatt_server_service_added_info_t;

/**
* GATT Server Included Service Added info
*/
typedef struct {
	int server_instance;               /** < GATT Server Instance Id */
	int service_handle;                /** < GATT Server Service handle */
	int included_service_handle;       /** < GATT Server Included service handle */
} bluetooth_gatt_server_included_service_added_info_t;


/**
 * GATT Server write attribute data received from Remote GATT client: Max Size is 600 bytes
 */
typedef struct {
	char data[BLUETOOTH_GATT_ATT_DATA_LENGTH_MAX + 1];
} bluetooth_gatt_server_write_data_t;


/**
* GATT Server Attribute Added Info
*/
typedef struct {
	int server_instance;	/** < GATT Server Instance Id */
	int service_handle;	/** < GATT Server Service handle */
	int attr_handle;	/** < GATT Server Attribute Handle (Characteristic or Descriptor) */
	char uuid[BLUETOOTH_UUID_STRING_MAX];	/** < GATT Server Atribute UUID */
} bluetooth_gatt_server_attribute_added_info_t;

/**
* GATT Server Service Info
*/
typedef struct {
	int server_instance;		/** < GATT Server Instance Id */
	int service_handle;		/** < GATT Server Service handle */
} bluetooth_gatt_server_service_status_t;

/**
* GATT Server Read Requested Info
*/
typedef struct {
	int connection_id;		/** < GATT Server Connection ID */
	int request_id;			/** < GATT Server Read Request ID */
	int attribute_handle;		/** < GATT Server att handle */
	int offset;			/** < GATT Server attribute offset */
	gboolean is_long;		/** < GATT Server Is Long attribute  */
	bluetooth_device_address_t device_address;	/**< Remote GATT client device address */
} bluetooth_gatt_server_read_requested_info_t;

/**
* GATT Server Write Requested Info
*/
typedef struct {
	int connection_id;                              /** < GATT Server Connection ID */
	int request_id;                                 /** < GATT Server Write Request ID */
	int attribute_handle;                           /** < GATT Server Write att handle */
	int offset;	                        /** < GATT Server attribute offset */
	gboolean need_resp;                                 /** < GATT Server Write req needs response  */
	gboolean is_prep_write;                             /** < GATT Server Is Prepared write  */
	int length;	                        /** < GATT Server Write attribute length  */
	bluetooth_gatt_server_write_data_t data;        /** < Write attribute data */
	bluetooth_device_address_t device_address;      /** < Remote GATT client device address */
} bluetooth_gatt_server_write_requested_info_t;

/**
* GATT Server Write Requested Info
*/
typedef struct {
	int connection_id;                              /** < GATT Server Connection ID */
	int request_id;                                 /** < GATT Server Write Request ID */
	int attribute_handle;                           /** < GATT Server Write att handle */
	int offset;	                        /** < GATT Server attribute offset */
	char address[BLUETOOTH_ADDRESS_STRING_LENGTH];
} bluetooth_gatt_server_acquire_write_info_t;;


/**
 * GATT Server Update Characteristic Value
 */
typedef struct {
	int attribute_handle;                           /** < GATT Server Write att handle */
	int length;                        /** < GATT Server Write attribute length  */
	bluetooth_gatt_server_write_data_t data;        /** < Write attribute data */
} bluetooth_gatt_server_update_value_t;

/**
 * GATT Server Exec write Requested Info
 */
typedef struct {
	int connection_id;	/** < GATT Server Connection ID */
	int request_id;		/** < GATT Server Exec Write Request ID */
	int exec_write;		/** < GATT Server Exec Write */
	bluetooth_device_address_t device_address;	/** < Remote GATT client device address */
} bluetooth_gatt_server_exec_write_requested_info_t;

/**
 * GATT Server Indicate confirmed Info
 */
typedef struct {
	int handle;		/** < GATT Attribute handle */
	gboolean completed;	/** < GATT Server Exec Write Request ID */
	bluetooth_device_address_t device_address;	/** < Remote GATT client device address */
} bluetooth_gatt_server_indicate_confirmed_info_t;

/**
 * GATT Server Indicate confirmed Info
 */
typedef struct {
	int handle;	/** < GATT Attribute handle */
	gboolean notification;                          /** < GATT CCCD Notification changed */
	bluetooth_device_address_t device_address;	/** < Remote GATT client device address */
} bluetooth_gatt_server_notification_changed_t;

/**
 * GATT Server response data
 */
typedef struct {
	bluetooth_gatt_att_request_type_e req_type;
	int request_id;
	int response_status;
	gboolean auth_req;
} bluetooth_gatt_server_response_params_t;

/**
 * GATT Server acquire response data
 */
typedef struct {
	bluetooth_gatt_att_request_type_e req_type;
	int request_id;
	int fd;
	int mtu;
} bluetooth_gatt_server_acquire_response_params_t;

/**
 * GATT Server Bluedroid Specific Permission Mask
 */
typedef enum {
	BLUETOOTH_PRODUCT_GATT_PERMISSION_READ = 0x01,
	BLUETOOTH_PRODUCT_GATT_PERMISSION_ENCRYPT_READ = 0x02,
	BLUETOOTH_PRODUCT_GATT_PERMISSION_ENCRYPT_READ_MITM = 0x04,
	BLUETOOTH_PRODUCT_GATT_PERMISSION_WRITE = 0x10,
	BLUETOOTH_PRODUCT_GATT_PERMISSION_ENCRYPT_WRITE = 0x20,
	BLUETOOTH_PRODUCT_GATT_PERMISSION_ENCRYPT_WRITE_MITM = 0x40,
	BLUETOOTH_PRODUCT_GATT_PERMISSION_WRITE_SIGNED = 0x80,
	BLUETOOTH_PRODUCT_GATT_PERMISSION_WRITE_SIGNED_MITM = 0x100,
	BLUETOOTH_PRODUCT_GATT_PERMISSION_INVALID = 0xffff,
} bt_product_gatt_permission_t;

typedef struct {
	int event;
	int result;
	void *param_data;
	void *user_data;
} gatt_server_event_param_t;

typedef void (*gatt_server_cb_func_ptr)(int, gatt_server_event_param_t *, void *);

/**
 * @fn int bluetooth_gatt_server_init(gatt_server_cb_func_ptr callback_ptr, void *user_data)
 * @brief Initialize GATT Server and register the callback
 *
 * This function is a synchronous call.
 *
 * @remark      None
 *
 */
int bluetooth_gatt_server_init(int *instance_id, gatt_server_cb_func_ptr callback_ptr,
		void *user_data);

//int bluetooth_gatt_register_application(int instance_id);
//int bluetooth_gatt_server_init(gatt_server_cb_func_ptr callback_ptr, void *user_data);

int bluetooth_gatt_server_deinit(void);

int bluetooth_gatt_server_register(const gchar *name, int *instance_id);

int bluetooth_gatt_server_unregister(int instance_id);

int bluetooth_gatt_server_connect(bluetooth_device_address_t *addr_hex, int instance_id);

int bluetooth_gatt_server_disconnect(bluetooth_device_address_t *addr_hex, int instance_id);

int bluetooth_gatt_server_add_service(const char *svc_uuid, int type, int numhandles, int instance_id, int *service_handle);

int bluetooth_gatt_server_add_included_service(int instance_id, int service_handle, int included_handle);

int bluetooth_gatt_server_add_new_characteristic(const char *char_uuid, const bluetooth_gatt_server_attribute_params_t *param, int *char_handle);


int bluetooth_gatt_server_add_descriptor(const char *desc_uuid, bt_gatt_permission_t permissions, int service_handle,
		int instance_id, int *descriptor_handle);

int bluetooth_gatt_server_start_service(int service_handle, int instance_id);

int bluetooth_gatt_server_stop_service(int service_handle, int instance_id);

int bluetooth_gatt_server_delete_service(int service_handle, int instance_id);

int bluetooth_gatt_server_send_indication(bluetooth_device_address_t *addr_hex,
		const bluetooth_gatt_server_indication_params_t *param,
		const bluetooth_gatt_att_data_t *att_value);

int bluetooth_gatt_server_send_response(const bluetooth_gatt_server_response_params_t *param,
		const bluetooth_gatt_att_data_t *value);

int bluetooth_gatt_server_update_multi_adv_instance(const gchar *unique_name, int instance_id);

int bluetooth_gatt_server_update_characteristic(int instance_id, const bluetooth_gatt_server_update_value_t *value);

#ifdef __cplusplus
}
#endif
#endif /* __BLUETOOTH_GATT_SERVER_API_H */
