/*
 * Copyright (c) 2015 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <syspopup_caller.h>
#include <bundle.h>
#include <bundle_internal.h>
#include <tethering.h>
#include <mime_type.h>
#include <app_control.h>

#include "plugin.h"
#include "bluetooth-api.h"

#define BT_AGENT_SYSPOPUP_MAX_ATTEMPT 3
#define BT_MAX_EVENT_STR_LENGTH	50
#define BT_AGENT_SYSPOPUP_TIMEOUT_FOR_MULTIPLE_POPUPS 200

#ifdef TIZEN_FEATURE_BT_DPM
int bt_launch_dpm_popup(char *mode)
{
	int ret = 0;
	bundle *b;

	b = bundle_create();
	retv_if(b == NULL, BLUETOOTH_ERROR_INTERNAL);

	bundle_add(b, "mode", mode);

	ret = syspopup_launch(BT_DPM_SYSPOPUP, b);

	if (ret < 0)
		BT_ERR("Popup launch failed: %d\n", ret);

	bundle_free(b);

	return ret;
}
#endif

static gboolean _bt_syspopup_timer_cb(gpointer user_data)
{
	int ret;
	bundle *b;
	retv_if(user_data == NULL, FALSE);

	b = (bundle *)user_data;

	ret = syspopup_launch("bt-syspopup", b);
	if (ret < 0) {
		BT_ERR("Sorry!! Cannot launch popup return = %d, Retrying...", ret);
	} else {
		BT_DBG("Hurray!!! Finally Popup launched");
		bundle_free(b);
	}
	return (ret < 0) ? TRUE : FALSE;
}


static gboolean _bt_agent_system_popup_timer_cb(gpointer user_data)
{
	int ret;
	static int retry_count;
	bundle *b = (bundle *)user_data;
	retv_if(user_data == NULL, FALSE);

	++retry_count;

	ret = syspopup_launch("bt-syspopup", b);
	if (ret < 0) {
		BT_ERR("Sorry! Can't launch popup, ret=%d, Re-try[%d] time..",
							ret, retry_count);
		if (retry_count >= BT_AGENT_SYSPOPUP_MAX_ATTEMPT) {
			BT_ERR("Sorry!! Max retry %d reached", retry_count);
			bundle_free(b);
			retry_count = 0;
			return FALSE;
		}
	} else {
		BT_DBG("Hurray!! Finally Popup launched");
		retry_count = 0;
		bundle_free(b);
	}

	return (ret < 0) ? TRUE : FALSE;
}

int bt_launch_system_popup(bt_agent_event_type_t event_type,
							const char *device_name,
							const unsigned char *auth_info,
							char *passkey,
							const char *filename,
							const char *agent_path)
{
	int ret;
	bundle *b;
	char event_str[BT_MAX_EVENT_STR_LENGTH + 1];

	b = bundle_create();
	if (!b) {
		BT_ERR("Launching system popup failed");
		return -1;
	}

	bundle_add(b, "device-name", device_name);
	bundle_add(b, "auth-info", (const char *)auth_info);
	bundle_add(b, "passkey", passkey);
	bundle_add(b, "file", filename);
	bundle_add(b, "agent-path", agent_path);

	switch (event_type) {
	case BT_AGENT_EVENT_PIN_REQUEST:
		g_strlcpy(event_str, "pin-request", sizeof(event_str));
		break;

	case BT_AGENT_EVENT_PASSKEY_CONFIRM_REQUEST:
		g_strlcpy(event_str, "passkey-confirm-request",
						sizeof(event_str));
		break;

	case BT_AGENT_EVENT_PASSKEY_REQUEST:
		g_strlcpy(event_str, "passkey-request", sizeof(event_str));
		break;

	case BT_AGENT_EVENT_PASSKEY_DISPLAY_REQUEST:
		g_strlcpy(event_str, "passkey-display-request",
						sizeof(event_str));
		break;

	case BT_AGENT_EVENT_AUTHORIZE_REQUEST:
		g_strlcpy(event_str, "authorize-request",
						sizeof(event_str));
		break;

	case BT_AGENT_EVENT_CONFIRM_MODE_REQUEST:
		g_strlcpy(event_str, "confirm-mode-request",
						sizeof(event_str));
		break;

	case BT_AGENT_EVENT_FILE_RECEIVED:
		g_strlcpy(event_str, "file-received", sizeof(event_str));
		break;

	case BT_AGENT_EVENT_KEYBOARD_PASSKEY_REQUEST:
		g_strlcpy(event_str, "keyboard-passkey-request",
						sizeof(event_str));
		break;

	case BT_AGENT_EVENT_TERMINATE:
		g_strlcpy(event_str, "terminate", sizeof(event_str));
		break;

	case BT_AGENT_EVENT_EXCHANGE_REQUEST:
		g_strlcpy(event_str, "exchange-request", sizeof(event_str));
		break;

	case BT_AGENT_EVENT_PBAP_REQUEST:
		g_strlcpy(event_str, "phonebook-request", sizeof(event_str));
		break;

	case BT_AGENT_EVENT_MAP_REQUEST:
		g_strlcpy(event_str, "message-request", sizeof(event_str));
		break;

	case BT_AGENT_EVENT_LEGACY_PAIR_FAILED_FROM_REMOTE:
		g_strlcpy(event_str, "remote-legacy-pair-failed", sizeof(event_str));
		break;

	default:
		BT_ERR("Invalid event type");
		bundle_free(b);
		return -1;

	}

	bundle_add(b, "event-type", event_str);

	ret = syspopup_launch("bt-syspopup", b);
	if (0 > ret) {
		BT_ERR("Popup launch failed...retry %d", ret);

		g_timeout_add(BT_AGENT_SYSPOPUP_TIMEOUT_FOR_MULTIPLE_POPUPS,
			      (GSourceFunc)_bt_agent_system_popup_timer_cb, b);
	} else {
		bundle_free(b);
	}

	BT_INFO("_bt_agent_launch_system_popup");
	return 0;
}

bool bt_launch_unable_to_pairing_syspopup(int result)
{
	BT_DBG("+");
	int ret = 0;
	bundle *b = NULL;

	b = bundle_create();
	if (b == NULL)
		return false;

	bundle_add(b, "event-type", "unable-to-pairing");

	if (result == BLUETOOTH_ERROR_TIMEOUT)
		bundle_add(b, "error", "timeout");
	else if (result == BLUETOOTH_ERROR_AUTHENTICATION_FAILED)
		bundle_add(b, "error", "authfailed");
	else
		bundle_add(b, "error", "error");

	ret = syspopup_launch("bt-syspopup", b);
	if (0 > ret) {
		BT_ERR("Popup launch failed...retry %d \n", ret);
		g_timeout_add(200, (GSourceFunc) _bt_syspopup_timer_cb,
				b);
	} else {
		bundle_free(b);
	}

	BT_DBG("-");
	return true;
}


bool bt_is_tethering_enabled(void)
{
	BT_DBG("+");

	tethering_h tethering = NULL;
	bool enabled = FALSE;
	int ret;

	ret = tethering_create(&tethering);

	if (ret != TETHERING_ERROR_NONE) {
		BT_ERR("Fail to create tethering: %d", ret);
		return FALSE;
	}

	enabled = tethering_is_enabled(tethering, TETHERING_TYPE_BT);

	if (enabled != true)
		BT_ERR("BT tethering is not enabled");

	ret = tethering_destroy(tethering);

	if (ret != TETHERING_ERROR_NONE)
		BT_ERR("Fail to destroy tethering: %d", ret);

	return enabled;
}

int bt_get_mime_type(char *file_name, char **mime_type)
{
	BT_DBG("+");

	int ret = MIME_TYPE_ERROR_NONE;

	if (file_name == NULL)
		return BLUETOOTH_ERROR_INVALID_PARAM;

	ret = mime_type_get_mime_type(file_name, (gchar **)mime_type);

	if (ret != MIME_TYPE_ERROR_NONE) {
		BT_ERR("Fail to get mime type: %d", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	return BLUETOOTH_ERROR_NONE;
}

void bt_destroy_popup_all(void)
{
	syspopup_destroy_all();
}

int bt_app_control_send_launch_request(char *absolute_path)
{
	BT_DBG("+");

	app_control_h app_control;
	int ret;

	app_control_create(&app_control);
	app_control_set_operation(app_control, APP_CONTROL_OPERATION_VIEW);
	app_control_set_uri(app_control, absolute_path);
	app_control_set_mime(app_control, "*/*");

	if (app_control_send_launch_request(app_control, NULL, NULL) == APP_CONTROL_ERROR_NONE)
		ret = BLUETOOTH_ERROR_NONE;
	else
		ret = BLUETOOTH_ERROR_INTERNAL;

	app_control_destroy(app_control);

	return ret;
}

extern struct bluetooth_headed_plugin_t headed_plugin
__attribute__ ((visibility("default")));
struct bluetooth_headed_plugin_t headed_plugin = {
#ifdef TIZEN_FEATURE_BT_DPM
	bt_launch_dpm_popup,
#else
		NULL,
#endif
	bt_launch_system_popup,
	bt_destroy_popup_all,
	bt_launch_unable_to_pairing_syspopup,
	bt_is_tethering_enabled,
	bt_get_mime_type
};

